/*  This extension adds a data gathering facility to <kview>
    Copyright (C) 1999  Richard Gooch
    This code is distributed under the GNU General Public License
*/
#include <stdio.h>
#include <math.h>
#ifdef BENCHMARK
#  include <sys/time.h>
#  include <unistd.h>
#endif
#include <errno.h>
#include <X11/Intrinsic.h>
#include <X11/Xos.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <karma_iarray.h>
#include <karma_event.h>
#include <karma_gls.h>
#include <karma_ds.h>
#include <karma.h>
#include <k_event_codes.h>


/*  Private functions  */
STATIC_FUNCTION (flag init_clear_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );
STATIC_FUNCTION (flag init_data_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );
STATIC_FUNCTION (flag add_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );
STATIC_FUNCTION (flag xy_event_func,
		 (KEvent event, Connection connection, void **f_info) );


/*  Private data  */
static iarray array = NULL;
#ifdef BENCHMARK
static struct timeval start_time;
#endif


/*  Public functions follow  */

flag Xt_AppInitialise (XtAppContext ctx)
{
    fprintf (stderr, "Gather extension loaded and running\n");
    ds_event_register_func (init_clear_event_func, NULL);
    ds_event_register_func (init_data_event_func, NULL);
    ds_event_register_func (add_event_func, NULL);
    event_register_event_func (xy_event_func, NULL);
    return (TRUE);
}   /*  End Function Xt_AppInitialise  */


/*  Private functions follow */

static flag init_clear_event_func (void *object, multi_array *multi_desc,
				   CONST char *domain, CONST char *name)
/*  [SUMMARY] Process a "GatherInitClear" event.
    [PURPOSE] This function will initialise the gather cube.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    unsigned long xlen, ylen, zlen;
    packet_desc *top_pack_desc;
    char *top_packet;
    double value[2];

    if (strcmp (name, "GatherInitClear") != 0) return (FALSE);
    if (array != NULL) iarray_dealloc (array);
    array = NULL;
    top_pack_desc = multi_desc->headers[0];
    top_packet = multi_desc->data[0];
    if ( !ds_get_unique_named_value (top_pack_desc, top_packet, "xlen", NULL,
				     value) ) return (TRUE);
    xlen = value[0];
    if ( !ds_get_unique_named_value (top_pack_desc, top_packet, "ylen", NULL,
				     value) ) return (TRUE);
    ylen = value[0];
    if ( !ds_get_unique_named_value (top_pack_desc, top_packet, "zlen", NULL,
				     value) ) return (TRUE);
    zlen = value[0];
    array = iarray_create_3D (zlen, ylen, zlen, K_FLOAT);
    fprintf (stderr, "%s event processed, size: %lu * %lu * %lu\n",
	     name, xlen, ylen, zlen);
    return (TRUE);
}   /*  End Function init_clear_event_func  */

static flag init_data_event_func (void *object, multi_array *multi_desc,
				  CONST char *domain, CONST char *name)
/*  [SUMMARY] Process a "GatherInitData" event.
    [PURPOSE] This function will initialise the gather cube, filling it with
    the received data and broadcasting the entire cube.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    iarray arr;
    extern char *data_type_names[NUMTYPES];

    if (strcmp (name, "GatherInitData") != 0) return (FALSE);
    if (array != NULL) iarray_dealloc (array);
    array = NULL;
    if ( ( arr = iarray_get_from_multi_array (multi_desc, NULL, 3, NULL,
					      NULL) ) == NULL )
    {
	fprintf (stderr, "Unable to find 3-dimensional array in %s event\n",
		 name);
	return (TRUE);
    }
    if (iarray_type (arr) != K_FLOAT)
    {
	fprintf (stderr, "Array data type is %s, must be float\n",
		 data_type_names[iarray_type (arr)]);
	iarray_dealloc (arr);
	return (TRUE);
    }
    ds_event_dispatch (multi_desc, domain, "Dataset2");
    array = arr;
    fprintf (stderr, "%s event processed, %u attachments\n",
	     name, multi_desc->attachments);
#ifdef BENCHMARK
    gettimeofday (&start_time, NULL);
#endif
    return (TRUE);
}   /*  End Function init_data_event_func  */

static flag add_event_func (void *object, multi_array *multi_desc,
			    CONST char *domain, CONST char *name)
/*  [SUMMARY] Process a "GatherAdd" event.
    [PURPOSE] This function will update the gather cube and then broadcasts the
    entire cube.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    iarray sub;
    unsigned long xstart, ystart, zstart, x, y, z, xlen, ylen, zlen;
#ifdef BENCHMARK
    struct timeval stop_time;
#endif
    double value[2];
    extern char *data_type_names[NUMTYPES];

    if (strcmp (name, "GatherAdd") != 0) return (FALSE);
    if (array == NULL) return (TRUE);
    if ( ( sub = iarray_get_from_multi_array (multi_desc, NULL, 3, NULL,
					      NULL) ) == NULL )
    {
	fprintf (stderr, "Unable to find 3-dimensional array in %s event\n",
		 name);
	return (TRUE);
    }
    if (iarray_type (sub) != K_FLOAT)
    {
	fprintf (stderr, "Array data type is %s, must be float\n",
		 data_type_names[iarray_type (sub)]);
	iarray_dealloc (sub);
	return (TRUE);
    }
    if ( !iarray_get_named_value (sub, "xstart", NULL, value) ) return (TRUE);
    xstart = value[0];
    if ( !iarray_get_named_value (sub, "ystart", NULL, value) ) return (TRUE);
    ystart = value[0];
    if ( !iarray_get_named_value (sub, "zstart", NULL, value) ) return (TRUE);
    zstart = value[0];
    xlen = iarray_dim_length (sub, 2);
    if (iarray_dim_length (array, 2) <= xstart + xlen)
	xlen = iarray_dim_length (array, 2) - xstart;
    ylen = iarray_dim_length (sub, 1);
    if (iarray_dim_length (array, 1) <= ystart + ylen)
	ylen = iarray_dim_length (array, 1) - ystart;
    zlen = iarray_dim_length (sub, 0);
    if (iarray_dim_length (array, 0) <= zstart + zlen)
	zlen = iarray_dim_length (array, 0) - zstart;
    for (z = 0; z < zlen; ++z) for (y = 0; y < ylen; ++y)
	for (x = 0; x < xlen; ++x)
	{
	    F3 (array, z + zstart, y + ystart, x + xstart) = F3 (sub, z, y, x);
	}
    iarray_copy_attached (array, sub, FALSE, TRUE);
    iarray_dealloc (sub);
    ds_event_dispatch (iarray_multi_array (array), domain, "Dataset2");
#ifdef BENCHMARK
    if (xstart >= 105)
    {
	signed long time_taken;

	gettimeofday (&stop_time, NULL);
	time_taken = (stop_time.tv_usec - start_time.tv_usec) / 1000;
	time_taken += (stop_time.tv_sec - start_time.tv_sec) * 1000;
	fprintf (stderr, "Time taken: %ld ms\n", time_taken);
    }
#endif
    return (TRUE);
}   /*  End Function add_event_func  */

static flag xy_event_func (KEvent event, Connection connection, void **f_info)
/*  [SUMMARY] Generic event callback.
    <event> The event data. This becomes invalid when the callback returns.
    <connection> The connection over which the event came.
    <f_info> The arbitrary function information pointer.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    double pos[2];

    if (event.type != K_CANVAS_EVENT_POINTER_MOVE) return (FALSE);
    if (!event.data.point.image_pixel.defined) return (FALSE);
    pos[0] = event.data.point.image_pixel.x;
    pos[1] = event.data.point.image_pixel.y;
    gls_send_plain_event ("PointerMoveEvent", pos, K_DOUBLE, 2,0, NULL);
    return (TRUE);
}   /*  End Function xy_event_func  */
