/*  This extension adds a waterfall data generator facility to <kvis>
    Copyright (C) 2004  Richard Gooch
    This code is distributed under the GNU General Public License
*/
#include <stdio.h>
#include <math.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <X11/Intrinsic.h>
#include <X11/Xos.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Label.h>
#include <karma.h>
#include <karma_iarray.h>
#include <karma_ds.h>
#include <karma_n.h>
#include <karma_m.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/SimpleSlider.h>
#include <Xkw/ExclusiveMenu.h>


#define SCROLL_INTERVAL  10  /*  milliseconds  */
#define IMAGE_SIZE      512
#define MAX_VALUE       100.0

#define MODE_RANDOM                0
#define MODE_V                     1
#define NUM_MODE_CHOICES           2

static char *mode_choices[NUM_MODE_CHOICES] =
{
    "random", "V",
};


/*  Private functions  */
STATIC_FUNCTION (void _kvis_waterfall_toggle_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _kvis_waterfall_scroll_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _kvis_waterfall_timer_cbk,
		 (XtPointer client_data, XtIntervalId *id) );
STATIC_FUNCTION (void _kvis_waterfall_mode_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );


/*  Private data  */
static iarray array = NULL;
static Widget popupshell = NULL;
static flag do_scroll = FALSE;
static int interval_ms = SCROLL_INTERVAL;
static Widget report_lbl = NULL;


/*  Public functions follow  */

flag Xt_AppInitialise (XtAppContext ctx, Widget main_shell)
{
    Widget form, w;
    /*static char function_name[] = "Xt_AppInitialise";*/

    popupshell = XtVaCreatePopupShell ("waterfallShell",
				       topLevelShellWidgetClass, main_shell,
				       XtNtitle, "Waterfall Control Panel",
				       NULL);
    form = XtVaCreateManagedWidget ("form", formWidgetClass, popupshell,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, form,
				 XtNlabel, "Continuous Scroll",
				 XkwNcrosses, False,
				 NULL);
    XtAddCallback (w, XtNcallback, _kvis_waterfall_toggle_cbk, NULL);
    w = XtVaCreateManagedWidget ("command", commandWidgetClass, form,
				 XtNlabel, "Scroll Once",
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, _kvis_waterfall_scroll_cbk, NULL);
    w = XtVaCreateManagedWidget ("modeMenu", exclusiveMenuWidgetClass, form,
				 XtNmenuName, "modeMenu",
				 XkwNchoiceName, "Mode",
				 XtNfromHoriz, w,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XkwNnumItems, NUM_MODE_CHOICES,
				 XkwNitemStrings, mode_choices,
				 NULL);
    XtAddCallback (w, XkwNselectCallback, _kvis_waterfall_mode_cbk, NULL);
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNhorizDistance, 0,
				 XtNlabel, "Update Interval (ms)",
				 XtNwidth, 430,
				 XtNfromVert, w,
				 XkwNminimum, 0,
				 XkwNmaximum, 400,
				 XkwNmodifier, 1,
				 XtNvalue, SCROLL_INTERVAL,
				 XkwNvaluePtr, &interval_ms,
				 NULL);
    w = XtVaCreateManagedWidget ("labelslider", labelWidgetClass, form,
				 XtNborderWidth, 0,
				 XtNwidth, 430,
				 XtNfromVert, w,
				 NULL);
    report_lbl = w;
    return (TRUE);
}   /*  End Function Xt_AppInitialise  */

flag Xt_InEventLoop (XtAppContext ctx, Widget main_shell)
{
    int mode = MODE_RANDOM;
    static char function_name[] = "Xt_InEventLoop";

    if ( ( array = iarray_create_2D (IMAGE_SIZE, IMAGE_SIZE, K_UBYTE) )
	 == NULL ) m_abort (function_name, "array");
    _kvis_waterfall_mode_cbk (NULL, NULL, &mode);
    ds_event_dispatch (iarray_multi_array (array), "WATERFALL", "waterfall");
    XtPopup (popupshell, XtGrabNone);
    return (TRUE);
}   /*  End Function Xt_InEventLoop  */


/*  Private functions follow */

static void _kvis_waterfall_toggle_cbk (Widget w, XtPointer client_data,
					XtPointer call_data)
{
    if (call_data && !do_scroll)
    {
	XtAppContext app_context = XtWidgetToApplicationContext (w);

	XtAppAddTimeOut (app_context, interval_ms,
			 _kvis_waterfall_timer_cbk, app_context);
    }
    do_scroll = (uaddr) call_data;
}   /*  End Function _kvis_waterfall_toggle_cbk  */

static void _kvis_waterfall_scroll_cbk (Widget w, XtPointer client_data,
					XtPointer call_data)
{
    struct timeval start_time, scroll_time, stop_time;
    static unsigned int updates_since_last_report = 0;
    static long scroll_accumulated_us = 0;
    static long update_accumulated_us = 0;
    static long total_accumulated_us = 0;

    gettimeofday (&start_time, NULL);
    iarray_scroll (array, 1, SCROLL_CIRCULAR);
    gettimeofday (&scroll_time, NULL);
    iarray_register_data_change (array, NULL);
    gettimeofday (&stop_time, NULL);
    ++updates_since_last_report;
    scroll_accumulated_us += (scroll_time.tv_sec -start_time.tv_sec) * 1000000;
    scroll_accumulated_us += scroll_time.tv_usec - start_time.tv_usec;
    update_accumulated_us += (stop_time.tv_sec - scroll_time.tv_sec) * 1000000;
    update_accumulated_us += stop_time.tv_usec - scroll_time.tv_usec;
    total_accumulated_us += (stop_time.tv_sec - start_time.tv_sec) * 1000000;
    total_accumulated_us += stop_time.tv_usec - start_time.tv_usec;
    if (total_accumulated_us > 100000)
    {
	char txt[64];

	sprintf (txt, "Times (ms): scroll: %-4.3g update: %ld",
		 (float) scroll_accumulated_us / updates_since_last_report
		 * 1e-3,
		 update_accumulated_us / updates_since_last_report / 1000);
	XtVaSetValues (report_lbl, XtNlabel, txt, NULL);
	scroll_accumulated_us = 0;
	update_accumulated_us = 0;
	total_accumulated_us = 0;
	updates_since_last_report = 0;
    }
}   /*  End Function _kvis_waterfall_scroll_cbk  */

static void _kvis_waterfall_timer_cbk (XtPointer client_data, XtIntervalId *id)
/*  This is the interval timer callback.
*/
{
    if (!do_scroll) return;
    _kvis_waterfall_scroll_cbk (NULL, NULL, NULL);
    XtAppAddTimeOut (client_data, interval_ms, _kvis_waterfall_timer_cbk,
		     client_data);
}   /*  End Function _kvis_waterfall_timer_cbk  */

static void _kvis_waterfall_mode_cbk (Widget w, XtPointer client_data,
				      XtPointer call_data)
{
    int new_mode = *(int *) call_data;
    unsigned int x, y;

    /*  Fill the array  */
    switch (new_mode)
    {
      case MODE_RANDOM:
	for (y = 0; y < IMAGE_SIZE; ++y) for (x = 0; x < IMAGE_SIZE; ++x)
	    UB2 (array, y, x) = n_uniform () * MAX_VALUE + 0.5;
	break;
      case MODE_V:
	for (y = 0; y < IMAGE_SIZE; ++y)
	{
	    signed int peak_x = y >> 1;  /*  Distance from centre  */

	    for (x = 0; x < IMAGE_SIZE; ++x)
	    {
		unsigned char val;
		signed int distance;

		distance = abs ( (IMAGE_SIZE >> 1) - (int) x );
		distance = 5 - abs (peak_x - distance);
		val = (distance > 0) ? (MAX_VALUE * distance / 5) :
		    MAX_VALUE / 2;
		UB2 (array, y, x) = val;
	    }
	}
	x = y = IMAGE_SIZE >> 1;
	UB2 (array, y - 1, x - 1) = 0;
	UB2 (array, y - 1, x) = 0;
	UB2 (array, y - 1, x + 1) = 0;
	UB2 (array, y, x - 1) = 0;
	UB2 (array, y, x) = 0;
	UB2 (array, y, x + 1) = 0;
	UB2 (array, y + 1, x - 1) = 0;
	UB2 (array, y + 1, x) = 0;
	UB2 (array, y + 1, x + 1) = 0;
	break;
    }
    iarray_register_data_change (array, NULL);
}   /*  End Function _kvis_waterfall_mode_cbk  */
