/*  BACKUP  Backup files.
    This program will perform backup operations on user files.
    The backup medium is given by a user specified pathname. This pathname
    may point to a directory or a logical link (such as through NFS) to a
    directory. It must not point to a regular file, symbolic link, etc.


    Written by      Richard Gooch   9-MAY-1989

    Updated by      Richard Gooch   5-MAR-1990

    Updated by      Richard Gooch   16-APR-1992: Added logging and owner
  options.

    Updated by      Richard Gooch   23-APR-1992: Added  symlink always  option.

    Updated by      Richard Gooch   13-MAY-1992: Do not remove and recreate
  symbolic links if not needed. Also log symbolic links which are backed up.

    Updated by     Richard Gooch   23-MAY-1992: Fixed bug which caused failure
  when backing up symbolic links which did not already exist.

    Updated by      Richard Gooch   24-MAY-1992: Added pruning facility.

    Updated by      Richard Gooch   15-DEC-1992: Added mode compare.

    Updated by      Richard Gooch   28-MAR-1993: Added SIGINT handler.

    Updated by      Richard Gooch   7-APR-1993: Added option to disable copying
  access modes and option to warn if destination file is newer.

    Updated by      Richard Gooch   12-APR-1993: Modified to report whether
  file is new.

    Updated by      Richard Gooch   14-APR-1993: Modified to print warnings if
  source and destination files have same age but different sizes.

    Updated by      Richard Gooch   19-JUN-1993: Added  -n  (no change) option.

    Updated by      Richard Gooch   4-APR-1995: Added  -d  (copy devices) and
  -S  (sparse) options.

    Updated by      Richard Gooch   9-JAN-1996: Changed to  getcwd  routine in
  order to avoid having to link with buggy UCB compatibility library in
  Slowaris 2.3

    Updated by      Richard Gooch   28-OCT-1997: Fixed bug in <back_entry>
  where temporary pathnames were declared too short.

    Updated by      Richard Gooch   18-JAN-2000: Internal restructure and
  cleanup.

    Updated by      Richard Gooch   19-JAN-2000: Removed -d (copy devices)
  option: it was being done automatically anyway. Implemented -S (sparse)
  option by simply calling <ch_open_file> with "s" mode type. Added -x (do not
  cross filesystem boundaries) option.

    Updated by      Richard Gooch   20-JAN-2000: Removed -u (backup unmodified)
  option: probably not useful. Removed -s (always backup symlinks) option: now
  always back up symlinks which point nowhere, which is probably the only
  useful behaviour. Added -c (compare checksums) option, for when modification
  times are useless (i.e. when users touch files).

    Updated by      Richard Gooch   24-JAN-2000: Changed -e (empty dest. dir)
  option to empty excess *after* backing up.

    Updated by      Richard Gooch   11-FEB-2000: Switched to ERRSTRING.

    Last updated by Richard Gooch   5-OCT-2001: Added #include <fcntl.h>.


    Usage:  backup [-raecflopmwSMxnvh] [sourcepath] destpath
*/
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/file.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/param.h>
#include <sys/time.h>
#include <signal.h>
#include <sys/resource.h>
#include <errno.h>
#include <karma.h>
#include <karma_ch.h>
#include <karma_md.h>
#include <karma_s.h>


#define OPT_RECURSIVE       0x00001
#define OPT_ALL             0x00002
#define OPT_EMPTYDIR        0x00004
#define OPT_USE_CHECKSUM    0x00008
#define OPT_FORCE           0x00010
#define OPT_LOG             0x00020
#define OPT_OWNER           0x00040
#define OPT_NOPRUNE         0x00080
#define OPT_NOCHMOD         0x00100
#define OPT_WARN            0x00200
#define OPT_SPARSE          0x00400
#define OPT_MD5             0x00800
#define OPT_NO_FS_CROSS     0x01000
#define OPT_NOOP            0x02000
#define OPT_VERBOSE         0x04000
#define OPT_PRINTHELP       0x08000

#define BUF_SIZE 65536

/*  External functions  */
#if defined(OS_MSDOS) || defined(OS_HPUX)
extern char *getcwd ();
#endif

/*  Private functions  */
static void printhelp (char *usage);
static flag back_dir (char *newdir, char *dirname, struct stat source_stat,
		      dev_t source_fs, unsigned int optmask);
static void empty_dir (char *dirname);
static void back_entry (char *newdir, char *dirname, char *entryname,
			dev_t source_fs, unsigned int optmask,
			flag backup_file);
static flag back_file (char *destfile,
		       char *sourcefile, struct stat source_stat,
		       unsigned int optmask);
static flag compare_files (char *destfile, struct stat dest_stat,
			   char *sourcefile, struct stat source_stat,
			   unsigned int optmask);
static flag compare_file_checksums (char *destfile, char *sourcefile,
				    unsigned int optmask);
static flag compute_md5_hash (char *file, unsigned char digest[16],
			      unsigned int optmask);
static flag file_copy (char *destfile, struct stat dest_stat,
		       char *sourcefile, unsigned int optmask);
static void back_symlink (char *destpath, char *sourcepath,
			  unsigned int optmask);


/*  Private data  */


/*  Public functions follow  */

int main (int argc, char **argv)
{
    unsigned int optmask = 0;
    unsigned int pos = 0;  /*  Initialised to keep gcc happy  */
    int argcount = 1;
    char opt_char;
    char usage_string[STRING_LENGTH];
    char sourcepath[MAXPATHLEN + 1];
    char destpath[MAXNAMLEN];
    struct stat statbuf;

    strcpy (usage_string,
	    "Usage:\tbackup [-raecflopmwSMxnvh] [sourcepath] destpath\n");
    /*  Set up control_c handler  */
    if ( (int) signal (SIGINT, s_int_handler) == -1 )
    {
	fprintf (stderr, "Error setting control_c handler\t%s%c\n",
		 ERRSTRING, BEL);
	exit (RV_SYS_ERROR);
    }
    if (argc < 2)
    {
	fprintf (stderr, usage_string);
        exit (RV_MISSING_PARAM);
    }
    /*  Find all options  */
    while ( (argv[argcount][0] == '-') && !(optmask & OPT_PRINTHELP) )
    {
	/*  User has supplied an option */
        pos = 1;
        while ( (opt_char = argv[argcount][pos]) != '\0' )
        {
	    switch (opt_char)
	    {
	      case 'r':
		optmask = optmask | OPT_RECURSIVE;
                break;
              case 'a':
                optmask = optmask | OPT_ALL;
                break;
              case 'e':
                optmask = optmask | OPT_EMPTYDIR;
                break;
              case 'c':
                optmask = optmask | OPT_USE_CHECKSUM;
                break;
	      case 'f':
		optmask = optmask | OPT_FORCE;
		break;
              case 'l':
                optmask = optmask | OPT_LOG;
                break;
              case 'o':
                optmask = optmask | OPT_OWNER;
                break;
	      case 'p':
		optmask = optmask | OPT_NOPRUNE;
		break;
	      case 'm':
		optmask = optmask | OPT_NOCHMOD;
		break;
              case 'w':
		optmask = optmask | OPT_WARN;
		break;
              case 'S':
		optmask = optmask | OPT_SPARSE;
		break;
              case 'M':
		optmask = optmask | OPT_MD5;
		break;
              case 'x':
		optmask = optmask | OPT_NO_FS_CROSS;
		break;
              case 'n':
		optmask = optmask | OPT_NOOP;
		break;
              case 'v':
                optmask = optmask | OPT_VERBOSE;
                break;
              case 'h':
                optmask = optmask | OPT_PRINTHELP;
                break;
              default:
                fprintf (stderr, usage_string);
                exit (RV_BAD_PARAM);
                break;
            }
            ++pos;
        }
        /*  This test is required else the test in the while loop
	    will look at unallocated memory.  */
        if ( !(optmask & OPT_PRINTHELP) )
        {
	    /*  User does not want to print help  */
            ++argcount;
        }
        if ( (argc == argcount) && !(optmask & OPT_PRINTHELP) )
        {
	    /*  User still must supply at least one filename  */
            fprintf (stderr, usage_string);
            exit (RV_MISSING_PARAM);
        }
    }
    if (optmask & OPT_PRINTHELP)
    {
	/*  User wants to print help  */
        puts ("BACKUP     Version 3.1    (C) 1989-2000     Richard Gooch.\n");
        puts ("This programme will backup files. Released under the GPL.\n");
        puts ("backup -h    for help information.\n");
        printhelp (usage_string);
        exit (RV_OK);
    }
    if (argc > argcount + 2)
    {
	/*  Too many arguments  */
        fprintf (stderr, usage_string);
        exit (RV_TOO_MANY_PARAM);
    }
    if (argcount + 2 == argc)
    {
	/*  Source and destination paths supplied by user  */
        strcpy (sourcepath, argv[argcount++]);
        strcpy (destpath, argv[argcount]);
    }
    else
    {
	/*  Only destination path supplied by user  */
        /*  Sourcepath is current working directory */
        if (getcwd (sourcepath, MAXPATHLEN) != sourcepath)
        {
	    /*  Error occured  */
            fprintf (stderr, "Error finding current working directory\n");
            fprintf (stderr, "Message: %s%c\n", sourcepath, BEL);
            exit (RV_SYS_ERROR);
        }
        strcpy (destpath, argv[argcount]);
    }
    /*  Make sure source path is a directory  */
    if (lstat (sourcepath, &statbuf) != 0)
    {
	fprintf (stderr, "Error getting stats on file: %s   %s%c\n",
		 sourcepath, ERRSTRING, BEL);
        exit (RV_SYS_ERROR);
    }
    switch (statbuf.st_mode & S_IFMT)
    {
      case S_IFDIR:
	break;
      case S_IFLNK:
	fprintf (stderr, "Path: %s must not be a symbolic link%c\n",
		 sourcepath, BEL);
	exit (RV_BAD_PARAM);
	/*break;*/
      default:
	fprintf (stderr,"Path: %s must be a directory%c\n", sourcepath, BEL);
	exit (RV_BAD_PARAM);
	/*break;*/
    }
    /*  Do the real work  */
    back_dir (destpath, sourcepath, statbuf, statbuf.st_dev, optmask);
    if (optmask & OPT_VERBOSE)
        fprintf (stderr, "Finished backup from  %s  to  %s\n",
		 sourcepath, destpath);
    exit (RV_OK);
}   /*  End Function main  */


/*  Private functions follow  */

static void printhelp (char *usage)
/*  [SUMMARY] Print help information.
    <usage> The usage string.
    [RETURNS] Nothing.
*/
{
    puts(usage);
    puts("Options are:");
    fprintf (stderr, "\n");
    puts("  r = recursively traverse subdirectories");
    puts("  a = all files (regardless of presence of .BACKUP_ENABLE file)");
    puts("  e = empty destination directories of entries not found in source");
    puts("  c = compare MD5 checksums instead of file modification times");
    puts("  f = forced continuation, even if files could not be accessed");
    puts("  l = log files which have changed");
    puts("  o = copy over owner and group");
    puts("  p = ignore presence of .BACKUP_PRUNE file");
    puts("  m = do not copy access modes");
    puts("  w = print warning messages");
    puts("  S = maintain holes in sparse files");
    puts("  M = perform MD5 checksum of files with same age and size");
    puts("  x = do not cross filesystem boundaries in source tree");
    puts("  n = no change (dry run)");
    puts("  v = verbose output");
    puts("  h = print help information");
    puts("\nThe current working directory is the default  sourcepath");
}   /*  End Function printhelp  */

static flag back_dir (char *newdir, char *dirname, struct stat source_stat,
		      dev_t source_fs, unsigned int optmask)
/*  [SUMMARY] Backup a directory, with possible recursion.
    <newdir> The destination directory name.
    <dirname> The directory to back up.
    <source_stat> The status information for the source directory.
    <source_fs> The device on which the source directory is mounted.
    <optmask> The option mask.
    [RETURNS] TRUE if any component of the directory tree was backed up, else
    FALSE.
*/
{
    char tmp_name[MAXPATHLEN];
    flag backup_files = FALSE;
    DIR *dirp;
    struct dirent *dp;
    struct stat dest_stat;

    if ( (optmask & OPT_NO_FS_CROSS) && (source_stat.st_dev != source_fs) )
    {
	if (optmask & OPT_VERBOSE)
	    fprintf (stderr,"Not descending: \"%s\" on different filesystem\n",
		     dirname);
	return (FALSE);
    }
    if ( !(optmask & OPT_NOPRUNE) )
    {
	/*  User did not supply  -p  (no prune) option  */
	strcpy (tmp_name, dirname);
	strcat (tmp_name, "/.BACKUP_PRUNE");
	if (access (tmp_name, F_OK) == 0)
	{
	    /*  .BACKUP_PRUNE  file is present  */
	    /*  Skip entire directory tree  */
	    if (optmask & OPT_VERBOSE)
		fprintf (stderr, "Pruning at: %s\n", dirname);
	    return (FALSE);
	}
    }
    if (optmask & OPT_VERBOSE)
        fprintf (stderr, "backup:    sourcedir: %s       dest_dir: %s\n",
		 dirname, newdir);
    if (optmask & OPT_ALL) backup_files = TRUE;
    if (!backup_files)
    {
	strcpy (tmp_name, dirname);
        strcat (tmp_name, "/.BACKUP_ENABLE");
        if (access (tmp_name, F_OK) == 0) backup_files = TRUE;
    }
    /*  Test for existence of backup directory.
        If it does not exist, make backup directory. Do it irrespective
        of whether there is anything to back up: too complicated to do
        tests on whether to backup files and any subdirectories. Besides,
	it's neater to preserve the directory structure.  */
    /*  Get stats on  newdir  */
    if (lstat (newdir, &dest_stat) == 0)
    {
	/*  Something called  newdir  exists  */
        if ( !S_ISDIR (dest_stat.st_mode) )
        {
	    /*  It is NOT a directory!  */
            fprintf (stderr, "File: %s is not a directory.\n", newdir);
            fprintf (stderr, "Directories can only overwrite directories.%c\n",
		     BEL);
	    if (optmask & OPT_FORCE) return (FALSE);
	    exit (RV_BAD_FILE_TYPE);
        }
    }
    else
    {
	dest_stat.st_mode = 0;
	if (errno != ENOENT)
	{
	    fprintf (stderr, "Error getting stats on file: %s   %s%c\n",
		     newdir, ERRSTRING, BEL);
	    if (optmask & OPT_FORCE) return (FALSE);
	    exit (RV_SYS_ERROR);
	}
	/*  Directory does not exist  */
	fprintf (stderr, "%s\tNEW DIRECTORY\n", dirname);
	if (optmask & OPT_NOOP) return (TRUE);
	if (mkdir (newdir, source_stat.st_mode & ~S_IFMT) != 0)
        {
	    /*  Directory could not be created  */
            fprintf (stderr, "Error creating directory: %s  %s%c\n",
		     newdir, ERRSTRING, BEL);
	    if (optmask & OPT_FORCE) return (FALSE);
	    exit (RV_SYS_ERROR);
        }
    }
    /*  Start to go through source directory  */
    if ( ( dirp = opendir (dirname) ) == NULL )
    {
	fprintf (stderr, "Error reading directory: \"%s\"\t%s\n",
		 dirname, ERRSTRING);
	/*  Return OK, else  newdir  will have a newer date  */
	if (optmask & OPT_FORCE) return (TRUE);
	exit (RV_SYS_ERROR);
    }
    while ( ( dp = readdir (dirp) ) != NULL )
    {
	if ( s_check_for_int () )
	{
	    fprintf (stderr,"User interrupt: exiting without partial files\n");
	    exit (RV_UNDEF_ERROR);
	}
        if (strcmp (".", dp->d_name) == 0) continue;
	if (strcmp ("..", dp->d_name) == 0) continue;
	/*  File is not a special directory */
	back_entry (newdir, dirname, dp->d_name, source_fs, optmask,
		    backup_files);
    }
    /*  Close source directory (and deallocate memory used)  */
    closedir (dirp);
    if ( !(optmask & OPT_EMPTYDIR) ) return (TRUE);
    /*  Go through destination directory, removing entries not in source  */
    if ( ( dirp = opendir (newdir) ) == NULL )
    {
	fprintf (stderr, "Error reading directory: \"%s\"\t%s\n",
		 newdir, ERRSTRING);
	/*  Return OK, else  newdir  will have a newer date  */
	if (optmask & OPT_FORCE) return (TRUE);
	exit (RV_SYS_ERROR);
    }
    while ( ( dp = readdir (dirp) ) != NULL )
    {
	if ( s_check_for_int () )
	{
	    fprintf (stderr,"User interrupt: exiting without partial files\n");
	    exit (RV_UNDEF_ERROR);
	}
        if (strcmp (".", dp->d_name) == 0) continue;
	if (strcmp ("..", dp->d_name) == 0) continue;
	/*  File is not a special directory */
	strcpy (tmp_name, dirname);
	strcat (tmp_name, "/");
	strcat (tmp_name, dp->d_name);
	if (access (tmp_name, F_OK) == 0) continue;
	/*  Nothing matching in the source directory: kill it  */
	strcpy (tmp_name, newdir);
	strcat (tmp_name, "/");
	strcat (tmp_name, dp->d_name);
	fprintf (stderr, "%s\tDELETING: not in source\n", tmp_name);
	if (optmask & OPT_NOOP) continue;
	if (lstat (tmp_name, &dest_stat) != 0)
	{
	    fprintf (stderr, "Error getting stats on file: %s   %s%c\n",
		     newdir, ERRSTRING, BEL);
	    closedir (dirp);
	    /*  Return OK, else  newdir  will have a newer date  */
	    if (optmask & OPT_FORCE) return (TRUE);
	    exit (RV_SYS_ERROR);
	}
	if ( S_ISDIR (dest_stat.st_mode) ) empty_dir (tmp_name);
	else unlink (tmp_name);
    }
    closedir (dirp);
    return (TRUE);
}   /*  End Function back_dir  */

static void empty_dir (char *dirname)
/*  This routine will remove the contents of the directory pointed to by
    dirname  .Any subdirectories are unaffected.
*/
{
    char filename[MAXPATHLEN+1];
    DIR *dirp;
    struct dirent *dp;
    struct stat statbuf;

    /*  Begin reading directory */
    errno = 0;
    if ( ( dirp = opendir (dirname) ) == NULL )
    {
	fprintf (stderr, "Error opening directory: %s   %s%c\n",
		 dirname, ERRSTRING, BEL);
        exit (RV_SYS_ERROR);
    }
    while ( ( dp = readdir (dirp) ) != NULL )
    {
	/*  Still another entry */
        strcpy (filename, dirname);
        strcat (filename, "/");
        strcat (filename, dp->d_name);
        /*  Get stats on file  */
        if (lstat (filename, &statbuf) != 0)
        {
	    fprintf (stderr, "Error getting stats on file: %s   %s%c\n",
		     filename, ERRSTRING, BEL);
            exit (RV_SYS_ERROR);
        }
        if ( (statbuf.st_mode & S_IFMT) != S_IFDIR )
        {
	    /*  Not a subdirectory  */
            if (unlink (filename) != 0)
            {
		fprintf (stderr, "Error unlinking file: %s  %s%c\n",
			 filename, ERRSTRING, BEL);
                exit (RV_SYS_ERROR);
            }
        }
    }
    /*  Close directory (and deallocate memory used)  */
    closedir (dirp);
}   /*  End Function empty_dir  */

static void back_entry (char *newdir, char *dirname, char *entryname,
			dev_t source_fs, unsigned int optmask,
			flag backup_file)
/*  [SUMMARY] Backup a directory entry.
    <newdir> The name of the destination directory.
    <dirname> The name of the source directory.
    <entryname> The name of the directory entry to be backed up.
    <source_fs> The device on which the source directory is mounted.
    <optmask> The option mask.
    <backup_file> If this is TRUE, then a regular file or a symbolic link to a
    regular file will be backed up, otherwise the entry won't be backed up.
    Other entry types are always backed up.
    [RETURNS] Nothing.
*/
{
    char sourcepath[MAXPATHLEN];
    char destpath[MAXPATHLEN];
    struct stat source_stat;
    struct timeval times[2];

    strcpy (sourcepath, dirname);
    strcat (sourcepath, "/");
    strcat (sourcepath, entryname);
    if (lstat (sourcepath, &source_stat) != 0)
    {
	fprintf (stderr, "Error getting stats on file: %s   %s%c\n",
		 sourcepath, ERRSTRING, BEL);
	if (optmask & OPT_FORCE) return;
	exit (RV_SYS_ERROR);
    }
    strcpy (destpath, newdir);
    strcat (destpath, "/");
    strcat (destpath, entryname);
    switch (source_stat.st_mode & S_IFMT)
    {
      case S_IFDIR:
	/*  File is a regular directory */
	/*  Directories are 'copied' if user supplied
	    -r  (recursive) option  */
	if (optmask & OPT_RECURSIVE)
	{
	    /*  Recursively search new directory  */
	    if ( !back_dir (destpath, sourcepath, source_stat, source_fs,
			    optmask) ) return;
	}
	break;
      case S_IFREG:
	/*  File is a regular file  */
	/*  backup_file flag must be TRUE to backup file  */
	if (!backup_file) return;
	if ( !back_file (destpath, sourcepath, source_stat, optmask) ) return;
	break;
      case S_IFLNK:
	/*  File is a symbolic link */
	back_symlink (destpath, sourcepath, optmask);
	return;
	/*break;*/
      case S_IFCHR:
      case S_IFBLK:
      case S_IFIFO:
	/*  File is a character special/block device file or pipe  */
	if (optmask & OPT_NOOP) return;
	if (mknod (destpath, source_stat.st_mode, source_stat.st_dev) != 0)
	{
	    fprintf (stderr, "Cannot create device file/pipe: \"%s\"\t%s\n",
		     destpath, ERRSTRING);
	    if (optmask & OPT_FORCE) return;
	    exit (RV_SYS_ERROR);
	}
	break;
      default:
	/*  Entry is not a regular file, directory or
	    symbolic link  */
	if (optmask & OPT_VERBOSE)
	    fprintf (stderr, "File: %s not backed up.\n", sourcepath);
	return;
	/*break;*/
    }
    /*  The system calls <chmod> and <utimes> affect the end of a symbolic
	link, not the start of the link, thus a symbolic link backup does not
        reach here  */
    if (optmask & OPT_NOOP) return;
    if (optmask & OPT_OWNER)
    {
	/*  Copy over owner and group  */
	if (chown (destpath, source_stat.st_uid, source_stat.st_gid) != 0)
	{
	    fprintf (stderr, "Error changing owner for file: \"%s\"  %s%c\n",
		     destpath, ERRSTRING, BEL);
	    if (optmask & OPT_FORCE) return;
	    exit (RV_SYS_ERROR);
	}
    }
    if ( !(optmask & OPT_NOCHMOD) )
    {
	/*  Copy over access modes to destination entry  */
	if (chmod (destpath, source_stat.st_mode) != 0)
	{
	    fprintf (stderr, "Error changing mode for file: %s  %s%c\n",
		     destpath, ERRSTRING, BEL);
	    if (optmask & OPT_FORCE) return;
	    exit (RV_SYS_ERROR);
	}
    }
    /*  Copy over file times. This must be done last (i.e. after file
	closure and access mode change), else the time of file closure
	or access mode change will be recorded  */
    times[0].tv_sec = source_stat.st_atime;
    times[0].tv_usec = 0L;
    times[1].tv_sec = source_stat.st_mtime;
    times[1].tv_usec = 0L;
    if (utimes (destpath, times) != 0)
    {
	fprintf (stderr, "Error changing times for file: %s %s%c\n",
		 destpath, ERRSTRING, BEL);
	if (optmask & OPT_FORCE) return;
	exit (RV_SYS_ERROR);
    }
}   /*  End Function back_entry  */

static flag back_file (char *destfile,
		       char *sourcefile, struct stat source_stat,
		       unsigned int optmask)
/*  This routine will backup the source file of name  sourcefile  to the
    destination file of name  destfile  .
    If the OPT_LOG bit of the user options  optmask  is set and the OPT_VERBOSE
    bit is clear, then the name of the source file is printed if the
    destination file is updated.
    If this bit is not set, then the file is only backed up if the current
    file named  destfile  does not exist or is older then the file
    sourcefile  .In other words, if  sourcefile  has been modified since it
    was last backed up to  destfile  .
    The routine returns the value TRUE if the file was indeed backed up, else
    it returns FALSE.
*/
{
    flag changed = FALSE;
    struct stat dest_stat;

    if (lstat (destfile, &dest_stat) != 0)
    {
	dest_stat.st_mode = 0;
	if (errno != ENOENT)
	{
	    fprintf (stderr, "Error getting stats on file: %s   %s%c\n",
		     destfile, ERRSTRING, BEL);
	    if (optmask & OPT_FORCE) return (FALSE);
	    exit (RV_SYS_ERROR);
	}
    }
    if ( compare_files (destfile, dest_stat, sourcefile, source_stat,optmask) )
    {
	if (optmask & OPT_NOOP) changed = TRUE;
	else changed = file_copy (destfile, dest_stat, sourcefile, optmask);
    }
    if (changed)
    {
	if ( optmask & (OPT_VERBOSE | OPT_LOG) )
	{
	    if (dest_stat.st_mode == 0)
		fprintf (stderr, "%s\t\tNEW FILE\n", sourcefile);
	    else fprintf (stderr, "%s\n", sourcefile);
	}
	return (TRUE);
    }
    /*  No update to file contents: check for required attribute changes  */
    if (dest_stat.st_mode == 0) return (FALSE);
    if ( (dest_stat.st_mode != source_stat.st_mode) &&
	 !(optmask & OPT_NOCHMOD) )
    {
	/*  Change mode  */
	if ( optmask & (OPT_VERBOSE | OPT_LOG) )
	    fprintf (stderr, "%s\t\tNEW MODE\n", sourcefile);
	return (TRUE);
    }
    if ( ( (dest_stat.st_uid != source_stat.st_uid) ||
	   (dest_stat.st_gid != source_stat.st_gid) ) &&
	 (optmask & OPT_OWNER) )
    {
	/*  Change owner/ group  */
	if ( optmask & (OPT_VERBOSE | OPT_LOG) )
	    fprintf (stderr, "%s\t\tNEW OWNER\n", sourcefile);
	return (TRUE);
    }
    return (FALSE);
}   /*  End Function back_file  */

static flag compare_files (char *destfile, struct stat dest_stat,
			   char *sourcefile, struct stat source_stat,
			   unsigned int optmask)
/*  [SUMMARY] Compare files to see if a backup is required.
    <destfile> The destination file.
    <sourcefile> The source file to be backed up.
    <optmask> The option mask.
    [RETURNS] TRUE if the source file should be backed up to the destination,
    else FALSE.
*/
{
    if ( !S_ISREG (dest_stat.st_mode) ) return (TRUE);
    if (optmask & OPT_USE_CHECKSUM)
    {
	if (dest_stat.st_size != source_stat.st_size) return (TRUE);
	return compare_file_checksums (destfile, sourcefile,
				       optmask) ? FALSE : TRUE;
    }
    /*  destfile  does exist: check dates  */
    if (dest_stat.st_mtime > source_stat.st_mtime)
    {
	if ( optmask & (OPT_WARN | OPT_VERBOSE) )
	{
	    /*  destfile  is newer than  sourcefile: print warning  */
	    fprintf (stderr,
		     "WARNING: destination: %s  is newer than source: %s\n",
		     destfile, sourcefile);
	}
	return (FALSE);
    }
    if (dest_stat.st_mtime < source_stat.st_mtime) return (TRUE);
    /*  Times are the same  */
    if (dest_stat.st_size != source_stat.st_size)
    {
	if ( optmask & (OPT_WARN | OPT_VERBOSE) )
	{
	    fprintf (stderr,
		     "WARNING: destination: %s  and source: %s  have different sizes\n",
		     destfile, sourcefile);
	}
	return (FALSE);
    }
    /*  Sizes are the same  */
    if (optmask & OPT_MD5)
    {
	if ( !compare_file_checksums (destfile, sourcefile, optmask) )
	{
	    fprintf (stderr,
		     "WARNING: destination: %s  and source: %s  have different contents\n",
		     destfile, sourcefile);
	    return (FALSE);
	}
    }
    if (optmask & OPT_VERBOSE)
	fprintf (stderr, "%s\t\tsame age%s\n",
		 destfile,
		 (optmask & OPT_MD5) ? ", size and contents" : " and size");
    return (FALSE);
}   /*  End Function compare_files  */

static flag compare_file_checksums (char *file1, char *file2,
				    unsigned int optmask)
/*  [SUMMARY] Compare the MD5 checksums of two files.
    <file1> The first file.
    <file2> The second file.
    <optmask> The option mask.
    [RETURNS] TRUE if the files are the same, else FALSE.
*/
{
    unsigned char digest1[16], digest2[16];

    if ( !compute_md5_hash (file1, digest1, optmask) ) return (FALSE);
    if ( !compute_md5_hash (file2, digest2, optmask) ) return (FALSE);
    return (memcmp (digest1, digest2, 16) == 0) ? TRUE : FALSE;
}   /*  End Function compare_file_checksums  */

static flag compute_md5_hash (char *file, unsigned char digest[16],
			      unsigned int optmask)
/*  [SUMMARY] Compute the MD5 hash of a file.
    <file> The file to compute the hash for.
    <digest> The message digest is written here.
    <optmask> The option mask.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    MD5Context ctx;
    Channel ch;
    uaddr bytes_read;
    CONST char *realbuf;
    char buffer[BUF_SIZE];

    if ( ( ch = ch_map_disc (file, K_CH_MAP_IF_AVAILABLE, FALSE, FALSE) )
	 == NULL )
    {
	fprintf (stderr, "Error opening file: \"%s\"\t%s\n", file, ERRSTRING);
	if (optmask & OPT_FORCE) return (FALSE);
	exit (RV_CANNOT_OPEN);
    }
    if ( ( ctx = md_md5_init () ) == NULL )
    {
	if (optmask & OPT_FORCE) return (FALSE);
	exit (RV_MEM_ERROR);
    }
    while ( ( bytes_read = ch_read2 (ch, buffer, &realbuf, 0, BUF_SIZE) ) > 0 )
    {
	md_md5_update (ctx, (CONST unsigned char *) realbuf, bytes_read);
	if (bytes_read < BUF_SIZE) break;
    }
    ch_close (ch);
    md_md5_final (ctx, digest);
    return (TRUE);
}   /*  End Function compute_md5_hash  */

static flag file_copy (char *destfile, struct stat dest_stat,
		       char *sourcefile, unsigned int optmask)
/*  This routine will copy the contents of the file pointed to by
    sourcefile  to the file pointed to by  destfile  .The file  destfile
    is first deleted if it exists already and then the destination file is
    created and the contents are copied.
    The routine will abort if an error occurs and the OPT_FORCE bit is not set
    in  optmask  .
    The routine returns TRUE if the file was sucessfully copied, else it
    returns FALSE.
*/
{
    Channel source_ch, dest_ch;

    /*  Try to open source file  */
    if ( ( source_ch = ch_map_disc (sourcefile, K_CH_MAP_IF_AVAILABLE, FALSE,
				    FALSE) ) == NULL )
    {
	fprintf (stderr, "Error opening file: %s for reading\t%s%c\n",
		 sourcefile, ERRSTRING, BEL);
	if (optmask & OPT_FORCE) return (FALSE);
	exit (RV_CANNOT_OPEN);
    }
    /*  Delete destination file if it exists  */
    if (dest_stat.st_mode != 0)
    {
	if (unlink (destfile) != 0)
        {
	    fprintf (stderr, "Error unlinking file: %s\t%s%c\n",
		     destfile, ERRSTRING, BEL);
	    if (optmask & OPT_FORCE) return (FALSE);
	    exit (RV_SYS_ERROR);
        }
    }
    /*  Try to open destination file  */
    if ( ( dest_ch = ch_open_file (destfile,
				   (optmask & OPT_SPARSE) ? "s" : "w") )
	 == NULL )
    {
	fprintf (stderr, "Error opening file: %s for writing\t%s%c\n",
		 destfile, ERRSTRING, BEL);
	if (optmask & OPT_FORCE) return (FALSE);
	exit (RV_CANNOT_OPEN);
    }
    /*  Copy file contents  */
    if (ch_copy (dest_ch, source_ch, 0) == 0)
    {
	if (unlink (destfile) != 0)
	    fprintf (stderr, "Error unlinking file: %s  %s%c\n",
		     destfile, ERRSTRING, BEL);
	if (optmask & OPT_FORCE) return (FALSE);
	exit (RV_WRITE_ERROR);
    }
    ch_close (source_ch);
    ch_close (dest_ch);
    if (dest_stat.st_mode != 0)
    {
	if (chmod (destfile, dest_stat.st_mode) != 0)
	{
	    fprintf (stderr,"Error initialising access mode for file: %s %s\n",
		     destfile, ERRSTRING);
	    if (unlink (destfile) != 0)
		fprintf (stderr, "Error unlinking file: %s  %s%c\n",
			 destfile, ERRSTRING, BEL);
	    if (optmask & OPT_FORCE) return (FALSE);
	    exit (RV_UNDEF_ERROR);
	}
    }
    return (TRUE);
}   /*  End Function file_copy  */

static void back_symlink (char *destpath, char *sourcepath,
			  unsigned int optmask)
/*  This routine will backup a symbolic link pointed to by  sourcepath  to the
    name pointed to by  destpath  .The routine will backup the symbolic link
    only if it points to a regular file or directory.
    If the OPT_LOG bit of the user options  optmask  is set or the OPT_VERBOSE
    bit is set, then the name of the source file is printed if the destination
    file is updated.
    The routine will abort on errors, unless the OPT_FORCE bit in  optmask  is
    set.
    The routine returns nothing.
*/
{
    flag new_link = FALSE;
    char source_symname[MAXPATHLEN];
    char dest_symname[MAXPATHLEN];
    int retlen;
    struct stat statbuf;

    /*  Symbolic link should be backed up  */
    if ( ( retlen = readlink (sourcepath, source_symname, MAXPATHLEN - 1) )
	< 0 )
    {
	fprintf (stderr, "Error reading symbolic link: %s   %s%c\n",
		 sourcepath, ERRSTRING, BEL);
	if (optmask & OPT_FORCE) return;
	exit (RV_SYS_ERROR);
    }
    source_symname[retlen] = '\0';
    /*  Get stats on what destination entry is  */
    if (lstat (destpath, &statbuf) != 0)
    {
	/*  Error getting stats  */
	if (errno != ENOENT)
	{
	    /*  destpath  exists but cannot get stats  */
	    fprintf (stderr, "Error getting stats on: %s  %s%c\n",
		     destpath, ERRSTRING, BEL);
	    if (optmask & OPT_FORCE) return;
	    exit (RV_SYS_ERROR);
	}
	/*  No existing link  */
	new_link = TRUE;
    }
    else
    {
	/*  Get stats:  destpath  exists  */
	if ( (statbuf.st_mode & S_IFMT) == S_IFLNK )
	{
	    /*  destpath  is also a symbolic link: see what it points to  */
	    if ( ( retlen = readlink (destpath, dest_symname, MAXPATHLEN - 1) )
		 < 0 )
	    {
		fprintf (stderr, "Error reading symbolic link: %s   %s%c\n",
			 destpath, ERRSTRING, BEL);
		if (optmask & OPT_FORCE) return;
		exit (RV_SYS_ERROR);
	    }
	    dest_symname[retlen] = '\0';
	    /*  Check if source and destination symbolic links are the same  */
	    if (strcmp (dest_symname, source_symname) == 0)
	    {
		/*  Identical: return  */
		if (optmask & OPT_VERBOSE)
		    fprintf (stderr, "%s  link unchanged\n", sourcepath);
		return;
	    }
	    /*  destpath  already exists: delete it first  */
	    if ( !(optmask & OPT_NOOP) )
	    {
		if (unlink (destpath) != 0)
		{
		    fprintf (stderr, "Error deleting file: %s   %s%c\n",
			     destpath, ERRSTRING, BEL);
		    if (optmask & OPT_FORCE) return;
		    exit (RV_SYS_ERROR);
		}
	    }
	}
	else
	{
	    /*  destpath  is not a symbolic link: don't overwrite entry  */
	    fprintf (stderr,
		     "Cannot overwrite file/ directory: \"%s\" with symbolic link\t%s\n",
		     destpath, ERRSTRING);
	    if (optmask & OPT_FORCE) return;
	    exit (RV_SYS_ERROR);
	}
    }
    /*  destpath  should no longer exist  */
    if ( !(optmask & OPT_NOOP) )
    {
	if (symlink (source_symname, destpath) != 0)
	{
	    fprintf (stderr, "Error creating symbolic link: %s  %s%c\n",
		     destpath, ERRSTRING, BEL);
	    if (optmask & OPT_FORCE) return;
	    exit (RV_SYS_ERROR);
	}
    }
    if ( optmask & (OPT_VERBOSE | OPT_LOG) )
	fprintf (stderr, "%s\t\t%s\n",
		 sourcepath, new_link ? "NEW LINK" : "LINK CHANGED");
}   /*  End Function back_symlink  */
