/*  widget2doc.c

    Source file for  widget2doc  (extract documentation from C code).

    Copyright (C) 1997  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma utility programme will read a C source file and extract
    documentation.


    Written by      Richard Gooch   4-MAY-1997

    Updated by      Richard Gooch   11-MAY-1997

    Last updated by Richard Gooch   16-JUN-1997: Support more information in
  header than just summary and purpose.


*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <karma.h>
#include <karma_chs.h>
#include <karma_ch.h>


STATIC_FUNCTION (flag process_file,
		 (Channel ch_doc, Channel ch_liblines,
		  Channel ch_csource, Channel ch_header,
		  CONST char *widget_name) );
STATIC_FUNCTION (flag skip_until, (Channel channel, CONST char *string) );
STATIC_FUNCTION (flag get_resources, (Channel ch_doc, Channel ch_csource,
				      CONST char *widget_name) );
STATIC_FUNCTION (int read_until_non_white, (Channel channel) );
STATIC_FUNCTION (flag read_until,
		 (Channel channel, char *buffer, unsigned int size, char ch) );
STATIC_FUNCTION (flag copy_until_blank, (Channel out, Channel inp) );


/*static char indent4[] = "    ";*/

int main (int argc, char *argv[])
{
    Channel ch_csource, ch_header, ch_doc, ch_liblines;
    int argcount;
    char *ptr;
    char name[STRING_LENGTH], txt[STRING_LENGTH];
    static char usage_string[] = "Usage:\twidget2doc doc_dir [infile...]";

    if (argc < 3)
    {
	fprintf (stderr, "%s\n", usage_string);
	exit (RV_MISSING_PARAM);
    }
    for (argcount = 2; argcount < argc; ++argcount)
    {
	strcpy (name, argv[argcount]);
	if ( ( ptr = strrchr (name, '.') ) == NULL )
	{
	    fprintf (stderr, "Bad filename: \"%s\"\n", argv[argcount]);
	    exit (RV_BAD_PARAM);
	}
	if (strcmp (ptr, ".c") != 0)
	{
	    fprintf (stderr, "Bad filename: \"%s\"\n", argv[argcount]);
	    exit (RV_BAD_PARAM);
	}
	*ptr = '\0';
	fprintf (stderr, "%s\n", argv[argcount]);
	if ( ( ch_csource = ch_open_file (argv[argcount], "r") ) == NULL )
	{
	    fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		     argv[argcount], ERRSTRING);
	    exit (RV_CANNOT_OPEN);
	}
	sprintf (txt, "doc/%s.header", name);
	if ( ( ch_header = ch_open_file (txt, "r") ) == NULL )
	{
	    fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		     txt, ERRSTRING);
	    exit (RV_CANNOT_OPEN);
	}
	sprintf (txt, "%s/%s.doc", argv[1], name);
	if ( ( ch_doc = ch_open_file (txt, "w") ) == NULL )
	{
	    fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		     txt, ERRSTRING);
	    exit (RV_CANNOT_OPEN);
	}
	sprintf (txt, "%s/liblines/%s", argv[1], name);
	if ( ( ch_liblines = ch_open_file (txt, "w") ) == NULL )
	{
	    fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		     txt, ERRSTRING);
	    exit (RV_CANNOT_OPEN);
	}
	if ( !process_file (ch_doc, ch_liblines, ch_csource, ch_header, name) )
	{
	    fprintf (stderr, "Error processing file\n");
	    exit (RV_UNDEF_ERROR);
	}
	ch_close (ch_csource);
	ch_close (ch_header);
	ch_close (ch_doc);
	ch_close (ch_liblines);
    }
    exit (RV_OK);
}   /*  End Function main  */

static flag process_file (Channel ch_doc, Channel ch_liblines,
			  Channel ch_csource, Channel ch_header,
			  CONST char *widget_name)
/*  [SUMMARY] Process a file.
    <ch_doc> The output documentation channel.
    <ch_liblines> The output liblines channel.
    <ch_csource> The input C source channel.
    <ch_header> The input header channel.
    <widget_name> The name of the widget.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    char *ptr;
    char txt[STRING_LENGTH];

    if ( !skip_until (ch_csource, "static XtResource") ) return (FALSE);
    if ( !chs_get_line (ch_csource, txt, STRING_LENGTH) ) return (FALSE);
    if (txt[0] != '{')
    {
	fprintf (stderr,
		 "No opening brace on line following resource variable\n");
	return (FALSE);
    }
    ch_printf (ch_doc, "$WIDGET           %s\n", widget_name);
    ch_getl (ch_header, txt, STRING_LENGTH);
    ch_puts (ch_doc, txt, TRUE);
    for (ptr = txt + 9; isspace (*ptr); ++ptr);
    ch_puts (ch_liblines, ptr, TRUE);
    ch_printf (ch_doc, "$PURPOSE\n");
    copy_until_blank (ch_doc, ch_header);
    ch_printf (ch_doc, "$LIBRARY          karmawidets\n");
    ch_printf (ch_doc, "$LINKWITH         -lkarmawidgets\n");
    ch_printf (ch_doc, "$END\n\n");
    ch_copy (ch_doc, ch_header, 0);
    ch_printf (ch_doc, "\n$TABLE            %s_Resources\n", widget_name);
    ch_printf (ch_doc, "$COLUMNS          4\n");
    ch_printf (ch_doc, "$SUMMARY          List of resources for %s widget\n",
	       widget_name);
    ch_printf (ch_doc, "$TABLE_DATA\n");
    ch_printf (ch_doc, "|.Name\t|,Class\t|,Rep. Type\t|,Default\n");
    if ( !get_resources (ch_doc, ch_csource, widget_name) ) return (FALSE);
    ch_printf (ch_doc, "$END\n");
    return (TRUE);
}  /*  End Function process_file  */

static flag skip_until (Channel channel, CONST char *string)
/*  [SUMMARY] Read lines until a string is found.
    <channel> The file to read from.
    <string> The string to search for.
    [RETURNS] TRUE if the string was found, else FALSE.
*/
{
    char txt[STRING_LENGTH];

    while ( chs_get_line (channel, txt, STRING_LENGTH) )
    {
	if (strstr (txt, string) != NULL) return (TRUE);
    }
    return (FALSE);
}   /*  End Function skip_until  */

static flag get_resources (Channel ch_doc, Channel ch_csource,
			   CONST char *widget_name)
/*  [SUMMARY] Get resources.
    <ch_doc> The output channel.
    <ch_csource> The input channel.
    <widget_name> The name of the widget.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    char ch;
    char txt[STRING_LENGTH], name[STRING_LENGTH], class[STRING_LENGTH];
    char rep_type[STRING_LENGTH], def_val[STRING_LENGTH];

    while ( ( ch = read_until_non_white (ch_csource) ) != -1 )
    {
	if (ch == '}') return (TRUE);
	if (ch != '{')
	{
	    fprintf (stderr, "No opening brace on first resource line\n");
	    return (FALSE);
	}
	if ( !read_until (ch_csource, name, STRING_LENGTH, ',') )
	    return (FALSE);
	if ( !read_until (ch_csource, class, STRING_LENGTH, ',') )
	    return (FALSE);
	if ( !read_until (ch_csource, rep_type, STRING_LENGTH, ',') )
	    return (FALSE);
	if ( !read_until (ch_csource, NULL, 0, ',') ) return (FALSE);
	if ( !read_until (ch_csource, NULL, 0, ',') ) return (FALSE);
	if ( !read_until (ch_csource, NULL, 0, ',') ) return (FALSE);
	if ( ( ch = read_until_non_white (ch_csource) ) == -1 ) return (FALSE);
	if (ch == '(')
	{
	    /*  Scan past cast  */
	    if ( !read_until (ch_csource, NULL, 0, ')') ) return (FALSE);
	    if ( !read_until (ch_csource, def_val, STRING_LENGTH, '}') )
		return (FALSE);
	}
	else
	{
	    def_val[0] = ch;
	    if ( !read_until (ch_csource, def_val + 1, STRING_LENGTH - 1,
			      '}') ) return (FALSE);
	}
	if ( !read_until (ch_csource, NULL, 0, ',') ) return (FALSE);
	ch_printf (ch_doc, "|.%s\t|,%s\t|,%s\t|,%s\n",
		   name, class, rep_type, def_val);
    }
    while ( chs_get_line (ch_csource, txt, STRING_LENGTH) )
    {
	if (strcmp (txt, "};") == 0) return (TRUE);
	if (txt[0] != '{')
	{
	    fprintf (stderr, "No opening brace on first resource line\n");
	    return (FALSE);
	}
    }
    return (FALSE);
}   /*  End Function get_resources  */

static int read_until_non_white (Channel channel)
/*  [SUMMARY] Read characters until not whitespace.
    <channel> The channel to read from.
    [RETURNS] The non-whitespace character on success, else -1.
*/
{
    char character;

    while (ch_read (channel, &character, 1) == 1)
    {
	if ( isspace (character) ) continue;
	return (character);
    }
    return (-1);
}   /*  End Function read_until_non_white  */

static flag read_until (Channel channel, char *buffer, unsigned int size,
			char ch)
/*  [SUMMARY] Read until a character is found.
    <channel> The channel to read from.
    <buffer> Characters are written here. This will be NULL terminated. If this
    is NULL, characters are discarded.
    <size> The length of the buffer.
    <ch> The character to read until.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    char in_char;
    unsigned int count = 0;

    if (buffer == NULL) size = 2;
    if (size < 1) return (TRUE);
    while (count < size - 1)
    {
	if (ch_read (channel, &in_char, 1) != 1) return (FALSE);
	if ( isspace (in_char) ) continue;
	if (in_char == ch)
	{
	    if (buffer != NULL) buffer[count] = '\0';
	    return (TRUE);
	}
	if (buffer != NULL)
	{
	    buffer[count] = in_char;
	    ++count;
	}
    }
    return (FALSE);
}   /*  End Function read_until  */

static flag copy_until_blank (Channel out, Channel inp)
/*  [SUMMARY] Copy lines until a blank line is found.
    <out> The output channel.
    <inp> The input channel.
    [RETURNS] TRUE on success, else FALSE
*/
{
    char txt[STRING_LENGTH];

    while ( ch_getl (inp, txt, STRING_LENGTH) )
    {
	if (strlen (txt) == 0) return (TRUE);
	if ( !ch_puts (out, txt, TRUE) )
	{
	    fprintf (stderr, "Error writing line\t%s\n", ERRSTRING);
	    return (FALSE);
	}
    }
    return (TRUE);
}   /*  End Function copy_until_blank  */
