/*LINTLIBRARY*/
/*  atm.c
    This code implements ATM support for Linux

    Copyright (C) 1998-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/
/*

    This file contains the various utility routines for manipulating Karma
    connections.


    Written by      Richard Gooch   25-DEC-1998: Copied from Linux/atm.c

    Updated by      Jane LLoyd      18-FEB-1999: Ported to FORE XTI.

    Updated by      Richard Gooch   19-FEB-1999: Various bugfixes and cleanups.

    Updated by      Richard Gooch   20-FEB-1999: Removed debugging messages.

    Last updated by Richard Gooch   3-OCT-2001: Added declaration of
  <<t_getname>> for Solaris.


*/
#ifdef HAS_FORE_ATM

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/param.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <fore_xti/xti_user_types.h>
#include <fore_xti/xti.h>
#include <fore_xti/xti_atm.h>
#include <fore_xti/netatm/atm.h>
#include <fore_xti/ans.h>
#include <karma.h>
#include <karma_r_def.h>
#include <karma_r.h>


#define DEVICE	          "/dev/xtisvc0"
#define T_ATM_AAL5        0x1
#define ATMNSAP_ADDR_LEN  20
#define DEFAULT_SDU       65535
#define BYTES_PER_CELL    48
#define BHLI_MAGIC        "FORE_ATM"
#define LINE_RATE         353207
#define OPT_FLAGS_CPID    0x1
#define OPT_FLAGS_PMP     0x2

/*  Private functions  */
STATIC_FUNCTION (int _r_clean_up, (int fd));
STATIC_FUNCTION (int _r_open_device, (char *device));
STATIC_FUNCTION (int _r_svc_bind_listener, (char *addr_string) );
STATIC_FUNCTION (void _r_atm_karma_to_atm,
		 (struct t_atm_traffic_substruct *atm, int32_t *size,
		  u_int8_t *traffic,
		  CONST struct karma_traffic_parameters *karma) );
STATIC_FUNCTION (void _r_atm_to_karma,
		 (struct karma_traffic_parameters *karma,
		  CONST struct t_atm_traffic_substruct *atm, int32_t size,
		  u_int8_t traffic) );
STATIC_FUNCTION (int _r_get_local_address, (int fd, unsigned char addr[]));
STATIC_FUNCTION (int _r_handle_error, (int fd, char *func));
STATIC_FUNCTION (char *_r_construct_options,
		 (int fd, struct t_atm_traffic *traffic,
		  struct t_atm_bearer *bearer,int flags, long *optsize));


/*  Public functions follow  */

/*UNPUBLISHED_FUNCTION*/
int r_atm_alloc (char *addr_string, unsigned int *addr_len)
/*  [SUMMARY] Create an ATMSVC socket and bind to a random port number.
    <addr_string> The textual representation of the ATM address is written here
    <addr_len> The length of the address string. This is updated.
    [NOTE] The close-on-exec flags of the socket is set such that the socket
    will close on a call to execve(2V).
    [RETURNS] The file descriptor of the socket on success, else -1.
*/
{
    int fd;	    
    static char function_name[]= "r_atm_alloc";

    /*  Check that the length of the address passed to function is greater than
	or equal to ATM forums NSAP address of an endpoint on a private network
	*/
    if (*addr_len < ATMNSAP_ADDR_LEN)
    {
	fprintf (stderr,
		 "%s: address length: %u is too short, must be at least: %u\n",
		 function_name, *addr_len, ATMNSAP_ADDR_LEN );
        return (-1);
    }
    /*  Bind local endpoint of svc listening process to a random SAP and set
	the value of addr_string  */
    if ( ( fd = _r_svc_bind_listener (addr_string) ) < 0 ) return (-1);
    /*  Set the value of addr_len to 20 bytes  */
    *addr_len = ATMNSAP_ADDR_LEN;
    /*  Set close-on exec flag  */
    if (fcntl (fd, F_SETFD, 1) == -1)
    {
	fprintf (stderr,
		 "Error setting close-on-exec flag for ATMSVC socket\t%s\n",
		 ERRSTRING);
	if (_r_clean_up (fd) != 0)
	{
	    fprintf (stderr, "Error closing ATMSVC socket\t%s\n", ERRSTRING);
	}
	exit (RV_SYS_ERROR);
    }
    return (fd);
}   /*  End Function r_atm_alloc  */

/*UNPUBLISHED_FUNCTION*/
int r_atm_connect (CONST char *addr_string, unsigned int addr_len,
		   struct karma_qos *qos)
/*  [SUMMARY] Make a connection using ATM SVC with AAL5.
    <addr_string> The textual representation of the ATM address to connect to.
    <addr_len> The length of the address string.
    <qos> The Quality of Service parameters. This will be updated to reflect
    the parameters provided. This may be NULL.
    [NOTE] The close-on-exec flags of the socket is set such that the socket
    will close on a call to execve(2V).
    [NOTE] This function is not implemented on platforms without ATM support.
    [RETURNS] The file descriptor of the opened connection on success, else -1.
*/
{
    int flags = OPT_FLAGS_CPID;
    ATMSAPAddress addr;
    struct karma_qos default_qos;
    struct t_atm_traffic atmqos;
    struct t_atm_aal5 SDUsize;
    struct t_atm_bearer bearer;
    struct t_call call_req;
    int fd;
    char *options;
    long optlen = 0;
    static char function_name[]= "r_atm_connect";

    /*Check that the length of the address passed to function is greater than or equal to 
     *ATM forums NSAP address of an endpoint on a private network
     */

    if (addr_len < ATMNSAP_ADDR_LEN)
    {
	fprintf (stderr,
                 "%s: address length: %u is too short, must be at least: %u\n",
                 function_name, addr_len, ATMNSAP_ADDR_LEN );
        return (-1);
    }/*End of if statement*/

    memset(&addr, 0, sizeof addr);

    addr.addressType = AF_ATM;

    memcpy( addr.sap.t_atm_sap_addr.address,addr_string, ATMNSAP_ADDR_LEN);

/*Set up the quality of service structure*/
    if (qos == NULL)
    {
	memset (&default_qos, 0, sizeof default_qos);
	default_qos.tx.traffic_class = K_ATM_TRAFFIC_CLASS_UBR;
	default_qos.rx.traffic_class = K_ATM_TRAFFIC_CLASS_UBR;
	qos = &default_qos;
    }
 
    memset(&atmqos,0,sizeof atmqos);
    memset(&SDUsize,0,sizeof SDUsize);
    memset(&bearer,0,sizeof bearer);

    /*Convert the Karma QOS parameters to ones recognised by ATM xti*/
    _r_atm_karma_to_atm (&atmqos.forward, &SDUsize.forward_max_SDU_size,
			 &bearer.traffic_type, &qos->tx);
    _r_atm_karma_to_atm (&atmqos.backward, &SDUsize.backward_max_SDU_size,
			 &bearer.traffic_type, &qos->rx);
    /*Open the device*/	
    fd = _r_open_device(DEVICE);
    if (fd < 0) {
	exit(1);
    }

    /* Binding the initiator*/

    if (t_bind(fd, (struct t_bind *) NULL, (struct t_bind *) NULL) < 0) {
	t_error("t_bind");
	exit(1);
    }

    /*Constructing the connection options*/
    options = _r_construct_options(fd,&atmqos,&bearer,flags,&optlen);

    /*Connect to the address passed to the function*/
    addr.addressType = (u_int16_t)AF_ATM;
    addr.sap.t_atm_sap_addr.SVE_tag_addr = (int8_t)T_ATM_PRESENT;
    addr.sap.t_atm_sap_addr.SVE_tag_selector = (int8_t)T_ATM_PRESENT;
    addr.sap.t_atm_sap_addr.address_format = 
	(u_int8_t)T_ATM_ENDSYS_ADDR;
    addr.sap.t_atm_sap_addr.address_length = 20;
    addr.sap.t_atm_sap_layer2.SVE_tag = (int8_t)T_ATM_ABSENT;
    addr.sap.t_atm_sap_layer3.SVE_tag = (int8_t)T_ATM_ABSENT;
    addr.sap.t_atm_sap_appl.SVE_tag = (int8_t)T_ATM_PRESENT;
    addr.sap.t_atm_sap_appl.ID_type = (u_int8_t)T_ATM_USER_APP_ID;
    memcpy(addr.sap.t_atm_sap_appl.ID.user_defined_ID, BHLI_MAGIC, 8);


    memset(&call_req, 0, sizeof(call_req));
    call_req.addr.len = sizeof(addr);
    call_req.addr.buf = (char *)&addr;

    if ((options != (char *) NULL) && (optlen > 0)) {
	call_req.opt.len = optlen;
	call_req.opt.buf = options;
    }
    /*Connecting to the remote address. If fails returns -1*/
    if (t_connect(fd, &call_req, (struct t_call *) NULL) < 0) {
	_r_handle_error(fd, "t_connect");
	return -1;
    }
    free (options);
    _r_atm_to_karma (&qos->tx, &atmqos.forward, SDUsize.forward_max_SDU_size,
		     bearer.traffic_type);
    _r_atm_to_karma (&qos->rx, &atmqos.backward, SDUsize.backward_max_SDU_size,
		     bearer.traffic_type);
    return (fd);
}   /*  End of function r_atm_connect  */

/*UNPUBLISHED_FUNCTION*/
int r_atm_accept (int dock, struct karma_qos *qos)
/*  [SUMMARY] Listen and accept a connection on an ATM dock.
    <dock> The dock.
    <qos> The QoS parameters will be written here.
    [RETURNS] The file descriptor of the opened connection on success, else -1.
*/
{
    struct t_call incoming;
    ATMSAPAddress remote_addr;
    struct t_info info;
    int fd;

    memset (&incoming, 0, sizeof(incoming));
    incoming.addr.maxlen = sizeof (remote_addr);
    incoming.addr.buf = (char *) &remote_addr;
    /* According to XNS Issue 5, we can set maxlen = 0 to ignore any
     * return options.  However, most OSes only implement Issue 4
     * currently, which doesn't have this feature.  Therefore, we 
     * always allocate space for incoming options data.
     */
    if (t_getinfo (dock, &info) < 0)
    {
	t_error ("t_getinfo");
	return -1;
    }
    incoming.opt.buf = (char *) malloc (info.options);
    if (incoming.opt.buf == (char *) NULL)
    {
	perror ("malloc (options buffer)");
	return -1;
    }
    incoming.opt.maxlen = info.options;
    /*  This function sets the value of the variable incoming. This information
	is needed for the accept function call later  */
    if (t_listen (dock, &incoming) < 0)
    {
	_r_handle_error (dock, "t_listen");
	free (incoming.opt.buf);
	return -1;
    }	
    /*   The address will only be present if the initiator has supplied the
	 calling party address (T_ATM_ORIG_ADDR option)  */
#if 0
    if (incoming.addr.len > 0)
    {
        fprintf (stderr, "Incoming call\n");
    }
#endif
    if ( ( fd = _r_open_device (DEVICE) ) < 0 )
    {
	fprintf (stderr, "Error opening connection endpoint\n");
	exit (1);
    }
    if (t_accept (dock, fd, &incoming) < 0)
    {
	_r_handle_error (dock, "t_accept");
	free (incoming.opt.buf);
	return -1;
    }
    free (incoming.opt.buf);
    qos->tx.sdu_max = DEFAULT_SDU;
    qos->rx.sdu_max = DEFAULT_SDU;
    return (fd);
}   /*  End Function r_atm_accept  */

/*UNPUBLISHED_FUNCTION*/
iaddr r_atm_read (int fd, char *buf, uaddr nbytes)
/*  [SUMMARY] Read bytes from a file descriptor.
    [PURPOSE] This routine is similar to the system <<read(2)>> call, except
    that interrupted system calls are restarted.
    <fd> The file descriptor to read from. This descriptor must NOT be set to
    non-blocking IO.
    <buf> The buffer in which to write the data.
    <nbytes> The number of bytes to read.
    [RETURNS] The number of bytes requested on success, the number of bytes
    read on end of file (or closure) and -1 on error.
*/
{
    uaddr total_bytes_read = 0;
    iaddr bytes_read;
    int flags = 0;

    while (total_bytes_read < nbytes)
    {
	errno = 0;
	if ( ( bytes_read = t_rcv (fd, buf + total_bytes_read,
				   nbytes - total_bytes_read, &flags) ) < 0 )
	{
	    /*  Error  */
	    if (errno != EINTR)
	    {
		/*  Unrecoverable error  */
		return (-1);
	    }
	    continue;
	}
	/*  No error condition  */
	total_bytes_read += bytes_read;
	return (total_bytes_read);
    }
    return (total_bytes_read);
}   /*  End Function r_atm_read  */

/*UNPUBLISHED_FUNCTION*/
iaddr r_atm_write (int fd, CONST char *buf, uaddr nbytes)
/*  [SUMMARY] Write bytes to a file descriptor.
    <fd> The file descriptor to write to. This descriptor must NOT be set to
    non-blocking IO.
    <buf> The buffer in which to write the data.
    <nbytes> The number of bytes to write.
    [RETURNS] The number of bytes requested on success, else -1 indicating
    error.
*/
{
    return t_snd (fd, (char *) buf, nbytes, 0);
}   /*  End Function r_atm_write  */

/*UNPUBLISHED_FUNCTION*/
flag r_atm_close (int fd)
/*  [SUMMARY] Close a file descriptor.
    <fd> The file descriptor to close.
    [RETURNS] TRUE on success, else FALSE.
*/
{ 

    int a;
    a = _r_clean_up(fd);
    if (a == 0)
     {
	return TRUE;
     }
     else
      {
	return FALSE;
      }	  

}   /*  End Function r_atm_close  */


/*  Private functions follow  */

/*
 * _r_clean_up(fd)
 *
 * Unbinds and closes the given endpoint.
 *
 * In:
 *	fd		fd to clean up
 *
 * Returns:	TRUE on success, else FALSE.
 */

static int _r_clean_up(int fd)
{
    int event=0;
    int result = 0;

    if (t_unbind(fd) < 0) {
        if ( t_errno == TLOOK ) {
            event = t_look(fd);
	}
    }

    if ( (event) && (event != T_DATA) ) {
        _r_handle_error(fd,"t_unbind");
    }

    result = t_close(fd);

    return result;
    
}/*End of function _r_clean_up*/   



/*
 * _r_open_device(device)
 *
 * Opens the transport device.
 *
 * In:
 *	device		pathname of device to open
 *
 * Returns:
 *	the return value from t_open()
 */

static int _r_open_device (char *device)
{
    int fd;
    struct t_info info;

    fd=t_open(device, O_RDWR, &info);
    if (fd < 0) {
	t_error("t_open");
    }
    else {
	/* max AAL5 PDU length = 65535 bytes */
	/* this is large enough to hold all possible options */
    }

    return fd;
}/*End  of function _r_open_device*/

/*
 * _r_svc_bind_listener (addr_string)
 *
 * Binds the local endpoint of an SVC listening process.
 *
 * In:
 *	addr_string	The binary representation of the ATM address is written here
 *
 * Returns:
 *	fd	on success
 *	-1	on failure
 *
 * Notes:
 *	The FORE CLIP listener consumes selector bytes 0x00-0x03.
 *	These should not be used by user applications.
 */

static int _r_svc_bind_listener (char *addr_string)
{
    int fd;
    struct t_bind bind_req;
    ATMSAPAddress addr_req;
    int count;
    static char function_name[] = "svc_bind_listener";

    /*Set the NSAP address to be zero*/
    memset(&addr_req,0,sizeof addr_req);	
    addr_req.addressType = AF_ATM;
    /*Open the device*/	
    fd = _r_open_device(DEVICE);
    if (fd < 0) {
	exit(1);
    }
    /*Fill the NSAP address field*/
    if (_r_get_local_address(fd, addr_req.sap.t_atm_sap_addr.address) < 0)
    {
	fprintf (stderr, "%s: error getting address\t%s\n",
		 function_name, ERRSTRING);
	return (-1);
    }
    /*
     * Note: this approach may be FORE implementation-specific.
     *
     * When we bind with tag_addr ABSENT and tag_selector PRESENT,
     * only the selector (i.e. address[19]) is used by the TP.  The rest
     * of the local address is filled in by the TP and can be obtained
     * via the 'ret' parameter or with t_getname()/t_getprotaddr().
     */
     
    addr_req.sap.t_atm_sap_addr.SVE_tag_addr = (int8_t)T_ATM_ABSENT;
    addr_req.sap.t_atm_sap_addr.SVE_tag_selector = (int8_t)T_ATM_PRESENT;
    addr_req.sap.t_atm_sap_addr.address_format = (u_int8_t)T_ATM_ENDSYS_ADDR;
    addr_req.sap.t_atm_sap_addr.address_length = 20;
    addr_req.sap.t_atm_sap_layer2.SVE_tag = (int8_t)T_ATM_ABSENT;
    addr_req.sap.t_atm_sap_layer3.SVE_tag = (int8_t)T_ATM_ABSENT;
    /* this is optional; we use BHLI (appl) ID here to avoid conflicts */
    addr_req.sap.t_atm_sap_appl.SVE_tag = (int8_t)T_ATM_PRESENT;
    addr_req.sap.t_atm_sap_appl.ID_type = (u_int8_t)T_ATM_USER_APP_ID;
    memcpy(addr_req.sap.t_atm_sap_appl.ID.user_defined_ID, BHLI_MAGIC, 8);
    memset(&bind_req, 0, sizeof(bind_req));
    bind_req.qlen = 1;
    bind_req.addr.len = sizeof(addr_req);
    bind_req.addr.buf = (char *) &addr_req;
    /*  Binding to a random port. On a fail the fd is unbound, closed and a new
	fd created. This is due to the fact when bind() is called it it seems
	that this function automatically binds the fd to an SAP regardless of
	whether that particular SAP is free or not  */
    for (count = 0; count < 256; ++count)
    {
	/*Try different SAPs*/
	addr_req.sap.t_atm_sap_addr.address[19] = count;
 
	/*Try to bind to an SAP - if successful break*/
	if ( t_bind (fd, &bind_req, (struct t_bind *) NULL) == 0) break;

       	/*  Check to see if the error is simply due to the fact that the SAP
	    is already in use - if so unbind and close the current fd before
	    creating a new one. Try to bind to another SAP using this new fd
	    */
	if (errno == EADDRINUSE)
	{
	    if (_r_clean_up(fd) <0) return -1;
	    fd =_r_open_device(DEVICE);
	    if (fd < 0) return -1;
	    continue;
	}
        fprintf (stderr, "%s: error binding ATMSVC socket\t%s\n",
                 function_name, ERRSTRING);
        exit (RV_SYS_ERROR);
    }
    /*  No free SAPs  */
    if (count >= 256)
    {
	fprintf (stderr, "%s: no free ATMSVC SAPs\n", function_name);
	errno = 0;
	_r_clean_up(fd);
	return (-1);
    }
    /*Sets the value of addr_string*/
    memcpy (addr_string,addr_req.sap.t_atm_sap_addr.address, ATMNSAP_ADDR_LEN);
    return (fd);
}   /*  End Function _r_svc_bind_listener  */

static void _r_atm_karma_to_atm (struct t_atm_traffic_substruct *atm,
				 int32_t *size, u_int8_t *traffic,
				 CONST struct karma_traffic_parameters *karma) 
/*  [SUMMARY] Convert traffic parameters from Karma to ATM.
    <atm> The ATM traffic parameters will be written here.
    <karma> The Karma traffic parameters.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_r_atm_karma_to_atm";

    switch (karma->traffic_class)
    {
      case K_ATM_TRAFFIC_CLASS_UBR:
	*traffic = 0;	/*UBR*/
	break;
      case K_ATM_TRAFFIC_CLASS_CBR:
	*traffic = T_ATM_CBR;
	break;
      default:	
                fprintf (stderr, "%s: illegal traffic class: %u\n",
		 function_name, karma->traffic_class);
	abort ();
	/*break;*/
    }
   
    atm->SCR_all_traffic = karma->rate_typ / BYTES_PER_CELL;
    atm->PCR_all_traffic = karma->rate_max / BYTES_PER_CELL;
    *size = karma->sdu_max;
    if (*size < 1) *size = DEFAULT_SDU;
    atm->MBS_all_traffic = karma->dv_max * BYTES_PER_CELL / 1000;
}   /*  End Function _r_atm_karma_to_atm  */

static void _r_atm_to_karma (struct karma_traffic_parameters *karma,
			     CONST struct t_atm_traffic_substruct *atm,
			     int32_t size, u_int8_t traffic)
/*  [SUMMARY] Convert traffic parameters from ATM to Karma.
    <karma> The Karma traffic parameters. This will be updated.
    <atm> The ATM traffic parameters.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_r_atm_to_karma";

    switch (traffic)
    {
      case 0:
	karma->traffic_class = K_ATM_TRAFFIC_CLASS_UBR;
	break;
      case T_ATM_CBR:
	karma->traffic_class = K_ATM_TRAFFIC_CLASS_CBR;
	break;
      default:
	fprintf (stderr, "%s: illegal traffic class: %u\n",
		 function_name, traffic);
	abort ();
	/*break;*/
    }
    karma->rate_typ = atm->SCR_all_traffic * BYTES_PER_CELL;
    karma->rate_max = atm->PCR_all_traffic * BYTES_PER_CELL;
    karma->sdu_max = size;
    karma->dv_max = atm->MBS_all_traffic * 1000 / BYTES_PER_CELL;
}   /*  End Function _r_atm_to_karma  */

/*
 * _r_get_local_address(fd, addr)
 *
 * Obtains the local ATM address
 *
 * In:
 *	fd		bound fd of transport endpoint to obtain address for
 *
 * Out:
 *	addr		local address
 *
 * Returns:
 *	0	on success
 *	-1	on failure
 */

/* Some OSes (namely, AIX) have implemented the v5 t_getprotaddr() command.
 * Others, like Irix, still use the v4 t_getname().  Both are offered here.
 */
#ifdef aix
#define T_GETPROTADDR(fd, laddr, raddr)	t_getprotaddr(fd, laddr, raddr)
#else
#define T_GETPROTADDR(fd, laddr, raddr) t_getname(fd, laddr.addr, LOCALNAME)
#endif /* aix */

#ifdef OS_Solaris
extern int t_getname ();
#endif

static int _r_get_local_address(int fd, unsigned char addr[])
{
    ATMSAPAddress local_addr;
    struct t_bind boundaddr;

    boundaddr.addr.maxlen = sizeof(local_addr);
    boundaddr.addr.buf = (char *) &local_addr;

    if (T_GETPROTADDR(fd, &boundaddr, NULL) < 0) {
	t_error("T_GETPROTADDR (local_address)");
	return -1;
    }

    memcpy(addr, local_addr.sap.t_atm_sap_addr.address, 20);

    return 0;
}/*End of function _r_get_local_address*/


/*
 * _r_handle_error(fd, func)
 *
 * In:
 *	fd		fd where error occurred
 *	func		name of function that caused the error
 *
 * Returns:
 *	0	if the error was that a disconnect was received
 *	-1 	if any other error occurred
 *
 * Notes:
 *	also handles disconnect indications by calling t_rcvdis().
 */

static int _r_handle_error(int fd, char *func)
{
    fprintf (stderr, "%s returns negative; ", func);

    if (t_errno == TLOOK) {
	int code;
	code = t_look(fd);
	if (code == T_DISCONNECT) {
	    fprintf (stderr, "t_look returns T_DISCONNECT, calling t_rcvdis()\n");
	    if (t_rcvdis(fd, (struct t_discon *) NULL) < 0) {
		t_error("t_rcvdis");
	    }
	    else {
		return 0;
	    }
	}
	else {
	    fprintf (stderr, "t_look returns %x (unexpected)\n", code);
	}
    }
    else {
	t_error(func);
    }

    return -1;
}/*End of function _r_handle_error*/



/*
 * _r_construct_options(fd, traffic, bearer, flags,optsize)
 *
 * Constructs a buffer with the requested options which can be passed
 * to t_optmgmt() or t_connect()
 *
 * In:
 *	fd		fd of bound trasport endpoint.  Used to obtain
 *			information and possibly the local address.
 *
 *	traffic		Specifies the type of traffic user wants (ie CBR or UBR)
 *			
 *	bearer		Specifies the max. SDU that the user wants.		
 *
 *	flags		if OPT_FLAGS_CPID is set, the T_ATM_ORIG_ADDR option 
 *			is created, corresponding to the Calling Party IE.
 *			if OPT_FLAGS_PMP is set, bearer and traffic IEs are
 *		       	created and set as appropriate for a PMP connection.
 *
 * Out:
 *	optsize		on return, holds the actual size of the options.
 *
 * Returns:
 *			pointer to buffer containing the options.  The
 *			caller must release the buffer using free()
 *
 *			returns NULL pointer on failure.
 */

static char *_r_construct_options(int fd, struct t_atm_traffic *traffic, struct t_atm_bearer *bearer,int flags, long *optsize)
{
    struct t_opthdr *popt;
    char *buf;
    int qos_cells;
    struct t_info info;

    if (t_getinfo(fd, &info) < 0) {
	t_error("t_getinfo");
	return (char *) NULL;
    }
 
    buf = (char *) malloc(info.options);
    if (buf == (char *) NULL) {
	fprintf(stderr,"Unable to allocate %ld bytes for options\n",
	    info.options);
	return (char *) NULL;
    }

    popt = (struct t_opthdr *) buf;

    if (flags & OPT_FLAGS_CPID) {
	/*
	 * this constructs the T_ATM_ORIG_ADDR option, which is used to
	 * signal the UNI 3.1 Calling Party ID Information Element.
	 */

	t_atm_addr *source_addr;

	popt->len	= sizeof(struct t_opthdr) + sizeof(t_atm_addr);
	popt->level	= T_ATM_SIGNALING;
	popt->name	= T_ATM_ORIG_ADDR;
	popt->status	= 0;

	source_addr = (t_atm_addr *)((char *)popt + sizeof(struct t_opthdr));


	source_addr->address_format = T_ATM_ENDSYS_ADDR;
	source_addr->address_length = 20;
	if (_r_get_local_address(fd, source_addr->address)) {
	    fprintf(stderr,"Can't get local address!\n");
	    free(buf);
	    return (char *) NULL;
	}

	popt = T_OPT_NEXTHDR(buf, info.options , popt);
    }

    /*
     * this constructs all options necessary (bearer cap., QoS, and 
     * Traffic Descriptor) to signal for a CBR connection with the
     * specified QoS in kbit/sec., and/or specify a PMP connection.
     *
     * for FORE 200e cards, the adapter shapes traffic to CBR with
     * rate equal to PCR CLP=0+1 (traffic.forward.PCR_all_traffic)
     */

	/*Choose the larger of the two values of PCR to set the variable qos_cells*/
	if ( (traffic->forward.PCR_all_traffic) > (traffic->backward.PCR_all_traffic))
	{
		qos_cells = traffic->forward.PCR_all_traffic/48;
	}
	else
	{
		qos_cells = traffic->backward.PCR_all_traffic/48;
	}


    if ((qos_cells > 0 && qos_cells < LINE_RATE) || (flags & OPT_FLAGS_PMP)) {

	/* T_ATM_BEARER_CAP: Broadband bearer capability */
	popt->len	= sizeof(struct t_opthdr) + 
	                  sizeof(struct t_atm_bearer);
	popt->level	= T_ATM_SIGNALING;
	popt->name	= T_ATM_BEARER_CAP;
	popt->status	= 0;

	bearer = (struct t_atm_bearer *)((char *)popt +
                                         sizeof(struct t_opthdr));
	bearer->bearer_class		 = T_ATM_CLASS_X;

	if (bearer->traffic_type==T_ATM_CBR) {
	    bearer->timing_requirements	 = T_ATM_END_TO_END;
	} else {
	    bearer->timing_requirements	 = 0;/* UBR */
	}
	bearer->clipping_susceptibility	 = T_ATM_NULL;
	if (flags & OPT_FLAGS_PMP)
	    bearer->connection_configuration = T_ATM_1_TO_MANY;
	else
	    bearer->connection_configuration = T_ATM_1_TO_1;

	popt = T_OPT_NEXTHDR(buf, info.options, popt);

	/* T_ATM_TRAFFIC: traffic descriptor */
	popt->len	= sizeof(struct t_opthdr) + 
                          sizeof(struct t_atm_traffic);
	popt->level	= T_ATM_SIGNALING;
	popt->name	= T_ATM_TRAFFIC;
	popt->status	= 0;

	traffic=(struct t_atm_traffic *)((char *)popt+sizeof(struct t_opthdr));

	traffic->forward.PCR_high_priority	= T_ATM_ABSENT;
	traffic->forward.SCR_high_priority	= T_ATM_ABSENT;
	traffic->forward.MBS_high_priority	= T_ATM_ABSENT;
	traffic->forward.tagging		= T_NO;


	traffic->backward.PCR_high_priority	= T_ATM_ABSENT;
	traffic->backward.SCR_high_priority	= T_ATM_ABSENT;
	traffic->backward.MBS_high_priority	= T_ATM_ABSENT;
	traffic->backward.tagging		= T_NO;
	
	traffic->best_effort			= qos_cells ? T_NO : T_YES;

	popt = T_OPT_NEXTHDR(buf, info.options, popt);
    }

    if (qos_cells > 0 && qos_cells < LINE_RATE) {
	struct t_atm_qos *qos;

	/* T_ATM_QOS: Quality of Service */
	popt->len	= sizeof(struct t_opthdr) + sizeof(struct t_atm_qos);
	popt->level	= T_ATM_SIGNALING;
	popt->name	= T_ATM_QOS;
	popt->status	= 0;

	qos = (struct t_atm_qos *)((char *)popt + sizeof(struct t_opthdr));
	qos->coding_standard	= T_ATM_ITU_CODING;
	qos->forward.qos_class	= T_ATM_QOS_CLASS_1;
	qos->backward.qos_class	= T_ATM_QOS_CLASS_1;

	popt = T_OPT_NEXTHDR(buf, info.options, popt);
    }

    /*
     * return actual size of options and option buffer to user
     */

    *optsize = (char *)popt - buf;

    return buf;
}   /*  End Function _r_construct_options  */


#else  /*  HAS_FORE_ATM  */

struct unused
{
    int a;
};

#endif  /*  !HAS_FORE_ATM  */
