/*  kvis-multibeam.cc

    Glish file for Intelligent Array scrolling buffer.

    Copyright (C) 2004  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This file contains code to implement a scrolling buffer using Intelligent
    Arrays. The data come from a Glish interpretor.


    Written by      Richard Gooch   24-APR-2004: Copied from
  modules/MultibeamView/glish.cc

    Updated by      Richard Gooch   25-APR-2004: First test version.

    Updated by      Richard Gooch   7-JUL-2004: Check KVIS_MULTIBEAM_VERBOSE
  environment variable.

    Last updated by Richard Gooch   8-JUL-2004: Added "scrollBufDestroy" event.

    $Id: kvis-multibeam_noDL.cc,v 2.4 2011/05/03 02:12:10 cal103 Exp $

*/
#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/Xaw/Label.h>
#include <Glish/Value.h>
#include <Glish/Client.h>
#include <karma.h>
#include <k_list.h>
#include <karma_dispdata.h>
#include <karma_viewimg.h>
#include <karma_iarray.h>
#include <karma_wcs.h>
#include <karma_gls.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_ex.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_r.h>
#include <karma_c.h>
#include <Xkw/ImageDisplay.h>


#define MAGIC_NUMBER      3959537415U
#define SCROLL_INCREMENT           1

#define VERIFY_SCROLLBUF(scrollbuf) if (scrollbuf == NULL) \
{fprintf (stderr, "NULL scrollbuf passed\n"); \
 return (FALSE); } \
if (scrollbuf->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid scrollbuf object\n"); \
 return (FALSE); }

#define VERIFY_SCROLLBUF2(scrollbuf) if (scrollbuf == NULL) \
{fprintf (stderr, "NULL scrollbuf passed\n"); \
 return; } \
if (scrollbuf->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid scrollbuf object\n"); \
 return; }


/*  Private structure types  */
struct header_pair_type
{
    const char *name;
    Value **value;
    flag required;
};

struct header_type
{
    Value *imageShape;
    Value *axisNames;
    Value *referenceValues;
    Value *referencePixels;
    Value *deltas;
    Value *name;
    Value *domain;
    Value *dataUnits;
    Value *blankVal;
};

struct scrolling_buffer_struct
{
    unsigned int magic_number;
    int ident;
    iarray cube;
    iarray ra_arr;
    iarray dec_arr;
    KDisplayDataObject dataobject;
    struct scrolling_buffer_struct *prev;
    struct scrolling_buffer_struct *next;
};


/*  Functions to export which need to be marked as C, to prevent C++ name
    mangling  */
EXTERN_FUNCTION (void multibeamvis_init, (void) );
EXTERN_FUNCTION (flag WidgetRealise, (Widget w) );


/*  Private functions  */
STATIC_FUNCTION (flag scroll_init_cbk,
		 (CONST void *client_ptr, void *object,
		  CONST char *name, CONST void *value_ptr) );
STATIC_FUNCTION (flag scroll_add_cbk,
		 (CONST void *client_ptr, void *object,
		  CONST char *name, CONST void *value_ptr) );
STATIC_FUNCTION (flag scroll_destroy_cbk,
		 (CONST void *client_ptr, void *object,
		  CONST char *name, CONST void *value_ptr) );
STATIC_FUNCTION (void copy_plane, (iarray array, unsigned long plane_index,
				   CONST void *in, double blank) );
STATIC_FUNCTION (void copy_line_of_doubles,
		 (iarray array, CONST double *data, uaddr ypos,double scale) );
STATIC_FUNCTION (void new_dataobject_func,
		 (void *info, KDisplayDataObject data) );
STATIC_FUNCTION (flag canvas_position_func,
		 (ViewableImage vimage, double x, double y,
		  void *value, unsigned int event_code,
		  void *e_info, void **f_info, double x_lin, double y_lin,
		  unsigned int value_type,
		  double x_im, double y_im, int x_pix, int y_pix) );


/*  Private data  */
static struct header_type header;
static struct header_pair_type header_pairs[] =
{
    {"imageShape",       &header.imageShape,       TRUE},
    {"axisNames",        &header.axisNames,        TRUE},
    {"referenceValues",  &header.referenceValues,  TRUE},
    {"referencePixels",  &header.referencePixels,  TRUE},
    {"deltas",           &header.deltas,           TRUE},
    {"name",             &header.name,             TRUE},
    {"domain",           &header.domain,           FALSE},
    {"dataUnits",        &header.dataUnits,        TRUE},
    {"blankVal",         &header.blankVal,         TRUE},
    {NULL, NULL}
};
static flag verbose = FALSE;
static struct scrolling_buffer_struct *first_scrollbuf = NULL;
static struct scrolling_buffer_struct *last_scrollbuf = NULL;


/*  Public functions follow  */

/*MAGIC_FUNCTION*/
void multibeamvis_init (void)
/*  [SUMMARY] Shared object initialisation function.
*/
{
    /*HACK*/fprintf (stderr, "Registering scroll buffer callbacks...\n");
    gls_register_glish_event_handler ( ( flag (*) () ) scroll_init_cbk,
				       "scrollBufBegin", NULL );
    gls_register_glish_event_handler ( ( flag (*) () ) scroll_add_cbk,
				       "scrollBufAddImage", NULL );
    gls_register_glish_event_handler ( ( flag (*) () ) scroll_destroy_cbk,
				       "scrollBufDestroy", NULL );
    if ( r_getenv ("KVIS_MULTIBEAM_VERBOSE") ) verbose = TRUE;
}   /*  End Function multibeamvis_init  */

/*MAGIC_FUNCTION*/
flag WidgetRealise (Widget w)
/*  [SUMMARY] WidgetRealise hook.
*/
{
    KWorldCanvas pc_canvas, tc_canvas;

    /*HACK*/fprintf (stderr, "Registering canvas event handlers...\n");
    XtVaGetValues (w,
		   XkwNpseudoColourCanvas, &pc_canvas,
		   XkwNtrueColourCanvas, &tc_canvas,
		   NULL);
    if (pc_canvas)
	viewimg_register_position_event_func (pc_canvas, canvas_position_func,
					      w);
    if (tc_canvas)
	viewimg_register_position_event_func (tc_canvas, canvas_position_func,
					      w);
    return (TRUE);
}   /*  End Function WidgetRealise  */


/*  Private functions follow  */

static flag scroll_init_cbk (CONST void *client_ptr, void *object,
			     CONST char *name, CONST void *value_ptr)
/*  [SUMMARY] This is a Glish event handler.
    <client_ptr> A pointer to the Client object.
    <object> An aribtrary object pointer.
    <name> The name of the Glish event.
    <value_ptr> A pointer to the Glish Value.
    [RETURNS] TRUE if the event was consumed, else FALSE.
*/
{
    KCallbackFunc callback;
    unsigned int  axis;
    unsigned long xlen, ylen, zlen, ktype, dim_index;
    double blank_val, min, max;
    struct header_pair_type *pair;
    void *data_ptr;
    char *axis_names, *ptr, *axis_name;
    CONST char *domain = "MultiBeam";
    Client *client = (Client *) client_ptr;
    Value *value = (Value *) value_ptr;
    Value *image_field, *count_field, *header_field, *RAvsBeam, *DECvsBeam;
    Value *data;
    struct scrolling_buffer_struct *scrollbuf;
    char txt[STRING_LENGTH];
    static char function_name[] = "scroll_init_cbk";

    if (verbose) fprintf (stderr, "event: \"%s\"\t", name);
    if ( ( scrollbuf = (struct scrolling_buffer_struct *)
	   m_calloc (sizeof *scrollbuf) ) == NULL )
	m_abort (function_name, "scrolling buffer structure");
    if (value->Type () != TYPE_RECORD)
    {
	fprintf (stderr, "%s: Glish event: \"%s\" is not a record\n",
		 function_name, name);
	return (TRUE);
    }
    image_field = value->NthField (1);
    if (image_field->Type () != TYPE_RECORD)
    {
	fprintf (stderr, "%s: first field is not a record\n", function_name);
	return (TRUE);
    }
    if ( ( count_field = value->NthField (2) ) == NULL )
    {
	fprintf (stderr, "%s: no second field\n", function_name);
	return (TRUE);
    }
    if (count_field->Type () != TYPE_INT)
    {
	fprintf (stderr, "%s: second field is not an int\n", function_name);
	return (TRUE);
    }
    zlen = count_field->IntVal ();
    if ( ( data = image_field->Field ("data") ) == NULL )
    {
	fprintf (stderr, "%s: no data field in image\n", function_name);
	return (TRUE);
    }
    if ( ( header_field = image_field->Field ("header") ) == NULL )
    {
	fprintf (stderr, "%s: no header field in image\n", function_name);
	return (TRUE);
    }
    if (header_field->Type () != TYPE_RECORD)
    {
	fprintf (stderr, "%s: header is not a record\n", function_name);
	return (TRUE);
    }
    /*  Get each header value  */
    for (pair = header_pairs; pair->name; ++pair)
    {
	if ( ( *pair->value = header_field->Field (pair->name) ) == NULL )
	{
	    if (!pair->required) continue;
	    fprintf (stderr, "%s: no \"%s\" field in header\n",
		     function_name, pair->name);
	    return (TRUE);
	}
    }
    if (header.referenceValues->Type () != TYPE_DOUBLE)
    {
	fprintf (stderr, "%s: referenceValues is not double type\n",
		 function_name);
	return (TRUE);
    }
    if (header.referencePixels->Type () != TYPE_DOUBLE)
    {
	fprintf (stderr, "%s: referencePixels is not double type\n",
		 function_name);
	return (TRUE);
    }
    if (header.deltas->Type () != TYPE_DOUBLE)
    {
	fprintf (stderr, "%s: deltas is not double type\n", function_name);
	return (TRUE);
    }
    if (header.dataUnits->Type () != TYPE_STRING)
    {
	fprintf (stderr, "%s: dataUnits is not string type\n", function_name);
	return (TRUE);
    }
    switch ( data->Type () )
    {
      case TYPE_ERROR:
	fprintf (stderr, "%s: TYPE_ERROR not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_REF:
	fprintf (stderr, "%s: TYPE_REF not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_SUBVEC_REF:
	fprintf (stderr, "%s: TYPE_SUBVEC_REF not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_BOOL:
	fprintf (stderr, "%s: TYPE_BOOL not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_BYTE:
	fprintf (stderr, "%s: TYPE_BYTE not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_SHORT:
	fprintf (stderr, "%s: TYPE_SHORT not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_INT:
	ktype = K_INT;
	data_ptr = data->IntPtr ();
	break;
      case TYPE_FLOAT:
	ktype = K_FLOAT;
	data_ptr = data->FloatPtr ();
	break;
      case TYPE_DOUBLE:
	ktype = K_DOUBLE;
	data_ptr = data->DoublePtr ();
	break;
      case TYPE_STRING:
	fprintf (stderr, "%s: TYPE_STRING not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_AGENT:
	fprintf (stderr, "%s: TYPE_AGENT not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_FUNC:
	fprintf (stderr, "%s: TYPE_FUNC not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_RECORD:
	fprintf (stderr, "%s: TYPE_RECORD not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_COMPLEX:
	ktype = K_COMPLEX;
	data_ptr = data->ComplexPtr ();
	break;
      case TYPE_DCOMPLEX:
	ktype = K_DCOMPLEX;
	data_ptr = data->DcomplexPtr ();
	break;
      case TYPE_FAIL:
	fprintf (stderr, "%s: TYPE_FAIL not supported\n", function_name);
	return (TRUE);
	/*break;*/
      default:
	fprintf ( stderr, "%s: unsupported data type: %u\n",
		  function_name, data->Type () );
	return (TRUE);
	/*break;*/
    }
    if ( (ktype == K_COMPLEX) || (ktype == K_DCOMPLEX) )
    {
	fprintf (stderr, "%s: complex type not supported\n", function_name);
	return (TRUE);
    }
    /*  Get rest of 3D buffer shape  */
    xlen = header.imageShape->IntVal (1);
    ylen = header.imageShape->IntVal (2);
    if (verbose) fprintf (stderr, "xlen: %lu  ylen: %lu  zlen: %lu\n",
			  xlen, ylen, zlen);
    /*  Check sanity of RA and DEC fields  */
    if ( ( RAvsBeam = image_field->Field ("RAvsBeam") ) != NULL )
    {
	if (RAvsBeam->Type () != TYPE_DOUBLE)
	{
	    fprintf (stderr, "%s: RAvsBeam field is not type double\n",
		     function_name);
	    return (TRUE);
	}
	if (RAvsBeam->Length () != ylen)
	{
	    fprintf (stderr, "%s: RAvsBeam field is not length: %lu\n",
		     function_name, ylen);
	    return (TRUE);
	}
    }
    if ( ( DECvsBeam = image_field->Field ("DECvsBeam") ) != NULL )
    {
	if (DECvsBeam->Type () != TYPE_DOUBLE)
	{
	    fprintf (stderr, "%s: DECvsBeam field is not type double\n",
		     function_name);
	    return (TRUE);
	}
	if (DECvsBeam->Length () != ylen)
	{
	    fprintf (stderr, "%s: DECvsBeam field is not length: %lu\n",
		     function_name, ylen);
	    return (TRUE);
	}
    }
    /*  Construct a 3D Intelligent Array for the scrolling data buffer  */
    if ( ( scrollbuf->cube = iarray_create_3D (zlen, ylen, xlen, ktype) )
	 == NULL )
    {
	m_error_notify (function_name, "scroll array");
	return (TRUE);
    }
    iarray_fill_double (scrollbuf->cube, TOOBIG);
    iarray_set_value_name (scrollbuf->cube,
			   header.dataUnits->StringVal (), TRUE);
    iarray_put_named_string ( scrollbuf->cube, "OBJECT",
			      header.name->StringVal (' ') );
    /*  Create 2D arrays for the scrolling RA and DEC buffers  */
    if ( ( scrollbuf->ra_arr = iarray_create_2D (zlen, ylen, K_DOUBLE) )
	 == NULL ) m_abort (function_name, "RA array");
    iarray_fill_double (scrollbuf->ra_arr, TOOBIG);
    if ( ( scrollbuf->dec_arr = iarray_create_2D (zlen, ylen, K_DOUBLE) )
	 == NULL ) m_abort (function_name, "DEC array");
    iarray_fill_double (scrollbuf->dec_arr, TOOBIG);
    axis_names = header.axisNames->StringVal (' ');
    ptr = axis_names;
    for (axis = 1; axis < 4; ++axis)
    {
	dim_index = 3 - axis;
	if (ptr == NULL)
	{
	    fprintf (stderr, "No more axis names!\n");
	    return (TRUE);
	}
	axis_name = ex_str (ptr, (CONST char **) &ptr);
	iarray_set_dim_name (scrollbuf->cube, dim_index, axis_name, TRUE);
	sprintf (txt, "CTYPE%u", axis);
	if ( !iarray_put_named_string (scrollbuf->cube, txt, axis_name) )
	    m_abort (function_name, "CTYPEn string value");
	m_free (axis_name);
	sprintf (txt, "CRVAL%u", axis);
	if ( !iarray_put_double ( scrollbuf->cube, txt,
				  header.referenceValues->DoubleVal (axis) ) )
	    m_abort (function_name, "CRVALn string value");
	sprintf (txt, "CRPIX%u", axis);
	if ( !iarray_put_double ( scrollbuf->cube, txt,
				  header.referencePixels->DoubleVal (axis) ) )
	    m_abort (function_name, "CRPIXn string value");
	sprintf (txt, "CDELT%u", axis);
	if ( !iarray_put_double ( scrollbuf->cube, txt,
				  header.deltas->DoubleVal (axis) ) )
	    m_abort (function_name, "CDELTn string value");
    }
    /*  Copy first image data  */
    blank_val = header.blankVal->DoubleVal ();
    copy_plane (scrollbuf->cube, 0, data_ptr, blank_val);
#if 0
    iarray_min_max (scrollbuf->cube, CONV_CtoR_REAL, &min, &max);
    if (max <= min)
    {
	fprintf (stderr,
		 "Constant data value: %e, setting range to: %e to %e\n",
		 min, min - 0.5, min + 0.5);
	min -= 0.5;
	max += 0.5;
    }
    else fprintf (stderr, "Initial data range: %e  to  %e\n", min, max);
#endif
    /*  Include RA and DEC data if available  */
    if (RAvsBeam)
	copy_line_of_doubles (scrollbuf->ra_arr, RAvsBeam->DoublePtr (),
			      0, 1.0 / PION180);
    if (DECvsBeam)
	copy_line_of_doubles (scrollbuf->dec_arr,
			      DECvsBeam->DoublePtr (), 0, 1.0 / PION180);
    ds_add_named_object (scrollbuf->cube->multi_desc,
			 scrollbuf->cube->multi_desc->num_arrays, "filename",
			 st_dup ("GLISH::scrollBuffer"),
			 ( void (*) (...) ) m_free);
    if (header.domain) domain = header.domain->StringVal (' ');
    scrollbuf->magic_number = MAGIC_NUMBER;
    if (last_scrollbuf == NULL) {
      scrollbuf->ident = 1;
    } else {
      scrollbuf->ident = last_scrollbuf->ident + 1;
    }
    KLIST_APPEND2 (first_scrollbuf, last_scrollbuf, scrollbuf);
    callback = dispdata_register_data_create_func
	(dispdata_array_create_dataclass (), new_dataobject_func, scrollbuf);
    ds_event_dispatch ( iarray_multi_array (scrollbuf->cube), domain,
			header.name->StringVal (' ') );
    c_unregister_callback (callback);

    /* Return the scroll buffer id (to pksmonitor.g). */
    Value reply_value ( scrollbuf->ident );
    client->PostEvent ("imageID", &reply_value);

    return (TRUE);
}   /*  End Function scroll_init_cbk  */

static flag scroll_add_cbk (CONST void *client_ptr, void *object,
			    CONST char *name, CONST void *value_ptr)
/*  [SUMMARY] This is a Glish event handler.
    <client_ptr> A pointer to the Client object.
    <object> An aribtrary object pointer.
    <name> The name of the Glish event.
    <value_ptr> A pointer to the Glish Value.
    [RETURNS] TRUE if the event was consumed, else FALSE.
*/
{
    KwcsAstro ap;
    unsigned int ktype;
    unsigned long ylen, zlen, index, num_to_scroll;
    int ident;
    double zval, zpos, crval3, crpix3, cdelt3;
    double blank_val;
    struct header_pair_type *pair;
    void *data_ptr;
    Value *value = (Value *) value_ptr;
    Value *image_field, *id_field, *header_field, *RAvsBeam, *DECvsBeam;
    Value *data;
    struct scrolling_buffer_struct *scrollbuf;
    char txt[STRING_LENGTH];
    static char function_name[] = "scroll_add_cbk";

    if (verbose) fprintf (stderr, "%s: event: \"%s\"\n", function_name, name);
    if (value->Type () != TYPE_RECORD)
    {
	fprintf (stderr, "%s: Glish event: \"%s\" is not a record\n",
		 function_name, name);
	return (TRUE);
    }
    id_field = value->NthField (1);
    if (id_field->Type () != TYPE_INT)
    {
	fprintf (stderr, "%s: ID field is not an int\n", function_name);
	return (TRUE);
    }
    if (verbose) fprintf ( stderr, "%s: image ID: %d\n",
			   function_name, id_field->IntVal () );

    /* Identify the scrollbuf. */
    ident = id_field->IntVal();
    for (scrollbuf = last_scrollbuf; scrollbuf; scrollbuf = scrollbuf->prev) {
        if (scrollbuf->ident == ident) break;
    }
    VERIFY_SCROLLBUF (scrollbuf);

    if ( ( image_field = value->NthField (2) ) == NULL )
    {
	fprintf (stderr, "%s: no second field\n", function_name);
	return (TRUE);
    }
    if (image_field->Type () != TYPE_RECORD)
    {
	fprintf (stderr, "%s: second field is not a record\n", function_name);
	return (TRUE);
    }
    if ( ( data = image_field->Field ("data") ) == NULL )
    {
	fprintf (stderr, "%s: no data field in image\n", function_name);
	return (TRUE);
    }
    if ( ( header_field = image_field->Field ("header") ) == NULL )
    {
	fprintf (stderr, "%s: no header field in image\n", function_name);
	return (TRUE);
    }
    if (header_field->Type () != TYPE_RECORD)
    {
	fprintf (stderr, "%s: header is not a record\n", function_name);
	return (TRUE);
    }
    ylen = iarray_dim_length (scrollbuf->cube, 1);
    /*  Check sanity of RA and DEC fields  */
    if ( ( RAvsBeam = image_field->Field ("RAvsBeam") ) != NULL )
    {
	if (RAvsBeam->Type () != TYPE_DOUBLE)
	{
	    fprintf (stderr, "%s: RAvsBeam field is not type double\n",
		     function_name);
	    return (TRUE);
	}
	if (RAvsBeam->Length () != ylen)
	{
	    fprintf (stderr, "%s: RAvsBeam field is not length: %lu\n",
		     function_name, ylen);
	    return (TRUE);
	}
    }
    if ( ( DECvsBeam = image_field->Field ("DECvsBeam") ) != NULL )
    {
	if (DECvsBeam->Type () != TYPE_DOUBLE)
	{
	    fprintf (stderr, "%s: DECvsBeam field is not type double\n",
		     function_name);
	    return (TRUE);
	}
	if (DECvsBeam->Length () != ylen)
	{
	    fprintf (stderr, "%s: DECvsBeam field is not length: %lu\n",
		     function_name, ylen);
	    return (TRUE);
	}
    }
    /*  Get each header value  */
    for (pair = header_pairs; pair->name; ++pair)
    {
	if ( ( *pair->value = header_field->Field (pair->name) ) == NULL )
	{
	    if (!pair->required) continue;
	    fprintf (stderr, "%s: no \"%s\" field in header\n",
		     function_name, pair->name);
	    return (TRUE);
	}
    }
    if (header.referenceValues->Type () != TYPE_DOUBLE)
    {
	fprintf (stderr, "%s: referenceValues is not double type\n",
		 function_name);
	return (TRUE);
    }
    switch ( data->Type () )
    {
      case TYPE_ERROR:
	fprintf (stderr, "%s: TYPE_ERROR not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_REF:
	fprintf (stderr, "%s: TYPE_REF not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_SUBVEC_REF:
	fprintf (stderr, "%s: TYPE_SUBVEC_REF not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_BOOL:
	fprintf (stderr, "%s: TYPE_BOOL not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_BYTE:
	fprintf (stderr, "%s: TYPE_BYTE not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_SHORT:
	fprintf (stderr, "%s: TYPE_SHORT not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_INT:
	ktype = K_INT;
	data_ptr = data->IntPtr ();
	break;
      case TYPE_FLOAT:
	ktype = K_FLOAT;
	data_ptr = data->FloatPtr ();
	break;
      case TYPE_DOUBLE:
	ktype = K_DOUBLE;
	data_ptr = data->DoublePtr ();
	break;
      case TYPE_STRING:
	fprintf (stderr, "%s: TYPE_STRING not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_AGENT:
	fprintf (stderr, "%s: TYPE_AGENT not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_FUNC:
	fprintf (stderr, "%s: TYPE_FUNC not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_RECORD:
	fprintf (stderr, "%s: TYPE_RECORD not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_COMPLEX:
	ktype = K_COMPLEX;
	data_ptr = data->ComplexPtr ();
	break;
      case TYPE_DCOMPLEX:
	ktype = K_DCOMPLEX;
	data_ptr = data->DcomplexPtr ();
	break;
      case TYPE_FAIL:
	fprintf (stderr, "%s: TYPE_FAIL not supported\n", function_name);
	return (TRUE);
	/*break;*/
      default:
	fprintf ( stderr, "%s: unsupported data type: %u\n",
		  function_name, data->Type () );
	return (TRUE);
	/*break;*/
    }
    if ( (ktype == K_COMPLEX) || (ktype == K_DCOMPLEX) )
    {
	fprintf (stderr, "%s: complex type not supported\n", function_name);
	return (TRUE);
    }
    if (iarray_type (scrollbuf->cube) != ktype)
    {
	fprintf (stderr, "%s: image data has changed type\n", function_name);
	return (TRUE);
    }
    /*  Determine index of plane  */
    zval = header.referenceValues->DoubleVal (3);
    if (verbose)
    {
	wcs_astro_format_time (txt, zval);
	fprintf (stderr, "Raw timestamp: %.2f   %s\n", zval, txt);
    }
    iarray_get_double (scrollbuf->cube, "CRVAL3", &crval3, TRUE);
    iarray_get_double (scrollbuf->cube, "CRPIX3", &crpix3, TRUE);
    iarray_get_double (scrollbuf->cube, "CDELT3", &cdelt3, TRUE);
    zlen = iarray_dim_length (scrollbuf->cube, 0);
    zpos = (zval - crval3) / cdelt3 + crpix3 - 1.0;
    blank_val = header.blankVal->DoubleVal ();
    if (zpos < -1e-3)
    {
	fprintf (stderr, "%s: illegal index: %g, zval: %g\n",
		 function_name, zpos, zval);
	fprintf (stderr, "  ref: %g  pix: %g  delt: %g\n",
		 crval3, crpix3, cdelt3);
	return (TRUE);
    }
    if (zpos < 0.0) index = 0;
    else index = (unsigned long) zpos;
    if (index < zlen)
    {
	if (verbose) fprintf (stderr, "%s: adding plane at index: %lu\t",
			      function_name, index);
	copy_plane (scrollbuf->cube, index, data_ptr, blank_val);
	/*  Include RA and DEC data if available  */
	if (RAvsBeam)
	    copy_line_of_doubles (scrollbuf->ra_arr, RAvsBeam->DoublePtr (),
				  index, 1.0 / PION180);
	if (DECvsBeam)
	    copy_line_of_doubles (scrollbuf->dec_arr, DECvsBeam->DoublePtr (),
				  index, 1.0 / PION180);
	if (verbose) fprintf (stderr, "done\n");
	iarray_register_data_change (scrollbuf->cube, NULL);
	return (TRUE);
    }
    /*  Have to scroll buffer  */
    ap = iarray_get_associated_wcs_astro (scrollbuf->cube);
    num_to_scroll = index + SCROLL_INCREMENT - zlen;
    if (num_to_scroll > zlen - 1)
    {
	/*  Can simply clear buffer and start from the beginning  */
	if (verbose) fprintf (stderr, "%s: filling...\t", function_name);
	iarray_fill_double (scrollbuf->cube, TOOBIG);
	iarray_put_double (scrollbuf->cube, "CRVAL3", zval);
	iarray_put_double (scrollbuf->cube, "CRPIX3", 1.0);
	wcs_astro_update_location (ap, 3, scrollbuf->cube->top_pack_desc,
				   *scrollbuf->cube->top_packet);
	if (verbose) fprintf (stderr, "adding first plane\t");
	copy_plane (scrollbuf->cube, 0, data_ptr, blank_val);
	iarray_fill_double (scrollbuf->ra_arr, TOOBIG);
	iarray_fill_double (scrollbuf->dec_arr, TOOBIG);
	/*  Include RA and DEC data if available  */
	if (RAvsBeam)
	    copy_line_of_doubles (scrollbuf->ra_arr, RAvsBeam->DoublePtr (), 0,
				  1.0 / PION180);
	if (DECvsBeam)
	    copy_line_of_doubles (scrollbuf->dec_arr, DECvsBeam->DoublePtr (),
				  0, 1.0 / PION180);
	if (verbose) fprintf (stderr, "done\n");
	iarray_register_data_change (scrollbuf->cube, NULL);
	return (TRUE);
    }
    if (verbose) fprintf (stderr, "%s: scrolling by: %lu...\t",
			  function_name, num_to_scroll);
    iarray_scroll (scrollbuf->cube, num_to_scroll, SCROLL_FILL_WITH_BLANKS);
    if (verbose) fprintf (stderr, "adding plane at: %lu\t",
			  index - num_to_scroll);
    copy_plane (scrollbuf->cube, index - num_to_scroll, data_ptr, blank_val);
    iarray_scroll (scrollbuf->ra_arr, num_to_scroll, SCROLL_FILL_WITH_BLANKS);
    iarray_scroll (scrollbuf->dec_arr, num_to_scroll, SCROLL_FILL_WITH_BLANKS);
    /*  Include RA and DEC data if available  */
    if (RAvsBeam)
	copy_line_of_doubles (scrollbuf->ra_arr, RAvsBeam->DoublePtr (),
			      index - num_to_scroll, 1.0 / PION180);
    if (DECvsBeam)
	copy_line_of_doubles (scrollbuf->dec_arr, DECvsBeam->DoublePtr (),
			      index - num_to_scroll, 1.0 / PION180);
    crpix3 -= (double) num_to_scroll;
    iarray_put_double (scrollbuf->cube, "CRPIX3", crpix3);
    wcs_astro_update_location (ap, 3, scrollbuf->cube->top_pack_desc,
			       *scrollbuf->cube->top_packet);
    if (verbose) fprintf (stderr, "done\n");
    iarray_register_data_change (scrollbuf->cube, NULL);
    return (TRUE);
}   /*  End Function scroll_add_cbk  */

static flag scroll_destroy_cbk (CONST void *client_ptr, void *object,
				CONST char *name, CONST void *value_ptr)
/*  [SUMMARY] This is a Glish event handler.
    <client_ptr> A pointer to the Client object.
    <object> An aribtrary object pointer.
    <name> The name of the Glish event.
    <value_ptr> A pointer to the Glish Value.
    [RETURNS] TRUE if the event was consumed, else FALSE.
*/
{
    int ident;
    Value *value = (Value *) value_ptr;
    struct scrolling_buffer_struct *scrollbuf;
    char txt[STRING_LENGTH];
    static char function_name[] = "scroll_destroy_cbk";

    if (verbose) fprintf (stderr, "%s: event: \"%s\"\n", function_name, name);
    if (value->Type () != TYPE_INT)
    {
	fprintf (stderr, "%s: Glish event: \"%s\" is not an int\n",
		 function_name, name);
	return (TRUE);
    }
    if (verbose) fprintf ( stderr, "%s: image ID: %d\n",
			   function_name, value->IntVal () );

    /* Identify the scrollbuf. */
    ident = value->IntVal();
    for (scrollbuf = last_scrollbuf; scrollbuf; scrollbuf = scrollbuf->prev) {
        if (scrollbuf->ident == ident) break;
    }
    VERIFY_SCROLLBUF (scrollbuf);

    KLIST_REMOVE2 (first_scrollbuf, last_scrollbuf, scrollbuf);
    dispdata_destroy_dataobject (scrollbuf->dataobject);
    iarray_dealloc2 (scrollbuf->cube);
    iarray_dealloc2 (scrollbuf->ra_arr);
    iarray_dealloc2 (scrollbuf->dec_arr);
    scrollbuf->magic_number = 0;
    m_free (scrollbuf);
    return (TRUE);
}   /*  End Function scroll_destroy_cbk  */

static void copy_plane (iarray array, unsigned long plane_index,
			CONST void *in, double blank)
/*  [SUMMARY] Copy a plane of data to an Intelligent Array.
    <array> The Intelligent Array to write to.
    <plain_index> The index of the plane to write.
    <in> A pointer to the input data.
    <blank> The blank value in the input.
    [RETURNS] Nothing.
*/
{
    signed int blank_i, val_i;
    unsigned long xlen, ylen, x, y;
    float blank_f, val_f;
    double val_d;
    CONST signed int *ptr_i;
    CONST float *ptr_f;
    CONST double *ptr_d;
    static char function_name[] = "copy_plane";

    xlen = iarray_dim_length (array, 2);
    ylen = iarray_dim_length (array, 1);
    switch ( iarray_type (array) )
    {
      case K_INT:
	blank_i = (signed int) blank;
	ptr_i = (CONST signed int *) in;
	for (y = 0; y < ylen; ++y) for (x = 0; x < xlen; ++x, ++ptr_i)
	{
	    if ( (val_i = *ptr_i) == blank_i )
	    {
		I3 (array, plane_index, y, x) = -2147483647 - 1;
	    }
	    else I3 (array, plane_index, y, x) = val_i;
	}
	break;
      case K_FLOAT:
	blank_f = blank;
	ptr_f = (CONST float *) in;
	for (y = 0; y < ylen; ++y) for (x = 0; x < xlen; ++x, ++ptr_f)
	{
	    if ( (val_f = *ptr_f) == blank_f )
	    {
		F3 (array, plane_index, y, x) = TOOBIG;
	    }
	    else F3 (array, plane_index, y, x) = val_f;
	}
	break;
      case K_DOUBLE:
	ptr_d = (CONST double *) in;
	for (y = 0; y < ylen; ++y) for (x = 0; x < xlen; ++x, ++ptr_d)
	{
	    if ( (val_d = *ptr_d) == blank )
	    {
		D3 (array, plane_index, y, x) = TOOBIG;
	    }
	    else D3 (array, plane_index, y, x) = val_d;
	}
	break;
      default:
	fprintf ( stderr, "Illegal data type: %u\n", iarray_type (array) );
	a_prog_bug (function_name);
	break;
    }
}   /*  End Function copy_plane  */

static void copy_line_of_doubles (iarray array, CONST double *data, uaddr ypos,
				  double scale)
/*  [SUMMARY] Copy a line of doubles to a 2D array.
    <array> The 2D array to write to.
    <data> The line of data to read.
    <ypos> The position of the line in the array.
    <scale> The scale value to multiply the input data by.
    [RETURNS] Nothing.
*/
{
    uaddr xpos, xlen;

    xlen = iarray_dim_length (array, 1);
    for (xpos = 0; xpos < xlen; ++xpos)
	D2 (array, ypos, xpos) = data[xpos] * scale;
}   /*  End Function copy_line_of_doubles  */

static void new_dataobject_func (void *info, KDisplayDataObject data)
{
    struct scrolling_buffer_struct *scrollbuf;
    static char function_name[] = "new_dataobject_func";

    scrollbuf = (struct scrolling_buffer_struct *) info;
    VERIFY_SCROLLBUF2 (scrollbuf);
    scrollbuf->dataobject = data;
}   /*  End Function new_dataobject_func  */

static flag canvas_position_func (ViewableImage vimage, double x, double y,
				  void *value, unsigned int event_code,
				  void *e_info, void **f_info,
				  double x_lin, double y_lin,
				  unsigned int value_type,
				  double x_im, double y_im,
				  int x_pix, int y_pix)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas
    which has a number of ViewableImage objects associated with it. Most
    co-ordinate values are quantised to the nearest image pixel centre.
    <viewimg> The active viewable image.
    <x> The quantised horizontal world co-ordinate of the event.
    <y> The quantised vertical world co-ordinate of the event.
    <value> A pointer to the data value in the viewable image corresponding
    to the event co-ordinates.
    <event_code> The arbitrary event code.
    <e_info> The arbitrary event information.
    <f_info> The arbitrary function information pointer.
    <x_lin> The quantised linear horizontal world co-ordinate (the co-ordinate
    prior to the transform function being called).
    <y_lin> The quantised linear vertical world co-ordinate (the co-ordinate
    prior to the transform function being called).
    <value_type> The type of the data value. This may be K_DCOMPLEX or
    K_UB_RGB.
    <x_im> The quantised horizontal image pixel co-ordinate. 0.0 is the pixel
    centre.
    <y_im> The quantised vertical image pixel co-ordinate. 0.0 is the pixel
    centre.
    <x_pix> The horizontal screen pixel co-ordinate.
    <y_pix> The vertical screen pixel co-ordinate.
    [NOTE] All co-ordinates are clipped to pixel boundaries.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KDisplayDataWindow window;
    KDisplayDataObject dataobject;
    int cube_y, cube_z;
    double x3, y3, z3, ra, dec;
    Widget image_display = (Widget) *f_info;
    Widget label;
    struct scrolling_buffer_struct *scrollbuf;
    char string[STRING_LENGTH], ra_str[STRING_LENGTH], dec_str[STRING_LENGTH];

    label = XtNameToWidget (image_display, "trackLabel3");
    window = XkwImageDisplayGetWindow (image_display);
    /*  Find current image dataobject  */
    dispdata_get_window_attributes (window,
				    KDISPLAYDATA_WINDOW_ATT_IMAGE, &dataobject,
				    KDISPLAYDATA_WINDOW_ATT_END);
    /*  Find matching scroll buffer (if any)  */
    for (scrollbuf = first_scrollbuf; scrollbuf; scrollbuf = scrollbuf->next)
    {
	if (scrollbuf->dataobject == dataobject) break;
    }
    if (!scrollbuf)
    {
	XtVaSetValues (label, XtNlabel, "", NULL);
	return (FALSE);
    }
    dispdata_array_convert_3coord (dataobject, window, &x3, &y3, &z3, x, y);
    cube_y = (int) y3;
    cube_z = (int) z3;
    if ( ( cube_y >= iarray_dim_length (scrollbuf->ra_arr, 1) )
	 || ( cube_z >= iarray_dim_length (scrollbuf->ra_arr, 0) ) )
	return (FALSE);
    ra = D2 (scrollbuf->ra_arr, cube_z, cube_y);
    dec = D2 (scrollbuf->dec_arr, cube_z, cube_y);
    if ( DOUBLE_IS_BLANK (ra) || DOUBLE_IS_BLANK (dec) ) string[0] = '\0';
    else
    {
	wcs_astro_format_hms (ra_str, ra / 15.0);
	wcs_astro_format_dms (dec_str, dec);
	sprintf (string, "Ra %s  Dec %s", ra_str, dec_str);
    }
    XtVaSetValues (label, XtNlabel, string, NULL);
    return (FALSE);
}   /*  End Function canvas_position_func  */
