/*  generic.c

    Generic file for  kview  (X11 image/movie display tool for Karma).

    Copyright (C) 1995-2003  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable on-screen imaging of a single arrayfile of
    the general data structure format. The arrayfile must contain a 2 or 3
    dimensional "Intelligent Array".
    This module runs on an X11 server.


    Written by      Richard Gooch   30-AUG-1995: Copied from  kmovie_main.c
  and  xraycast_3d/load_disp.c

    Updated by      Richard Gooch   11-JUN-1996: Added magnifier support.

    Updated by      Richard Gooch   12-JUN-1996: Fiddle minimum and maximum if
  equal.

    Updated by      Richard Gooch   15-JUN-1996: Created <coord_transform_func>

    Updated by      Richard Gooch   19-JUN-1996: Fixed bug in <display_file>
  when displaying TrueColour images with only PseudoColour canvas available for
  magnifier.

    Updated by      Richard Gooch   27-JUN-1996: Removed <coord_transform_func>

    Updated by      Richard Gooch   22-JUL-1996: Made use of
  <viewimg_create_sequence_from_iarray> routine.

    Updated by      Richard Gooch   17-SEP-1996: Modified <display_file> to
  return PseudoColour array minimum and maximum.

    Updated by      Richard Gooch   27-OCT-1996: Added hostname and port number
  to title.

    Updated by      Richard Gooch   1-JAN-1997: Added <<reset_intensity>>
  parameter to <display_data> function.

    Updated by      Richard Gooch   27-JAN-1997: Changed to Dataset structure.

    Updated by      Richard Gooch   7-AUG-1997: Show min. & max.

    Updated by      Richard Gooch   29-OCT-1997: Added hack to fiddle minimum
  if min == max.

    Updated by      Richard Gooch   20-FEB-1998: Added trap for min or max
  outside {-TOOBIG,TOOBIG}.

    Updated by      Richard Gooch   26-FEB-1998: Added Karma library version
  to title.

    Updated by      Richard Gooch   16-MAY-1998: Use stated range if available
  (i.e. "DATAMIN" and "DATAMAX") in preference to computing range.

    Updated by      Richard Gooch   28-JUL-1998: Ignore stated range if bogus
  (i.e. min >= max) in <process_set>.

    Updated by      Richard Gooch   5-SEP-1998: Removed comms. setup code.

    Updated by      Richard Gooch   14-OCT-1998: Handle negative data scaling
  in <process_set>.

    Updated by      Richard Gooch   5-JUL-1999: Make use of <<quiet>> global.

    Updated by      Richard Gooch   2-AUG-1999: Tolerate NULL <<multi_desc>>
  in function <process_set>.

    Updated by      Richard Gooch   26-APR-2000: Trap data with computed min >=
  max.

    Last updated by Richard Gooch   9-FEB-2003: Improved diagnostic message.


*/
#include <stdio.h>
#include <math.h>
#include <errno.h>
#include <X11/Xlib.h>
#include <karma.h>
#include <karma_viewimg.h>
#include <karma_iarray.h>
#include <karma_conn.h>
#include <karma_ds.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_r.h>
#include "kview.h"


/*  Private functions  */
STATIC_FUNCTION (void dealloc_set, (Dataset *set) );


/*  Private data  */


/*  Public functions follow  */

flag process_set (multi_array *multi_desc, Dataset *set)
/*  [PURPOSE] This routine will display data.
    <multi_desc> The data to display. If this is NULL, the old dataset is
    deallocated and no further action is taken.
    <set> The dataset to update.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int cmap_index;
    extern flag quiet;
    static flag first_time = TRUE;
    /*static char function_name[] = "process_set";*/

    if (first_time)
    {
	first_time = FALSE;
	conn_attempt_connection ("unix", r_get_def_port ("khogd", ":0.0"),
				 "hog_request");
    }
    dealloc_set (set);
    if (multi_desc == NULL) return (TRUE);
    /*  Try to get 2-D image  */
    if ( !iarray_get_image_from_multi (multi_desc, &set->array_pseudo,
				       &set->array_red, &set->array_green,
				       &set->array_blue, &cmap_index) )
    {
	if ( !iarray_get_movie_from_multi (multi_desc, &set->array_pseudo,
					   &set->array_red, &set->array_green,
					   &set->array_blue, &cmap_index) )
	{
	    fprintf (stderr, "Error getting image/movie\n");
	    dealloc_set (set);
	    return (FALSE);
	}
    }
    if (set->array_pseudo)
    {
	flag compute_range = TRUE;
	flag scaled;
	double scale, offset, tmp;

	scaled = iarray_get_data_scaling (set->array_pseudo, &scale, &offset);
	if ( iarray_get_data_range (set->array_pseudo, &set->min, &set->max) )
	{
	    /*  Range was provided  */
	    if (!quiet) fprintf (stderr, "Stated scaled min: %g  max: %g\n",
				 set->min, set->max);
	    if (set->min >= set->max)
	    {
		fprintf (stderr,"Min is not less than max: ignoring header\n");
	    }
	    else
	    {
		compute_range = FALSE;
		if (scaled)
		{
		    set->min = (set->min - offset) / scale;
		    set->max = (set->max - offset) / scale;
		    if (scale < 0.0) swap (set->min, set->max, tmp);
		    if (!quiet)
			fprintf (stderr, "Stated unscaled min: %g  max: %g\n",
				 set->min, set->max);
		}
	    }
	}
	if (compute_range)
	{
	    if ( !iarray_min_max (set->array_pseudo, CONV1_REAL,
				  &set->min, &set->max) )
	    {
		fprintf (stderr, "Error computing min-max\n");
		dealloc_set (set);
		return (FALSE);
	    }
	    if (!quiet)
	    {
		if (scaled) fprintf (stderr,
				     "Computed scaled min: %g  max: %g\n",
				     set->min * scale + offset,
				     set->max * scale + offset);
		fprintf (stderr, "Computed unscaled min: %g  max: %g\n",
			 set->min, set->max);
	    }
	    iarray_set_data_range (set->array_pseudo,
				   set->min * scale + offset,
				   set->max * scale + offset);
	}
	if (set->min <= -TOOBIG)
	{
	    fprintf (stderr, "min: %e  is not greater than %e\n",
		     set->min, -TOOBIG);
	    dealloc_set (set);
	    return (FALSE);
	}
	if ( DOUBLE_IS_BLANK (set->max) )
	{
	    fprintf (stderr, "max: %e  is not less than %e\n",
		     set->max, TOOBIG);
	    dealloc_set (set);
	    return (FALSE);
	}
	if (set->min >= set->max)
	{
	    fprintf (stderr, "Min is not less than max: bad data\n");
	    dealloc_set (set);
	    return (FALSE);
	}
	if (set->min == set->max)
	{
	    double step = 1.0;

	    fprintf (stderr,
		     "WARNING: min: %e is same as max: may cause problems\n",
		     set->min);
	    while (set->min == set->max)
	    {
		set->min -= step;
		step *= 10.0;
	    }
	}
	return (TRUE);
    }
    return (TRUE);
}   /*  End Function process_set  */


/*  Private functions follow  */

static void dealloc_set (Dataset *set)
/*  [SUMMARY] Deallocate a dataset.
    <set> The dataset.
    [RETURNS] Nothing.
*/
{
    if (set->array_pseudo) iarray_dealloc (set->array_pseudo);
    set->array_pseudo = NULL;
    if (set->array_red) iarray_dealloc (set->array_red);
    set->array_red = NULL;
    if (set->array_green) iarray_dealloc (set->array_green);
    set->array_green = NULL;
    if (set->array_blue) iarray_dealloc (set->array_blue);
    set->array_blue = NULL;
}   /*  End Function dealloc_set  */
