/*  glish.cc

    Glish file for  MultibeamView  (X11 image/movie display tool for Karma).

    Copyright (C) 1997-2004  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable on-screen imaging.
    This module runs on an X11 server.


    Written by      Richard Gooch   9-JAN-1997

    Updated by      Richard Gooch   30-JAN-1997: Fixed bugs in scrolling.

    Updated by      Richard Gooch   2-FEB-1997: Added "filename" association to
  scrolling buffer.

    Updated by      Richard Gooch   5-FEB-1997: Post "imageID" event once
  scrolling buffer is set up.

    Updated by      Richard Gooch   4-MAR-1997: Update KwcsAstro object when
  scrolling.

    Updated by      Richard Gooch   9-MAR-1997: Made use of verbose global.

    Updated by      Richard Gooch   18-APR-1997: Added scrolling RA&DEC arrays.

    Updated by      Richard Gooch   23-APR-1997: Converted RA&DEC data from
  radians to degrees.

    Updated by      Richard Gooch   25-NOV-1997: Removed trap for opaque data
  type since Glish 2.6 no longer defines it.

    Updated by      Richard Gooch   22-SEP-1998: Removed unnecessary casts.

    Updated by      Richard Gooch   19-OCT-1998: Improved debugging in
  <scroll_add_cbk>.

    Updated by      Richard Gooch   28-OCT-1998: Display raw timestamp in
  <scroll_add_cbk> when in verbose mode.

    Updated by      Richard Gooch   10-MAY-2000: Attempted to make ecgs happy:
  failed.

    Updated by      Richard Gooch   28-JUN-2001: Removed duplicate #include.

    Updated by      Richard Gooch   6-SEP-2001: Made use of experimental
  function <iarray_set_value_name>.

    Updated by      Richard Gooch   15-OCT-2002: Shut up gcc 3.1.1.

    Last updated by Richard Gooch   14-APR-2004: Switched to new experimental
  function <iarray_scroll>.

    $Id: glish.cc,v 2.1 2011/04/21 03:00:56 cal103 Exp $

*/
#include <stdio.h>
#include <Glish/Value.h>
#include <Glish/Client.h>
/*#include <iostream.h>*/  /*  Disabled because gcc 3.1.1 complains, and is
                               apparently not needed anyway  */
#include <karma.h>
#include <karma_iarray.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_ex.h>
#include <karma_m.h>
#include <karma_a.h>
#include "kview.h"
#include "MultibeamView.h"

#define SCROLL_INCREMENT 1


/*  Private structure types  */
struct header_pair_type
{
    const char *name;
    Value **value;
};

struct header_type
{
    Value *imageShape;
    Value *axisNames;
    Value *referenceValues;
    Value *referencePixels;
    Value *deltas;
    Value *name;
    Value *dataUnits;
    Value *blankVal;
};

struct image_type
{
    Value *data;
    struct header_type header;
    Value *RAvsBeam;
    Value *DECvsBeam;
};


/*  Private functions  */


/*  Private data  */
static struct image_type image;
static struct header_pair_type header_pairs[] =
{
    {"imageShape", &image.header.imageShape},
    {"axisNames", &image.header.axisNames},
    {"referenceValues", &image.header.referenceValues},
    {"referencePixels", &image.header.referencePixels},
    {"deltas", &image.header.deltas},
    {"name", &image.header.name},
    {"dataUnits", &image.header.dataUnits},
    {"blankVal", &image.header.blankVal},
    {NULL, NULL}
};


/*  Public functions follow  */

flag scroll_init_cbk (CONST void *client_ptr, void *object, CONST char *name,
		      CONST void *value_ptr)
/*  [SUMMARY] This is a Glish event handler.
    <client_ptr> A pointer to the Client object.
    <object> An aribtrary object pointer.
    <name> The name of the Glish event.
    <value_ptr> A pointer to the Glish Value.
    [RETURNS] TRUE if the event was consumed, else FALSE.
*/
{
    unsigned int  axis;
    unsigned long xlen, ylen, zlen, ktype, dim_index;
    double blank_val;
    struct header_pair_type *pair;
    void *data_ptr;
    char *axis_names, *ptr, *axis_name;
    Client *client = (Client *) client_ptr;
    Value *value = (Value *) value_ptr;
    Value *image_field, *count_field, *header_field;
    char txt[STRING_LENGTH];
    extern flag verbose;
    extern iarray ra_arr, dec_arr;
    extern Dataset set1;
    static char function_name[] = "scroll_init_cbk";

    if (verbose) fprintf (stderr, "event: \"%s\"\t", name);
    /*  Destroy old arrays if they exist  */
    if (set1.array_pseudo) iarray_dealloc (set1.array_pseudo);
    m_clear (&set1, sizeof set1);
    if (ra_arr) iarray_dealloc (ra_arr);
    ra_arr = NULL;
    if (dec_arr) iarray_dealloc (dec_arr);
    dec_arr = NULL;
    if (value->Type () != TYPE_RECORD)
    {
	fprintf (stderr, "%s: Glish event: \"%s\" is not a record\n",
		 function_name, name);
	return (TRUE);
    }
    image_field = value->NthField (1);
    if (image_field->Type () != TYPE_RECORD)
    {
	fprintf (stderr, "%s: first field is not a record\n", function_name);
	return (TRUE);
    }
    if ( ( count_field = value->NthField (2) ) == NULL )
    {
	fprintf (stderr, "%s: no second field\n", function_name);
	return (TRUE);
    }
    if (count_field->Type () != TYPE_INT)
    {
	fprintf (stderr, "%s: second field is not an int\n", function_name);
	return (TRUE);
    }
    zlen = count_field->IntVal ();
    if ( ( image.data = image_field->Field ("data") ) == NULL )
    {
	fprintf (stderr, "%s: no data field in image\n", function_name);
	return (TRUE);
    }
    if ( ( header_field = image_field->Field ("header") ) == NULL )
    {
	fprintf (stderr, "%s: no header field in image\n", function_name);
	return (TRUE);
    }
    if (header_field->Type () != TYPE_RECORD)
    {
	fprintf (stderr, "%s: header is not a record\n", function_name);
	return (TRUE);
    }
    /*  Get each header value  */
    for (pair = header_pairs; pair->name; ++pair)
    {
	if ( ( *pair->value = header_field->Field (pair->name) ) == NULL )
	{
	    fprintf (stderr, "%s: no \"%s\" field in header\n",
		     function_name, pair->name);
	    return (TRUE);
	}
    }
    if (image.header.referenceValues->Type () != TYPE_DOUBLE)
    {
	fprintf (stderr, "%s: referenceValues is not double type\n",
		 function_name);
	return (TRUE);
    }
    if (image.header.referencePixels->Type () != TYPE_DOUBLE)
    {
	fprintf (stderr, "%s: referencePixels is not double type\n",
		 function_name);
	return (TRUE);
    }
    if (image.header.deltas->Type () != TYPE_DOUBLE)
    {
	fprintf (stderr, "%s: deltas is not double type\n", function_name);
	return (TRUE);
    }
    if (image.header.dataUnits->Type () != TYPE_STRING)
    {
	fprintf (stderr, "%s: dataUnits is not string type\n", function_name);
	return (TRUE);
    }
    switch ( image.data->Type () )
    {
      case TYPE_ERROR:
	fprintf (stderr, "%s: TYPE_ERROR not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_REF:
	fprintf (stderr, "%s: TYPE_REF not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_SUBVEC_REF:
	fprintf (stderr, "%s: TYPE_SUBVEC_REF not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_BOOL:
	fprintf (stderr, "%s: TYPE_BOOL not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_BYTE:
	fprintf (stderr, "%s: TYPE_BYTE not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_SHORT:
	fprintf (stderr, "%s: TYPE_SHORT not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_INT:
	ktype = K_INT;
	data_ptr = image.data->IntPtr ();
	break;
      case TYPE_FLOAT:
	ktype = K_FLOAT;
	data_ptr = image.data->FloatPtr ();
	break;
      case TYPE_DOUBLE:
	ktype = K_DOUBLE;
	data_ptr = image.data->DoublePtr ();
	break;
      case TYPE_STRING:
	fprintf (stderr, "%s: TYPE_STRING not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_AGENT:
	fprintf (stderr, "%s: TYPE_AGENT not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_FUNC:
	fprintf (stderr, "%s: TYPE_FUNC not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_RECORD:
	fprintf (stderr, "%s: TYPE_RECORD not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_COMPLEX:
	ktype = K_COMPLEX;
	data_ptr = image.data->ComplexPtr ();
	break;
      case TYPE_DCOMPLEX:
	ktype = K_DCOMPLEX;
	data_ptr = image.data->DcomplexPtr ();
	break;
      case TYPE_FAIL:
	fprintf (stderr, "%s: TYPE_FAIL not supported\n", function_name);
	return (TRUE);
	/*break;*/
      default:
	fprintf ( stderr, "%s: unsupported data type: %u\n",
		  function_name, image.data->Type () );
	return (TRUE);
	/*break;*/
    }
    if ( (ktype == K_COMPLEX) || (ktype == K_DCOMPLEX) )
    {
	fprintf (stderr, "%s: complex type not supported\n", function_name);
	return (TRUE);
    }
    /*  Get rest of 3D buffer shape  */
    xlen = image.header.imageShape->IntVal (1);
    ylen = image.header.imageShape->IntVal (2);
    if (verbose) fprintf (stderr, "xlen: %lu  ylen: %lu  zlen: %lu\n",
			  xlen, ylen, zlen);
    /*  Check sanity of RA and DEC fields  */
    if ( ( image.RAvsBeam = image_field->Field ("RAvsBeam") ) != NULL )
    {
	if (image.RAvsBeam->Type () != TYPE_DOUBLE)
	{
	    fprintf (stderr, "%s: RAvsBeam field is not type double\n",
		     function_name);
	    return (TRUE);
	}
	if (image.RAvsBeam->Length () != ylen)
	{
	    fprintf (stderr, "%s: RAvsBeam field is not length: %lu\n",
		     function_name, ylen);
	    return (TRUE);
	}
    }
    if ( ( image.DECvsBeam = image_field->Field ("DECvsBeam") ) != NULL )
    {
	if (image.DECvsBeam->Type () != TYPE_DOUBLE)
	{
	    fprintf (stderr, "%s: DECvsBeam field is not type double\n",
		     function_name);
	    return (TRUE);
	}
	if (image.DECvsBeam->Length () != ylen)
	{
	    fprintf (stderr, "%s: DECvsBeam field is not length: %lu\n",
		     function_name, ylen);
	    return (TRUE);
	}
    }
    /*  Construct a 3D Intelligent Array for the scrolling data buffer  */
    if ( ( set1.array_pseudo = iarray_create_3D (zlen, ylen, xlen, ktype) )
	 == NULL )
    {
	m_error_notify (function_name, "scroll array");
	return (TRUE);
    }
    iarray_fill_double (set1.array_pseudo, TOOBIG);
    iarray_set_value_name (set1.array_pseudo,
			   image.header.dataUnits->StringVal (), TRUE);
    iarray_put_named_string ( set1.array_pseudo, "OBJECT",
			      image.header.name->StringVal (' ') );
    /*  Create 2D arrays for the scrolling RA and DEC buffers  */
    if ( ( ra_arr = iarray_create_2D (zlen, ylen, K_DOUBLE) ) == NULL )
	m_abort (function_name, "RA array");
    iarray_fill_double (ra_arr, TOOBIG);
    if ( ( dec_arr = iarray_create_2D (zlen, ylen, K_DOUBLE) ) == NULL )
	m_abort (function_name, "DEC array");
    iarray_fill_double (dec_arr, TOOBIG);
    axis_names = image.header.axisNames->StringVal (' ');
    ptr = axis_names;
    for (axis = 1; axis < 4; ++axis)
    {
	dim_index = 3 - axis;
	if (ptr == NULL)
	{
	    fprintf (stderr, "No more axis names!\n");
	    return (TRUE);
	}
	axis_name = ex_str (ptr, (CONST char **) &ptr);
	iarray_set_dim_name (set1.array_pseudo, dim_index, axis_name, TRUE);
	sprintf (txt, "CTYPE%u", axis);
	if ( !iarray_put_named_string (set1.array_pseudo, txt, axis_name) )
	    m_abort (function_name, "CTYPEn string value");
	m_free (axis_name);
	sprintf (txt, "CRVAL%u", axis);
	if ( !iarray_put_double ( set1.array_pseudo, txt,
				  image.header.referenceValues->DoubleVal (axis) ) )
	    m_abort (function_name, "CRVALn string value");
	sprintf (txt, "CRPIX%u", axis);
	if ( !iarray_put_double ( set1.array_pseudo, txt,
				  image.header.referencePixels->DoubleVal (axis) ) )
	    m_abort (function_name, "CRPIXn string value");
	sprintf (txt, "CDELT%u", axis);
	if ( !iarray_put_double ( set1.array_pseudo, txt,
				  image.header.deltas->DoubleVal (axis) ) )
	    m_abort (function_name, "CDELTn string value");
    }
    /*  Copy first image data  */
    blank_val = image.header.blankVal->DoubleVal ();
    copy_plane (set1.array_pseudo, 0, data_ptr, blank_val);
    iarray_min_max (set1.array_pseudo, CONV_CtoR_REAL, &set1.min, &set1.max);
    if (set1.max <= set1.min)
    {
	fprintf (stderr,
		 "Constant data value: %e, setting range to: %e to %e\n",
		 set1.min, set1.min - 0.5, set1.min + 0.5);
	set1.min -= 0.5;
	set1.max += 0.5;
    }
    else fprintf (stderr, "Initial data range: %e  to  %e\n",
		  set1.min, set1.max);
    /*  Include RA and DEC data if available  */
    if (image.RAvsBeam)
	copy_line_of_doubles (ra_arr, image.RAvsBeam->DoublePtr (), 0,
			      1.0 / PION180);
    if (image.DECvsBeam)
	copy_line_of_doubles (dec_arr, image.DECvsBeam->DoublePtr (), 0,
			      1.0 / PION180);
    ds_add_named_object (set1.array_pseudo->multi_desc,
			 set1.array_pseudo->multi_desc->num_arrays, "filename",
			 st_dup ("GLISH::scrollBuffer"),
			 ( void (*) (...) ) m_free);
    new_scroll_array ();
    /*  Clear temporary structure  */
    m_clear (&image, sizeof image);
    Value reply_value (1);
    client->PostEvent ("imageID", &reply_value);
    return (TRUE);
}   /*  End Function scroll_init_cbk  */

flag scroll_add_cbk (CONST void *client_ptr, void *object, CONST char *name,
		     CONST void *value_ptr)
/*  [SUMMARY] This is a Glish event handler.
    <client_ptr> A pointer to the Client object.
    <object> An aribtrary object pointer.
    <name> The name of the Glish event.
    <value_ptr> A pointer to the Glish Value.
    [RETURNS] TRUE if the event was consumed, else FALSE.
*/
{
    KwcsAstro ap;
    unsigned int ktype;
    unsigned long ylen, zlen, index, num_to_scroll;
    double zval, zpos, crval3, crpix3, cdelt3;
    double blank_val;
    struct header_pair_type *pair;
    void *data_ptr;
    Value *value = (Value *) value_ptr;
    Value *image_field, *id_field, *header_field;
    char txt[STRING_LENGTH];
    extern flag verbose;
    extern iarray ra_arr, dec_arr;
    extern Dataset set1;
    static char function_name[] = "scroll_add_cbk";

    if (verbose) fprintf (stderr, "%s: event: \"%s\"\n", function_name, name);
    if (set1.array_pseudo == NULL)
    {
	fprintf (stderr, "No scrolling buffer set up yet\n");
	return (TRUE);
    }
    if (value->Type () != TYPE_RECORD)
    {
	fprintf (stderr, "%s: Glish event: \"%s\" is not a record\n",
		 function_name, name);
	return (TRUE);
    }
    id_field = value->NthField (1);
    if (id_field->Type () != TYPE_INT)
    {
	fprintf (stderr, "%s: ID field is not an int\n", function_name);
	return (TRUE);
    }
    if ( ( image_field = value->NthField (2) ) == NULL )
    {
	fprintf (stderr, "%s: no second field\n", function_name);
	return (TRUE);
    }
    if (image_field->Type () != TYPE_RECORD)
    {
	fprintf (stderr, "%s: second field is not a record\n", function_name);
	return (TRUE);
    }
    if ( ( image.data = image_field->Field ("data") ) == NULL )
    {
	fprintf (stderr, "%s: no data field in image\n", function_name);
	return (TRUE);
    }
    if ( ( header_field = image_field->Field ("header") ) == NULL )
    {
	fprintf (stderr, "%s: no header field in image\n", function_name);
	return (TRUE);
    }
    if (header_field->Type () != TYPE_RECORD)
    {
	fprintf (stderr, "%s: header is not a record\n", function_name);
	return (TRUE);
    }
    ylen = iarray_dim_length (set1.array_pseudo, 1);
    /*  Check sanity of RA and DEC fields  */
    if ( ( image.RAvsBeam = image_field->Field ("RAvsBeam") ) != NULL )
    {
	if (image.RAvsBeam->Type () != TYPE_DOUBLE)
	{
	    fprintf (stderr, "%s: RAvsBeam field is not type double\n",
		     function_name);
	    return (TRUE);
	}
	if (image.RAvsBeam->Length () != ylen)
	{
	    fprintf (stderr, "%s: RAvsBeam field is not length: %lu\n",
		     function_name, ylen);
	    return (TRUE);
	}
    }
    if ( ( image.DECvsBeam = image_field->Field ("DECvsBeam") ) != NULL )
    {
	if (image.DECvsBeam->Type () != TYPE_DOUBLE)
	{
	    fprintf (stderr, "%s: DECvsBeam field is not type double\n",
		     function_name);
	    return (TRUE);
	}
	if (image.DECvsBeam->Length () != ylen)
	{
	    fprintf (stderr, "%s: DECvsBeam field is not length: %lu\n",
		     function_name, ylen);
	    return (TRUE);
	}
    }
    /*  Get each header value  */
    for (pair = header_pairs; pair->name; ++pair)
    {
	if ( ( *pair->value = header_field->Field (pair->name) ) == NULL )
	{
	    fprintf (stderr, "%s: no \"%s\" field in header\n",
		     function_name, pair->name);
	    return (TRUE);
	}
    }
    if (image.header.referenceValues->Type () != TYPE_DOUBLE)
    {
	fprintf (stderr, "%s: referenceValues is not double type\n",
		 function_name);
	return (TRUE);
    }
    switch ( image.data->Type () )
    {
      case TYPE_ERROR:
	fprintf (stderr, "%s: TYPE_ERROR not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_REF:
	fprintf (stderr, "%s: TYPE_REF not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_SUBVEC_REF:
	fprintf (stderr, "%s: TYPE_SUBVEC_REF not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_BOOL:
	fprintf (stderr, "%s: TYPE_BOOL not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_BYTE:
	fprintf (stderr, "%s: TYPE_BYTE not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_SHORT:
	fprintf (stderr, "%s: TYPE_SHORT not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_INT:
	ktype = K_INT;
	data_ptr = image.data->IntPtr ();
	break;
      case TYPE_FLOAT:
	ktype = K_FLOAT;
	data_ptr = image.data->FloatPtr ();
	break;
      case TYPE_DOUBLE:
	ktype = K_DOUBLE;
	data_ptr = image.data->DoublePtr ();
	break;
      case TYPE_STRING:
	fprintf (stderr, "%s: TYPE_STRING not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_AGENT:
	fprintf (stderr, "%s: TYPE_AGENT not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_FUNC:
	fprintf (stderr, "%s: TYPE_FUNC not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_RECORD:
	fprintf (stderr, "%s: TYPE_RECORD not supported\n", function_name);
	return (TRUE);
	/*break;*/
      case TYPE_COMPLEX:
	ktype = K_COMPLEX;
	data_ptr = image.data->ComplexPtr ();
	break;
      case TYPE_DCOMPLEX:
	ktype = K_DCOMPLEX;
	data_ptr = image.data->DcomplexPtr ();
	break;
      case TYPE_FAIL:
	fprintf (stderr, "%s: TYPE_FAIL not supported\n", function_name);
	return (TRUE);
	/*break;*/
      default:
	fprintf ( stderr, "%s: unsupported data type: %u\n",
		  function_name, image.data->Type () );
	return (TRUE);
	/*break;*/
    }
    if ( (ktype == K_COMPLEX) || (ktype == K_DCOMPLEX) )
    {
	fprintf (stderr, "%s: complex type not supported\n", function_name);
	return (TRUE);
    }
    if (iarray_type (set1.array_pseudo) != ktype)
    {
	fprintf (stderr, "%s: image data has changed type\n", function_name);
	return (TRUE);
    }
    /*  Determine index of plane  */
    zval = image.header.referenceValues->DoubleVal (3);
    if (verbose)
    {
	wcs_astro_format_time (txt, zval);
	fprintf (stderr, "Raw timestamp: %.2f   %s\n", zval, txt);
    }
    iarray_get_double (set1.array_pseudo, "CRVAL3", &crval3, TRUE);
    iarray_get_double (set1.array_pseudo, "CRPIX3", &crpix3, TRUE);
    iarray_get_double (set1.array_pseudo, "CDELT3", &cdelt3, TRUE);
    zlen = iarray_dim_length (set1.array_pseudo, 0);
    zpos = (zval - crval3) / cdelt3 + crpix3 - 1.0;
    blank_val = image.header.blankVal->DoubleVal ();
    if (zpos < -1e-3)
    {
	fprintf (stderr, "%s: illegal index: %g, zval: %g\n",
		 function_name, zpos, zval);
	fprintf (stderr, "  ref: %g  pix: %g  delt: %g\n",
		 crval3, crpix3, cdelt3);
	return (TRUE);
    }
    if (zpos < 0.0) index = 0;
    else index = (unsigned long) zpos;
    if (index < zlen)
    {
	if (verbose) fprintf (stderr, "%s: adding plane at index: %lu\t",
			      function_name, index);
	copy_plane (set1.array_pseudo, index, data_ptr, blank_val);
	/*  Include RA and DEC data if available  */
	if (image.RAvsBeam)
	    copy_line_of_doubles (ra_arr, image.RAvsBeam->DoublePtr (), index,
				  1.0 / PION180);
	if (image.DECvsBeam)
	    copy_line_of_doubles (dec_arr, image.DECvsBeam->DoublePtr (),
				  index, 1.0 / PION180);
	if (verbose) fprintf (stderr, "done\n");
	updated_scroll_array (index);
	return (TRUE);
    }
    /*  Have to scroll buffer  */
    ap = iarray_get_associated_wcs_astro (set1.array_pseudo);
    num_to_scroll = index + SCROLL_INCREMENT - zlen;
    if (num_to_scroll > zlen - 1)
    {
	/*  Can simply clear buffer and start from the beginning  */
	if (verbose) fprintf (stderr, "%s: filling...\t", function_name);
	iarray_fill_double (set1.array_pseudo, TOOBIG);
	iarray_put_double (set1.array_pseudo, "CRVAL3", zval);
	iarray_put_double (set1.array_pseudo, "CRPIX3", 1.0);
	wcs_astro_update_location (ap, 3, set1.array_pseudo->top_pack_desc,
				   *set1.array_pseudo->top_packet);
	if (verbose) fprintf (stderr, "adding first plane\t");
	copy_plane (set1.array_pseudo, 0, data_ptr, blank_val);
	iarray_fill_double (ra_arr, TOOBIG);
	iarray_fill_double (dec_arr, TOOBIG);
	/*  Include RA and DEC data if available  */
	if (image.RAvsBeam)
	    copy_line_of_doubles (ra_arr, image.RAvsBeam->DoublePtr (), 0,
				  1.0 / PION180);
	if (image.DECvsBeam)
	    copy_line_of_doubles (dec_arr, image.DECvsBeam->DoublePtr (), 0,
				  1.0 / PION180);
	if (verbose) fprintf (stderr, "done\n");
	updated_scroll_array (-1);
	return (TRUE);
    }
    if (verbose) fprintf (stderr, "%s: scrolling by: %lu...\t",
			  function_name, num_to_scroll);
    iarray_scroll (set1.array_pseudo, num_to_scroll, SCROLL_FILL_WITH_BLANKS);
    if (verbose) fprintf (stderr, "adding plane at: %lu\t",
			  index - num_to_scroll);
    copy_plane (set1.array_pseudo, index - num_to_scroll, data_ptr, blank_val);
    iarray_scroll (ra_arr, num_to_scroll, SCROLL_FILL_WITH_BLANKS);
    iarray_scroll (dec_arr, num_to_scroll, SCROLL_FILL_WITH_BLANKS);
    /*  Include RA and DEC data if available  */
    if (image.RAvsBeam)
	copy_line_of_doubles (ra_arr, image.RAvsBeam->DoublePtr (),
			      index - num_to_scroll, 1.0 / PION180);
    if (image.DECvsBeam)
	copy_line_of_doubles (dec_arr, image.DECvsBeam->DoublePtr (),
			      index - num_to_scroll, 1.0 / PION180);
    crpix3 -= (double) num_to_scroll;
    iarray_put_double (set1.array_pseudo, "CRPIX3", crpix3);
    wcs_astro_update_location (ap, 3, set1.array_pseudo->top_pack_desc,
			       *set1.array_pseudo->top_packet);
    if (verbose) fprintf (stderr, "done\n");
    updated_scroll_array (-1);
    return (TRUE);
}   /*  End Function scroll_add_cbk  */


/*  Private functions follow  */
