/*  karma2fits.c

    Source file for  karma2fits  (Karma to FITS conversion module).

    Copyright (C) 1996-2001  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will read in a Karma data file and convert the
    n-dimensional array to a FITS file.


    Written by      Richard Gooch   6-AUG-1996

    Updated by      Richard Gooch   2-DEC-1996: Added trap if output file has
  the same name as the input file.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package.

    Updated by      Richard Gooch   12-MAY-2000: Added some history.

    Last updated by Richard Gooch   21-DEC-2001: Included k_version.h directly.


*/
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <errno.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_module.h>
#include <karma_dsxfr.h>
#include <karma_panel.h>
#include <karma_ch.h>
#include <karma_ds.h>
#include <karma_ex.h>
#include <karma_im.h>
#include <karma_m.h>
#include <k_version.h>


STATIC_FUNCTION (flag karma2fits, (CONST char *command, FILE *fp) );

#define VERSION "1.0"

int main (int argc, char **argv)
{
    KControlPanel panel;
    static char function_name[] = "main";

    im_register_lib_version (KARMA_VERSION);
    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
	m_abort (function_name, "control panel");
    }
    panel_push_onto_stack (panel);
    module_run (argc, argv, "karma2fits", VERSION, karma2fits, -1, -1, FALSE);
    return (RV_OK);
}   /*  End Function main   */

static flag karma2fits (CONST char *command, FILE *fp)
{
    Channel out;
    int len;
    multi_array *multi_desc;
    char *inp_filename, *out_filename;
    char txt[STRING_LENGTH];
    extern char module_name[STRING_LENGTH + 1];
    extern char module_version_date[STRING_LENGTH + 1];
    extern char module_lib_version[STRING_LENGTH + 1];
    extern char karma_library_version[STRING_LENGTH + 1];

    if ( ( inp_filename = ex_word (command, &command) ) == NULL )
    {
	fprintf (stderr, "Must supply input and output filenames\n");
	return (TRUE);
    }
    if ( ( out_filename = ex_word (command, &command) ) == NULL )
    {
	fprintf (stderr, "Must also supply output filename\n");
	m_free (inp_filename);
	return (TRUE);
    }
    /*  Append ".kf" if doesn't already exist  */
    strcpy (txt, inp_filename);
    len = strlen (txt);
    if (strcmp (txt + len - 3, ".kf") != 0) strcat (txt, ".kf");
    if (strcmp (txt, out_filename) == 0)
    {
	fprintf (stderr, "Input file would be overwritten!\n");
	m_free (inp_filename);
	m_free (out_filename);
	return (TRUE);
    }
    /*  Read input file  */
    if ( ( multi_desc = dsxfr_get_multi (inp_filename, FALSE,
					 K_CH_MAP_LOCAL, FALSE) )
	== NULL )
    {
	fprintf (stderr, "Error getting arrayfile: \"%s\"\n",
		 inp_filename);
	m_free (inp_filename);
	m_free (out_filename);
	return (TRUE);
    }
    /*  Append a bit of history  */
    sprintf (txt, "%s: Module version %s  Karma v%s  compiled with v%s",
	     module_name, module_version_date,
	     karma_library_version, module_lib_version);
    ds_history_append_string (multi_desc, txt, TRUE);
    sprintf (txt, "%s: infile: %s  outfile: %s",
	     module_name, inp_filename, out_filename);
    ds_history_append_string (multi_desc, txt, TRUE);
    if ( ( out = ch_open_file (out_filename, "w") ) == NULL )
    {
	fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		 out_filename, ERRSTRING);
	m_free (inp_filename);
	m_free (out_filename);
	ds_dealloc_multi (multi_desc);
	return (TRUE);
    }
    if ( !foreign_fits_write (out, multi_desc,
			      FA_FITS_WRITE_END) )
    {
	fprintf (stderr, "Error writing FITS file\n");
	unlink (out_filename);
    }
    ch_close (out);
    m_free (inp_filename);
    m_free (out_filename);
    ds_dealloc_multi (multi_desc);
    return (TRUE);
}   /*  End Function karma2fits  */
