/*  karma2ppm.c

    Source file for  karma2ppm  (Karma to PPM image conversion module).

    Copyright (C) 1995-2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will read in a Karma data file and convert the image into
    a PPM file.


    Written by      Richard Gooch   15-APR-1995

    Updated by      Richard Gooch   30-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   19-JUN-1997: Support PseudoColour images
  and movies. Support colourmap file for Greyscale images and movies.

    Updated by      Richard Gooch   5-MAY-1998: Support blank-filling to
  minimum image size. Support output filter.

    Updated by      Richard Gooch   6-MAY-1998: Support fixed-width sequence
  numbers.

    Updated by      Richard Gooch   7-MAY-1998: Changed minimum image parameter
  names to output parameters to reflect changed behaviour. Count from 1 by
  default.

    Updated by      Richard Gooch   13-MAY-1998: Changed sequence number
  separator from '.' to '_' to accommodate M$ crud.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package.

    Updated by      Richard Gooch   30-NOV-1998: Fixed bug in <open_file>:
  fixed width serial numbers was screwed causing a segmentation fault.

    Updated by      Richard Gooch   15-JUN-1999: Took account of change to
  <foreign_ppm_write_pseudo>.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Updated by      Richard Gooch   21-DEC-2001: Included k_version.h directly.

    Last updated by Richard Gooch   17-APR-2002: Took account of change to
  <foreign_ppm_write_pseudo>.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_iarray.h>
#include <karma_module.h>
#include <karma_dsxfr.h>
#include <karma_panel.h>
#include <karma_ch.h>
#include <karma_ds.h>
#include <karma_ex.h>
#include <karma_im.h>
#include <karma_m.h>
#include <karma_a.h>
#include <k_version.h>


STATIC_FUNCTION (flag karma2ppm, (CONST char *command, FILE *fp) );
STATIC_FUNCTION (multi_array *get_cmap,
		 (multi_array *multi_desc, CONST char *cmap_file,
		  CONST unsigned short **cmap, unsigned int *cmap_size) );
STATIC_FUNCTION (Channel open_file, (CONST char *subname,
				     int image_num, int max_images) );


static flag binary = TRUE;
static char *cmap_file = NULL;
static unsigned int out_width = 0;
static unsigned int out_height = 0;
static char *extension = NULL;
static char *converter = NULL;
static flag fixed_width_sequence = TRUE;
static flag count_from_zero = FALSE;


#define VERSION "1.7"

int main (int argc, char **argv)
{
    KControlPanel panel;
    static char function_name[] = "main";

    im_register_lib_version (KARMA_VERSION);
    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
	m_abort (function_name, "control panel");
    }
    panel_add_item (panel, "out_width", "output image width",
		    K_INT, &out_width,
		    PIA_END);
    panel_add_item (panel, "out_height", "output image height",
		    K_INT, &out_height,
		    PIA_END);
    panel_add_item (panel, "extension", "converter extension name", K_VSTRING,
		    &extension,
		    PIA_END);
    panel_add_item (panel, "fixed_width_sequence",
		    "keep sequence numbers same width", PIT_FLAG,
		    &fixed_width_sequence,
		    PIA_END);
    panel_add_item (panel, "cmapfile", "default colourmap filename", K_VSTRING,
		    &cmap_file,
		    PIA_END);
    panel_add_item (panel, "count_from_zero", "count images from 0",
		    PIT_FLAG, &count_from_zero,
		    PIA_END);
    panel_add_item (panel, "converter", "converter name", K_VSTRING,
		    &converter,
		    PIA_END);
    panel_add_item (panel, "binary", "whether to write binary mode files",
		    PIT_FLAG, &binary,
		    PIA_END);
    panel_push_onto_stack (panel);
    panel_set_extra_comment_string (panel, "\nGeneral usage: infile outfile");
    module_run (argc, argv, "karma2ppm", VERSION, karma2ppm, -1, -1, FALSE);
    return (RV_OK);
}   /*  End Function main   */

static flag karma2ppm (CONST char *command, FILE *fp)
{
    flag ok;
    Channel out;
    iarray image_pseudo, image_red, image_green, image_blue;
    iarray movie_pseudo, movie_red, movie_green, movie_blue;
    unsigned int cmap_index, cmap_size, count, num_frames;
    double min, max;
    multi_array *multi_desc, *cmap_multi_desc;
    char *inp_filename, *out_filename, *ptr;
    CONST unsigned short *cmap;
    char fname[STRING_LENGTH];
    extern char *data_type_names[NUMTYPES];

    if ( ( inp_filename = ex_word (command, &command) ) == NULL )
    {
	fprintf (stderr, "Must supply input and output filenames\n");
	return (TRUE);
    }
    if ( ( out_filename = ex_word (command, &command) ) == NULL )
    {
	fprintf (stderr, "Must also supply output filename\n");
	m_free (inp_filename);
	return (TRUE);
    }
    strcpy (fname, out_filename);
    m_free (out_filename);
    if ( ( ptr = strstr (fname, ".ppm") ) != NULL ) *ptr = '\0';
    if ( (converter != NULL) && (*converter != '\0') )
    {
	fprintf (stderr,
		 "Using converter programme: \"%s\" to write files with extension: \"%s\"\n",
		 converter, extension);
    }
    /*  Read input file  */
    /*  Don't mmap as colourmap might be re-ordered  */
    if ( ( multi_desc = dsxfr_get_multi (inp_filename, FALSE, K_CH_MAP_LOCAL,
					 FALSE) ) == NULL )
    {
	fprintf (stderr, "Error getting arrayfile: \"%s\"\n", inp_filename);
	m_free (inp_filename);
	return (TRUE);
    }
    m_free (inp_filename);
    /*  Try to get 2-D image  */
    if ( iarray_get_image_from_multi (multi_desc, &image_pseudo,
				      &image_red, &image_green, &image_blue,
				      &cmap_index) )
    {
	/*  Easy! Open output file and write  */
	/*  Don't need Intelligent Arrays  */
	if ( ( out = open_file (fname, -1, 0) ) == NULL )
	{
	    ds_dealloc_multi (multi_desc);
	    return (TRUE);
	}
	if (image_pseudo != NULL)
	{
	    if ( ( cmap_multi_desc = get_cmap (multi_desc, cmap_file, &cmap,
					       &cmap_size) )
		 == NULL )
	    {
		fprintf (stderr,  "Greyscale images are not supported\n");
		iarray_dealloc (image_pseudo);
		unlink (fname);
		return (TRUE);
	    }
	    iarray_min_max (image_pseudo, CONV_CtoR_REAL, &min, &max);
	    ok = foreign_ppm_write_pseudo
		(out, binary, (CONST char *) image_pseudo->data,
		 iarray_type (image_pseudo),
		 image_pseudo->offsets[1],
		 image_pseudo->offsets[0],
		 iarray_dim_length (image_pseudo, 1),
		 iarray_dim_length (image_pseudo, 0),
		 out_width, out_height,
		 cmap, cmap + 1, cmap + 2, cmap_size, 0, 0, cmap_size - 1, 3,
		 min, max, NULL, NULL);
	    ds_dealloc_multi (cmap_multi_desc);
	    iarray_dealloc (image_pseudo);
	}
	else
	{
	    ok = foreign_ppm_write_rgb
		(out, binary,
		 (CONST unsigned char *) image_red->data,
		 (CONST unsigned char *) image_green->data,
		 (CONST unsigned char *) image_blue->data,
		 image_red->offsets[1], image_red->offsets[0],
		 iarray_dim_length (image_red, 1),
		 iarray_dim_length (image_red, 0),
		 out_width, out_height,
		 NULL, NULL, NULL, 0);
	    iarray_dealloc (image_red);
	    iarray_dealloc (image_green);
	    iarray_dealloc (image_blue);
	}
	if (!ok)
	{
	    fprintf (stderr, "Error writing PPM file\n");
	    unlink (fname);
	}
	ch_close (out);
	ds_dealloc_multi (multi_desc);
	return (TRUE);
    }
    /*  Try to get movie  */
    if ( !iarray_get_movie_from_multi (multi_desc, &movie_pseudo,
				       &movie_red, &movie_green, &movie_blue,
				       &cmap_index) )
    {
	fprintf (stderr, "Error getting movie\n");
	ds_dealloc_multi (multi_desc);
	return (TRUE);
    }
    ds_dealloc_multi (multi_desc);
    if (movie_pseudo != NULL)
    {
	multi_desc = movie_pseudo->multi_desc;
	if ( ( cmap_multi_desc = get_cmap (multi_desc, cmap_file, &cmap,
					   &cmap_size) )
	     == NULL )
	{
	    fprintf (stderr,  "Greyscale movies are not supported\n");
	    iarray_dealloc (movie_pseudo);
	    return (TRUE);
	}
	iarray_min_max (movie_pseudo, CONV_CtoR_REAL, &min, &max);
	num_frames = iarray_dim_length (movie_pseudo, 0);
	for (count = 0; count < num_frames; ++count)
	{
	    if ( ( out = open_file (fname, count, num_frames) ) == NULL )
	    {
		iarray_dealloc (movie_pseudo);
		ds_dealloc_multi (cmap_multi_desc);
		return (TRUE);
	    }
	    if ( !foreign_ppm_write_pseudo
		 (out, binary,
		  (CONST char *) movie_pseudo->data +
		  movie_pseudo->offsets[0][count],
		  iarray_type (movie_pseudo),
		  movie_pseudo->offsets[2],
		  movie_pseudo->offsets[1],
		  iarray_dim_length (movie_pseudo, 2),
		  iarray_dim_length (movie_pseudo, 1),
		  out_width, out_height,
		  cmap, cmap + 1, cmap + 2, cmap_size, 0, 0, cmap_size - 1, 3,
		  min, max, NULL, NULL) )
	    {
		fprintf (stderr, "Error writing PPM file: \"%s\"\n", fname);
		iarray_dealloc (movie_pseudo);
		ds_dealloc_multi (cmap_multi_desc);
		ch_close (out);
		return (TRUE);
	    }
	    ch_close (out);
	}
	iarray_dealloc (movie_pseudo);
	ds_dealloc_multi (cmap_multi_desc);
	return (TRUE);
    }
    /*  Sanity checking  */
    if (iarray_type (movie_red) != K_UBYTE)
    {
	fprintf (stderr, "Array data is of type: %s, must be ubyte\n",
		 data_type_names[iarray_type (movie_red)]);
	iarray_dealloc (movie_red);
	iarray_dealloc (movie_green);
	iarray_dealloc (movie_blue);
	return (TRUE);
    }
    num_frames = iarray_dim_length (movie_red, 0);
    for (count = 0; count < num_frames; ++count)
    {
	if ( ( out = open_file (fname, count, num_frames) ) == NULL )
	{
	    iarray_dealloc (movie_red);
	    iarray_dealloc (movie_green);
	    iarray_dealloc (movie_blue);
	    return (TRUE);
	}
	if ( !foreign_ppm_write_rgb (out, binary,
				     (CONST unsigned char *) movie_red->data +
				     movie_red->offsets[0][count],
				     (CONST unsigned char *) movie_green->data+
				     movie_red->offsets[0][count],
				     (CONST unsigned char *) movie_blue->data +
				     movie_red->offsets[0][count],
				     movie_red->offsets[2],
				     movie_red->offsets[1],
				     iarray_dim_length (movie_red, 2),
				     iarray_dim_length (movie_red, 1),
				     out_width, out_height,
				     NULL, NULL, NULL, 0) )
	{
	    fprintf (stderr, "Error writing PPM file: \"%s\"\n", fname);
	    iarray_dealloc (movie_red);
	    iarray_dealloc (movie_green);
	    iarray_dealloc (movie_blue);
	    ch_close (out);
	    return (TRUE);
	}
	ch_close (out);
    }
    iarray_dealloc (movie_red);
    iarray_dealloc (movie_green);
    iarray_dealloc (movie_blue);
    return (TRUE);
}   /*  End Function karma2ppm  */

static multi_array *get_cmap (multi_array *multi_desc, CONST char *cmap_file,
			      CONST unsigned short **cmap,
			      unsigned int *cmap_size)
/*  [SUMMARY] Load a colourmap and append to data structure.
    [PURPOSE] This routine will load a colourmap from a file and append it
    to a data structure. If the data structure already has a colourmap, the
    routine does nothing.
    <multi_desc> The data structure.
    <cmap_file> The file to load the colourmap from.
    <cmap> The colourmap pointer is written here.
    <cmap_size> The size of the colourmap is written here.
    [RETURNS] A pointer to the multi_array structure containing the colourmap
    on success, else NULL.
*/
{
    unsigned int dummy, cmap_index;
    multi_array *cmap_multi_desc = NULL;
    static char function_name[] = "get_cmap";

    /*  Check for existing colourmap  */
    switch ( dummy = ds_f_array_name (multi_desc, "RGBcolourmap",
				      (char **) NULL, &cmap_index) )
    {
      case IDENT_NOT_FOUND:
	if ( (cmap_file == NULL) || (*cmap_file == '\0') )
	{
	    fprintf (stderr, "%s: no colourmap file specified\n",
		     function_name);
	    return (NULL);
	}
	if ( ( cmap_multi_desc = dsxfr_get_multi (cmap_file, FALSE,
						  K_CH_MAP_LOCAL, FALSE) )
	     == NULL )
	{
	    fprintf (stderr, "Error getting arrayfile: \"%s\"\n", cmap_file);
	    return (NULL);
	}
	if (cmap_multi_desc->num_arrays != 1)
	{
	    fprintf (stderr,"%s: multiple data structures in colourmap file\n",
		     function_name);
	    ds_dealloc_multi (cmap_multi_desc);
	    return (NULL);
	}
	cmap_index = 0;
	break;
      case IDENT_GEN_STRUCT:
	/*  Got it!  */
	++multi_desc->attachments;
	cmap_multi_desc = multi_desc;
	break;
      case IDENT_MULTIPLE:
	fprintf (stderr, "Multiple RGBcolourmap structures found\n");
	return (NULL);
	/*break;*/
      default:
	fprintf (stderr,"Illegal return value: %u from: ds_f_array_name\n",
		 dummy);
	a_prog_bug (function_name);
	break;
    }
    if ( ( *cmap =
	   ds_cmap_find_colourmap (cmap_multi_desc->headers[cmap_index],
				   cmap_multi_desc->data[cmap_index],
				   cmap_size, NULL, NULL, NULL, 0) )
	 == NULL )
    {
	fprintf (stderr, "%s: no colourmap found in colourmap file\n",
		 function_name);
	ds_dealloc_multi (cmap_multi_desc);
	return (NULL);
    }
    return (cmap_multi_desc);
}   /*  End Function get_cmap  */

static Channel open_file (CONST char *subname, int image_num, int max_images)
/*  [SUMMARY] Open an output file, with possible conversion.
    <subname> The output sub-filename.
    <image_num> The image number. If this is less than 0 no image sequence
    number is used.
    <max_images> The maximum number of images.
    [RETURNS] The channel object on success, else NULL.
*/
{
    Channel channel;
    char *ext = extension;
    char fname[STRING_LENGTH];
    char format[STRING_LENGTH];

    if ( (converter == NULL) || (*converter == '\0') ) ext = "ppm";
    if (image_num < 0) sprintf (fname, "%s.%s", subname, ext);
    else if (!fixed_width_sequence)
	sprintf (fname, "%s_%d.%s",
		 subname, image_num + (count_from_zero ? 0 : 1), ext);
    else
    {
	max_images = log10 ( (double) max_images );
	++max_images;
	sprintf (format, "%%s_%%0%dd.%%s", max_images);
	sprintf (fname, format,
		 subname, image_num + (count_from_zero ? 0 : 1), ext);
    }
    if ( ( channel = ch_open_file (fname, "w") ) == NULL )
    {
	fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		 fname, ERRSTRING);
	return (NULL);
    }
    fprintf (stderr, "Writing file: \"%s\"\n", fname);
    if ( (converter == NULL) || (*converter == '\0') ) return (channel);
    if (ch_add_write_filter (channel, converter) == NULL)
    {
	ch_close (channel);
	return (NULL);
    }
    return (channel);
}   /*  End Function open_file  */
