/*  kdecimate.c

    Source file for  kdecimate  (module to decimate files).

    Copyright (C) 1997-2001  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will decimate an imput file.


    Written by      Richard Gooch   11-APR-1997

    Updated by      Richard Gooch   12-APR-1997

    Updated by      Richard Gooch   8-AUG-1997: Took account of change to
  <foreign_miriad_create_data_context>

    Updated by      Richard Gooch   19-MAY-1998: Trap illegal values.

    Last updated by Richard Gooch   21-DEC-2001: Included k_version.h directly.


*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_module.h>
#include <karma_panel.h>
#include <karma_dsrw.h>
#include <karma_ch.h>
#include <karma_ds.h>
#include <karma_im.h>
#include <karma_m.h>
#include <karma_a.h>
#include <k_version.h>
#include "lib.h"


#define VERSION "1.0"
#define MAX_DIMENSIONS 100

int main (int argc, char **argv)
{
    KControlPanel panel;
    extern unsigned int skip_lengths[MAX_DIMENSIONS];
    extern unsigned int num_skip_lengths;
    static char function_name[] = "main";

    im_register_lib_version (KARMA_VERSION);
    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
	m_abort (function_name, "control panel");
    }
    panel_add_item (panel, "lengths", "decimate lengths", K_UINT, skip_lengths,
		    PIA_ARRAY_LENGTH, &num_skip_lengths,
		    PIA_ARRAY_MAX_LENGTH, MAX_DIMENSIONS,
		    PIA_END);
    panel_push_onto_stack (panel);
    panel_set_extra_comment_string (panel, "\nGeneral usage: infile outfile");
    module_run (argc, argv, "kdecimate", VERSION, command_parse, -1, 0, FALSE);
    return (RV_OK);
}   /*  End Function main   */

void generate_file (char *infile, char *arrayfile)
/*  [PURPOSE] This routine will read in a FITS file and will convert it to a
    Karma arrayfile with a multi-dimensional array of a single atomic element.
    <infilename> The name of the imput file.
    <arrayfile> The name of the output file.
    [RETURNS] Nothing.
*/
{
    Channel channel = NULL;
    Channel karma_ch;
    KMiriadDataContext miriad_context = NULL;
    unsigned int elem_type, elem_size;
    unsigned int ftype, count, axis;
    uaddr xlen_in, ylen_in, zlen_in, xlen_out, ylen_out, zlen_out, xmax, ymax;
    uaddr z;
    char *ptr;
    unsigned char *countbuf;
    float *inbuf, *outbuf;
    multi_array *multi_desc;
    packet_desc *top_pack_desc;
    packet_desc *pack_desc;
    array_desc *arr_desc;
    char fname[STRING_LENGTH], txt[STRING_LENGTH];
    extern unsigned int num_skip_lengths;
    extern unsigned int skip_lengths[MAX_DIMENSIONS];
    extern char host_type_sizes[NUMTYPES];
    static char function_name[] = "generate_file";

    if (num_skip_lengths < 3)
    {
	fprintf (stderr, "Insufficient skip lenghts\n");
	return;
    }
    /*  Determine format  */
    ftype = foreign_guess_format_from_filename (infile);
    if (ftype == FOREIGN_FILE_FORMAT_UNKNOWN) return;
    switch (ftype)
    {
      case FOREIGN_FILE_FORMAT_FITS:
	/*  Read FITS file  */
	if ( ( channel = ch_open_file (infile, "r") ) == NULL )
	{
	    fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		     infile, ERRSTRING);
	    return;
	}
	if ( ( multi_desc =
	       foreign_fits_read_header (channel, FALSE, TRUE, TRUE,
					 FA_FITS_READ_HEADER_END) )
	     == NULL )
	{
	    fprintf (stderr, "Error reading FITS file header\n");
	    ch_close (channel);
	    return;
	}
	break;
      case FOREIGN_FILE_FORMAT_MIRIAD:
	/*  Read Miriad Image file  */
	sprintf (fname, "%s/header", infile);
	if ( ( channel = ch_open_file (fname, "r") ) == NULL )
	{
	    fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		     fname, ERRSTRING);
	    return;
	}
	if ( ( multi_desc =
	       foreign_miriad_read_header (channel, FALSE, TRUE,
					   FA_MIRIAD_READ_HEADER_END) )
	     == NULL )
	{
	    fprintf (stderr, "Error reading Miriad Image header\n");
	    ch_close (channel);
	    return;
	}
	if ( !foreign_miriad_read_history (infile, multi_desc) )
	{
	    ds_dealloc_multi (multi_desc);
	    ch_close (channel);
	    return;
	}
	ch_close (channel);
	channel = NULL;
	if ( ( miriad_context =
	       foreign_miriad_create_data_context (infile, K_CH_MAP_NEVER,
						   FALSE) ) == NULL )
	{
	    ds_dealloc_multi (multi_desc);
	    return;
	}
	break;
      case FOREIGN_FILE_FORMAT_GIPSY:
	/*  Read GIPSY file  */
	strcpy (fname, infile);
	if ( ( ptr = strrchr (fname, '.') ) == NULL ) return;
	strcpy (ptr, ".descr");
	if ( ( channel = ch_open_file (fname, "r") ) == NULL )
	{
	    fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		     fname, ERRSTRING);
	    return;
	}
	if ( ( multi_desc =
	       foreign_gipsy_read_header (channel, FALSE, TRUE,
					  FA_GIPSY_READ_HEADER_END) )
	     == NULL )
	{
	    fprintf (stderr, "Error reading GIPSY header\n");
	    ch_close (channel);
	    return;
	}
	ch_close (channel);
	strcpy (ptr, ".image");
	if ( ( channel = ch_open_file (fname, "r") ) == NULL )
	{
	    fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		     fname, ERRSTRING);
	    return;
	}
	break;
      default:
	fprintf (stderr, "File format unsupported\n");
	return;
	/*break;*/
    }
    /*  Sanity checks on data structure  */
    if (multi_desc->num_arrays != 1)
    {
	fprintf (stderr, "Multiple data structures found!\n");
	a_prog_bug (function_name);
    }
    top_pack_desc = multi_desc->headers[0];
    if (top_pack_desc->element_types[0] != K_ARRAY)
    {
	fprintf (stderr, "Array not found!\n");
	a_prog_bug (function_name);
    }
    arr_desc = (array_desc *) top_pack_desc->element_desc[0];
    if (arr_desc->num_dimensions != 3)
    {
	fprintf (stderr, "%u dimensions not supported\n",
		 arr_desc->num_dimensions);
	ds_dealloc_multi (multi_desc);
	return;
    }
    pack_desc = arr_desc->packet;
    if (pack_desc->num_elements != 1)
    {
	fprintf (stderr, "%u array packet elements not supported\n",
		 pack_desc->num_elements);
	ds_dealloc_multi (multi_desc);
	return;
    }
    elem_type = pack_desc->element_types[0];
    if (elem_type != K_FLOAT)
    {
	fprintf (stderr, "Array type: %u not supported\n",
		 pack_desc->element_types[0]);
	ds_dealloc_multi (multi_desc);
	return;
    }
    elem_size = host_type_sizes[elem_type];
    xlen_in = arr_desc->dimensions[2]->length;
    ylen_in = arr_desc->dimensions[1]->length;
    zlen_in = arr_desc->dimensions[0]->length;
    xlen_out = xlen_in / skip_lengths[0];
    ylen_out = ylen_in / skip_lengths[1];
    zlen_out = zlen_in / skip_lengths[2];
    /*  Fiddle descriptor  */
    for (count = 0; count < 3; ++count)
    {
	double skip = skip_lengths[2 - count];
	double crpix[2], cdelt[2], naxis[2];

	if ( ( axis = ds_get_fits_axis (multi_desc->headers[0],
					multi_desc->data[0],
					arr_desc->dimensions[count]->name) )
	     < 1 )
	{
	    fprintf (stderr, "No FITS axis for dimension: \"%s\"\n",
		     arr_desc->dimensions[count]->name);
	    ds_dealloc_multi (multi_desc);
	    return;
	}
	arr_desc->dimensions[count]->length /= skip;
	arr_desc->lengths[count] = arr_desc->dimensions[count]->length;
	arr_desc->dimensions[count]->last_coord = arr_desc->lengths[count] - 1;
	sprintf (txt, "NAXIS%u", axis);
	naxis[0] = arr_desc->dimensions[count]->length;
	naxis[1] = 0.0;
	if ( !ds_put_unique_named_value (multi_desc->headers[0],
					 multi_desc->data, txt, K_INT, naxis,
					 TRUE) )
	{
	    fprintf (stderr, "Error setting keyword %s\n", txt);
	    ds_dealloc_multi (multi_desc);
	    return;
	}
	sprintf (txt, "CRPIX%u", axis);
	if ( !ds_get_unique_named_value (multi_desc->headers[0],
					 multi_desc->data[0], txt, NULL,
					 crpix) )
	{
	    fprintf (stderr, "No %s keyword\n", txt);
	    ds_dealloc_multi (multi_desc);
	    return;
	}
	crpix[0] = (crpix[0] - 1.0 + skip * 0.5) / skip;
	if ( !ds_put_unique_named_value (multi_desc->headers[0],
					 multi_desc->data, txt, K_DOUBLE,
					 crpix, TRUE) )
	{
	    fprintf (stderr, "Error setting keyword %s\n", txt);
	    ds_dealloc_multi (multi_desc);
	    return;
	}
	sprintf (txt, "CDELT%u", axis);
	if ( !ds_get_unique_named_value (multi_desc->headers[0],
					 multi_desc->data[0], txt, NULL,
					 cdelt) )
	{
	    fprintf (stderr, "No %s keyword\n", txt);
	    ds_dealloc_multi (multi_desc);
	    return;
	}
	cdelt[0] *= skip;
	if ( !ds_put_unique_named_value (multi_desc->headers[0],
					 multi_desc->data, txt, K_DOUBLE,
					 cdelt, TRUE) )
	{
	    fprintf (stderr, "Error setting keyword %s\n", txt);
	    ds_dealloc_multi (multi_desc);
	    return;
	}
    }
    if ( ( inbuf = (float *) m_alloc (sizeof *inbuf * xlen_in * ylen_in) )
	 == NULL )
    {
	m_error_notify (function_name, "input image buffer");
	ds_dealloc_multi (multi_desc);
	return;
    }
    if ( ( outbuf = (float *) m_alloc (sizeof *outbuf * xlen_out * ylen_out) )
	 == NULL )
    {
	m_error_notify (function_name, "output image buffer");
	m_free ( (char *) inbuf );
	ds_dealloc_multi (multi_desc);
	return;
    }
    if ( ( countbuf = (unsigned char *)
	   m_alloc (sizeof *countbuf * xlen_out * ylen_out) ) == NULL )
    {
	m_error_notify (function_name, "output image count buffer");
	m_free ( (char *) inbuf );
	m_free ( (char *) outbuf );
	ds_dealloc_multi (multi_desc);
	return;
    }
    if ( ( karma_ch = open_karma (arrayfile, multi_desc) ) == NULL )
    {
	m_free ( (char *) inbuf );
	m_free ( (char *) outbuf );
	m_free ( (char *) countbuf );
	ds_dealloc_multi (multi_desc);
	return;
    }
    /*  Loop through input planes  */
    xmax = xlen_in - xlen_in % skip_lengths[0];
    ymax = ylen_in - ylen_in % skip_lengths[1];
    zlen_in -= zlen_in % skip_lengths[2];
    m_clear ( (char *) outbuf, xlen_out * ylen_out * sizeof *outbuf );
    m_clear ( (char *) countbuf, xlen_out * ylen_out );
    for (z = 0; z < zlen_in; ++z)
    {
	unsigned char ubval;
	uaddr xin, yin, xout, yout, tot;
	float fval;
	float toobig = TOOBIG;
	unsigned char *countline;
	float *inpline, *outline;

	switch (ftype)
	{
	    case FOREIGN_FILE_FORMAT_FITS:
	      if ( !foreign_fits_read_data (channel, multi_desc,
					    (char *) inbuf, xlen_in * ylen_in,
					    FA_FITS_READ_DATA_END) )
	      {
		  fprintf (stderr, "Error reading FITS data\n");
		  return;
	      }
	      break;
	    case FOREIGN_FILE_FORMAT_MIRIAD:
	      if ( !foreign_miriad_read_data (miriad_context, multi_desc,
					      (char *) inbuf, xlen_in *ylen_in,
					      FA_MIRIAD_READ_DATA_END) )
	      {
		  fprintf (stderr, "Error reading Miriad data\n");
		  return;
	      }
	      break;
	    case FOREIGN_FILE_FORMAT_GIPSY:
	      if ( !foreign_gipsy_read_data (channel, multi_desc,
					     (char *) inbuf, xlen_in * ylen_in,
					     FA_GIPSY_READ_DATA_END) )
	      {
		  fprintf (stderr, "Error reading GIPSY data\n");
		  return;
	      }
	      break;
	}
	/*  Accumulate  */
	inpline = inbuf;
	for (yin = 0; yin < ymax; ++yin, inpline += xlen_in)
	{
	    yout = yin / skip_lengths[1];
	    outline = outbuf + yout * xlen_out;
	    countline = countbuf + yout * xlen_out;
	    for (xin = 0; xin < xmax; ++xin)
	    {
		xout = xin / skip_lengths[0];
		fval = inpline[xin];
		if ( DOUBLE_IS_BLANK (fval) ) continue;
		outline[xout] += fval;
		++countline[xout];
	    }
	}
	if ( (z + 1) % skip_lengths[2] == 0)
	{
	    /*  Average  */
	    tot = xlen_out * ylen_out;
	    for (xout = 0; xout < tot; ++xout)
	    {
		if ( (ubval = countbuf[xout]) == 0 ) fval = toobig;
		else fval = outbuf[xout] / (float) ubval;
		outbuf[xout] = fval;
	    }
	    if ( !dsrw_write_packets (karma_ch, arr_desc->packet,
				      (char *) outbuf, xlen_out * ylen_out) )
	    {
		fprintf (stderr, "Error writing Karma file\n");
		return;
	    }
	    m_clear ( (char *) outbuf, xlen_out * ylen_out * sizeof *outbuf );
	    m_clear ( (char *) countbuf, xlen_out * ylen_out );
	}
    }
    write_tail (karma_ch, multi_desc, infile, arrayfile);
    ch_close (karma_ch);
    ds_dealloc_multi (multi_desc);
}   /*  End Function generate_file  */


/*  Private functions follow  */


/*  Put globals here to force functions to be explicit  */
unsigned int skip_lengths[MAX_DIMENSIONS] = {2, 2, 2};
unsigned int num_skip_lengths = 3;
