/*  kgen.c

    Source file for  kgen  (data structure generation module).

    Copyright (C) 1993-1999  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will generate a Karma data file, using the general data
    structure format.


    Written by      Richard Gooch   3-OCT-1992

    Updated by      Richard Gooch   13-FEB-1993

    Updated by      Richard Gooch   28-MAR-1993: Took account of changes to
  list_header  structure and replaced reading routines with  dsra_  package.

    Updated by      Richard Gooch   26-APR-1993: Fixed bug in call to
  dsra_multi_data  .

    Updated by      Richard Gooch   24-JUL-1993: Changed to  ch_open_file  .

    Updated by      Richard Gooch   13-OCT-1993: Changed over to  panel_
  package for command line user interface and moved  main  into this file.

    Updated by      Richard Gooch   23-NOV-1993: Changed to use of
  <ex_word_skip>.

    Updated by      Richard Gooch   30-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package.

    Last updated by Richard Gooch   30-JAN-1999: Switched to new experimental
  function <dsxfr_read_split>.


*/
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <sys/file.h>
#include <karma.h>
#include <karma_module.h>
#include <karma_panel.h>
#include <karma_dsxfr.h>
#include <karma_dsrw.h>
#include <karma_dsra.h>
#include <karma_ds.h>
#include <karma_ch.h>
#include <karma_ex.h>
#include <karma_m.h>

#define VERSION "1.2"

EXTERN_FUNCTION (flag kgen, (CONST char *command, FILE *fp) );

STATIC_FUNCTION (void generate_file, (char *filename) );


#define DATA_FORMAT_ASCII 0
#define DATA_FORMAT_BINARY 1
#define NUM_DATA_FORMATS 2

static char *data_format_alternatives[] =
{
    "ascii",
    "binary"
};
static int data_format = DATA_FORMAT_ASCII;

int main (int argc, char **argv)
{
    KControlPanel panel;
    static char function_name[] = "main";

    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
	m_abort (function_name, "control panel");
    }
    panel_add_item (panel, "data_format", "choice", PIT_CHOICE_INDEX,
		    &data_format,
		    PIA_NUM_CHOICE_STRINGS, NUM_DATA_FORMATS,
		    PIA_CHOICE_STRINGS, data_format_alternatives,
		    PIA_END);
    panel_push_onto_stack (panel);
    module_run (argc, argv, "kgen", VERSION, kgen, -1, 0, FALSE);
    return (RV_OK);
}   /*  End Function main   */

flag kgen (CONST char *command, FILE *fp)
{
    char filename[STRING_LENGTH+1];

    for ( ; command; command = ex_word_skip (command) )
    {
	sscanf (command, "%s", filename);
        generate_file (filename);
    }
    return (TRUE);
}   /*  End Function kgen  */

void generate_file (char *filename)
/*  This routine will generate a Karma data file using the  filename  specified
    as a basis.
    The routine returns nothing.
*/
{
    char dat_file[STRING_LENGTH];
    multi_array *multi_desc;
    extern int data_format;
    /*static char function_name[] = "generate_file";*/

    /*  Generate filenames  */
    sprintf (dat_file, "%s.dat", filename);
    /*  Test readability of data file  */
    if (access (dat_file, R_OK) != 0)
    {
	fprintf (stderr, "File: %s is not readable\n", dat_file);
        return;
    }
    if ( ( multi_desc =
	   dsxfr_read_split (dat_file,
			     (data_format == DATA_FORMAT_ASCII) ? TRUE : FALSE,
			     FALSE) ) == NULL ) return;
    fprintf (stderr, "File: %s read in\n", filename);
    if (dsxfr_put_multi (filename, multi_desc) != TRUE)
    {
	fprintf (stderr, "Error transmitting data structure\n");
    }
    ds_dealloc_multi (multi_desc);
}   /*  End Function generate_file  */
