/*  ktranslate.c

    Source file for  ktranslate  (module to convert between file formats).

    Copyright (C) 2001-2003  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will convert an input file of any format to an output
    file of any format.


    Written by      Richard Gooch   10-MAY-2001

    Updated by      Richard Gooch   15-AUG-2001: First implementation.

    Updated by      Richard Gooch   16-AUG-2001: Removed duplicate history
  when writing Miriad files.

    Updated by      Richard Gooch   21-DEC-2001: Included k_version.h directly.

    Updated by      Richard Gooch   17-APR-2002: Support .sfits input format.

    Updated by      Richard Gooch   4-FEB-2003: Added history for piecemeal
  Miriad and AIPS input.

    Last updated by Richard Gooch   22-APR-2003: Took account of API change for
  DRAO data.


*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_module.h>
#include <karma_dsxfr.h>
#include <karma_panel.h>
#include <karma_pio.h>
#include <karma_ch.h>
#include <karma_ds.h>
#include <karma_ex.h>
#include <karma_st.h>
#include <karma_im.h>
#include <karma_m.h>
#include <karma_a.h>
#include <k_version.h>


#define VERSION "1.1"
#define BUF_SIZE 1048576


/*  Private data for user interface  */

static unsigned int ftype_values[] =
{   /*  Formats which can be written  */
    FOREIGN_FILE_FORMAT_UNKNOWN,
    FOREIGN_FILE_FORMAT_KARMA,
    FOREIGN_FILE_FORMAT_PPM,
    FOREIGN_FILE_FORMAT_FITS,
    FOREIGN_FILE_FORMAT_SUNRAS,
    FOREIGN_FILE_FORMAT_MIRIAD,
    FOREIGN_FILE_FORMAT_GIPSY,
    /*  Formats which can only be read  */
    FOREIGN_FILE_FORMAT_PGM,
    FOREIGN_FILE_FORMAT_DRAO,
    FOREIGN_FILE_FORMAT_AIPS1,
    FOREIGN_FILE_FORMAT_PNM,
    FOREIGN_FILE_FORMAT_SIMPLE_FITS,
    FOREIGN_FILE_FORMAT_IDF,
};
#define NUM_OUTPUT_FTYPES  7
static char *ftype_alternatives[] =
{   /*  Formats which can be written  */
    "guess",
    "karma",
    "ppm",
    "fits",
    "sunras",
    "miriad",
    "gipsy",
    /*  Formats which can only be read  */
    "pgm",
    "drao",
    "aips1",
    "pnm",
    "sfits",
    "idf",
    NULL
};
static char *ftype_comments[] =
{   /*  Formats which can be written  */
    "guess format from filename",
    "Karma",
    "Portable Pixel Map",
    "FITS",
    "Sun Rasterfile",
    "Miriad",
    "GIPSY",
    /*  Formats which can only be read  */
    "Portable Grey Map",
    "DRAO",
    "Classic AIPS",
    "Portable aNy Map",
    "Simple FITS",
    "IDF",
    NULL
};
static unsigned int in_ftype_index = 0;
static unsigned int out_ftype_index = 0;


/*  Private data  */


/*  Private functions  */
STATIC_FUNCTION (flag ktranslate, (CONST char *command, FILE *fp) );
STATIC_FUNCTION (void generate_file,
		 (CONST char *inname, CONST char *outname) );


int main (int argc, char **argv)
{
    KControlPanel panel;
    static char function_name[] = "main";

    im_register_lib_version (KARMA_VERSION);
    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
	m_abort (function_name, "control panel");
    }
    panel_add_item (panel, "outtype", "output filetype",
		    PIT_CHOICE_INDEX, &out_ftype_index,
		    PIA_CHOICE_STRINGS, ftype_alternatives,
		    PIA_CHOICE_COMMENTS, ftype_comments,
		    PIA_NUM_CHOICE_STRINGS, NUM_OUTPUT_FTYPES,
		    PIA_END);
    panel_add_item (panel, "intype", "input filetype",
		    PIT_CHOICE_INDEX, &in_ftype_index,
		    PIA_CHOICE_STRINGS, ftype_alternatives,
		    PIA_CHOICE_COMMENTS, ftype_comments,
		    PIA_END);
    panel_push_onto_stack (panel);
    module_run (argc, argv, "ktranslate", VERSION, ktranslate, -1, 0,
		FALSE);
    return (RV_OK);
}   /*  End Function main   */

static flag ktranslate (CONST char *command, FILE *fp)
{
    char *output_filename;
    char *input_filename;

    for ( ; command; command = ex_word_skip (command) )
    {
	if ( ( input_filename = ex_str (command, &command) ) == NULL )
	{
	    fprintf (stderr, "Error extracting input filename\n");
	    return (TRUE);
	}
	if ( ( output_filename = ex_str (command, &command) ) == NULL )
	{
	    fprintf (stderr, "Error extracting output filename\n");
	    m_free (input_filename);
	    return (TRUE);
	}
	generate_file (input_filename, output_filename);
	m_free (input_filename);
	m_free (output_filename);
    }
    return (TRUE);
}   /*  End Function ktranslate  */

static void generate_file (CONST char *inname, CONST char *outname)
/*  [PURPOSE] This routine will convert a file from one format to another.
    <inname> The name of the input dataset.
    <outname> The name of the output dataset.
    [RETURNS] Nothing.
*/
{
    Channel inp_ch = NULL;
    Channel out_ch = NULL;
    KMiriadDataContext miriad_ctx = NULL;
    KDRAODataContext drao_ctx = NULL;
    KAIPS1DataContext aips1_ctx = NULL;
    flag do_piecemeal = FALSE;
    flag ok = FALSE;
    flag sfits_in = TRUE;
    unsigned int array_size, elem_size;
    unsigned int array_count, num_values;
    unsigned int inp_ftype = ftype_values[in_ftype_index];
    unsigned int out_ftype = ftype_values[out_ftype_index];
    unsigned long toobig_count;
    unsigned long toobig_count_tmp;
    multi_array *inp_multi_desc = NULL;
    multi_array *out_multi_desc;
    packet_desc *pack_desc;
    array_desc *arr_desc;
    char *ptr;
    char fname[STRING_LENGTH], txt[STRING_LENGTH];
    extern char module_name[STRING_LENGTH + 1];
    extern char module_version_date[STRING_LENGTH + 1];
    extern char module_lib_version[STRING_LENGTH + 1];
    extern char karma_library_version[STRING_LENGTH + 1];
    static char *buffer = NULL;
    static char function_name[] = "generate_file";

    if (buffer == NULL)
    {
	if ( ( buffer = m_alloc (BUF_SIZE) ) == NULL )
	    m_abort (function_name, "buffer");
    }
    if (inp_ftype == FOREIGN_FILE_FORMAT_UNKNOWN)
	inp_ftype = foreign_guess_format_from_filename (inname);
    if (inp_ftype == FOREIGN_FILE_FORMAT_UNKNOWN)
    {
	fprintf (stderr, "Unknown file format for input: \"%s\"\n", inname);
	return;
    }
    if (out_ftype == FOREIGN_FILE_FORMAT_UNKNOWN)
	out_ftype = foreign_guess_format_from_filename (outname);
    if (out_ftype == FOREIGN_FILE_FORMAT_UNKNOWN) out_ftype = inp_ftype;
    switch (out_ftype)  /*  Some output formats can be done piecemeal  */
    {
      case FOREIGN_FILE_FORMAT_FITS:
      case FOREIGN_FILE_FORMAT_GIPSY:
	do_piecemeal = TRUE;
	break;
    }
    /*  Read input header, setup for reading data and possibly read data  */
    switch (inp_ftype)
    {
      case FOREIGN_FILE_FORMAT_KARMA:
	do_piecemeal = FALSE;
	inp_multi_desc = dsxfr_get_multi (inname, FALSE, K_CH_MAP_IF_AVAILABLE,
					  FALSE);
	break;
      case FOREIGN_FILE_FORMAT_PNM:
      case FOREIGN_FILE_FORMAT_PGM:
      case FOREIGN_FILE_FORMAT_PPM:
	do_piecemeal = FALSE;
	if ( ( inp_ch = ch_open_file (inname, "r") ) == NULL )
	{
	    fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		     function_name, inname, ERRSTRING);
	    return;
	}
	inp_multi_desc = foreign_pnm_read (inp_ch);
	ch_close (inp_ch);
	inp_ch = NULL;
	break;
      case FOREIGN_FILE_FORMAT_FITS:
	sfits_in = FALSE;
	/*  Fall through  */
      case FOREIGN_FILE_FORMAT_SIMPLE_FITS:
	if (do_piecemeal)
	{
	    if ( ( inp_ch = ch_open_file (inname, "rz") ) == NULL )
	    {
		fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
			 function_name, inname, ERRSTRING);
		return;
	    }
	    inp_multi_desc =
		foreign_fits_read_header (inp_ch, FALSE, FALSE, TRUE,
					  FA_FITS_READ_HEADER_SIMPLE, sfits_in,
					  FA_FITS_READ_HEADER_END);
	}
	else inp_multi_desc = foreign_fits_read (inname, FALSE, TRUE,
						 FA_FITS_READ_SIMPLE, sfits_in,
						 FA_FITS_READ_END);
	break;
      case FOREIGN_FILE_FORMAT_SUNRAS:
	do_piecemeal = FALSE;
	if ( ( inp_ch = ch_open_file (inname, "r") ) == NULL )
	{
	    fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		     function_name, inname, ERRSTRING);
	    return;
	}
	inp_multi_desc = foreign_sunras_read (inp_ch, FA_SUNRAS_READ_END);
	ch_close (inp_ch);
	inp_ch = NULL;
	break;
      case FOREIGN_FILE_FORMAT_MIRIAD:
	if (do_piecemeal)
	{
	    sprintf (fname, "%s/header", inname);
	    if ( ( inp_ch = ch_open_file (fname, "r") ) == NULL )
	    {
		fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
			 function_name, fname, ERRSTRING);
		return;
	    }
	    inp_multi_desc =
		foreign_miriad_read_header (inp_ch, FALSE, TRUE,
					    FA_MIRIAD_READ_HEADER_END);
	    ch_close (inp_ch);
	    inp_ch = NULL;
	    if (inp_multi_desc == NULL) break;
	    foreign_miriad_read_history (inname, inp_multi_desc);
	    if ( ( miriad_ctx =
		   foreign_miriad_create_data_context (inname, K_CH_MAP_NEVER,
						       FALSE) ) == NULL )
	    {
		fprintf (stderr, "%s: error opening Miriad data\t%s\n",
			 function_name, ERRSTRING);
		ds_dealloc_multi (inp_multi_desc);
		return;
	    }
	}
	else inp_multi_desc = foreign_miriad_read (inname, TRUE,
						   FA_MIRIAD_READ_END);
	break;
      case FOREIGN_FILE_FORMAT_GIPSY:
	if (do_piecemeal)
	{
	    strcpy (fname, inname);
	    if ( ( ptr = strrchr (fname, '.') ) == NULL ) return;
	    strcpy (ptr, ".descr");
	    if ( ( inp_ch = ch_open_file (fname, "r") ) == NULL )
	    {
		fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
			 function_name, fname, ERRSTRING);
		return;
	    }
	    inp_multi_desc =
		foreign_gipsy_read_header (inp_ch, FALSE, TRUE,
					   FA_GIPSY_READ_HEADER_END);
	    ch_close (inp_ch);
	    strcpy (ptr, ".image");
	    if ( ( inp_ch = ch_open_file (fname, "r") ) == NULL )
	    {
		fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
			 function_name, fname, ERRSTRING);
		ds_dealloc_multi (inp_multi_desc);
		return;
	    }
	}
	else inp_multi_desc = foreign_gipsy_read (inname, TRUE,
						  FA_GIPSY_READ_END);
	break;
      case FOREIGN_FILE_FORMAT_DRAO:
	if (do_piecemeal)
	    inp_multi_desc = foreign_drao_read_header(inname, FALSE, FALSE,
						      &drao_ctx,
						      FA_DRAO_READ_HEADER_END);
	else inp_multi_desc = foreign_drao_read (inname, FALSE,
						 FA_DRAO_READ_END);
	break;
      case FOREIGN_FILE_FORMAT_AIPS1:
	if (do_piecemeal)
	{
	    inp_multi_desc = foreign_aips1_read_header (inname, FALSE,
							&aips1_ctx);
	    foreign_aips1_read_history (aips1_ctx, inp_multi_desc);
	}
	else inp_multi_desc = foreign_aips1_read (inname);
	break;
      case FOREIGN_FILE_FORMAT_IDF:
	do_piecemeal = FALSE;  /*  More code needed  */
	if (do_piecemeal)
	{
	    if ( ( inp_ch = ch_open_file (inname, "r") ) == NULL )
	    {
		fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
			 function_name, inname, ERRSTRING);
		return;
	    }
	    inp_multi_desc = foreign_idf_read_header (inp_ch, FALSE, FALSE,
						      FA_IDF_READ_HEADER_END);
	    ch_close (inp_ch);
	    inp_ch = NULL;
	}
	else inp_multi_desc = foreign_idf_read (inname, TRUE, FA_IDF_READ_END);
	break;
    }
    if (inp_multi_desc == NULL)
    {
	fprintf (stderr, "Error reading header/dataset\n");
	return;
    }
    switch (out_ftype)
    {
      case FOREIGN_FILE_FORMAT_FITS:
      case FOREIGN_FILE_FORMAT_GIPSY:
	/*  Create FITS-style header  */
	if ( !foreign_fits_generate_header (&out_multi_desc, inp_multi_desc,
					    FA_FITS_GENERATE_HEADER_END) )
	    m_abort (function_name, "temporary header");
	break;
      default:
	out_multi_desc = inp_multi_desc;
	++out_multi_desc->attachments;
	sprintf (txt, "%s: Module version %s  Karma v%s  compiled with v%s",
		 module_name, module_version_date,
		 karma_library_version, module_lib_version);
	ds_history_append_string (out_multi_desc, txt, TRUE);
	break;
    }
    /*  Append a bit of history  */
    panel_put_history_with_stack (out_multi_desc, FALSE);
    sprintf (txt, "%s: infile: %s  outfile: %s", module_name, inname, outname);
    ds_history_append_string (out_multi_desc, txt, TRUE);
    switch (out_ftype)
    {
      case FOREIGN_FILE_FORMAT_KARMA:
	ok = dsxfr_put_multi (outname, out_multi_desc);
	break;
      case FOREIGN_FILE_FORMAT_PPM:
	if ( ( out_ch = ch_open_file (outname, "w") ) == NULL )
	{
	    fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		     function_name, outname, ERRSTRING);
	    ok = FALSE;
	    break;
	}
	ok = foreign_ppm_write (out_ch, out_multi_desc, TRUE,FA_PPM_WRITE_END);
	break;
      case FOREIGN_FILE_FORMAT_FITS:
	if ( ( out_ch = ch_open_file (outname, "w") ) == NULL )
	{
	    fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		     function_name, outname, ERRSTRING);
	    ok = FALSE;
	    break;
	}
	ok = foreign_fits_write_header (out_ch, out_multi_desc,
					inp_multi_desc->first_hist);
	if (!ok) break;
	if (!do_piecemeal)
	    ok = foreign_fits_write_data (out_ch, inp_multi_desc,
					  out_multi_desc->headers[0],
					  out_multi_desc->data[0],
					  NULL, 0,
					  FA_FITS_WRITE_END);
	break;
      case FOREIGN_FILE_FORMAT_SUNRAS:
	if ( ( out_ch = ch_open_file (outname, "w") ) == NULL )
	{
	    fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		     function_name, outname, ERRSTRING);
	    ok = FALSE;
	    break;
	}
	ok = foreign_sunras_write (out_ch, out_multi_desc, TRUE,
				   FA_SUNRAS_WRITE_END);
	break;
      case FOREIGN_FILE_FORMAT_MIRIAD:
	ok = foreign_miriad_write (outname, out_multi_desc,
				   FA_MIRIAD_WRITE_EXTRA_HISTORY, FALSE,
				   FA_MIRIAD_WRITE_END);
	break;
      case FOREIGN_FILE_FORMAT_GIPSY:
	strcpy (fname, inname);
	if ( ( ptr = strrchr (fname, '.') ) == NULL )
	    ptr = fname + strlen (fname);
	strcpy (ptr, ".descr");
	/*  Write the GIPSY header  */
	if ( ( out_ch = ch_open_file (fname, "w") ) == NULL )
	{
	    fprintf (stderr, "Error opening GIPSY header file\t%s\n",
		     ERRSTRING);
	    ok = FALSE;
	    break;
	}
	ok = foreign_gipsy_write_header (out_ch,
					 out_multi_desc->headers[0],
					 out_multi_desc->data[0],
					 FA_GIPSY_WRITE_HEADER_END);
	if (!ok) break;
	ok = ch_close (out_ch);
	out_ch = NULL;
	if (!ok) break;
	/*  Create the GIPSY image file  */
	strcpy (ptr, ".image");
	if ( (out_ch = ch_open_file (fname, "w") ) == NULL )
	{
	    fprintf (stderr, "Error opening GIPSY data file\t%s\n",
		     ERRSTRING);
	    ok = FALSE;
	    break;
	}
	if (!do_piecemeal)
	    ok = foreign_gipsy_write_data (out_ch, inp_multi_desc,
					   out_multi_desc->headers[0],
					   out_multi_desc->data[0],
					   NULL, 0,
					   FA_GIPSY_WRITE_DATA_END);
	break;
    }
    pack_desc = inp_multi_desc->headers[0];
    arr_desc = (array_desc *) pack_desc->element_desc[0];
    array_size = ds_get_array_size (arr_desc);
    elem_size = ds_get_packet_size (arr_desc->packet);
    toobig_count = 0;
    if (ok && do_piecemeal)  /*  Read and write in blocks  */
    {
	/*  Start converting data. Do it in blocks so that we don't have to
	    allocate possibly huge amounts of virtual memory  */
	for (array_count = 0; ok && (array_count < array_size);
	     array_count += num_values)
	{
	    num_values = array_size - array_count;
	    if (num_values * elem_size > BUF_SIZE)
		num_values = BUF_SIZE / elem_size;
	    /*  Read a block of data into memory  */
	    toobig_count_tmp = 0;
	    switch (inp_ftype)
	    {
	      case FOREIGN_FILE_FORMAT_FITS:
	      case FOREIGN_FILE_FORMAT_SIMPLE_FITS:
		ok = foreign_fits_read_data (inp_ch, inp_multi_desc,
					     buffer, num_values,
					     FA_FITS_READ_DATA_ASCII, sfits_in,
					     FA_FITS_READ_DATA_END);
		break;
	      case FOREIGN_FILE_FORMAT_MIRIAD:
		ok = foreign_miriad_read_data (miriad_ctx, inp_multi_desc,
					       buffer, num_values,
					       FA_MIRIAD_READ_DATA_NUM_BLANKS,
					       &toobig_count_tmp,
					       FA_MIRIAD_READ_DATA_END);
		break;
	      case FOREIGN_FILE_FORMAT_GIPSY:
		ok = foreign_gipsy_read_data (inp_ch, inp_multi_desc,
					      buffer, num_values,
					      FA_GIPSY_READ_DATA_END);
		break;
	      case FOREIGN_FILE_FORMAT_DRAO:
		ok = foreign_drao_read_data (drao_ctx, buffer, num_values);
		break;
	      case FOREIGN_FILE_FORMAT_AIPS1:
		ok = foreign_aips1_read_data (aips1_ctx, buffer, num_values);
		break;
	      case FOREIGN_FILE_FORMAT_IDF:
		ok = foreign_idf_read_data (inp_ch, inp_multi_desc, buffer,
					    num_values);
		break;
	    }
	    if (!ok) break;
	    toobig_count += toobig_count_tmp;
	    /*  Write a block of data out  */
	    switch (out_ftype)
	    {
	      case FOREIGN_FILE_FORMAT_FITS:
		ok = foreign_fits_write_data (out_ch, inp_multi_desc,
					      out_multi_desc->headers[0],
					      out_multi_desc->data[0],
					      buffer, num_values,
					      FA_FITS_WRITE_DATA_END);
		break;
	      case FOREIGN_FILE_FORMAT_GIPSY:
		 ok = foreign_gipsy_write_data (out_ch, out_multi_desc,
						out_multi_desc->headers[0],
						out_multi_desc->data[0],
						buffer, num_values,
						FA_GIPSY_WRITE_DATA_END);
		 break;
	    }
	}
	
    }
    if (ok && do_piecemeal)  /*  Finish up  */
    {
	switch (out_ftype)
	{
	  case FOREIGN_FILE_FORMAT_FITS:
	    ok = foreign_fits_write_padding (out_ch);
	    break;
	}
    }
    /*  Clean up  */
    if (inp_ch != NULL) ch_close (inp_ch);
    if (miriad_ctx != NULL) foreign_miriad_close_data_context (miriad_ctx);
    if (drao_ctx != NULL) foreign_drao_close_data_context (drao_ctx);
    if (aips1_ctx != NULL) foreign_aips1_close_data_context (aips1_ctx);
    ds_dealloc_multi (inp_multi_desc);
    ds_dealloc_multi (out_multi_desc);
    if (out_ch != NULL) ch_close (out_ch);
    if (toobig_count > 0)
    {
	fprintf (stderr, "Number of blank values: %lu\n", toobig_count);
    }
}   /*  End Function generate_file  */
