/*  raw_rgb2karma.c

    Source file for  raw_rgb2karma  (convert raw data to Karma format).

    Copyright (C) 2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will generate a multi-dimensional array of RGB data from
    a raw data file containing values and will write the data to a Karma data
    file.
    The raw data file must be in packed RGB (24 bit) format.


    Written by      Richard Gooch   8-MAR-2002

    Last updated by Richard Gooch   8-MAR-2002


*/
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <karma.h>
#include <karma_module.h>
#include <karma_panel.h>
#include <karma_dsxfr.h>
#include <karma_ds.h>
#include <karma_ch.h>
#include <karma_ex.h>
#include <karma_m.h>


#define VERSION "1.0"

#define MAX_DIMENSIONS (unsigned int) 100


STATIC_FUNCTION (flag raw_rgb2karma, (CONST char *command, FILE *fp) );
STATIC_FUNCTION (void generate_file,
		 (CONST char *inp_filename, CONST char *arrayname) );


/*  Private data  */
static unsigned int num_dimensions = 0;
static char *names[MAX_DIMENSIONS];
static unsigned int lengths[MAX_DIMENSIONS];
static double minima[MAX_DIMENSIONS];
static double maxima[MAX_DIMENSIONS];
static flag invert_flag = FALSE;


int main (int argc, char **argv)
{
    KControlPanel panel, group;
    static char function_name[] = "main";

    if ( ( group = panel_create_group () ) == NULL )
    {
	m_abort (function_name, "group panel");
    }
    panel_add_item (group, "maximum", "", K_DOUBLE, (char *) maxima,
		    PIA_END);
    panel_add_item (group, "minimum", "", K_DOUBLE, (char *) minima,
		    PIA_END);
    panel_add_item (group, "length", "", K_UINT, (char *) lengths,
		    PIA_END);
    panel_add_item (group, "name", "", K_VSTRING, (char *) names,
		    PIA_END);
    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
	m_abort (function_name, "control panel");
    }
    panel_add_item (panel, "dimensions", "", PIT_GROUP, (char *) group,
		    PIA_ARRAY_MAX_LENGTH, MAX_DIMENSIONS,
		    PIA_ARRAY_LENGTH, (char *) &num_dimensions,
		    PIA_END);
    panel_add_item (panel, "invert_image", "flag", PIT_FLAG, &invert_flag,
		    PIA_END);
    panel_push_onto_stack (panel);
    module_run (argc, argv, "raw_rgb2karma", VERSION, raw_rgb2karma,
		-1, 0, FALSE);
    return (RV_OK);
}   /*  End Function main   */

static flag raw_rgb2karma (CONST char *command, FILE *fp)
{
    char *arrayfile, *input_filename;
    extern unsigned int num_dimensions;

    if (num_dimensions < 1)
    {
	fprintf (fp, "Must have at least one dimension to create\n");
	return (TRUE);
    }
    for ( ; command; command = ex_command_skip (command) )
    {
	if ( ( input_filename = ex_str (command, &command) ) == NULL )
	{
	    fprintf (fp, "Error extracting input filename\n");
	    return (TRUE);
	}
	if ( ( arrayfile = ex_str (command, &command) ) == NULL )
	{
	    fprintf (fp, "Error extracting arrayfile name\n");
	    m_free (input_filename);
	    return (TRUE);
	}
	generate_file (input_filename, arrayfile);
	m_free (input_filename);
	m_free (arrayfile);
    }
    return (TRUE);
}   /*  End Function raw_rgb2karma  */

static void generate_file (CONST char *inp_filename, CONST char *arrayname)
/*  This routine will generate a Karma arrayfile with a multi-dimensional array
    of multiple atomic elements. The data will be converted from the input
    files.
    The names of the input files must be pointed to by  inp_filenames  .
    The name of the Karma arrayfile must be pointed to by  arrayname  .
    The type of the data elements must be in  elem_type  .
    The routine returns nothing.
*/
{
    Channel inp_ch;
    array_pointer arrayp;
    int row_step;
    unsigned int array_size, row_size, dim_count, row_count;
    char *array;
    multi_array *multi_desc;
    array_desc *arr_desc;
    uaddr ulengths[MAX_DIMENSIONS];
    extern flag invert_flag;
    extern unsigned int num_dimensions;
    extern unsigned int lengths[MAX_DIMENSIONS];
    extern double minima[MAX_DIMENSIONS];
    extern double maxima[MAX_DIMENSIONS];
    extern char *names[MAX_DIMENSIONS];
    static unsigned int elem_types[3] = {K_UBYTE, K_UBYTE, K_UBYTE};
    static char *elem_names[3] = {"Red Intensity", "Green Intensity",
				  "Blue Intensity"};
    /*static char function_name[] = "generate_file";*/

    /*  Determine array size (in co-ordinates)  */
    array_size = 1;
    for (dim_count = 0; dim_count < num_dimensions; ++dim_count)
    {
	if (lengths[dim_count] < 1)
	{
	    fprintf (stderr,
		     "Dimension: %u has length: %d: must be greater than 0\n",
		     dim_count, lengths[dim_count]);
	    return;
	}
	array_size *= (unsigned int) lengths[dim_count];
    }
    if ( ( inp_ch = ch_open_file (inp_filename, "rz") ) == NULL )
    {
	fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		 inp_filename, ERRSTRING);
	return;
    }
    for (dim_count = 0; dim_count < num_dimensions; ++dim_count)
	ulengths[dim_count] = lengths[dim_count];
    if ( ( arr_desc =
	   ds_easy_alloc_array_desc (num_dimensions, ulengths, minima, maxima,
				     (CONST double **) NULL,
				     (CONST char **) names,
				     3, elem_types,
				     (CONST char **) elem_names) )
	 == NULL )
    {
	ch_close (inp_ch);
	return;
    }
    if ( ( multi_desc = ds_easy_alloc_array_from_array_desc (arr_desc, NULL,
							     FALSE) ) == NULL )
    {
	ds_dealloc_array_desc (arr_desc);
	ch_close (inp_ch);
	return;
    }
    m_copy (&arrayp, multi_desc->data[0], sizeof arrayp);
    /*  Determine if inversion of image is needed  */
    row_size = array_size * 3 / lengths[0];
    if (invert_flag)
    {
	array = arrayp.array + row_size * (lengths[0] - 1);
	row_step = -1;
    }
    else
    {
	array = arrayp.array;
	row_step = 1;
    }
    /*  Read entire array  */
    for (row_count = 0; row_count < lengths[0];
	 ++row_count, array += row_size * row_step)
    {
	/*  Read one "row" (highest dimension)  */
	if (ch_read (inp_ch, array, row_size) < row_size)
	{
	    fprintf (stderr, "Error reading file: \"%s\"\t%s\n",
		     inp_filename, ERRSTRING);
	    ds_dealloc_multi (multi_desc);
	    ch_close (inp_ch);
	    return;
	}
    }
    ch_close (inp_ch);
    /*  Write to arrayfile  */
    if ( !dsxfr_put_multi (arrayname, multi_desc) )
    {
	fprintf (stderr, "Error putting array\n");
    }
    /*  Write comments to new array  */
    ds_dealloc_multi (multi_desc);
}   /*  End Function generate_file  */
