/*  kregrid.c

    Source file for  kregrid  (Regrid file according to another file's grid).

    Copyright (C) 1996-2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will read two 2-dimensional arrays. One file will be used
  to compute a new grid for the other file.


    Written by      Richard Gooch   13-JUL-1996

    Updated by      Richard Gooch   20-NOV-1996: Trap for pixels outside
  defined co-ordinate system.

    Updated by      Richard Gooch   22-NOV-1996: Write blanks when outside
  input image boundaries, rather than clipping to "nearest" input value.

    Updated by      Richard Gooch   23-NOV-1996: Added mosaicing support.

    Updated by      Richard Gooch   24-NOV-1996: Improved speed by only
  regridding area(s) corresponding to input image(s). Especially good for
  mosaicing.

    Updated by      Richard Gooch   26-NOV-1996: Added manual grid
  specification.

    Updated by      Richard Gooch   29-DEC-1996: Switched to
  <iarray_put_double>.

    Updated by      Richard Gooch   14-FEB-1997: Switched to
  <ds_convert_coordinates> and did co-ordinate conversions for 2D regridding in
  one hit, as this is faster, especially for threaded co-ordinate
  transformations (up to 10x under Linux).

    Updated by      Richard Gooch   16-APR-1997: Compute min&max of output and
  print warning if the output is uniform value.

    Updated by      Richard Gooch   28-MAY-1997: Display output file.

    Updated by      Richard Gooch   27-JUN-1997: Restricted manual grid to 2
  dimensions since that is all that is really supported.

    Updated by      Richard Gooch   5-AUG-1997: Test to ensure images overlap.

    Updated by      Richard Gooch   17-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Updated by      Richard Gooch   15-JUN-2000: Moved <iarray_regrid_2D> and
  <iarray_regrid_3D> to lib/karma/packages/iarray/regrid.c

    Updated by      Richard Gooch   21-DEC-2001: Included k_version.h directly.

    Last updated by Richard Gooch   18-MAR-2002: Made use of new experimental
  function <foreign_guess_write>.


*/
#include <stdio.h>
#include <math.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_module.h>
#include <karma_iarray.h>
#include <karma_panel.h>
#include <karma_dsxfr.h>
#include <karma_wcs.h>
#include <karma_chs.h>
#include <karma_ds.h>
#include <karma_ch.h>
#include <karma_ex.h>
#include <karma_im.h>
#include <karma_m.h>
#include <karma_a.h>
#include <k_version.h>


#define VERSION "1.8.2"
#define MIN_DIMENSIONS 2
#define MAX_DIMENSIONS 2


/*  Private data for user interface  */
static char *grid_file = NULL;
static flag mosaic = FALSE;

#define NUM_SAMPLE_OPTIONS 2
static unsigned int sample_option_values[NUM_SAMPLE_OPTIONS] =
{
    REGRID_SAMPLE_DATA_COPY,
    REGRID_SAMPLE_BILINEAR_INTERPOLATION,
};
static char *sample_option_alternatives[NUM_SAMPLE_OPTIONS] =
{
    "data_copy",
    "linear_interpolation",
};
static char *sample_option_comments[NUM_SAMPLE_OPTIONS] =
{
    "copy nearest data value",
    "bi-linear interpolation",
};
static unsigned int sample_option = 0;

static unsigned int num_dimensions = 2;
static char *names[MAX_DIMENSIONS] = {"RA---SIN", "DEC--SIN"};
static unsigned long lengths[MAX_DIMENSIONS] = {512, 512};
static double crval[MAX_DIMENSIONS] = {20.0, -34.0};
static double crpix[MAX_DIMENSIONS] = {256.0, 256.0};
static double cdelt[MAX_DIMENSIONS] = {-0.1, 0.1};
static double crota = 0.0;
static char *unit_name = "JY/BEAM";
static flag manual_grid = FALSE;

/*  Private data for other purposes  */
static float min;
static float max;


/*  Private functions  */
STATIC_FUNCTION (flag kregrid, (CONST char *command, FILE *fp) );
STATIC_FUNCTION (flag regrid_one_file,
		 (CONST char *infile, CONST char *grid_file) );
STATIC_FUNCTION (flag regrid_many_files,
		 (CONST char *infile, CONST char *grid_file) );
STATIC_FUNCTION (iarray create_from_gridfile,
		 (CONST char *file, iarray in, KwcsAstro *out_ap) );
STATIC_FUNCTION (iarray create_manual_grid, (iarray in, KwcsAstro *out_ap) );
STATIC_FUNCTION (flag compare_coord_systems, (iarray out_arr, iarray in_arr) );


/*  Public functions follow  */

int main (int argc, char **argv)
{
    KControlPanel panel, group;
    static char function_name[] = "main";

    if ( ( group = panel_create_group () ) == NULL )
    {
	m_abort (function_name, "group panel");
    }
    panel_add_item (group, "cdelt", "pixel increment", K_DOUBLE,
		    (char *) cdelt,
		    PIA_END);
    panel_add_item (group, "crpix", "reference pixel position", K_DOUBLE,
		    (char *) crpix,
		    PIA_END);
    panel_add_item (group, "crval", "reference value", K_DOUBLE,
		    (char *) crval,
		    PIA_END);
    panel_add_item (group, "length", "length of axis", K_ULONG,
		    (char *) lengths,
		    PIA_END);
    panel_add_item (group, "name", "name of axis", K_VSTRING, (char *) names,
		    PIA_END);
    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
	m_abort (function_name, "control panel");
    }
    panel_add_item (panel, "unit_name", "used for manual grid specification",
		    K_VSTRING, &unit_name,
		    PIA_END);
    panel_add_item (panel, "sample_option", "how to sample data",
		    PIT_CHOICE_INDEX, &sample_option,
		    PIA_NUM_CHOICE_STRINGS, NUM_SAMPLE_OPTIONS,
		    PIA_CHOICE_STRINGS, sample_option_alternatives,
		    PIA_CHOICE_COMMENTS, sample_option_comments,
		    PIA_END);
    panel_add_item (panel, "mosaic", "input file contains a list of files",
		    PIT_FLAG, &mosaic,
		    PIA_END);
    panel_add_item (panel, "manual_grid", "use the manual grid specification",
		    PIT_FLAG, &manual_grid,
		    PIA_END);
    panel_add_item (panel, "grid_file", "file containing desired grid",
		    K_VSTRING, (char *) &grid_file,
		    PIA_END);
    panel_add_item (panel, "crota", "rotation angle", K_DOUBLE,
		    (char *) &crota,
		    PIA_END);
    panel_add_item (panel, "dimensions", "manual grid specification",
		    PIT_GROUP, (char *) group,
		    PIA_ARRAY_MIN_LENGTH, MIN_DIMENSIONS,
		    PIA_ARRAY_MAX_LENGTH, MAX_DIMENSIONS,
		    PIA_ARRAY_LENGTH, (char *) &num_dimensions,
		    PIA_END);
    panel_push_onto_stack (panel);
    panel_set_extra_comment_string (panel, "\nGeneral usage: infile");
    im_register_lib_version (KARMA_VERSION);
    module_run (argc, argv, "kregrid", VERSION, kregrid, -1, 0, FALSE);
    return (RV_OK);
}   /*  End Function main   */

static flag kregrid (CONST char *command, FILE *fp)
{
    flag ok;
    char *infile;

    for ( ; command != NULL; command = ex_word_skip (command) )
    {
	if ( ( infile = ex_str (command, &command) ) == NULL )
	{
	    fprintf (fp, "Error extracting arrayfile name\n");
	    return (TRUE);
	}
	min = TOOBIG;
	max = -TOOBIG;
	if (mosaic) ok = regrid_many_files (infile, grid_file);
	else ok = regrid_one_file (infile, grid_file);
	m_free (infile);
	if (!ok) continue;
	if ( DOUBLE_IS_BLANK (min) )
	{
	    fprintf (stderr,
		     "WARNING: output file: \"kregrid_%s.kf\" is all blanks\n",
		     infile);
	    fprintf (stderr,
		     "This may be because the input image lies outside the ");
	    fprintf (stderr, "new co-ordinate system\n");
	}
	else if (min >= max)
	{
	    fprintf (stderr,
		     "WARNING: output file: \"kregrid_%s.kf\" contains only: ",
		     infile);
	    fprintf (stderr, "%e\n", min);
	    fprintf (stderr,
		     "This may be because the input image lies outside the ");
	    fprintf (stderr, "new co-ordinate system\n");
	}
    }
    return (TRUE);
}   /*  End Function kregrid  */

static flag regrid_one_file (CONST char *infile, CONST char *grid_file)
/*  [SUMMARY] Regrid a file to the grid of another file.
    <infile> The file to regrid.
    <grid_file> The file containing the desired grid.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    KwcsAstro in_ap = NULL;
    KwcsAstro out_ap;
    iarray in_arr = NULL;
    iarray out_arr;
    flag ok;
    unsigned int ftype, in_dim;
    char txt[STRING_LENGTH];

    /*  Read input array  */
    if ( !foreign_read_and_setup (infile, K_CH_MAP_IF_AVAILABLE, FALSE,
				  &ftype, TRUE, 0, K_FLOAT, TRUE,
				  &in_arr, NULL, NULL,
				  TRUE, &in_ap) ) return (FALSE);
    if (in_ap == NULL)
    {
	fprintf (stderr, "File: \"%s\" has no astronomical projection\n",
		 infile);
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    in_dim = iarray_num_dim (in_arr);
    if ( (in_dim < 2) || (in_dim > 3) )
    {
	fprintf (stderr, "Unsupported number of dimensions: %u\n", in_dim);
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    if (manual_grid) out_arr = create_manual_grid (in_arr, &out_ap);
    else out_arr = create_from_gridfile (grid_file, in_arr, &out_ap);
    if (out_arr == NULL)
    {
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    if ( !compare_coord_systems (out_arr, in_arr) )
    {
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	wcs_astro_destroy (out_ap);
	return (FALSE);
    }
    if (in_dim == 3)
	ok = iarray_regrid_3D (out_arr, out_ap, in_arr, in_ap,
			       sample_option_values[sample_option],
			       &min, &max);
    else ok = iarray_regrid_2D (out_arr, out_ap, in_arr, in_ap,
				sample_option_values[sample_option],
				&min, &max);
    if (ok)
    {
	sprintf (txt, "kregrid_%s", infile);
	fprintf (stderr, "Writing %s\n", txt);
	foreign_guess_write (txt, iarray_multi_array (out_arr),
			     FA_GUESS_WRITE_END);
    }
    iarray_dealloc (in_arr);
    iarray_dealloc (out_arr);
    wcs_astro_destroy (out_ap);
    return (ok);
}   /*  End Function regrid_one_file  */

static flag regrid_many_files (CONST char *infile, CONST char *grid_file)
/*  [SUMMARY] Regrid many files to the grid of another file.
    <infile> The file containing the list of file to regrid.
    <grid_file> The file containing the desired grid.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    KwcsAstro in_ap = NULL;
    KwcsAstro out_ap;
    Channel ch;
    iarray in_arr = NULL;
    iarray out_arr = NULL;
    flag ok;
    unsigned int ftype, num_dim;
    char txt[STRING_LENGTH];

    if ( ( ch = ch_open_file (infile, "r") ) == NULL ) return (FALSE);
    num_dim = 0;
    while (chs_get_value (ch, txt, STRING_LENGTH) > 0)
    {
	fprintf (stderr, "Processing file: \"%s\"\n", txt);
	if ( !foreign_read_and_setup (txt, K_CH_MAP_IF_AVAILABLE, FALSE,
				      &ftype, TRUE, num_dim, K_FLOAT, TRUE,
				      &in_arr, NULL, NULL,
				      TRUE, &in_ap) )
	{
	    ch_close (ch);
	    iarray_dealloc (out_arr);
	    wcs_astro_destroy (out_ap);
	    return (FALSE);
	}
	if (out_arr == NULL)
	{
	    if (manual_grid) out_arr = create_manual_grid (in_arr, &out_ap);
	    else out_arr = create_from_gridfile (grid_file, in_arr, &out_ap);
	    if (out_arr == NULL)
	    {
		return (FALSE);
	    }
	    num_dim = iarray_num_dim (out_arr);
	}
	if (in_ap == NULL)
	{
	    fprintf (stderr, "File: \"%s\" has no astronomical projection\n",
		     txt);
	    ch_close (ch);
	    iarray_dealloc (out_arr);
	    wcs_astro_destroy (out_ap);
	    iarray_dealloc (in_arr);
	    return (FALSE);
	}
	if (iarray_num_dim (in_arr) != num_dim)
	{
	    fprintf (stderr, "File: \"%s\" has %u dimensions, should be: %u\n",
		     txt, iarray_num_dim (in_arr), num_dim);
	    ch_close (ch);
	    iarray_dealloc (out_arr);
	    iarray_dealloc (in_arr);
	    wcs_astro_destroy (out_ap);
	    return (FALSE);
	}
	if ( !compare_coord_systems (out_arr, in_arr) )
	{
	    ch_close (ch);
	    iarray_dealloc (out_arr);
	    iarray_dealloc (in_arr);
	    wcs_astro_destroy (out_ap);
	    return (FALSE);
	}
	if (num_dim == 3)
	    ok = iarray_regrid_3D (out_arr, out_ap, in_arr,in_ap,
				   sample_option_values[sample_option],
				   &min, &max);
	else ok = iarray_regrid_2D (out_arr, out_ap, in_arr, in_ap,
				    sample_option_values[sample_option],
				    &min, &max);
	if (!ok)
	{
	    ch_close (ch);
	    iarray_dealloc (out_arr);
	    iarray_dealloc (in_arr);
	    wcs_astro_destroy (out_ap);
	    return (FALSE);
	}
	iarray_dealloc (in_arr);
	in_arr = NULL;
	in_ap = NULL;
    }
    wcs_astro_destroy (out_ap);
    sprintf (txt, "kregrid_%s", infile);
    fprintf (stderr, "Writing %s.kf\n", txt);
    iarray_write (out_arr, txt);
    iarray_dealloc (out_arr);
    return (TRUE);
}   /*  End Function regrid_many_files  */

static iarray create_from_gridfile (CONST char *file, iarray in,
				    KwcsAstro *out_ap)
/*  [SUMMARY] Create an array using a gridfile as a template.
    <file> The template grid file.
    <in> The input file used to control the z dimension (if it exists). This
    may be NULL.
    <out_ap> The KwcsAstro object for the output array is written here.
    [RETURNS] The array on success, else NULL.
*/
{
    KwcsAstro ap = NULL;
    iarray new;
    iarray grid = NULL;
    unsigned int dummy, in_dim, grid_dim;
    double first_coord, last_coord;
    unsigned long dim_lengths[3];
    CONST char *dim_names[3];

    if ( (file == NULL) || (*file == '\0') )
    {
	fprintf (stderr, "No grid_file specified\n");
	return (NULL);
    }
    if ( !foreign_read_and_setup (file, K_CH_MAP_IF_AVAILABLE, FALSE, &dummy,
				  TRUE, 0, NONE, FALSE, &grid, NULL, NULL,
				  FALSE, &ap) ) return (NULL);
    if (ap == NULL)
    {
	fprintf (stderr, "Grid file has no astronomical projection\n");
	iarray_dealloc (grid);
	return (NULL);
    }
    grid_dim = iarray_num_dim (grid);
    if ( (grid_dim != 2) && (grid_dim != 3) )
    {
	fprintf (stderr, "Unsupported number of grid dimensions: %u\n",
		 grid_dim);
	iarray_dealloc (grid);
	return (NULL);
    }
    if ( (in == NULL) || (iarray_num_dim (in) == 2) )
    {
	/*  Take all from grid file  */
	fprintf (stderr, "creating...");
	new = iarray_create_from_template (grid, K_FLOAT, TRUE, TRUE, TRUE);
	iarray_dealloc (grid);
	fprintf (stderr, "\tfilling...");
	iarray_fill_float (new, TOOBIG);
	fprintf (stderr, "\tfilled\n");
	if ( ( *out_ap = wcs_astro_setup (new->top_pack_desc,
					  *new->top_packet) ) == NULL )
	{
	    iarray_dealloc (new);
	    return (NULL);
	}
	return (new);
    }
    /*  Take x & y dimensions from grid file and z dimension from input file */
    in_dim = iarray_num_dim (in);
    /*  Create output array with same grid as grid file  */
    dim_lengths[in_dim - 2] = iarray_dim_length (grid, grid_dim - 2);
    dim_lengths[in_dim - 1] = iarray_dim_length (grid, grid_dim - 1);
    dim_names[in_dim - 2] = iarray_dim_name (grid, grid_dim - 2);
    dim_names[in_dim - 1] = iarray_dim_name (grid, grid_dim - 1);
    if (in_dim == 3)
    {
	dim_lengths[0] = iarray_dim_length (in, 0);
	dim_names[0] = iarray_dim_name (in, 0);
    }
    fprintf (stderr, "creating...");
    if ( ( new = iarray_create (K_FLOAT, in_dim, dim_names, dim_lengths,
				iarray_value_name (grid), grid) ) == NULL )
    {
	fprintf (stderr, "Error creating output array\n");
	iarray_dealloc (grid);
	return (NULL);
    }
    iarray_dealloc (grid);
    fprintf (stderr, "\tfilling...");
    iarray_fill_float (new, TOOBIG);
    fprintf (stderr, "\tfilled\n");
    if (in_dim == 3)
    {
	iarray_get_world_coords (in, 0, &first_coord, &last_coord);
	iarray_set_world_coords (new, 0, first_coord, last_coord);
    }
    if ( ( *out_ap = wcs_astro_setup (new->top_pack_desc,
				      *new->top_packet) ) == NULL )
    {
	iarray_dealloc (new);
	return (NULL);
    }
    return (new);
}   /*  End Function create_from_gridfile  */

static iarray create_manual_grid (iarray in, KwcsAstro *out_ap)
/*  [SUMMARY] Create an array using a manual grid as a template.
    <in> The input file used to control the z dimension (if it exists). This
    may be NULL.
    <out_ap> The KwcsAstro object for the output array is written here.
    [RETURNS] The array on success, else NULL.
*/
{
    iarray new;
    unsigned int in_dim = 0;  /*  Initialised to keep compiler happy  */
    unsigned int count, new_dim;
    double first_coord, last_coord;
    unsigned long dim_lengths[3];
    CONST char *dim_names[3];
    char txt[STRING_LENGTH];

    if (in == NULL) new_dim = num_dimensions;
    else
    {
	in_dim = iarray_num_dim (in);
	new_dim = in_dim;
    }
    /*  Setup the x,y (RA,DEC) dimensions  */
    dim_lengths[new_dim - 1] = lengths[0];
    dim_lengths[new_dim - 2] = lengths[1];
    dim_names[new_dim - 1] = names[0];
    dim_names[new_dim - 2] = names[1];
    if (new_dim == 3)
    {
	dim_lengths[0] = iarray_dim_length (in, 0);
	dim_names[0] = iarray_dim_name (in, 0);
    }
    fprintf (stderr, "creating...");
    if ( ( new = iarray_create (K_FLOAT, new_dim, dim_names, dim_lengths,
				unit_name, NULL) ) == NULL )
    {
	fprintf (stderr, "Error creating output array\n");
	return (NULL);
    }
    fprintf (stderr, "\tfilling...");
    iarray_fill_float (new, TOOBIG);
    fprintf (stderr, "\tfilled\n");
    if ( (in != NULL) && (in_dim == 3) )
    {
	iarray_get_world_coords (in, 0, &first_coord, &last_coord);
	iarray_set_world_coords (new, 0, first_coord, last_coord);
    }
    /*  Add header information  */
    for (count = 0; (count < new_dim) && (count < num_dimensions); ++count)
    {
	sprintf (txt, "CTYPE%u", count + 1);
	if ( !iarray_put_named_string (new, txt, names[count]) )
	{
	    iarray_dealloc (new);
	    return (NULL);
	}
	sprintf (txt, "CRVAL%u", count + 1);
	if ( !iarray_put_double (new, txt, crval[count]) )
	{
	    iarray_dealloc (new);
	    return (NULL);
	}
	sprintf (txt, "CRPIX%u", count + 1);
	if ( !iarray_put_double (new, txt, crpix[count]) )
	{
	    iarray_dealloc (new);
	    return (NULL);
	}
	sprintf (txt, "CDELT%u", count + 1);
	if ( !iarray_put_double (new, txt, cdelt[count]) )
	{
	    iarray_dealloc (new);
	    return (NULL);
	}
    }
    if ( !iarray_put_double (new, "CROTA2", crota) )
    {
	iarray_dealloc (new);
	return (NULL);
    }
    if ( ( *out_ap = wcs_astro_setup (new->top_pack_desc,
				      *new->top_packet) ) == NULL )
    {
	fprintf (stderr, "Error creating KwcsAstro object for output\n");
	iarray_dealloc (new);
	return (NULL);
    }
    return (new);
}   /*  End Function create_manual_grid  */

static flag compare_coord_systems (iarray out_arr, iarray in_arr)
/*  [SUMMARY] Test if two co-ordinate systems are compatible.
    <out_arr> The array with the output co-ordinate system.
    <in_arr> The array with the input co-ordinate system.
    [RETURNS] TRUE if the co-ordinate systems are compatible, else FALSE.
*/
{
    CONST char *in_dim_h, *in_dim_v, *out_dim_h, *out_dim_v;

    if (iarray_num_dim (in_arr) == 3)
    {
	in_dim_h = iarray_dim_name (in_arr, 2);
	in_dim_v = iarray_dim_name (in_arr, 1);
	out_dim_h = iarray_dim_name (out_arr, 2);
	out_dim_v = iarray_dim_name (out_arr, 1);
    }
    else
    {
	in_dim_h = iarray_dim_name (in_arr, 1);
	in_dim_v = iarray_dim_name (in_arr, 0);
	out_dim_h = iarray_dim_name (out_arr, 1);
	out_dim_v = iarray_dim_name (out_arr, 0);
    }
    if ( ( wcs_astro_test_if_ra (NULL, in_dim_h) !=
	   wcs_astro_test_if_ra (NULL, out_dim_h) ) ||
	 ( wcs_astro_test_if_ra (NULL, in_dim_v) !=
	   wcs_astro_test_if_ra (NULL, out_dim_v) ) ||
	 ( wcs_astro_test_if_dec (NULL, in_dim_h) !=
	   wcs_astro_test_if_dec (NULL, out_dim_h) ) ||
	 ( wcs_astro_test_if_dec (NULL, in_dim_v) !=
	   wcs_astro_test_if_dec (NULL, out_dim_v) ) ||
	 ( wcs_astro_test_if_glon (NULL, in_dim_h) !=
	   wcs_astro_test_if_glon (NULL, out_dim_h) ) ||
	 ( wcs_astro_test_if_glon (NULL, in_dim_v) !=
	   wcs_astro_test_if_glon (NULL, out_dim_v) ) ||
	 ( wcs_astro_test_if_glat (NULL, in_dim_h) !=
	   wcs_astro_test_if_glat (NULL, out_dim_h) ) ||
	 ( wcs_astro_test_if_glat (NULL, in_dim_v) !=
	   wcs_astro_test_if_glat (NULL, out_dim_v) ) )
    {
	fprintf (stderr, "Co-ordinate systems do not match\n");
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function compare_coord_systems  */
