/*  krotate.c

    Source file for  krotate  (rotate file so longitude axis is horizontal).

    Copyright (C) 2000-2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will rotate a 2 or 3 dimensional array so that the
    minor (X) axis corresponds to longitude.


    Written by      Richard Gooch   16-JUN-2000

    Updated by      Richard Gooch   17-JUN-2000: Applied cos (lat) term.

    Updated by      Richard Gooch   20-JUN-2000: Added rotation control.

    Updated by      Richard Gooch   9-MAY-2001: Recognise ".FITS" output name.

    Updated by      Richard Gooch   9-MAY-2001: Added grid option.

    Updated by      Richard Gooch   15-MAY-2001: Added background and bgvalue
 options.

    Updated by      Richard Gooch   11-AUG-2001: Copy over keywords for axes
  with index > 2.

    Updated by      Richard Gooch   5-OCT-2001: Added #ifdef OS_SunOS.

    Updated by      Richard Gooch   6-DEC-2001: Took account of interface
  change to <wcs_astro_transform*>.

    Updated by      Richard Gooch   21-DEC-2001: Included k_version.h directly.

    Last updated by Richard Gooch   18-MAR-2002: Made use of new experimental
  function <foreign_guess_write>.


*/
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_module.h>
#include <karma_iarray.h>
#include <karma_panel.h>
#include <karma_wcs.h>
#include <karma_ch.h>
#include <karma_ex.h>
#include <karma_im.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_a.h>
#include <k_version.h>


#ifdef OS_SunOS
#  define strtoul (unsigned long) strtol
#endif


#define VERSION "1.0.6"


/*  Private data for user interface  */

static unsigned int sample_option_values[] =
{
    REGRID_SAMPLE_BILINEAR_INTERPOLATION,
    REGRID_SAMPLE_DATA_COPY,
};
static char *sample_option_alternatives[] =
{
    "linear_interpolation",
    "data_copy",
    NULL
};
static char *sample_option_comments[] =
{
    "bi-linear interpolation",
    "copy nearest data value",
    NULL
};
static unsigned int sample_option = 0;

static char *projection_option_alternatives[] =
{
    "input",
    "SIN",
    "TAN",
    "NCP",
    "FLT",
    "ARC",
    "ZEA",
    "GLS",
    NULL
};
static char *projection_option_comments[] =
{
    "use projection of input file",
    "Orthographic",
    "Gnomonic",
    "North Celestial Pole",
    "Flat",
    "Rectangular",
    "Zenithal Equal-Area",
    "Global Sinusoidal",
    NULL
};
static unsigned int projection_option = 0;

static char *grid_option_alternatives[] =
{
    "rescale",
    "pad",
    "truncate",
    NULL
};
static char *grid_option_comments[] =
{
    "preserve image size by rescaling",
    "preserve pixel size by padding",
    "preserve pix&image size by truncating",
    NULL
};
#define GRID_RESCALE              0
#define GRID_PAD                  1
#define GRID_TRUNCATE             2
static unsigned int grid_option = GRID_RESCALE;

static char *background_option_alternatives[] =
{
    "blank",
    "bgvalue",
    NULL
};
static char *background_option_comments[] =
{
    "use blank value",
    "use the bgvalue parameter",
    NULL
};
#define BACKGROUND_BLANK          0
#define BACKGROUND_CONSTANT       1
static unsigned int background_option = BACKGROUND_BLANK;
static double background_constant = 0.0;

static double rotation = 0.0;

static char *keywords[] =
{
    "OBJECT",
    "DATE-OBS",
    "OBSRA",
    "OBSDEC",
    "EPOCH",
    "TELESCOP",
    "INSTRUME",
    "EQUINOX",
    "BTYPE",
    "BMAJ",
    "BMIN",
    "BPA",
    "RESTFREQ",
    "VELREF",
    "PBFWHM",
    NULL
};

static char *axis_keywords[] =
{
    "NAXIS",
    "CTYPE",
    "CRVAL",
    "CRPIX",
    "CDELT",
    "CUNIT",
    NULL
};

/*  Private data for other purposes  */
static float min;
static float max;


/*  Private functions  */
STATIC_FUNCTION (flag krotate, (CONST char *command, FILE *fp) );
STATIC_FUNCTION (flag do_rotate, (CONST char *infile, CONST char *outfile) );
STATIC_FUNCTION (flag copy_header_info, (iarray out, iarray in,
					 unsigned int min_axis) );


/*  Public functions follow  */

int main (int argc, char **argv)
{
    KControlPanel panel;
    static char function_name[] = "main";

    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
	m_abort (function_name, "control panel");
    }
    panel_add_item (panel, "sample_option", "how to sample data",
		    PIT_CHOICE_INDEX, &sample_option,
		    PIA_CHOICE_STRINGS, sample_option_alternatives,
		    PIA_CHOICE_COMMENTS, sample_option_comments,
		    PIA_END);
    panel_add_item (panel, "rotation", "degrees", K_DOUBLE, &rotation,
		    PIA_END);
    panel_add_item (panel, "projection", "output projection",
		    PIT_CHOICE_INDEX, &projection_option,
		    PIA_CHOICE_STRINGS, projection_option_alternatives,
		    PIA_CHOICE_COMMENTS, projection_option_comments,
		    PIA_END);
    panel_add_item (panel, "grid", "output grid",
		    PIT_CHOICE_INDEX, &grid_option,
		    PIA_CHOICE_STRINGS, grid_option_alternatives,
		    PIA_CHOICE_COMMENTS, grid_option_comments,
		    PIA_END);
    panel_add_item (panel, "bgvalue", "value to use for background",
		    K_DOUBLE, &background_constant,
		    PIA_END);
    panel_add_item (panel, "background", "values outside input image",
		    PIT_CHOICE_INDEX, &background_option,
		    PIA_CHOICE_STRINGS, background_option_alternatives,
		    PIA_CHOICE_COMMENTS, background_option_comments,
		    PIA_END);
    panel_push_onto_stack (panel);
    panel_set_extra_comment_string (panel, "\nGeneral usage: infile outfile");
    im_register_lib_version (KARMA_VERSION);
    module_run (argc, argv, "krotate", VERSION, krotate, -1, 0, FALSE);
    return (RV_OK);
}   /*  End Function main   */

static flag krotate (CONST char *command, FILE *fp)
{
    for ( ; command != NULL; command = ex_word_skip (command) )
    {
	flag ok;
	char *infile, *outfile;

	if ( ( infile = ex_str (command, &command) ) == NULL )
	{
	    fprintf (fp, "Error extracting input filename\n");
	    return (TRUE);
	}
	if ( ( outfile = ex_str (command, &command) ) == NULL )
	{
	    fprintf (fp, "Error extracting output filename\n");
	    return (TRUE);
	}
	min = TOOBIG;
	max = -TOOBIG;
	ok = do_rotate (infile, outfile);
	m_free (infile);
	m_free (outfile);
	if (!ok) continue;
	if ( DOUBLE_IS_BLANK (min) )
	{
	    fprintf (stderr,
		     "WARNING: output file: \"krotate_%s.kf\" is all blanks\n",
		     infile);
	    fprintf (stderr,
		     "This may be because the input image lies outside the ");
	    fprintf (stderr, "new co-ordinate system\n");
	}
	else if (min >= max)
	{
	    fprintf (stderr,
		     "WARNING: output file: \"krotate_%s.kf\" contains only: ",
		     infile);
	    fprintf (stderr, "%e\n", min);
	    fprintf (stderr,
		     "This may be because the input image lies outside the ");
	    fprintf (stderr, "new co-ordinate system\n");
	}
    }
    return (TRUE);
}   /*  End Function krotate  */

static flag do_rotate (CONST char *infile, CONST char *outfile)
/*  [SUMMARY] Rotate a file.
    <infile> The file to rotate.
    <outfile> The output filename
    [RETURNS] TRUE on success, else FALSE.
*/
{
    KwcsAstro in_ap = NULL;
    KwcsAstro out_ap;
    iarray in_arr = NULL;
    iarray out_arr;
    flag ok;
    flag galactic = FALSE;
    unsigned int ftype, num_dim, count, dim_h_index, dim_v_index;
    double tmp1, tmp2, out_crval1, out_crval2, out_cdelt1, out_cdelt2;
    double lon_min = TOOBIG;
    double lon_max = -TOOBIG;
    double lat_min = TOOBIG;
    double lat_max = -TOOBIG;
    CONST dim_desc *in_dim_h, *in_dim_v, *in_dim_lon, *in_dim_lat;
    CONST char *in_projection, *out_projection;
    char txt[STRING_LENGTH], lon[STRING_LENGTH], lat[STRING_LENGTH];
    unsigned long dim_lengths[3];
    double x[4], y[4], lon_coords[4], lat_coords[4];
    CONST char *dim_names[3];
    extern char module_name[STRING_LENGTH + 1];

    /*  Read input array  */
    if ( !foreign_read_and_setup (infile, K_CH_MAP_IF_AVAILABLE, FALSE,
				  &ftype, TRUE, 0, K_FLOAT, TRUE,
				  &in_arr, NULL, NULL,
				  TRUE, &in_ap) ) return (FALSE);
    if (in_ap == NULL)
    {
	fprintf (stderr, "File: \"%s\" has no astronomical projection\n",
		 infile);
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    num_dim = iarray_num_dim (in_arr);
    if ( (num_dim < 2) || (num_dim > 3) )
    {
	fprintf (stderr, "Unsupported number of dimensions: %u\n", num_dim);
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    /*  Construct output grid  */
    if (iarray_num_dim (in_arr) == 3)
    {
	dim_h_index = 2;
	dim_v_index = 1;
	in_dim_h = iarray_get_dim_desc (in_arr, dim_h_index);
	in_dim_v = iarray_get_dim_desc (in_arr, dim_v_index);
    }
    else
    {
	dim_h_index = 1;
	dim_v_index = 0;
	in_dim_h = iarray_get_dim_desc (in_arr, dim_h_index);
	in_dim_v = iarray_get_dim_desc (in_arr, dim_v_index);
    }
    if ( !wcs_astro_test_if_lon (in_ap, in_dim_h->name) &&
	 !wcs_astro_test_if_lon (in_ap, in_dim_v->name) )
    {
	fprintf (stderr, "No longitude axis found\n");
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    if ( !wcs_astro_test_if_lat (in_ap, in_dim_h->name) &&
	 !wcs_astro_test_if_lat (in_ap, in_dim_v->name) )
    {
	fprintf (stderr, "No latitude axis found\n");
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    if ( !wcs_astro_test_if_lon (NULL, in_dim_h->name) &&
	 !wcs_astro_test_if_lat (NULL, in_dim_h->name) )
	in_projection = "SIN";  /*   Unknown projection (such as DSS)  */
    else
    {
	char *ptr = strrchr (in_dim_h->name, '-');
	if (ptr == NULL) in_projection = "FLT";
	else in_projection = ptr + 1;
    }
    if (projection_option == 0) out_projection = in_projection;
    else out_projection = projection_option_alternatives[projection_option];
    if ( wcs_astro_test_if_glon (in_ap, in_dim_h->name) ||
	 wcs_astro_test_if_glat (in_ap, in_dim_h->name) ) galactic = TRUE;
    for (count = 0; count < num_dim; ++count)
	dim_lengths[count] = iarray_dim_length (in_arr, count);
    sprintf (lon, "%s---", galactic ? "GLON" : "RA");
    strcpy (lon + 5, out_projection);
    sprintf (lat, "%s--", galactic ? "GLAT" : "DEC");
    strcpy (lat + 5, out_projection);
    dim_names[dim_h_index] = lon;
    dim_names[dim_v_index] = lat;
    if (num_dim == 3) dim_names[0] = iarray_dim_name (in_arr, 0);
    /*  Find the corners of the input image in world co-ordinates  */
    x[0] = 0.0;
    y[0] = 0.0;
    x[1] = in_dim_h->length - 1;
    y[1] = 0.0;
    x[2] = in_dim_h->length - 1;
    y[2] = in_dim_v->length - 1;
    x[3] = 0.0;
    y[3] = in_dim_v->length - 1;
    wcs_astro_transform3 (in_ap, FALSE, 4,
			  in_dim_h->name, x, FALSE,
			  in_dim_v->name, y, FALSE,
			  NULL, NULL, FALSE,
			  0, NULL, NULL);
    if ( wcs_astro_test_if_lon (in_ap, in_dim_h->name) )
    {
	m_copy (lon_coords, x, sizeof x);
	m_copy (lat_coords, y, sizeof y);
	in_dim_lon = in_dim_h;
	in_dim_lat = in_dim_v;
    }
    else
    {
	m_copy (lon_coords, y, sizeof y);
	m_copy (lat_coords, x, sizeof x);
	in_dim_lon = in_dim_v;
	in_dim_lat = in_dim_h;
    }
    for (count = 0; count < 4; ++count)
    {
	if (lon_coords[count] < lon_min) lon_min = lon_coords[count];
	if (lon_coords[count] > lon_max) lon_max = lon_coords[count];
	if (lat_coords[count] < lat_min) lat_min = lat_coords[count];
	if (lat_coords[count] > lat_max) lat_max = lat_coords[count];
    }
    /*  Compute header information  */
    /*  Output reference pixel is always centre of rotation/image  */
    out_crval1 = (lon_min + lon_max) * 0.5;
    out_crval2 = (lat_min + lat_max) * 0.5;
    /*  Select two points at same latitude to determine longitude increment  */
    x[0] = out_crval1 - (lon_max - lon_min) * 0.25;
    y[0] = out_crval2;
    x[1] = out_crval1 + (lon_max - lon_min) * 0.25;
    y[1] = out_crval2;
    wcs_astro_transform3 (in_ap, FALSE, 2,
			  in_dim_lon->name, x, TRUE,
			  in_dim_lat->name, y, TRUE,
			  NULL, NULL, TRUE,
			  0, NULL, NULL);
    out_cdelt1 = 0.5 * (lon_max - lon_min) /
	sqrt ( (x[1] - x[0]) * (x[1] - x[0]) + (y[1] - y[0]) * (y[1] - y[0]) );
    if (strcmp (in_projection, "FLT") != 0)
	out_cdelt1 *= cos (out_crval2 * PION180);
    /*  Select two points at same longitude to determine latitude increment  */
    x[0] = out_crval1;
    y[0] = out_crval2 - (lat_max - lat_min) * 0.25;
    x[1] = out_crval1;
    y[1] = out_crval2 + (lat_max - lat_min) * 0.25;
    wcs_astro_transform3 (in_ap, FALSE, 2,
			  in_dim_lon->name, x, TRUE,
			  in_dim_lat->name, y, TRUE,
			  NULL, NULL, TRUE,
			  0, NULL, NULL);
    out_cdelt2 = 0.5 * (lat_max - lat_min) /
	sqrt ( (x[1] - x[0]) * (x[1] - x[0]) + (y[1] - y[0]) * (y[1] - y[0]) );
    switch (grid_option)
    {
      case GRID_RESCALE:
	out_cdelt1 *= sqrt (2.0);
	out_cdelt2 *= sqrt (2.0);
	break;
      case GRID_PAD:
	dim_lengths[dim_h_index] = (sqrt (2.0) * dim_lengths[dim_h_index]);
	dim_lengths[dim_v_index] = (sqrt (2.0) * dim_lengths[dim_v_index]);
	break;
      case GRID_TRUNCATE:
	break;
    }
    fprintf (stderr, "creating...");
    if ( ( out_arr = iarray_create2 (K_FLOAT, num_dim, dim_names, dim_lengths,
				     iarray_value_name (in_arr), NULL, FALSE) )
	 == NULL )
    {
	fprintf (stderr, "Error creating output array\n");
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    fprintf (stderr, "\tfilling...");
    switch (background_option)
    {
      case BACKGROUND_BLANK:
	iarray_fill_float (out_arr, TOOBIG);
	break;
      case BACKGROUND_CONSTANT:
	iarray_fill_float (out_arr, background_constant);
	break;
      default:
	fprintf (stderr, "Unsupported boundary option: %u\n", background_option);
	fprintf (stderr, "Using random values\n");
	break;
    }
    fprintf (stderr, "\tfilled\n");
    if (num_dim == 3)
    {
	iarray_get_world_coords (in_arr, 0, &tmp1, &tmp2);
	iarray_set_world_coords (out_arr, 0, tmp1, tmp2);
    }
    if ( !iarray_put_named_string (out_arr, "CTYPE1", lon) )
    {
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    if ( !iarray_put_double (out_arr, "CRVAL1", out_crval1) )
    {
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    if ( !iarray_put_double (out_arr, "CRPIX1",
			     (double) dim_lengths[dim_h_index] * 0.5) )
    {
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    if ( !iarray_put_double (out_arr, "CDELT1", -out_cdelt1) )
    {
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    if ( !iarray_put_named_string (out_arr, "CTYPE2", lat) )
    {
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    if ( !iarray_put_double (out_arr, "CRVAL2", out_crval2) )
    {
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    if ( !iarray_put_double (out_arr, "CRPIX2",
			     (double) dim_lengths[dim_v_index] * 0.5) )
    {
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    if ( !iarray_put_double (out_arr, "CDELT2", out_cdelt2) )
    {
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    if (rotation != 0.0)
    {
	if ( !iarray_put_double (out_arr, "CROTA2", rotation) )
	{
	    iarray_dealloc (in_arr);
	    iarray_dealloc (out_arr);
	    return (FALSE);
	}
    }
    if ( !copy_header_info (out_arr, in_arr, 3) )
    {
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    if ( !ds_history_copy ( iarray_multi_array (out_arr),
			    iarray_multi_array (in_arr) ) )
    {
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    if ( !panel_put_history_with_stack (iarray_multi_array (out_arr), TRUE) )
    {
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    sprintf (txt, "%s: infile: %s  outfile: %s",
	     module_name, infile, outfile);
    ds_history_append_string (iarray_multi_array (out_arr), txt, TRUE);
    if ( ( out_ap = iarray_get_associated_wcs_astro (out_arr) ) == NULL )
    {
	fprintf (stderr, "Error creating KwcsAstro object for output\n");
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    if (num_dim == 3)
	ok = iarray_regrid_3D (out_arr, out_ap, in_arr, in_ap,
			       sample_option_values[sample_option],
			       &min, &max);
    else ok = iarray_regrid_2D (out_arr, out_ap, in_arr, in_ap,
				sample_option_values[sample_option],
				&min, &max);
    if (ok) foreign_guess_write (outfile, iarray_multi_array (out_arr),
				 FA_GUESS_WRITE_END);
    iarray_dealloc (in_arr);
    iarray_dealloc (out_arr);
    return (ok);
}   /*  End Function do_rotate  */

static flag copy_header_info (iarray out, iarray in, unsigned int min_naxis)
/*  [SUMMARY] Copy some header information from one array to another.
    <out> The output array.
    <in> The input array.
    <min_naxis> The minimum value for an axis index if related information is
    to be copied.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int count, elem_count;
    packet_desc *pack_desc;
    char *packet;
    multi_array *multi_desc = iarray_multi_array (in);
    static char function_name[] = "copy_header_info";

    pack_desc = multi_desc->headers[iarray_multi_index (in)];
    packet = multi_desc->data[iarray_multi_index (in)];
    for (elem_count = 0; elem_count < pack_desc->num_elements; ++elem_count)
    {
	flag copy = FALSE;
	char *name = pack_desc->element_desc[elem_count];

	if ( !ds_element_is_named (pack_desc->element_types[elem_count]) )
	    continue;
	/*  Check if it's in the list of copyable keywords  */
	for (count = 0; keywords[count] != NULL; ++count)
	{
	    if (strcmp (name, keywords[count]) != 0) continue;
	    copy = TRUE;
	    break;
	}
	for (count = 0; !copy && (axis_keywords[count] != NULL); ++count)
	{
	    int len = strlen (axis_keywords[count]);

	    if (strncmp (name, axis_keywords[count], len) != 0) continue;
	    if ( !isdigit (name[len]) ) continue;
	    if (strtoul (name + len, NULL, 10) < min_naxis) continue;
	    copy = TRUE;
	    break;
	}
	if (!copy) continue;
	if ( !iarray_copy_named_element (out, in, name, FALSE, FALSE, TRUE) )
	{
	    fprintf (stderr, "%s: Failed to copy header keyword %s\n",
		     function_name, name);
	    return (FALSE);
	}
    }
    return (TRUE);
}   /*  End Function copy_header_info  */
