/*  event.c

    Event file for  kpolar  (X11 cartesian->polar tool for Karma).

    Copyright (C) 1997-2000  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable interactive generation of an astronomical
    co-ordinate system on a polar image, using a cartesian image.
    This module runs on an X11 server.


    Written by      Richard Gooch   6-AUG-1997

    Updated by      Richard Gooch   24-SEP-1997: Added co-ordinate information.

    Updated by      Richard Gooch   17-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   14-APR-1999: Trap small radii.

    Updated by      Richard Gooch   21-APR-1999: Added call to new function
  <update_polar_dataclip>.

    Last updated by Richard Gooch   12-JUL-2000: Added #include <karma_kwin.h>
  and <karma_canvas.h>.


*/
#include <stdio.h>
#include <math.h>
#include <karma.h>
#define NEW_WIN_SCALE
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_canvas.h>
#include <karma_iarray.h>
#include <karma_kwin.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_a.h>
#include <karma_m.h>
#include "kpolar.h"


/*  Private data  */
static double centre_wx = TOOBIG;
static double centre_wy = TOOBIG;
static double major_wx = TOOBIG;
static double major_wy = TOOBIG;
static double minor_wx = TOOBIG;
static double minor_wy = TOOBIG;
static double rotation = TOOBIG;  /*  In pixel canvas  */


/*  Private functions  */


/*  Public functions follow  */

flag geom_func (KPixCanvasEditorContext context, KPixCanvas canvas,
		unsigned int object_type, unsigned int num_coords,
		CONST int *x, CONST int *y,
		unsigned int raw_event_code, void *raw_e_info, void **f_info)
/*  [SUMMARY] Process a geometry editor event on a pixel canvas.
    <context> The KPixCanvasEditorContext.
    <canvas> The pixel canvas on which the event occurred.
    <object_type> The type of the geometric object.
    <num_coords> The number of co-ordinates in the geometric object.
    <x> The array of horizontal co-ordinates for the geometric object.
    <y> The array of vertical co-ordinates for the geometric object.
    <raw_event_code> The raw arbitrary event code.
    <raw_e_info> A pointer to the raw arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KWorldCanvas worldcanvas = (KWorldCanvas) *f_info;
    unsigned int xlen, ylen;
    unsigned long cx, cy, rx, ry;
    double min, max;
    double centre_lx = TOOBIG;
    double centre_ly = TOOBIG;
    double major_lx = TOOBIG;
    double major_ly = TOOBIG;
    double minor_lx = TOOBIG;
    double minor_ly = TOOBIG;
    dim_desc *in_hdim, *in_vdim;
    extern KWorldCanvas pol_canvas;
    extern iarray car_array, pol_array;
    static ViewableImage vimage = NULL;
    static char function_name[] = "geom_func";

    if (raw_event_code != K_CANVAS_EVENT_RIGHT_MOUSE_RELEASE) return (FALSE);
    if (car_array == NULL) return (FALSE);
    centre_lx = x[0];
    centre_ly = y[0];
    canvas_convert_to_canvas_coords (worldcanvas, FALSE, 1,
				     &centre_lx, &centre_ly,
				     &centre_lx, &centre_ly,
				     &centre_wx, &centre_wy);
    major_lx = x[0] + x[1];
    major_ly = y[0];
    canvas_convert_to_canvas_coords (worldcanvas, FALSE, 1,
				     &major_lx, &major_ly,
				     &major_lx, &major_ly,
				     &major_wx, &major_wy);
    minor_lx = x[0];
    minor_ly = y[0] + y[1];
    canvas_convert_to_canvas_coords (worldcanvas, FALSE, 1,
				     &minor_lx, &minor_ly,
				     &minor_lx, &minor_ly,
				     &minor_wx, &minor_wy);
    rotation = (double) x[2] / 64.0;
    canvas_resize (worldcanvas, NULL, FALSE);
    /*  Convert linear co-ordinates to dimension indices  */
    in_hdim = iarray_get_dim_desc (car_array, 1);
    in_vdim = iarray_get_dim_desc (car_array, 0);
    cx = ds_get_coord_num (in_hdim, centre_lx, SEARCH_BIAS_CLOSEST);
    cy = ds_get_coord_num (in_vdim, centre_ly, SEARCH_BIAS_CLOSEST);
    rx = ds_get_coord_num (in_hdim, major_lx, SEARCH_BIAS_CLOSEST) - cx;
    ry = cy - ds_get_coord_num (in_vdim, minor_ly, SEARCH_BIAS_CLOSEST);
    if ( (rx < 5) || (ry < 5) )
    {
	fprintf (stderr, "Radius (%lu, %lu) too small\n", rx, ry);
	return (TRUE);
    }
#if 0
    xlen = (rx > ry) ? rx : ry;
    ylen = TWOPI * (double) xlen;
    xlen *= 3;
#endif
    xlen = 512;
    ylen = 512;
    if (pol_array != NULL)
    {
	if ( (iarray_dim_length (pol_array, 1) != xlen) ||
	     (iarray_dim_length (pol_array, 0) != ylen) )
	{
	    iarray_dealloc (pol_array);
	    viewimg_destroy (vimage);
	    pol_array = NULL;
	    vimage = NULL;
	}
    }
    if (pol_array == NULL)
    {
	if ( ( pol_array = iarray_create_2D (ylen, xlen, K_FLOAT) ) == NULL )
	{
	    m_abort (function_name, "polar image");
	}
    }
    iarray_set_world_coords (pol_array, 1, 0.0, rx);
    iarray_set_dim_name (pol_array, 1, "Radius (pixels)", TRUE);
    iarray_set_world_coords (pol_array, 0, 0.0, 360.0);
    iarray_set_dim_name (pol_array, 0, "ANGLE", TRUE);
    iarray_rect_to_polar (pol_array, car_array, cx, cy, rx, ry, rotation);
    iarray_min_max (pol_array, CONV_CtoR_REAL, &min, &max);
    canvas_set_attributes (pol_canvas,
			   CANVAS_ATT_VALUE_MIN, min,
			   CANVAS_ATT_VALUE_MAX, max,
			   CANVAS_ATT_END);
    if (vimage == NULL)
    {
	if ( ( vimage = viewimg_create_from_iarray (pol_canvas, pol_array,
						    FALSE) ) == NULL )
	{
	    m_abort (function_name, "polar vimage");
	}
	viewimg_make_active (vimage);
    }
    else viewimg_register_data_change (vimage);
    update_polar_dataclip (pol_array, min, max);
    return (TRUE);
}   /*  End Function geom_func  */

void car_refresh_func (KWorldCanvas canvas, int width, int height,
		       struct win_scale_type *win_scale,
		       Kcolourmap cmap, flag cmap_resize, void **info,
		       PostScriptPage pspage,
		       unsigned int num_areas,
		       KPixCanvasRefreshArea *areas,
		       flag *honoured_areas)
/*  [SUMMARY] Refresh event callback.
    [PURPOSE] This routine is a refresh event consumer for a world canvas.
    <canvas> The world canvas being refreshed.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The colourmap associated with the canvas.
    <cmap_resize> TRUE if the refresh function was called as a result of a
    colourmap resize, else FALSE.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    double cx, cy, mjx, mjy, mix, miy;
    extern unsigned long car_pixel_value;

    if ( DOUBLE_IS_BLANK (centre_wx) ) return;
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
				       &centre_wx, &centre_wy, &cx, &cy);
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
				       &major_wx, &major_wy, &mjx, &mjy);
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
				       &minor_wx, &minor_wy, &mix, &miy);
    kwin_draw_arc (canvas_get_pixcanvas (canvas),
		   cx, cy, mjx - cx, miy - cy, 0.0, 360.0, rotation,
		   car_pixel_value, FALSE);
}   /*  End Function car_refresh_func  */


/*  Private functions follow  */
