/*  generic.c

    Generic file for  kpolar  (X11 co-ordinate generator tool for Karma).

    Copyright (C) 1997-2000  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable interactive generation of an astronomical
    co-ordinate system on a target image, using a reference image.
    This module runs on an X11 server.


    Written by      Richard Gooch   5-AUG-1997: Copied from koords/generic.c

    Updated by      Richard Gooch   5-SEP-1998: Removed comms. setup code.

    Last updated by Richard Gooch   12-JUL-2000: Added
  #include <karma_canvas.h>


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <errno.h>
#include <karma.h>
#include <X11/Xlib.h>
#include <karma_viewimg.h>
#include <karma_canvas.h>
#include <karma_iarray.h>
#include <karma_conn.h>
#include <karma_ds.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_r.h>
#include "kpolar.h"


/*  Private functions  */
STATIC_FUNCTION (void destroy_all_vimages,
		 (ViewableImage *image, ViewableImage *magnified_image) );


/*  Private data  */
static char *keywords[] =
{
    "OBJECT",
    "DATE-OBS",
    "EPOCH",
    "BUNIT",
    "TELESCOP",
    "INSTRUME",
    "EQUINOX",
    "BTYPE",
    "BMAJ",
    "BMIN",
    "BPA",
    "RESTFREQ",
    "VELREF",
    "PBFWHM",
    NULL
};


/*  Public functions follow  */

flag process_image (multi_array *multi_desc, iarray *array, KwcsAstro *ap,
		    KWorldCanvas pseudo_canvas, KWorldCanvas mag_pseudo_canvas,
		    ViewableImage *image, ViewableImage *magnified_image,
		    double *min, double *max)
/*  [PURPOSE] This routine will load a file and display it.
    <multi_desc> The multi_array descriptor.
    <array> The image array will be written here.
    <ap> The KwcsAstro object will be written here.
    <pseudo_canvas> The PseudoColour canvas.
    <image> If an image is loaded the ViewableImage is written here. If no
    image is loaded, NULL is written here. The value written here must be
    preserved between calls.
    <magnified_image> If an image is loaded the magnified ViewableImage is
    written here. If no image is loaded, NULL is written here. The value
    written here must be preserved between calls.
    <min> The minimum image value is written here.
    <max> The maximum image value is written here.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    /*static char function_name[] = "process_image";*/

    if ( !iarray_create_and_setup (array, multi_desc, TRUE, 2, NONE,
				   min, max, TRUE, ap) )
    {
	return (FALSE);
    }
    destroy_all_vimages (image, magnified_image);
    if ( ( *image = viewimg_create_from_iarray (pseudo_canvas, *array,
						FALSE) ) == NULL )
    {
	fprintf (stderr, "Error getting ViewableImage from Iarray\n");
	iarray_dealloc (*array);
	*array = NULL;
	return (FALSE);
    }
    if ( ( *magnified_image =
	   viewimg_create_from_iarray (mag_pseudo_canvas, *array,
				       FALSE) ) == NULL )
    {
	fprintf (stderr, "Error getting ViewableImage from Iarray\n");
	destroy_all_vimages (image, magnified_image);
	iarray_dealloc (*array);
	*array = NULL;
	return (FALSE);
    }
    canvas_set_attributes (pseudo_canvas,
			   CANVAS_ATT_VALUE_MIN, *min,
			   CANVAS_ATT_VALUE_MAX, *max,
			   CANVAS_ATT_END);
    canvas_set_attributes (mag_pseudo_canvas,
			   CANVAS_ATT_VALUE_MIN, *min,
			   CANVAS_ATT_VALUE_MAX, *max,
			   CANVAS_ATT_END);
    viewimg_make_active (*image);
    viewimg_make_active (*magnified_image);
    return (TRUE);
}   /*  End Function process_image  */

flag copy_header_info (iarray out, iarray in)
/*  [SUMMARY] Copy some header information from one array to another.
    <out> The output array.
    <in> The input array.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int count;
    static char function_name[] = "copy_header_info";

    for (count = 0; keywords[count] != NULL; ++count)
    {
	if ( !iarray_copy_named_element (out, in, keywords[count],
					 FALSE, FALSE, TRUE) )
	{
	    fprintf (stderr, "%s: Failed to copy header keyword %s\n",
		     function_name, keywords[count]);
	    return (FALSE);
	}
    }
    return (TRUE);
}   /*  End Function copy_header_info  */


/*  Private functions follow  */

static void destroy_all_vimages (ViewableImage *image,
				 ViewableImage *magnified_image)
/*  [PURPOSE] This routine will destroy all ViewableImages and clear the
    appropriate pointers.
    [RETURNS] Nothing.
*/
{
    if (*image != NULL) viewimg_destroy (*image);
    *image = NULL;
    if (*magnified_image != NULL) viewimg_destroy (*magnified_image);
    *magnified_image = NULL;
}   /*  End Function destroy_all_vimages  */
