/*  kpvget.c

    Source file for  kpvget  (command-line PV slice extraction module).

    Copyright (C) 2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will read in a data file and allows extraction of a PV
    slice.


    Written by      Richard Gooch   5-JAN-2002: Moved disabled parts of pvget.c

    Last updated by Richard Gooch   7-JAN-2002: Fixed world co-ordinates in
  generated header.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_module.h>
#include <karma_iarray.h>
#include <karma_panel.h>
#include <karma_dsxfr.h>
#include <karma_wcs.h>
#include <karma_ex.h>
#include <karma_ds.h>
#include <karma_im.h>
#include <karma_ch.h>
#include <karma_a.h>
#include <karma_s.h>
#include <karma_m.h>
#include <karma_n.h>
#include <karma_r.h>
#include <k_version.h>
#include "pvget.h"


#define VERSION "0.1"


/*STATIC_FUNCTION (flag kpvget, (CONST char *command, FILE *fp) );*/
STATIC_FUNCTION (flag fix_slice_header,
		 (iarray cube_arr, iarray image,
		  unsigned int num_points,
		  double slice_centre_ra, double slice_centre_dec,
		  double slice_position_angle) );


/*  Private data  */
static char *keywords[] =
{
    "OBJECT",
    "DATE-OBS",
    "EPOCH",
    "BUNIT",
    "TELESCOP",
    "INSTRUME",
    "EQUINOX",
    "BTYPE",
    "BMAJ",
    "BMIN",
    "BPA",
    "RESTFREQ",
    "VELREF",
    "PBFWHM",
    NULL
};

#if 0
int main (int argc, char **argv)
{
    KControlPanel panel;
    static char function_name[] = "main";

    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
	m_abort (function_name, "control panel");
    }
    panel_add_item (panel, "transmit_rate","maximum transmit rate (bytes/sec)",
		    K_ULONG, (char *) &transmit_rate,
		    PIA_END);
    panel_add_item (panel, "do_spray", "bytes [packet_size]",
		    PIT_FUNCTION, (void *) do_spray,
		    PIA_END);
    panel_add_item (panel, "synchronous_spray", "wait for sync at end",
		    PIT_FLAG,
		    &synchronous_spray,
		    PIA_END);
    panel_add_item (panel, "spray_data", "choice of spray data",
		    PIT_CHOICE_INDEX, (char *) &spray_type,
		    PIA_NUM_CHOICE_STRINGS, NUM_SPRAY_DATA_CHOICES,
		    PIA_CHOICE_STRINGS, spray_data_choices,
		    PIA_END);
    panel_add_item (panel, "ping", "ping specified number of times",
		    PIT_FUNCTION, (void *) ping,
		    PIA_END);
    panel_add_item (panel, "pattern", "spray pattern", K_UBYTE, spray_pattern,
		    PIA_ARRAY_MIN_LENGTH, 1,
		    PIA_ARRAY_MAX_LENGTH, MAX_PATTERN_LENGTH,
		    PIA_ARRAY_LENGTH, (char *) &pattern_length,
		    PIA_END);
    panel_add_item (panel, "num_iterations", "number", K_INT, &num_iterations,
		    PIA_END);
    panel_add_item (panel, "animate", "flag", PIT_FLAG, &animate_flag,
		    PIA_END);
    panel_push_onto_stack (panel);
    im_register_lib_version (KARMA_VERSION);
    panel_set_extra_comment_string (panel, "\nGeneral usage: infile");
    module_run (argc, argv, "kpvget", VERSION, kpvget, -1, 0, FALSE);
    return (RV_OK);
}   /*  End Function main   */

static flag kpvget (CONST char *command, FILE *fp)
{
    char filename[STRING_LENGTH+1];
    extern int animate_flag;

    for ( ; command != NULL; command = ex_word_skip (command) )
    {
	sscanf (command, "%s", filename);
	if (animate_flag) run_animate_file (filename);
	else transmit_arrayfile (filename);
    }
    return (TRUE);
}   /*  End Function kpvget  */
#else
int main (int argc, char **argv)
{
#define MAXNB 5
    multi_array  *multi_desc;
    unsigned int ftype;
    iarray       cube, pvmap;
    double       fwhm;
    double       *x_locus, *y_locus;
    char         *cubefile, *pvmapfile;
    int          x1, y1, x2, y2;
    double       ra[3],dec[3],ra_off,dec_off,first_coord,last_coord,position_angle;
    dim_desc     *xdim, *ydim;
    Channel      fits_ch;
    KwcsAstro    cube_ap;
    int xpix,ypix,zpix,lpix;

    if (argc==7 || argc==8 ) {
	cubefile  = argv[1];
	pvmapfile = argv[2];
	x1        = atoi (argv[3]);
	y1        = atoi (argv[4]);
	x2        = atoi (argv[5]);
	y2        = atoi (argv[6]);
	fwhm      = 0.0;
	if (argc == 8) {fwhm = strtod (argv[7], (char **) NULL); }
    }else{
	printf("Usage: <cube> <o/p slice> <x1> <y1> <x2> <y2>\n");
	printf("       All inputs must be in PIXEL units.\n");
	exit (1);
    }

    /* Read in any kind of file; FITS, MIRIAD, KARMA */
    if ( ( multi_desc = foreign_guess_and_read (cubefile, K_CH_MAP_LOCAL,
						FALSE, &ftype,
						FA_GUESS_READ_END) ) == NULL )
    {
	fprintf (stderr, "Error reading file: \"%s\"\n", cubefile);
	exit(1);
    }
    cube_ap = wcs_astro_get_associated (multi_desc, 0);
    /* if (wcs_astro_test_radec(cube_ap)) {
       fprintf(stderr," cube has ra, dec ok\n");
       }
       if (wcs_astro_test_velocity(cube_ap)) {
       fprintf(stderr," cube has velocity ok\n");
       }
       */
    /* extract cube from karma general data structure */
    if ( (cube = iarray_get_from_multi_array(multi_desc,NULL,3,NULL,NULL))
	 == NULL )
    {
        fprintf (stderr, "Could not find file %s\n",cubefile);
        exit(1);
    }

    /* check the inputs */
    /* don't! check_bounds makes sure the endpoints are legal
       if ( x1 < 0 || x2 < 0 || y1 < 0 || y2 < 0 ) {
       fprintf (stderr, "Negative pixel coordinates not allowed.\n");
       fprintf (stderr, "Please use absolute coordinates.\n");
       exit(1);
       }
       */
    /* extract useful numbers from the cube */
    zpix = iarray_dim_length (cube, 0);
    ypix = iarray_dim_length (cube, 1);
    xpix = iarray_dim_length (cube, 2);
    if ( !(xpix*ypix*zpix > 0) ) {
	fprintf (stderr, "PVGET: problem with input cube dimensions\n");
	exit(1);
    }
 
    /* get the slice locus */
    lpix = get_slice(cube,x1,y1,x2,y2,&x_locus,&y_locus);

    fprintf (stderr, "got the slice locus, %d points\n",lpix);

    if (lpix < 1) exit (1);
    /* extract slice from cube */
    pvmap = pvslice(cube,lpix,x_locus,y_locus);
    fprintf (stderr, "slice extracted, fixing headers...\n");

    if ( wcs_astro_test_radec(cube_ap) ) {
	/* get an approximate estimate of the offset in degrees & the p.a.,
	   using cube linear world coordinates. How do I xform to proper world? */
	/* there's something wrong with getting the RA, DECs */
	xdim   = iarray_get_dim_desc (cube, 2);
	ydim   = iarray_get_dim_desc (cube, 1);
	ra[0]  = ds_get_coordinate (xdim, x_locus[0]);
	dec[0] = ds_get_coordinate (ydim, y_locus[0]);
	ra[1]  = ds_get_coordinate (xdim, x_locus[(lpix-1)]);
	dec[1] = ds_get_coordinate (ydim, y_locus[(lpix-1)]);
	ra[2]  = 0.5 * (ra[0] + ra[1]);
	dec[2] = 0.5 * (dec[0] + dec[1]);
	wcs_astro_transform (cube_ap, FALSE, 3,
			     ra, FALSE, dec, FALSE, NULL, FALSE,
			     0, NULL, NULL);
	fprintf(stderr, "loc  %f, %f;  %f, %f\n",
		x_locus[0],y_locus[0],x_locus[(lpix-1)],y_locus[(lpix-1)]);
	fprintf(stderr, "ra,dec  %f, %f;  %f, %f\n",ra[0],dec[0],ra[1],dec[1]);
	ra_off = (ra[0] - ra[2]) * cos (dec[2] * PION180);
	dec_off = dec[1] - dec[2];
	position_angle = atan2 (-ra_off, dec_off) / PION180 - 90.0;
	if (position_angle < 0 ) position_angle += 360.0;
	last_coord  = sqrt (ra_off * ra_off + dec_off * dec_off);
	first_coord = -last_coord;
      
	iarray_set_world_coords(pvmap, 1, first_coord, last_coord);
	/* in dimension descriptor, fix up `reference' field */

	if ( ! fix_slice_header(cube, pvmap, lpix, ra[2], dec[2], position_angle) ) {
	    /* do nothing fprintf(stderr, "Some problem with headers\n"); */
	}
    }else{
	fprintf(stderr,"No world coordinate info in cube. O/p will be in pixel coords\n");
    }
    m_free( (char *) x_locus);
    m_free( (char *) y_locus);

    fprintf (stderr, "writing...\n");
    /* Write pvmap to disk, as KARMA */
    /* iarray_write (pvmap, pvmapfile); */
    /* FITS standard axis descriptions and units */
    /* clean up */
    if ( ( fits_ch = ch_open_file(pvmapfile, "w") ) == NULL ) {
	fprintf(stderr, "Error opening output file '%s'\n", pvmapfile);
	ch_close(fits_ch);
	unlink(pvmapfile);
	exit(1);
    }
    fprintf(stderr,"opened file ok\n");
    if ( !foreign_fits_write_iarray(fits_ch, pvmap, FA_FITS_WRITE_END) ) {
	fprintf(stderr,"Failed to write file '%s'\n", pvmapfile );
	unlink(pvmapfile);
	exit(1);
    }
    ch_close(fits_ch);
    fprintf(stderr,"File '%s' sucessfully written\n", pvmapfile );

    iarray_dealloc(pvmap);
    iarray_dealloc(cube);

    exit(0);
    return (RV_OK);
}
#endif

static flag fix_slice_header (iarray cube_arr, iarray image,
			      unsigned int num_points,
			      double slice_centre_ra, double slice_centre_dec,
			      double slice_position_angle)
/*  [SUMMARY] Add relevant info from the cube to the slice image
    [RETURNS] TRUE on success, FALSE if there was a problem.
*/
{
    flag           ok = TRUE;
    char           tmp_str[STRING_LENGTH];
    char           hist_str[STRING_LENGTH];
    unsigned int   i;
    static char    function_name[] = "fix_slice_header";

    /* NB:
       EPOCH is used by FITS to mean EQUINOX, ie the equinox the coords
       are referred to. The epoch of observation is stored in DATE-OBS.
       The revised FITS standard ecourages EQUINOX & deprecates EPOCH, but
       I don't attempt to fix headers that use the old way.

      */
    for (i = 0; keywords[i] != NULL; ++i)
    {
	if ( !iarray_copy_named_element (image, cube_arr, keywords[i],
					 FALSE, FALSE, TRUE) )
	{
	    fprintf (stderr, "%s: Failed to copy header keyword %s\n",
		     function_name, keywords[i]);
	    ok = FALSE;
	}
    }
    /* add history */
    sprintf (hist_str, "kgetpvslice: Version %s\n", VERSION);
    if ( !iarray_append_history_string (image, hist_str, TRUE) ) ok = FALSE;

    wcs_astro_format_hms (tmp_str, slice_centre_ra);
    sprintf (hist_str, "kgetpvslice: Slice Centre RA %s\n", tmp_str);
    if ( !iarray_append_history_string (image, hist_str, TRUE) ) ok = FALSE;

    wcs_astro_format_dms (tmp_str, slice_centre_dec);
    sprintf (hist_str, "kgetpvslice: Slice Centre DEC %s\n", tmp_str);
    if ( !iarray_append_history_string (image, hist_str, TRUE) ) ok = FALSE;

    wcs_astro_format_dms (tmp_str, slice_position_angle);
    sprintf (hist_str, "kgetpvslice: Slice PA %s\n", tmp_str);
    if ( !iarray_append_history_string (image, hist_str, TRUE) ) ok = FALSE;

    return (ok);
}   /*  End Function fix_slice_header  */
