/*  event.c

    Event file for  kshell  (X11 ellipse integrator tool for Karma).

    Copyright (C) 1996-2001  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable interactive integration along the
    circumference of concentric ellipses.
    This module runs on an X11 server.


    Written by      Richard Gooch   24-SEP-1996: Copied from kpvslice module.

    Updated by      Richard Gooch   26-SEP-1996: Setup frequency/velocity
  co-ordinates.

    Updated by      Richard Gooch   28-SEP-1996: Setup ANGLE co-ordinates.

    Updated by      Richard Gooch   2-OCT-1996: Made <compute_ellipses>
  tolerate missing "CTYPEn" keyword for cube z axis and more robust trapping of
  too small radii. Don't apply cos (dec) correction to horizontal axis if a
  projection object is not available.

    Updated by      Richard Gooch   4-MAR-1997: Automatically recompute
  projection upon cube reload.

    Updated by      Richard Gooch   13-APR-1997: Increased number of radii bins
  by 10% in <integrate_ellipses> in order to prevent aliasing.

    Updated by      Richard Gooch   12-NOV-1997: Improved boundary test in
  <integrate_ellipses> so that roundoff errors would not prevent last radial
  bin from being computed.

    Updated by      Richard Gooch   7-DEC-1997: Removed code to set pan centre
  for magnifier as it is now done automatically by the MagnifierPopup widget.

    Updated by      Richard Gooch   17-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   26-MAY-1998: Switch to <wcs_astro_format>.

    Updated by      Richard Gooch   12-AUG-1998: Update cursor position on all
  canvas events.

    Updated by      Richard Gooch   8-APR-1999: Removed unnecessary casts.

    Updated by      Richard Gooch   10-APR-1999: Added zero clipping.

    Updated by      Richard Gooch   13-APR-1999: Created function
  <transform_position_func>.

    Updated by      Richard Gooch   20-APR-1999: Threaded function
  <integrate_ellipses>.

    Updated by      Richard Gooch   22-APR-1999: Preserve Karma co-ordinate
  information for cube Z axis if FITS data not available.

    Updated by      Richard Gooch   9-SEP-1999: Do not create the KwcsAstro
  object for the image: leave it to the <viewimg> package. Fixed bug in
  <main_position_func> where first undraw would wipe image.

    Updated by      Richard Gooch   12-JUL-2000: Added #include <karma_kwin.h>.

    Updated by      Richard Gooch   14-JUN-2001: Trap for computed images with
  zero range. Add module version and circle centre to image history.

    Last updated by Richard Gooch   6-DEC-2001: Took account of interface
  change to <wcs_astro_transform*>.


*/
#include <unistd.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <errno.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_iarray.h>
#include <karma_canvas.h>
#include <karma_kwin.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_mt.h>
#include <karma_m.h>
#include <karma_a.h>
#include "kshell.h"


#define ELLIPSE_RADIUS  5
#define NUM_SEGMENTS    60

typedef struct
{
    int xstart, xstop;
    int num_radii_bins;
    float rad_scale;
    double centre_x, centre_y;
    double xscale;
    iarray count_arr;
} info_type;


/*  Private functions  */
STATIC_FUNCTION (void compute_area_from_ellipse,
		 (KWorldCanvas canvas, KPixCanvasRefreshArea *area,
		  double centre_x, double centre_y, double radius) );
STATIC_FUNCTION (unsigned long get_colour,
		 (KWorldCanvas canvas, CONST char *colourname) );
STATIC_FUNCTION (void integrate_ellipses,
		 (KWorldCanvas canvas, double centre_x, double centre_y,
		  double radius) );
STATIC_FUNCTION (void job_func,
		 (void *pool_info, void *call_info1, void *call_info2,
		  void *call_info3, void *call_info4, void *thread_info) );


/*  Private data  */
static iarray image_arr = NULL;


/*  Public functions follow  */

flag main_position_func (KWorldCanvas canvas, double x, double y,
			 unsigned int event_code, void *e_info,
			 void **f_info, double x_lin, double y_lin)
/*  [SUMMARY] Position event callback.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    [NOTE] These values will have been transformed by the registered
    transform function (see [<canvas_register_transform_func>]).
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KWorldCanvas aux_canvas = (KWorldCanvas) *f_info;
    KPixCanvasRefreshArea area;
    double tmp;
    char *xlabel = NULL;
    char *ylabel = NULL;
    char window_string[STRING_LENGTH];
    char x_str[STRING_LENGTH], y_str[STRING_LENGTH];
    extern KwcsAstro main_ap;
    extern iarray cube_arr;
    extern double centre_x, centre_y, radius;
    /*static char function_name[] = "main_position_func";*/

    m_clear (&area, sizeof area);
    if ( DOUBLE_IS_VALID (centre_x) && DOUBLE_IS_VALID (radius) )
    {
	/*  Compute area to be refreshed when old ellipse is to be erased  */
	compute_area_from_ellipse (canvas, &area, centre_x, centre_y, radius);
    }
    else
    {
	area.startx = 0;
	area.endx = -1;
	area.starty = 0;
	area.endy = -1;
	area.clear = FALSE;
    }
    canvas_get_specification (canvas, &xlabel, &ylabel, NULL, NULL, NULL);
    switch (event_code)
    {
      case K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK:
	/*  Erase the old ellipse  */
	centre_x = TOOBIG;
	viewimg_partial_refresh (canvas, 1, &area);
	centre_x = x;
	centre_y = y;
	return (TRUE);
	/*break;*/
      case K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE:
	tmp = x - centre_x;
	if ( wcs_astro_test_if_lon (main_ap, xlabel) )
	    tmp *= cos (centre_y * PION180);
	radius = sqrt ( tmp * tmp + (y - centre_y) * (y - centre_y) );
	break;
      default:
	return (FALSE);
	/*break;*/
    }
    /*  Force a refresh, which will in turn draw the ellipse  */
    viewimg_partial_refresh (canvas, 1, &area);
    if (event_code != K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE) return (TRUE);
    if (main_ap != NULL)
    {
	wcs_astro_format (main_ap, xlabel, x_str, centre_x);
	wcs_astro_format (main_ap, ylabel, y_str, centre_y);
	sprintf (window_string, "Centre: %s  %s", x_str, y_str);
	print_transform_info (window_string);
    }
    if (cube_arr == NULL) return (TRUE);
    integrate_ellipses (aux_canvas, centre_x, centre_y, radius);
    return (TRUE);
}   /*  End Function main_position_func  */

flag transform_position_func (KWorldCanvas canvas, double x, double y,
			      unsigned int event_code, void *e_info,
			      void **f_info, double x_lin, double y_lin,
			      int x_pix, int y_pix)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    <x_pix> The horizontal pixel co-ordinate.
    <y_pix> The vertical pixel co-ordinate.
    [NOTE] The world co-ordinate values will have been transformed by the
    registered transform function (see [<canvas_register_transform_func>]).
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    unsigned int count;
    double angle, ra, dec;
    double xfactor = 1.0;
    CONST char *string = e_info;
    extern KwcsAstro main_ap;
    extern double centre_x, centre_y;

    if (event_code != K_CANVAS_EVENT_PLAIN_KEY_PRESS) return (FALSE);
    if (strcmp (string, "c") != 0) return (FALSE);
    if (main_ap != NULL) xfactor = 1.0 / cos (centre_y * PION180);
    printf ("# RA  DEC  VEL\n");
    for (count = 0; count < NUM_SEGMENTS; ++count)
    {
	angle = TWOPI * (double) count / (double) NUM_SEGMENTS;
	ra = centre_x + cos (angle) * x * xfactor;
	dec = centre_y + sin (angle) * x;
	printf ("%.9g %.9g %.9g\t", ra, dec, y);
	angle = TWOPI * (double) (count + 1) / (double) NUM_SEGMENTS;
	ra = centre_x + cos (angle) * x * xfactor;
	dec = centre_y + sin (angle) * x;
	printf ("%.9g %.9g %.9g\n", ra, dec, y);
    }
    return (TRUE);
}   /*  End Function transform_position_func  */

void main_refresh_func (KWorldCanvas canvas, int width, int height,
			struct win_scale_type *win_scale,
			Kcolourmap cmap, flag cmap_resize, void **info,
			PostScriptPage pspage,
			unsigned int num_areas,
			KPixCanvasRefreshArea *areas, flag *honoured_areas)
/*  [SUMMARY] Refresh event callback.
    [PURPOSE] This routine is a refresh event consumer for a world canvas.
    <canvas> The world canvas being refreshed.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The colourmap associated with the canvas.
    <cmap_resize> TRUE if the refresh function was called as a result of a
    colourmap resize, else FALSE.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    unsigned long pixel_value;
    char *xlabel = NULL;
    char *ylabel = NULL;
    double px[2], py[2];
    extern KwcsAstro main_ap;
    extern double centre_x, centre_y, radius;

    if ( DOUBLE_IS_BLANK (centre_x) ) return;
    px[0] = centre_x;
    py[0] = centre_y;
    canvas_get_specification (canvas, &xlabel, &ylabel, NULL, NULL, NULL);
    if ( wcs_astro_test_if_lon (main_ap, xlabel) )
	px[1] = centre_x + radius * cos (centre_y * PION180);
    else
	px[1] = centre_x + radius;
    py[1] = centre_y + radius;
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 2, px, py, px,py);
    px[1] = fabs (px[1] - px[0]);
    py[1] = fabs (py[1] - py[0]);
    pixel_value = get_colour (canvas, "green");
    kwin_draw_ellipse (canvas_get_pixcanvas (canvas),
		       px[0], py[0], py[1], py[1], pixel_value);
}   /*  End Function main_refresh_func  */

void update_projection (KWorldCanvas canvas)
/*  [SUMMARY] Update the projection window.
    <canvas> The world canvas.
    [RETURNS] Nothing.
*/
{
    extern double centre_x, centre_y, radius;

    integrate_ellipses (canvas, centre_x, centre_y, radius);
}   /*  End Function update_projection  */


/*  Private functions follow  */

static void compute_area_from_ellipse (KWorldCanvas canvas,
				       KPixCanvasRefreshArea *area,
				       double centre_x, double centre_y,
				       double radius)
/*  [SUMMARY] Compute a pixel area from ellipse.
    <canvas> The world canvas.
    <area> The area will be written here.
    <centre_x> The horizontal world co-ordinate of the centre of the ellipse.
    <centre_y> The vertical world co-ordinate of the centre of the ellipse.
    <radius> The radius in world co-ordinates.
    [RETURNS] Nothing.
*/
{
    char *xlabel = NULL;
    char *ylabel = NULL;
    double px[2], py[2];
    extern KwcsAstro main_ap;

    px[0] = centre_x;
    py[0] = centre_y;
    canvas_get_specification (canvas, &xlabel, &ylabel, NULL, NULL, NULL);
    if ( wcs_astro_test_if_lon (main_ap, xlabel) )
	px[1] = centre_x + radius * cos (centre_y * PION180);
    else
	px[1] = centre_x + radius;
    py[1] = centre_y + radius;
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 2, px, py, px,py);
    px[1] = fabs (px[1] - px[0]);
    py[1] = fabs (py[1] - py[0]);
    area->startx = px[0] - py[1] - 1.0;
    area->endx = px[0] + py[1] + 1.0;
    area->starty = py[0] - py[1] - 1.0;
    area->endy = py[0] + py[1] + 1.0;
    area->clear = TRUE;
}   /*  End Function compute_area_from_ellipse  */

static unsigned long get_colour (KWorldCanvas canvas, CONST char *colourname)
/*  [SUMMARY] Get colour.
    <canvas> The world canvas.
    <colourname> The name of the colour.
    [RETURNS] The specified colour if available, else a default colour.
*/
{
    unsigned long pixel_value;

    if ( canvas_get_colour (canvas, colourname, &pixel_value, NULL,NULL,NULL) )
    {
	return (pixel_value);
    }
    if ( canvas_get_colour (canvas, "white", &pixel_value, NULL, NULL, NULL) )
    {
	fprintf (stderr,
		 "Error allocating colour: \"%s\" defaulting to white\n",
		 colourname);
	return (pixel_value);
    }
    fprintf (stderr, "Error allocating colour: \"%s\" defaulting to 0\n",
	     colourname);
    return (0);
}   /*  End Function get_colour  */

static void integrate_ellipses (KWorldCanvas canvas,
				double centre_x, double centre_y,double radius)
/*  [SUMMARY] Integrate around the circumferences of ellipses.
    <canvas> The KWorldCanvas to draw the new image to.
    <centre_x> The horizontal world co-ordinate of the centre of the ellipse.
    <centre_y> The vertical world co-ordinate of the centre of the ellipse.
    <radius> The radius in world co-ordinates.
    [RETURNS] Nothing.
*/
{
    KThreadPool pool;
    info_type info;
    flag is_astro = FALSE;
    int xlen, ylen, zlen, ystart, ystop, x, y, ystep;
    int val_rad;
    unsigned int vel_axis, num_threads, count;
    float val;
    double sq_radius;
    double min = TOOBIG;
    double max = -TOOBIG;
    dim_desc *dim;
    char txt[STRING_LENGTH], line[STRING_LENGTH];
    unsigned long dim_lengths[2];
    double dx[2], dy[2];
    double crval[2], crpix[2], cdelt[2];
    CONST char *dim_names[2];
    extern flag zero_clip;
    extern KwcsAstro cube_ap;
    extern iarray cube_arr;
    static ViewableImage vimage = NULL;
    extern KwcsAstro main_ap;
    extern char module_name[STRING_LENGTH + 1];
    extern char module_version_date[STRING_LENGTH + 1];
    extern char module_lib_version[STRING_LENGTH + 1];
    extern char karma_library_version[STRING_LENGTH + 1];
    static char function_name[] = "integrate_ellipses";

    /*  Clean up from before  */
    if (image_arr != NULL) iarray_dealloc (image_arr);
    image_arr = NULL;
    if (vimage != NULL) viewimg_destroy (vimage);
    vimage = NULL;
    if (radius <= 0.0) return;
    if ( DOUBLE_IS_BLANK (centre_x) || DOUBLE_IS_BLANK (centre_y) ) return;
    if (cube_arr == NULL)
    {
	kwin_refresh_if_visible (canvas_get_pixcanvas (canvas), TRUE);
	return;
    }
    is_astro = wcs_astro_test_if_lon ( main_ap,
				       iarray_dim_name (cube_arr, 2) );
    vel_axis = iarray_get_fits_axis (cube_arr, 0);
    /*  Create new image: first determine ellipse limits in cube pixel
	co-ordinates  */
    dx[0] = centre_x - radius;
    dy[0] = centre_y - radius;
    dx[1] = centre_x + radius;
    dy[1] = centre_y + radius;
    if (cube_ap != NULL) wcs_astro_transform (cube_ap, FALSE, 2,
					      dx, TRUE, dy, TRUE, NULL, FALSE,
					      0, NULL, NULL);
    dim = iarray_get_dim_desc (cube_arr, 2);
    dx[0] = ds_get_coord_num (dim, dx[0], SEARCH_BIAS_CLOSEST);
    dx[1] = ds_get_coord_num (dim, dx[1], SEARCH_BIAS_CLOSEST);
    dim = iarray_get_dim_desc (cube_arr, 1);
    dy[0] = ds_get_coord_num (dim, dy[0], SEARCH_BIAS_CLOSEST);
    dy[1] = ds_get_coord_num (dim, dy[1], SEARCH_BIAS_CLOSEST);
    if (dx[0] < dx[1])
    {
	info.xstart = dx[0];
	info.xstop = dx[1];
    }
    else
    {
	info.xstart = dx[1];
	info.xstop = dx[0];
    }
    if (dy[0] < dy[1])
    {
	ystart = dy[0];
	ystop = dy[1];
    }
    else
    {
	ystart = dy[1];
	ystop = dy[0];
    }
    ++info.xstop;
    ++ystop;
    xlen = (info.xstop - info.xstart) / 2;
    ylen = (ystop - ystart) / 2;
    zlen = iarray_dim_length (cube_arr, 0);
    info.num_radii_bins = sqrt (xlen * xlen + ylen * ylen) * 1.1;
    if (info.num_radii_bins < 2) return;
    info.rad_scale = (float) (info.num_radii_bins - 1) / radius;
    dim_lengths[0] = zlen;
    dim_lengths[1] = info.num_radii_bins;
    dim_names[0] = iarray_dim_name (cube_arr, 0);
    dim_names[1] = is_astro ? "ANGLE" : "Radius (Pixels)";
    if ( ( image_arr = iarray_create (K_FLOAT, 2, dim_names, dim_lengths,
				      iarray_value_name (cube_arr), NULL) )
	 == NULL )
    {
	m_abort (function_name, "image");
    }
    if (vel_axis > 0)
    {
	/*  Set the "CTYPE2" keyword with the vertical dimension name so that
	    the <wcs> package can deal with it  */
	iarray_put_named_string ( image_arr, "CTYPE2",
				  iarray_dim_name (image_arr, 0) );
	/*  Copy over frequency/velocity co-ordinate information from cube  */
	sprintf (txt, "CRVAL%u", vel_axis);
	if ( iarray_get_named_value (cube_arr, txt, NULL, crval) )
	    iarray_put_named_value (image_arr, "CRVAL2", K_DOUBLE, crval);
	sprintf (txt, "CRPIX%u", vel_axis);
	if ( iarray_get_named_value (cube_arr, txt, NULL, crpix) )
	    iarray_put_named_value (image_arr, "CRPIX2", K_DOUBLE, crpix);
	sprintf (txt, "CDELT%u", vel_axis);
	if ( iarray_get_named_value (cube_arr, txt, NULL, cdelt) )
	    iarray_put_named_value (image_arr, "CDELT2", K_DOUBLE, cdelt);
    }
    else
    {
	/*  No FITS data: use native Karma information  */
	double first, last;

	iarray_get_world_coords (cube_arr, 0, &first, &last);
	iarray_set_world_coords (image_arr, 0, first, last);
    }
    /*  Set the co-ordinate scale for the radius/angle dimension  */
    /*  Set the "CTYPE1" keyword with the horizontal dimension name so that the
	<wcs> package can deal with it  */
    iarray_put_named_string ( image_arr, "CTYPE1",
			      iarray_dim_name (image_arr, 1) );
    iarray_put_double (image_arr, "CRVAL1", 0.0);
    iarray_put_double (image_arr, "CRPIX1", 1.0);
    iarray_put_double (image_arr, "CDELT1",
		       radius / (double) (info.num_radii_bins - 1));
    if ( ( info.count_arr = iarray_create_2D (zlen, info.num_radii_bins,
					      K_UINT) ) == NULL )
    {
	m_abort (function_name, "count array");
    }
    sq_radius = radius * radius;
    pool = mt_get_shared_pool ();
    info.centre_x = centre_x;
    info.centre_y = centre_y;
    info.xscale = is_astro ? cos (centre_y * PION180) : 1.0;
    num_threads = mt_num_threads (pool);
    ylen = ystop - ystart;
    ystep = ylen / num_threads;
    if (ystep * num_threads < ylen) ++ystep;
    for (count = 0, y = ystart; count < num_threads; ++count)
    {
	int ystop_job;

	ystop_job = y + ystep;
	if (ystop_job >= ystop) ystop_job = ystop;
	mt_launch_job (pool, job_func, &info, (void *) y, (void *) ystop_job,
		       NULL);
	y = ystop_job;
    }
    mt_wait_for_all_jobs (pool);
    /*  Divide sums by counts (average computation) and maybe zero clip  */
    for (y = 0; y < zlen; ++y) for (x = 0; x < info.num_radii_bins; ++x)
    {
	val_rad = UI2 (info.count_arr, y, x);
	val = F2 (image_arr, y, x);
	if ( zero_clip && (val < 0.0) ) val = 0.0;
	else if (val_rad > 0) val /= (float) val_rad;
	F2 (image_arr, y, x) = val;
	if (val < min) min = val;
	if (val > max) max = val;
    }
    iarray_dealloc (info.count_arr);
    if (max <= min)
    {
	fprintf (stderr,
		 "No data to show: perhaps image and cube don't overlap?\n");
	return;
    }
    if ( ( vimage = viewimg_create_from_iarray (canvas, image_arr, FALSE) )
	 == NULL )
    {
	m_abort (function_name, "ViewableImage");
    }
    canvas_set_attributes (canvas,
			   CANVAS_ATT_VALUE_MIN, min,
			   CANVAS_ATT_VALUE_MAX, max,
			   CANVAS_ATT_END);
    viewimg_make_active (vimage);
    update_transform_dataclip (image_arr, min, max);
    sprintf (line, "%s: Module version %s  Karma v%s  compiled with v%s",
	     module_name, module_version_date,
	     karma_library_version, module_lib_version);
    iarray_append_history_string (image_arr, line, TRUE);
    /*  Add circle centre to history  */
    strcpy (line, "Circle centre: ");
    wcs_astro_format (cube_ap, iarray_dim_name (cube_arr, 2), txt, centre_x);
    strcat (line, txt);
    strcat (line, "  ");
    wcs_astro_format (cube_ap, iarray_dim_name (cube_arr, 1), txt, centre_y);
    strcat (line, txt);
    iarray_append_history_string (image_arr, line, TRUE);
}   /*  End Function integrate_ellipses  */

static void job_func (void *pool_info, void *call_info1, void *call_info2,
		      void *call_info3, void *call_info4, void *thread_info)
/*  [SUMMARY] Perform a job.
    <pool_info> The arbitrary pool information pointer.
    <call_info1> The first arbitrary call information pointer.
    <call_info2> The second arbitrary call information pointer.
    <call_info3> The third arbitrary call information pointer.
    <call_info4> The fourth arbitrary call information pointer.
    <thread_info> A pointer to arbitrary, per thread, information. This
    information is private to the thread.
    [RETURNS] Nothing.
*/
{
    CONST info_type *info = call_info1;
    int ystart = (int) call_info2;
    int ystop = (int) call_info3;
    int x, y, z, zlen, val_rad;
    float distance, val;
    double wx, wy;
    extern iarray cube_arr, cube_ra_coords, cube_dec_coords;

    zlen = iarray_dim_length (cube_arr, 0);
    for (y = ystart; y < ystop; ++y)
    {
	for (x = info->xstart; x < info->xstop; ++x)
	{
	    /*  Test if this pixel is within the ellipse  */
	    wx = (D2 (cube_ra_coords, y, x) - info->centre_x) * info->xscale;
	    wy = D2 (cube_dec_coords, y, x) - info->centre_y;
	    distance = sqrt (wx * wx + wy * wy);
	    val_rad = (distance * info->rad_scale);
	    if (val_rad >= info->num_radii_bins) continue;
	    /*  Loop through channels  */
	    for (z = 0; z < zlen; ++z)
	    {
		val = F3 (cube_arr, z, y, x);
		if ( DOUBLE_IS_BLANK (val) ) continue;
		F2 (image_arr, z, val_rad) += val;
		++UI2 (info->count_arr, z, val_rad);
	    }
	}
    }
}   /*  End Function job_func  */
