/*  generic.c

    Generic file for  kshell  (X11 ellipse integrator tool for Karma).

    Copyright (C) 1996-2001  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable interactive integration along the
    circumference of concentric ellipses.
    This module runs on an X11 server.


    Written by      Richard Gooch   24-SEP-1996: Copied from kpvslice module.

    Updated by      Richard Gooch   28-OCT-1996: Added hostname and port number
  to title.

    Updated by      Richard Gooch   27-DEC-1996: Switched to
  <iarray_create_and_setup>.

    Updated by      Richard Gooch   31-DEC-1996: Now return min&max from
  <process_image>.

    Updated by      Richard Gooch   2-FEB-1997: No longer destroy KwcsAstro
  objects.

    Updated by      Richard Gooch   9-FEB-1997: Fixed bug in <process_image>
  where <main_ap> was not correctly updated.

    Updated by      Richard Gooch   5-SEP-1998: Removed comms. setup code.

    Last updated by Richard Gooch   6-DEC-2001: Took account of interface
  change to <wcs_astro_transform*>.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <errno.h>
#include <X11/Xlib.h>
#include <karma.h>
#include <karma_viewimg.h>
#include <karma_iarray.h>
#include <karma_conn.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_r.h>
#include "kshell.h"


/*  Private functions  */
STATIC_FUNCTION (void destroy_all_vimages,
		 (ViewableImage *image, ViewableImage *magnified_image) );


/*  Private data  */
static char *keywords[] =
{
    "OBJECT",
    "DATE-OBS",
    "OBSRA",
    "OBSDEC",
    "EPOCH",
    "BUNIT",
    "TELESCOP",
    "INSTRUME",
    "EQUINOX",
    "BTYPE",
    "BMAJ",
    "BMIN",
    "BPA",
    "RESTFREQ",
    "VELREF",
    "PBFWHM",
    NULL
};


/*  Public functions follow  */

flag process_image (multi_array *multi_desc, iarray *image_arr,
		    double *min, double *max,
		    KWorldCanvas pseudo_canvas, KWorldCanvas mag_pseudo_canvas,
		    ViewableImage *image, ViewableImage *magnified_image)
/*  [PURPOSE] This routine will process an image and display it.
    <multi_desc> The multi_array descriptor.
    <image_arr> The image array is written here.
    <min> The minimum image value is written here.
    <max> The maximum image value is written here.
    <pseudo_canvas> The PseudoColour canvas.
    <image> If an image is loaded the ViewableImage is written here. If no
    image is loaded, NULL is written here. The value written here must be
    preserved between calls.
    <magnified_image> If an image is loaded the magnified ViewableImage is
    written here. If no image is loaded, NULL is written here. The value
    written here must be preserved between calls.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    extern KwcsAstro loaded_image_ap, main_ap;
    extern unsigned int image_mode;
    /*static char function_name[] = "process_image";*/

    if ( !iarray_create_and_setup (image_arr, multi_desc, TRUE, 2, K_FLOAT,
				   min, max, TRUE, &loaded_image_ap) )
    {
	return (FALSE);
    }
    if (image_mode == IMAGE_MODE_LOADED) main_ap = NULL;
    destroy_all_vimages (image, magnified_image);
    if ( ( *image = viewimg_create_from_iarray (pseudo_canvas, *image_arr,
						FALSE) ) == NULL )
    {
	fprintf (stderr, "Error getting ViewableImage from Iarray\n");
	iarray_dealloc (*image_arr);
	*image_arr = NULL;
	loaded_image_ap = NULL;
	return (FALSE);
    }
    if ( ( *magnified_image =
	   viewimg_create_from_iarray (mag_pseudo_canvas, *image_arr,
				       FALSE) ) == NULL )
    {
	fprintf (stderr, "Error getting ViewableImage from Iarray\n");
	iarray_dealloc (*image_arr);
	*image_arr = NULL;
	destroy_all_vimages (image, magnified_image);
	loaded_image_ap = NULL;
	return (FALSE);
    }
    viewimg_set_attributes (*image,
			    VIEWIMG_VATT_VALUE_MIN, *min,
			    VIEWIMG_VATT_VALUE_MAX, *max,
			    VIEWIMG_ATT_END);
    viewimg_set_attributes (*magnified_image,
			    VIEWIMG_VATT_VALUE_MIN, *min,
			    VIEWIMG_VATT_VALUE_MAX, *max,
			    VIEWIMG_ATT_END);
    if (image_mode == IMAGE_MODE_LOADED)
    {
	if ( !viewimg_make_active (*image) )
	{
	    fprintf (stderr, "Error making ViewableImage active\n");
	    iarray_dealloc (*image_arr);
	    *image_arr = NULL;
	    destroy_all_vimages (image, magnified_image);
	    loaded_image_ap = NULL;
	    return (FALSE);
	}
	if ( !viewimg_make_active (*magnified_image) )
	{
	    fprintf (stderr, "Error making ViewableImage active\n");
	    iarray_dealloc (*image_arr);
	    *image_arr = NULL;
	    destroy_all_vimages (image, magnified_image);
	    loaded_image_ap = NULL;
	    return (FALSE);
	}
	main_ap = loaded_image_ap;
    }
    return (TRUE);
}   /*  End Function process_image  */

flag process_cube (multi_array *multi_desc, iarray *cube_arr,
		   double *min, double *max)
/*  [PURPOSE] This routine will load a cube.
    <multi_desc> The multi_array descriptor.
    <cube_arr> The cube array is written here.
    <min> The minimum cube value is written here.
    <max> The maximum cube value is written here.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int xlen, ylen, x, y;
    extern KwcsAstro cube_ap;
    extern iarray cube_ra_coords, cube_dec_coords;
    static char function_name[] = "process_cube";

    if ( !iarray_create_and_setup (cube_arr, multi_desc, TRUE, 3, K_FLOAT,
				   min, max, TRUE, &cube_ap) )
    {
	return (FALSE);
    }
    /*  Setup cube co-ordinate arrays which will convert from x-y pixels to
	proper RA and DEC  */
    if (cube_ra_coords != NULL) iarray_dealloc (cube_ra_coords);
    if (cube_dec_coords != NULL) iarray_dealloc (cube_dec_coords);
    xlen = iarray_dim_length (*cube_arr, 2);
    ylen = iarray_dim_length (*cube_arr, 1);
    if ( ( cube_ra_coords = iarray_create_2D (ylen, xlen, K_DOUBLE) ) == NULL )
    {
	m_abort (function_name, "cube RA co-ordinate array");
    }
    if ( ( cube_dec_coords = iarray_create_2D (ylen, xlen,K_DOUBLE) ) == NULL )
    {
	m_abort (function_name, "cube DEC co-ordinate array");
    }
    /*  Fill in linear co-ordinates  */
    for (y = 0; y < ylen; ++y) for (x = 0; x < xlen; ++x)
    {
	D2 (cube_ra_coords, y, x) = iarray_get_coordinate (*cube_arr, 2, x);
	D2 (cube_dec_coords, y, x) = iarray_get_coordinate (*cube_arr, 1, y);
    }
    fprintf (stderr,
	     "Filled in linear co-ordinates in cube co-ordinate array\n");
    if (cube_ap != NULL)
    {
	/*  Convert to proper RA and DEC  */
	wcs_astro_transform (cube_ap, TRUE, xlen * ylen,
			     (double *) cube_ra_coords->data, FALSE,
			     (double *) cube_dec_coords->data, FALSE,
			     NULL, FALSE,
			     0, NULL, NULL);
    }
    fprintf (stderr, "Created cube co-ordinate array\n");
    return (TRUE);
}   /*  End Function process_cube  */

flag copy_header_info (iarray out, iarray in)
/*  [SUMMARY] Copy some header information from one array to another.
    <out> The output array.
    <in> The input array.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int count;
    static char function_name[] = "copy_header_info";

    for (count = 0; keywords[count] != NULL; ++count)
    {
	if ( !iarray_copy_named_element (out, in, keywords[count],
					 FALSE, FALSE, TRUE) )
	{
	    fprintf (stderr, "%s: Failed to copy header keyword %s\n",
		     function_name, keywords[count]);
	    return (FALSE);
	}
    }
    return (TRUE);
}   /*  End Function copy_header_info  */

void convert_lin_world_coords (double *xout, double *yout,
			       KwcsAstro out_ap,
			       CONST double *xin, CONST double *yin,
			       KwcsAstro in_ap, unsigned int num_coords)
/*  [SUMMARY] Convert linear world co-ordinates in two reference frames.
    [PURPOSE] This routine will convert linear world co-ordinates in one
    reference frame to linear world co-ordinates in another frame. If
    necessary, linear to non-linear and non-linear to linear conversions are
    performed.
    <xout> The output horizontal co-ordinates are written here.
    <yout> The output vertical co-ordinates are written here.
    <out_ap> The output reference frame.
    <xin> The input horizontal co-ordinates are written here.
    <yin> The input vertical co-ordinates are written here.
    <in_ap> The input reference frame.
    [RETURNS] Nothing.
*/
{
    unsigned int count;

    /*  First copy to the output arrays  */
    for (count = 0; count < num_coords; ++count)
    {
	xout[count] = xin[count];
	yout[count] = yin[count];
    }
    if ( !wcs_astro_test_radec (in_ap) && !wcs_astro_test_radec (out_ap) )
    {
	/*  Both reference frames are linear: plain copy  */
	return;
    }
    if ( wcs_astro_test_radec (in_ap) && wcs_astro_test_radec (out_ap) )
    {
	/*  Both reference frames are non-linear: convert back and forth  */
	wcs_astro_transform (in_ap, TRUE, num_coords,
			     xout, FALSE, yout, FALSE, NULL, FALSE,
			     0, NULL, NULL);
	wcs_astro_transform (out_ap, TRUE, num_coords,
			     xout, TRUE, yout, TRUE, NULL, FALSE,
			     0, NULL, NULL);
	return;
    }
    if ( wcs_astro_test_radec (in_ap) )
    {
	/*  Input reference frame is non-linear but output is linear: convert
	    back to linear  */
	wcs_astro_transform (in_ap, TRUE, num_coords,
			     xout, TRUE, yout, TRUE, NULL, FALSE,
			     0, NULL, NULL);
	return;
    }
    /*  Output reference frame is non-linear but input is linear: do nothing */
}   /*  End Function convert_lin_world_coords  */


/*  Private functions follow  */

static void destroy_all_vimages (ViewableImage *image,
				 ViewableImage *magnified_image)
/*  [PURPOSE] This routine will destroy all ViewableImages and clear the
    appropriate pointers.
    [RETURNS] Nothing.
*/
{
    if (*image != NULL) viewimg_destroy (*image);
    *image = NULL;
    if (*magnified_image != NULL) viewimg_destroy (*magnified_image);
    *magnified_image = NULL;
}   /*  End Function destroy_all_vimages  */


/*  Private routines follow  */
