/*  main_xt.c

    Main file for  kslice_3d  (X11 3D slicer display tool for Karma).

    Copyright (C) 1994-2004  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable on-screen viewing of orthogonal slices in a
    3 dimensional "Intelligent Array".
    This module runs on an X11 server.


    Written by      Richard Gooch   28-JAN-1995: Copied from  kview_2d_main.c

    Updated by      Richard Gooch   28-JAN-1995

    Updated by      Richard Gooch   4-FEB-1995: Tidying up.

    Updated by      Richard Gooch   15-APR-1995: Added call to
  <im_register_lib_version>.

    Updated by      Richard Gooch   17-APR-1995: Added "-private_cmap" option.

    Updated by      Richard Gooch   23-APR-1995: Added FITS file support.

    Updated by      Richard Gooch   31-JUL-1995: Made use of ThreeDeeSlice
  widget.

    Updated by      Richard Gooch   1-SEP-1995: Changed to <kcmap_va_create>.

    Updated by      Richard Gooch   28-SEP-1995: Added Miriad Image support.

    Updated by      Richard Gooch   22-OCT-1995: Print number of colours.

    Updated by      Richard Gooch   28-OCT-1995: Search for and require 8 bit
  PseudoColour visual.

    Updated by      Richard Gooch   28-DEC-1995: Set canvas background to black

    Updated by      Richard Gooch   10-JAN-1996: Added #include for iarray.

    Updated by      Richard Gooch   21-APR-1996: Made use of
  XkwNsimpleColourbar resource for Cmapwinpopup widget.

    Updated by      Richard Gooch   4-MAY-1996: Switched to KtoggleWidget
  resources.

    Updated by      Richard Gooch   5-MAY-1996: Switched to <xtmisc_popup_cbk>
  routine and improved documentation.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   22-JUN-1996: Made use of
  <XkwFilewinStandardFileTester_nD> and <XkwFilewinStandardFileTester_3D>
  routines.

    Updated by      Richard Gooch   15-SEP-1996: Made use of new <xv_*>
  routines.

    Updated by      Richard Gooch   29-OCT-1996: Added hostname and port number
  to title.

    Updated by      Richard Gooch   8-NOV-1996: Added font resource for track
  labels.

    Updated by      Richard Gooch   12-JAN-1997: Made use of new
  XkwNtrapDirectoryDatasets resource for FilewinWidgetClass.

    Updated by      Richard Gooch   21-JAN-1997: Switched to <dmx> package and
  switched to <conn_initialise>.

    Updated by      Richard Gooch   22-APR-1997: Made use of <ds_event_*>
  system.

    Updated by      Richard Gooch   31-DEC-1997: Made use of function
  <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   1-MAR-1998: Made use of new filewinWidget
  resource.

    Updated by      Richard Gooch   5-SEP-1998: Used new comms. setup.

    Updated by      Richard Gooch   8-SEP-1998: Made use of new function
  <xc_get_location2>.

    Updated by      Richard Gooch   10-NOV-1998: Added support for TrueColour
  only displays.

    Updated by      Richard Gooch   30-AUG-1999: Added dirname font resource.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Updated by      Richard Gooch   11-MAY-2001: Changed dirname font resource.

    Updated by      Richard Gooch   3-OCT-2001: Added translations for Filewin.

    Updated by      Richard Gooch   21-DEC-2001: Included k_version.h directly.

    Updated by      Richard Gooch   28-JAN-2002: Added "Print" button and
  PostScript control widget. Disabled temporarily while waiting for graphics
  library to support PostScript output and child pixel canvases together.

    Updated by      Richard Gooch   22-DEC-2002: Added #include <stdlib.h>

    Last updated by Richard Gooch   30-MAR-2004: Resource changes.


*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <karma.h>
#include <k_event_codes.h>
#include <k_version.h>
#ifndef X11
#  define X11
#endif
#include <karma_iarray.h>
#include <karma_xtmisc.h>
#include <karma_dsxfr.h>
#include <karma_kcmap.h>
#include <karma_kwin.h>
#include <karma_conn.h>
#include <karma_xv.h>
#include <karma_ds.h>
#include <karma_im.h>
#include <karma_hi.h>
#include <karma_xc.h>
#include <karma_a.h>
#include <karma_r.h>
#include <Xkw/ThreeDeeSlice.h>
#include <Xkw/Filepopup.h>
#include <Xkw/Cmapwinpopup.h>
#include <Xkw/Filewin.h>
#include <Xkw/Postscript.h>
#include <Xkw/Canvas.h>


#define DEFAULT_COLOURMAP_NAME "Greyscale1"
#define NUM_COLOURS (unsigned int) 200

#define VERSION "1.9.1"


/*  Private functions  */
STATIC_FUNCTION (flag cube_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );
STATIC_FUNCTION (void quit_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (Kcolourmap get_kcmap,
		 (Widget shell, Visual **visual, Colormap *xcmap) );


/*  Private data  */
String fallback_resources[] =
{
    "Kslice_3d*Canvas*background:                        black",
    "Kslice_3d*pseudoColourCanvas*foreground:            white",
    "Kslice_3d*directColourCanvas*foreground:            white",
    "Kslice_3d*Command*background:                       grey70",
    "Kslice_3d*Ktoggle*background:                       grey80",
    "Kslice_3d*Value*background:                         #d0a0a0",
    "Kslice_3d*closeButton*background:                   grey90",
    "Kslice_3d*ChoiceMenu.background:                    turquoise",
    "Kslice_3d*topForm*quit*background:                  orange",
    "Kslice_3d*ScrollableList.highlightPixel:            aquamarine3",
    "Kslice_3d*ChoiceMenu.background:                    turquoise",
    "Kslice_3d*ExclusiveMenu.background:                 turquoise",
    "Kslice_3d*SimpleSlider.foreground:                  Sea Green",
    "Kslice_3d*SimpleProgressMeter.background:           turquoise3",
    "Kslice_3d*SimpleProgressMeter.foreground:           grey60",
    "Kslice_3d*ThreeDeeSlice*trackLabel0*font:           8x13bold",
    "Kslice_3d*ThreeDeeSlice*trackLabel1*font:           8x13bold",
    "Kslice_3d*SimpleSlider.borderWidth:                 0",
    "Kslice_3d*Filewin.dirname.textSink.font:            6x13",
    "Kslice_3d*font:                                     9x15bold",
    "Kslice_3d*background:                               aquamarine2",
    "Kslice_3d*foreground:                               black",
    NULL
};
static XrmOptionDescRec Options[] =
{
    {"-private_cmap", ".topForm.slicer.canvas.forceNewCmap", XrmoptionNoArg,
     (XtPointer) "True"},
};
static Widget slicer = NULL;
static iarray cube = NULL;
static char title_name[STRING_LENGTH];


int main (int argc, char **argv)
{
    Kcolourmap kcmap;
    /*KPixCanvas pixcanvas;*/
    unsigned int num_ccels;
    Colormap xcmap;
    XtAppContext app_context;
    Widget main_shell, form, filewin, filepopup, colourmapwinpopup;
    Widget files_btn, cmap_btn, precomp_btn, print_btn, quit_btn;
    unsigned long *pixel_values;
    Visual *visual;
    /*static char function_name[] = "main";*/

    /*  Initialise module  */
    im_register_module_name ("kslice_3d");
    im_register_module_version_date (VERSION);
    im_register_lib_version (KARMA_VERSION);
    /*  Start up Xt  */
    main_shell = xtmisc_init_app_initialise (&app_context, "Kslice_3d",
					     Options, XtNumber (Options),
					     &argc, argv, fallback_resources,
					     XTMISC_INIT_ATT_COMMS_SETUP, TRUE,
					     XTMISC_INIT_ATT_BECOME_SERVER,
					     TRUE,
					     XTMISC_INIT_ATT_TITLE, title_name,
					     XTMISC_INIT_ATT_END,
					     NULL);
    dsxfr_register_connection_limits (1, -1);
/*
    conn_register_client_protocol ("3D_cursor_position", 0, 1,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL,
				   read_3d_cursor,
				   ( void (*) () ) NULL);
*/
    kcmap = get_kcmap (main_shell, &visual, &xcmap);
    num_ccels = kcmap_get_pixels (kcmap, &pixel_values);
    fprintf (stderr, "num colours: %u\n", num_ccels);
    form = XtVaCreateManagedWidget ("topForm", formWidgetClass, main_shell,
				    XtNborderWidth, 0,
				    NULL);
    filepopup = XtVaCreatePopupShell ("filewinPopup", filepopupWidgetClass,
				      form,
				      XtNtitle, "kslice_3d File Selector",
				      XtNx, 0,
				      XtNy, 0,
				      XkwNautoPopdown, TRUE,
				      XkwNloadFiles, TRUE,
				      XkwNdomainName, "CUBE",
				      XkwNfilenameTester,
				          XkwFilewinStandardFileTester_3D,
				      NULL);
    /*  Register callback which will process file selection events  */
    ds_event_register_func (cube_event_func, NULL);
    /*  Register callback which will process directory selection events  */
    XtVaGetValues (filepopup, XkwNfilewinWidget, &filewin, NULL);
    XtVaSetValues (filewin,
		   XkwNtrapDirectoryDatasets, True,
		   NULL);
    /*  Now start creating sundry buttons  */
    files_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
					 XtNlabel, "Files",
					 NULL);
    XtAddCallback (files_btn, XtNcallback, xtmisc_popup_cbk, filepopup);
    cmap_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
					XtNlabel, "Colourmap",
					XtNfromHoriz, files_btn,
					NULL);
    precomp_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
					   XtNlabel, "Precompute",
					   XtNfromHoriz, cmap_btn,
					   NULL);
#if 0
    print_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
					 XtNlabel, "Print",
					 XtNfromHoriz, precomp_btn,
					 NULL);
#else
    print_btn = precomp_btn;
#endif
    quit_btn = XtVaCreateManagedWidget ("quit", commandWidgetClass, form,
					XtNlabel, "Quit",
					XtNfromHoriz, print_btn,
					NULL);
    XtAddCallback (quit_btn, XtNcallback, quit_cbk, NULL);
    slicer = XtVaCreateManagedWidget ("slicer", threeDeeSliceWidgetClass, form,
				      XtNborderWidth, 0,
				      XtNhorizDistance, 0,
				      XtNvertDistance, 0,
				      XtNfromVert, quit_btn,
				      XkwNkarmaColourmap, kcmap,
				      XkwNcanvasVisual, visual,
				      XkwNcanvasColourmap, xcmap,
				      NULL);
    XtAddCallback (precomp_btn, XtNcallback, XkwThreeDeeSlicePrecompute,
		   (XtPointer) slicer);
    XtRealizeWidget (main_shell);
    colourmapwinpopup = XtVaCreatePopupShell ("cmapwinpopup",
					      cmapwinpopupWidgetClass, form,
					      XkwNkarmaColourmap, kcmap,
					      XkwNsimpleColourbar, True,
					      NULL);
    XtAddCallback (cmap_btn, XtNcallback, xtmisc_popup_cbk, colourmapwinpopup);
#if 0
    w = XtVaCreatePopupShell ("postscriptwinpopup", postscriptWidgetClass,form,
			      XtNtitle, "Postscript Window",
			      XkwNautoIncrement, TRUE,
			      NULL);
    XtAddCallback (print_btn, XtNcallback, xtmisc_popup_cbk, w);
    cnv = XtNameToWidget (slicer, "canvas");
    XtVaGetValues (cnv, XkwNmonoPixCanvas, &pixcanvas, NULL);
    XkwPostscriptRegisterImageAndName (w, pixcanvas, NULL);
#endif
    if ( !conn_controlled_by_cm_tool () ) XtPopup (filepopup, XtGrabNone);
    XtAppMainLoop (app_context);
    return (RV_OK);
}   /*  End Function main   */


/*  Private routines follow  */

static flag cube_event_func (void *object, multi_array *multi_desc,
			     CONST char *domain, CONST char *name)
/*  [SUMMARY] Process a new cube event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    if (cube != NULL) iarray_dealloc (cube);
    cube = iarray_get_from_multi_array (multi_desc, NULL, 3, NULL, NULL);
    if (cube == NULL ) return (TRUE);
    XtVaSetValues (slicer,
		   XkwNiarray, cube,
		   NULL);
    return (TRUE);
}   /*  End Function cube_event_func  */

static void quit_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the quit button callback.
*/
{
    exit (RV_OK);
}   /*  End Function quit_cbk   */

static Kcolourmap get_kcmap (Widget shell, Visual **visual, Colormap *xcmap)
/*  [SUMMARY] Get Kcolourmap object.
    <shell> The shell widget.
    <visual> The visual is written here.
    <xcmap> The X colourmap is written here.
    [RETURNS] The Kcolourmap.
*/
{
    Kdisplay dpy_handle;
    Kcolourmap kcmap;
    Display *dpy;
    Screen *screen;
    XVisualInfo *vinfo;
    Visual *pc_visual, *tc_visual;
    Visual *root_visual;
    static char function_name[] = "get_kcmap";

    dpy = XtDisplay (shell);
    screen = XtScreen (shell);
    /*  Get visual information  */
    root_visual = XDefaultVisualOfScreen (screen);
    xv_get_visuals (screen, &pc_visual, &tc_visual, NULL);
    *visual = (pc_visual == NULL) ? tc_visual : pc_visual;
    if (*visual == NULL)
    {
	fprintf (stderr, "No PseudoColour or TrueColour visual available\n");
	exit (RV_UNDEF_ERROR);
    }
    if (*visual == root_visual)
    {
	/*  Use the default colourmap  */
	XtVaGetValues (shell,
		       XtNcolormap, xcmap,
		       NULL);
    }
    else
    {
	/*  Different visual: create colourmap  */
	if ( ( *xcmap = XCreateColormap (dpy, XRootWindowOfScreen (screen),
					 *visual, AllocNone) ) ==
	     (Colormap) NULL )
	{
	    fprintf (stderr, "Could not create colourmap\n");
	    exit (1);
	}
	XSync (dpy, False);
	fprintf (stderr, "Created colourmap: 0x%lx\n", *xcmap);
    }
    if (*visual == tc_visual)
    {
	vinfo = xv_get_visinfo_for_visual (dpy, tc_visual);
	if ( ( kcmap = kcmap_create_scmap (DEFAULT_COLOURMAP_NAME,
					   NUM_COLOURS, TRUE,
					   vinfo->red_mask, vinfo->green_mask,
					   vinfo->blue_mask) )
	     == NULL )
	{
	    fprintf (stderr,
		     "Error creating software Karma PseudoColour cmap\n");
	    a_prog_bug (function_name);
	}
	XFree (vinfo);
	return (kcmap);
    }
    if ( ( dpy_handle = xc_get_dpy_handle (dpy, *xcmap) ) == NULL )
    {
	fprintf (stderr, "Error getting display handle\n");
	a_prog_bug (function_name);
    }
    /*  Initialise colourmap  */
    if ( ( kcmap = kcmap_va_create (DEFAULT_COLOURMAP_NAME,
				    NUM_COLOURS, TRUE,
				    dpy_handle,
				    xc_alloc_colours, xc_free_colours,
				    xc_store_colours, xc_get_location2,
				    KCMAP_ATT_END) )
	== NULL )
    {
	fprintf (stderr, "Error creating main colourmap\n");
	a_prog_bug (function_name);
    }
    return (kcmap);
}   /*  End Function get_kcmap  */
