/*  send.c

    Source file for  send  (overlay list test transmitter module).

    Copyright (C) 1993-2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*


    Written by      Richard Gooch   6-DEC-1993

    Updated by      Richard Gooch   11-DEC-1993

    Updated by      Richard Gooch   29-FEB-1996: Added spiral.

    Updated by      Richard Gooch   27-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   16-MAR-1997: Added single ellipse.

    Updated by      Richard Gooch   2-SEP-1998: Added random arcs.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package.

    Last updated by Richard Gooch   22-DEC-2002: Added #include <stdlib.h>


*/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/file.h>
#include <karma.h>
#include <karma_overlay.h>
#include <karma_module.h>
#include <karma_panel.h>
#include <karma_dsxfr.h>
#include <karma_conn.h>
#include <karma_ex.h>
#include <karma_ch.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_n.h>


#define VERSION "1.4"

#define NUM_ELLIPSES 100
#define NUM_SEGMENTS 100
#define NUM_VECTORS 100

STATIC_FUNCTION (flag send, (CONST char *command, FILE *fp) );
STATIC_FUNCTION (void server_exited, (Connection connection, void *info) );
static void add_to_list ();
STATIC_FUNCTION (void empty_list, (CONST char *p) );
STATIC_FUNCTION (void send_ellipses, (CONST char *p) );
STATIC_FUNCTION (void send_segments, (CONST char *p) );
STATIC_FUNCTION (void send_vectors, (CONST char *p) );
STATIC_FUNCTION (void send_spiral, (CONST char *p) );
STATIC_FUNCTION (void send_ellipse, (CONST char *p) );
STATIC_FUNCTION (void send_avector, (CONST char *p) );
STATIC_FUNCTION (void send_random_arcs, (CONST char *p) );

static KOverlayList olist = NULL;
static unsigned int num_segments = NUM_SEGMENTS;
static flag one_object = FALSE;
static unsigned int coord_type = OVERLAY_COORD_PIXEL;
static unsigned int num_arcs = 1000;
static double minimum_x = -1.0;
static double maximum_x = 1.0;
static double minimum_y = -1.0;
static double maximum_y = 1.0;
static double radius_x = 1e-2;
static double radius_y = 1e-2;
static double rotation = 0.0;


int main (int argc, char **argv)
{
    KControlPanel panel;
    char *coord_type_choices[5];
    extern KOverlayList olist;
    static char function_name[] = "main";

    coord_type_choices[OVERLAY_COORD_PIXEL] = "pixel";
    coord_type_choices[OVERLAY_COORD_RELATIVE] = "relative";
    coord_type_choices[OVERLAY_COORD_WORLD] = "world";
    coord_type_choices[OVERLAY_COORD_LAST] = "last";
    coord_type_choices[OVERLAY_COORD_LINEAR] = "linear";
    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
	m_abort (function_name, "control panel");
    }
    panel_add_item (panel, "spiral", "sends spiral", PIT_FUNCTION,
		    (void *) send_spiral,
		    PIA_END);
    panel_add_item (panel, "send", "sends objects", PIT_FUNCTION,
		    (void *) add_to_list,
		    PIA_END);
    panel_add_item (panel, "ellipses", "sends ellipses",
		    PIT_FUNCTION, (void *) send_ellipses,
		    PIA_END);
    panel_add_item (panel, "one_ellipse", "sends ellipse",
		    PIT_FUNCTION, (void *) send_ellipse,
		    PIA_END);
    panel_add_item (panel, "segments", "sends segments",
		    PIT_FUNCTION, (void *) send_segments,
		    PIA_END);
    panel_add_item (panel, "vectors", "sends vectors",
		    PIT_FUNCTION, (void *) send_vectors,
		    PIA_END);
    panel_add_item (panel, "random_arcs", "random arcs",
		    PIT_FUNCTION, (void *) send_random_arcs,
		    PIA_END);
    panel_add_item (panel, "num_arcs", "number of random arcs",
		    K_UINT, &num_arcs,
		    PIA_END);
    panel_add_item (panel, "yradius", "y radius", K_DOUBLE, &radius_y,
		    PIA_END);
    panel_add_item (panel, "ymax", "random y maximum",
		    K_DOUBLE, &maximum_y,
		    PIA_END);
    panel_add_item (panel, "ymin", "random y minimum",
		    K_DOUBLE, &minimum_y,
		    PIA_END);
    panel_add_item (panel, "xradius", "x radius", K_DOUBLE, &radius_x,
		    PIA_END);
    panel_add_item (panel, "xmax", "random x maximum",
		    K_DOUBLE, &maximum_x,
		    PIA_END);
    panel_add_item (panel, "xmin", "random x minimum",
		    K_DOUBLE, &minimum_x,
		    PIA_END);
    panel_add_item (panel, "rotation", "ellipse rotation",
		    K_DOUBLE, &rotation,
		    PIA_END);
    panel_add_item (panel, "empty", "number of objects", PIT_FUNCTION,
		    (void *) empty_list,
		    PIA_END);
    panel_add_item (panel, "one_object", "flag", PIT_FLAG, &one_object,
		    PIA_END);
    panel_add_item (panel, "num_segments", "number of segments", K_UINT,
		    &num_segments,
		    PIA_MAX_VALUE, (double) NUM_SEGMENTS,
		    PIA_END);
    panel_add_item (panel, "coord_type", "coordinate type", PIT_CHOICE_INDEX,
		    &coord_type,
		    PIA_NUM_CHOICE_STRINGS,
		    sizeof coord_type_choices / sizeof *coord_type_choices,
		    PIA_CHOICE_STRINGS, coord_type_choices,
		    PIA_END);
    panel_add_item (panel, "avector", "type1 x1 y1 type2 x2 y2",
		    PIT_FUNCTION, (void *) send_avector,
		    PIA_END);
    panel_push_onto_stack (panel);
    if ( ( olist = overlay_va_create_list (NULL, "mainOverlayList",
					   OVERLAY_ATT_END) ) == NULL )
    {
	exit (1);
    }
    conn_register_client_protocol ("ping_server", 0, 0,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL,
				   ( void (*) () ) NULL);
    conn_register_client_protocol ("server_exit", 0, 0,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL,
				   server_exited);
    module_run (argc, argv, "send-overlays", VERSION, send, -1, 0, FALSE);
    return (RV_OK);
}   /*  End Function main   */

static flag send (CONST char *command, FILE *fp)
{
    return (TRUE);
}   /*  End Function send  */

static void add_to_list (CONST char *p)
/*  This routine will flush the accumulated arrays.
*/
{
    unsigned int count;
    double x;
    double y;
    unsigned int types[3];
    double x_arr[3];
    double y_arr[3];
    /*static char function_name[] = "add_to_list";*/

    overlay_line (olist, OVERLAY_COORD_RELATIVE, 0.0, 1.0,
		  OVERLAY_COORD_RELATIVE, 1.0, 0.0, "blue");
    overlay_text (olist, "Hello there",
		  OVERLAY_COORD_RELATIVE, 0.1, 0.5, "red", "10x20",
		  FALSE);
    overlay_text (olist, "Gidday!!!",
		  OVERLAY_COORD_RELATIVE, 0.1, 0.6, "red", "12x24",
		  TRUE);
    overlay_ellipse (olist, OVERLAY_COORD_WORLD, 265.0, 245.0,
		     OVERLAY_COORD_WORLD, 21.0, 9.0, "green", FALSE);
    overlay_ellipse (olist, OVERLAY_COORD_WORLD, 265.0, 245.0,
		     OVERLAY_COORD_RELATIVE, 0.02, 0.01, "blue", FALSE);
    overlay_ellipse (olist, OVERLAY_COORD_RELATIVE, 1.0, 0.0,
		     OVERLAY_COORD_PIXEL, 10, 10, "red", TRUE);
    types[0] = types[1] = types[2] = OVERLAY_COORD_WORLD;
    x_arr[0] = 300.0;
    y_arr[0] = 186.0;
    x_arr[1] = 296.0;
    y_arr[1] = 179.0;
    x_arr[2] = 304.0;
    y_arr[2] = 179.0;
    overlay_filled_polygon (olist, 3, types, x_arr, y_arr, "green");
    for (count = 0; count < NUM_ELLIPSES; ++count)
    {
	x = (double) count / (double) NUM_ELLIPSES;
	y = x;
	overlay_ellipse (olist, OVERLAY_COORD_RELATIVE, x, x,
			 OVERLAY_COORD_PIXEL, 10, 10, "red", FALSE);
    }
}   /*  End Function add_to_list  */

static void empty_list (CONST char *p)
/*  This routine will flush the accumulated arrays.
*/
{
    int num_objects;
    /*static char function_name[] = "empty_list";*/

    num_objects = ex_int (p, &p);
    overlay_remove_objects (olist, (unsigned int) num_objects);
}   /*  End Function empty_list  */

static void send_ellipses (CONST char *p)
{
    unsigned int count;
    unsigned int ctypes[NUM_ELLIPSES];
    unsigned int rtypes[NUM_ELLIPSES];
    double cx[NUM_ELLIPSES];
    double cy[NUM_ELLIPSES];
    double rx[NUM_ELLIPSES];
    double ry[NUM_ELLIPSES];
    /*static char function_name[] = "send_ellipses";*/

    for (count = 0; count < NUM_ELLIPSES; ++count)
    {
	ctypes[count] = OVERLAY_COORD_RELATIVE;
	cx[count] = (double) count / (double) NUM_ELLIPSES;
	cy[count] = cx[count];
	rtypes[count] = OVERLAY_COORD_PIXEL;
	rx[count] = 10.0;
	ry[count] = 10.0;
    }
    overlay_ellipses (olist, NUM_ELLIPSES,
		      ctypes, cx, cy, rtypes, rx, ry, "red", FALSE);
}   /*  End Function send_ellipses  */

static void send_segments (CONST char *p)
{
    unsigned int count;
    unsigned int types0[NUM_SEGMENTS];
    unsigned int types1[NUM_SEGMENTS];
    double x0[NUM_SEGMENTS];
    double y0[NUM_SEGMENTS];
    double x1[NUM_SEGMENTS];
    double y1[NUM_SEGMENTS];
    /*static char function_name[] = "send_segments";*/

    for (count = 0; count < NUM_SEGMENTS; ++count)
    {
	types0[count] = OVERLAY_COORD_RELATIVE;
	x0[count] = (double) count / (double) NUM_SEGMENTS;
	y0[count] = x0[count];
	types1[count] = OVERLAY_COORD_RELATIVE;
	x1[count] = 0.0;
	y1[count] = 1.0;
    }
    overlay_segments (olist, NUM_SEGMENTS,
		      types0, x0, y0, types1, x1, y1, "red");
}   /*  End Function send_segments  */

static void send_vectors (CONST char *p)
{
    unsigned int count;
    unsigned int stypes[NUM_VECTORS];
    unsigned int dtypes[NUM_VECTORS];
    double sx[NUM_VECTORS];
    double sy[NUM_VECTORS];
    double dx[NUM_VECTORS];
    double dy[NUM_VECTORS];
    /*static char function_name[] = "send_vectors";*/

    for (count = 0; count < NUM_VECTORS; ++count)
    {
	stypes[count] = OVERLAY_COORD_RELATIVE;
	sx[count] = (double) count / (double) NUM_VECTORS;
	sy[count] = sx[count];
	dtypes[count] = OVERLAY_COORD_PIXEL;
	dx[count] = 10.0;
	dy[count] = 10.0;
    }
    overlay_vectors (olist, NUM_VECTORS,
		     stypes, sx, sy, dtypes, dx, dy, "red");
}   /*  End Function send_vectors  */

static void server_exited (Connection connection, void *info)
{
    fprintf (stderr, "Server exited\n");
}   /*  End Function server_exited  */

static void send_spiral (CONST char *p)
/*  [PURPOSE] This routine will send a spiral.
    <p> Optional arguments. Ignored.
    [RETURNS] Nothing.
*/
{
    int seg_count;
    int segs_per_rot;
    int seg_in_rot;
    double angle;
    double centre_x = 0.5;
    double centre_y = 0.5;
    double radius_x = 0.1;
    double radius_y = 0.1;
    double radius_increment;
    unsigned int types[NUM_SEGMENTS];
    double x_arr[NUM_SEGMENTS];
    double y_arr[NUM_SEGMENTS];

    fprintf (stderr, "Drawing spiral with %u segments as %s\n",
	     num_segments, one_object ? "one object" : "many objects");
    segs_per_rot = num_segments / 10;
    radius_increment = 0.35 / (double) num_segments;
    for (seg_count = 0; seg_count < num_segments; ++seg_count)
    {
	seg_in_rot = seg_count % segs_per_rot;
	angle = TWOPI * (double) seg_in_rot / (double) segs_per_rot;
	x_arr[seg_count] = centre_x + radius_x * sin (angle);
	y_arr[seg_count] = centre_y + radius_y * cos (angle);
	types[seg_count] = OVERLAY_COORD_RELATIVE;
	radius_x += radius_increment;
	radius_y += radius_increment;
    }
    if (one_object)
    {
	overlay_lines (olist, num_segments, types, x_arr, y_arr, "green");
    }
    else
    {
	for (seg_count = 0; seg_count < num_segments - 1; ++seg_count)
	{
	    overlay_line (olist, types[seg_count],
			  x_arr[seg_count], y_arr[seg_count],
			  types[seg_count + 1],
			  x_arr[seg_count + 1], y_arr[seg_count + 1],
			  "red");
	}
    }
}   /*  End Function send_spiral  */

static void send_ellipse (CONST char *p)
{
    /*static char function_name[] = "send_ellipse";*/

    overlay_ellipse (olist, OVERLAY_COORD_PIXEL, 50, 50,
		     OVERLAY_COORD_PIXEL, 50, 25, "red", FALSE);
}   /*  End Function send_ellipse  */

static void send_avector (CONST char *p)
{
    char *ptr;
    unsigned int type1, type2;
    double x1, y1, x2, y2;
    /*static char function_name[] = "send_avector";*/

    if ( ( ptr = ex_str (p, &p) ) == NULL )
    {
	fprintf (stderr, "Must supply type1\n");
	return;
    }
    if ( (st_icmp (ptr, "p") == 0) || (st_icmp (ptr, "pixel") == 0) )
    {
	type1 = OVERLAY_COORD_PIXEL;
    }
    else if ( (st_icmp (ptr, "r") == 0) || (st_icmp (ptr, "relative") == 0) )
    {
	type1 = OVERLAY_COORD_RELATIVE;
    }
    else if ( (st_icmp (ptr, "w") == 0) || (st_icmp (ptr, "world") == 0) )
    {
	type1 = OVERLAY_COORD_WORLD;
    }
    else if ( (st_icmp (ptr, "l") == 0) || (st_icmp (ptr, "last") == 0) )
    {
	type1 = OVERLAY_COORD_LAST;
    }
    else
    {
	fprintf (stderr, "Type string: \"%s\" illegal\n", ptr);
	m_free (ptr);
	return;
    }
    m_free (ptr);
    x1 = ex_float (p, &p);
    y1 = ex_float (p, &p);
    if ( ( ptr = ex_str (p, &p) ) == NULL )
    {
	fprintf (stderr, "Must supply type2\n");
	return;
    }
    if ( (st_icmp (ptr, "p") == 0) || (st_icmp (ptr, "pixel") == 0) )
    {
	type2 = OVERLAY_COORD_PIXEL;
    }
    else if ( (st_icmp (ptr, "r") == 0) || (st_icmp (ptr, "relative") == 0) )
    {
	type2 = OVERLAY_COORD_RELATIVE;
    }
    else if ( (st_icmp (ptr, "w") == 0) || (st_icmp (ptr, "world") == 0) )
    {
	type2 = OVERLAY_COORD_WORLD;
    }
    else if ( (st_icmp (ptr, "l") == 0) || (st_icmp (ptr, "last") == 0) )
    {
	type2 = OVERLAY_COORD_LAST;
    }
    else
    {
	fprintf (stderr, "Type string: \"%s\" illegal\n", ptr);
	m_free (ptr);
	return;
    }
    m_free (ptr);
    x2 = ex_float (p, &p);
    y2 = ex_float (p, &p);
    overlay_vector (olist, type1, x1, y1, type2, x2, y2, "red");
}   /*  End Function send_avector  */

static void send_random_arcs (CONST char *p)
{
    unsigned int count;
    double cx, cy, xrange, yrange;

    xrange = maximum_x - minimum_x;
    yrange = maximum_y - minimum_y;
    for (count = 0; count < num_arcs; ++count)
    {
	cx = minimum_x + n_uniform () * xrange;
	cy = minimum_y + n_uniform () * yrange;
	overlay_arc (olist, coord_type, cx, cy, coord_type, radius_x, radius_y,
		     rotation, "green", FALSE);
    }
}   /*  End Function send_random_arcs  */
