/*  kchangehead.c

    Source file for  kchangehead  (change header in Karma format file).

    Copyright (C) 1997-2001  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will change the header information contained in a Karma
  format file.


    Written by      Richard Gooch   12-FEB-1997

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package.

    Last updated by Richard Gooch   21-DEC-2001: Included k_version.h directly.


*/
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_module.h>
#include <karma_panel.h>
#include <karma_dsxfr.h>
#include <karma_dmp.h>
#include <karma_ds.h>
#include <karma_im.h>
#include <karma_ex.h>
#include <karma_m.h>
#include <karma_a.h>
#include <k_version.h>


#define VERSION "1.0"


/*  Private functions  */
STATIC_FUNCTION (flag kchangehead, (CONST char *command, FILE *fp) );
STATIC_FUNCTION (void load_file, (char *p) );
STATIC_FUNCTION (void save_file, (char *p) );
STATIC_FUNCTION (void unload_file, (char *p) );
STATIC_FUNCTION (void show_header, (char *p) );
STATIC_FUNCTION (void set_keyword, (char *p) );


/*  Private data  */
static flag show_history = TRUE;
static multi_array *multi_desc = NULL;
static flag changes_made = FALSE;
static unsigned int type = K_DOUBLE;
static double value = 0.0;


/*  Public functions follow  */

int main (int argc, char **argv)
{
    KControlPanel panel;
    extern char *data_type_names[NUMTYPES];
    static char function_name[] = "main";

    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
	m_abort (function_name, "control panel");
    }
    panel_add_item (panel, "value", "new keyword value", K_DOUBLE, &value,
		    PIA_END);
    panel_add_item (panel, "unload", "unload loaded file",
		    PIT_FUNCTION, (void *) unload_file,
		    PIA_END);
    panel_add_item (panel, "type", "new keyword data type", PIT_CHOICE_INDEX,
		    &type,
		    PIA_NUM_CHOICE_STRINGS, NUMTYPES,
		    PIA_CHOICE_STRINGS, data_type_names,
		    PIA_END);
    panel_add_item (panel, "set_keyword", "create or change specified keyword",
		    PIT_FUNCTION, (void *) set_keyword,
		    PIA_END);
    panel_add_item (panel, "save", "save loaded file to specified name",
		    PIT_FUNCTION, (void *) save_file,
		    PIA_END);
    panel_add_item (panel, "load", "load specified file",
		    PIT_FUNCTION, (void *) load_file,
		    PIA_END);
    panel_add_item (panel, "history", "show history with header", PIT_FLAG,
		    &show_history,
		    PIA_END);
    panel_add_item (panel, "header", "display header",
		    PIT_FUNCTION, (void *) show_header,
		    PIA_END);
    panel_push_onto_stack (panel);
    panel_set_extra_comment_string (panel, "\nGeneral usage: NONE");
    im_register_lib_version (KARMA_VERSION);
    module_run (argc, argv, "kchangehead", VERSION, kchangehead, -1, -1,FALSE);
    return (RV_OK);
}   /*  End Function main   */

static flag kchangehead (CONST char *command, FILE *fp)
{
    char *infile;

    for ( ; command; command = ex_command_skip (command) )
    {
	if ( ( infile = ex_str (command, &command) ) == NULL )
	{
	    fprintf (fp, "Error extracting infile name\n");
	    return (TRUE);
	}
	/*process_file (infile);*/
	m_free (infile);
    }
    return (TRUE);
}   /*  End Function kchangehead  */

static void load_file (char *p)
/*  [PURPOSE] This routine will load a file.
    <p> A string containing the name of the file to load.
    [RETURNS] Nothing.
*/
{
    unsigned int ftype;

    if (multi_desc != NULL)
    {
	if (changes_made)
	{
	    fprintf (stderr, "Existing file not yet saved or unloaded\n");
	    return;
	}
	unload_file (NULL);
    }
    if (p == NULL)
    {
	fprintf (stderr, "Must specify file to load\n");
	return;
    }
    /*  Read file in whatever format  */
    if ( ( multi_desc =
	   foreign_guess_and_read (p, K_CH_MAP_IF_AVAILABLE, FALSE, &ftype,
				   FA_GUESS_READ_END) ) == NULL )
    {
	fprintf (stderr, "Error reading\n");
	return;
    }
    changes_made = FALSE;
}   /*  End Function load_file  */

static void save_file (char *p)
/*  [PURPOSE] This routine will save the loaded file.
    <p> A string containing the name of the file to save.
    [RETURNS] Nothing.
*/
{
    if (multi_desc == NULL)
    {
	fprintf (stderr, "Nothing loaded yet\n");
	return;
    }
    if (p == NULL)
    {
	fprintf (stderr, "Must specify file to save\n");
	return;
    }
    if (changes_made)
    {
	dsxfr_put_multi (p, multi_desc);
    }
    else
    {
	fprintf (stderr, "No changes were made: simply unloading\n");
    }
    unload_file (NULL);
}   /*  End Function save_file  */

static void unload_file (char *p)
/*  [PURPOSE] This routine will unload the loaded file.
    <p> Ignored.
    [RETURNS] Nothing.
*/
{
    if (multi_desc == NULL)
    {
	fprintf (stderr, "Nothing loaded yet\n");
	return;
    }
    ds_dealloc_multi (multi_desc);
    multi_desc = NULL;
    changes_made = FALSE;
}   /*  End Function unload_file  */

static void show_header (char *p)
/*  [SUMMARY] Show the header of a structure in virtual memory.
    <p> Ignored.
    [RETURNS] Nothing.
*/
{
    unsigned int count, type;
    packet_desc *top_pack_desc;
    history *hist;
    char *top_packet;

    if (multi_desc == NULL)
    {
	fprintf (stderr, "Nothing loaded yet\n");
	return;
    }
    /* Assume all the interesting information is in the first data structure */
    top_pack_desc = multi_desc->headers[0];
    top_packet = multi_desc->data[0];
    for (count = 0; count < top_pack_desc->num_elements; ++count)
    {
	type = top_pack_desc->element_types[count];
	if (type == K_ARRAY)
	{
	    dmp_array_desc (stdout,
			    (array_desc *) top_pack_desc->element_desc[count],
			    FALSE);
	}
	else if (type == LISTP)
	{
	    puts ("Linked List header");
	}
	else if ( ds_element_is_named (type) )
	{
	    printf ("Element: \"%s\"\t\t", top_pack_desc->element_desc[count]);
	    dmp_element (stdout, type, top_pack_desc->element_desc[count],
			 top_packet +
			 ds_get_element_offset (top_pack_desc, count),
			 FALSE);
	}
	else
	{
	    printf ("Unknown element type: %u\n", type);
	}
    }
    if (!show_history) return;
    /*  Show history  */
    if (multi_desc->first_hist != NULL) puts ("");
    for (hist = multi_desc->first_hist; hist != NULL; hist = hist->next)
    {
	printf ("HISTORY: %s\n", hist->string);
    }
}   /*  End Function show_header  */

static void set_keyword (char *p)
/*  [PURPOSE] This routine will set a keyword.
    <p> A string containing the name of the keyword to set.
    [RETURNS] Nothing.
*/
{
    double dval[2];

    if (multi_desc == NULL)
    {
	fprintf (stderr, "Nothing loaded yet\n");
	return;
    }
    if (p == NULL)
    {
	fprintf (stderr, "Must specify keyword to set\n");
	return;
    }
    dval[0] = value;
    dval[1] = 0.0;
    if ( !ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				     p, type, dval, TRUE) )
    {
	fprintf (stderr, "Error setting keyword\n");
	return;
    }
    changes_made = TRUE;
}   /*  End Function set_keyword  */
