/*  klincoords.c

    Source file for klincoords

    Copyright (C) 2000-2001  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module transform a data cube to linear kpc scale */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_module.h>
#include <karma_iarray.h>
#include <karma_panel.h>
#include <karma_wcs.h>
#include <karma_ch.h>
#include <karma_ex.h>
#include <karma_im.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_a.h>
#include <k_version.h>


#define VERSION ".001"

/*  Private data for user interface  */

static double position_angle = 0.0;
static double origin[2] = {0.0, 0.0};
static unsigned int origin_array_length = 0;
static double distance = 0.0;

static char *keywords[] =
{
    "OBJECT",
    "DATE-OBS",
    "OBSRA",
    "OBSDEC",
    "TELESCOP",
    "INSTRUME",
    "BTYPE",
    "RESTFREQ",
    "VELREF",
    "PBFWHM",
    NULL
};

static char *deletable_keywords[] =
{
    "EPOCH",
    "EQUINOX",
    "BMAJ",
    "BMIN",
    "BPA",
    NULL
};

/*  Private functions  */
STATIC_FUNCTION (flag klincoords, (CONST char *command, FILE *fp) );
STATIC_FUNCTION (flag do_klincoords,
		 (CONST char *infile, CONST char *outfile) );
STATIC_FUNCTION (flag copy_header_info, (iarray out, iarray in) );


/*  Public functions follow  */

int main (int argc, char **argv)
{
    KControlPanel panel;
    static char function_name[] = "main";

    if ( ( panel = panel_create (FALSE) ) == NULL )
	m_abort (function_name, "control panel");
    panel_add_item (panel, "pa", "position angle (degrees)",
		    K_DOUBLE, &position_angle,
		    PIA_END);
    panel_add_item (panel, "origin", "centre of galaxy in decimal degrees",
		    K_DOUBLE, origin,
		    PIA_ARRAY_MIN_LENGTH, 2,
		    PIA_ARRAY_MAX_LENGTH, 2,
		    PIA_ARRAY_LENGTH, &origin_array_length,
		    PIA_END);
    panel_add_item (panel, "distance", "Mpc", K_DOUBLE, &distance, PIA_END);
    panel_push_onto_stack (panel);
    panel_set_extra_comment_string (panel, "\nGeneral usage: infile outfile");
    im_register_lib_version (KARMA_VERSION);
    module_run (argc, argv, "klincoords", VERSION, klincoords, -1, 0, FALSE);
    return (RV_OK);
}   /*  End Function main   */


/*  Private functions follow  */

static flag klincoords (CONST char *command, FILE *fp)
{
    for ( ; command != NULL; command = ex_word_skip (command) )
    {
	flag ok;
	char *infile, *outfile;

	if ( ( infile = ex_str (command, &command) ) == NULL )
	{
	    fprintf (fp, "Error extracting input filename\n");
	    return (TRUE);
	}
	if ( ( outfile = ex_str (command, &command) ) == NULL )
	{
	    fprintf (fp, "Error extracting output filename\n");
	    return (TRUE);
	}
	ok = do_klincoords (infile, outfile);
	m_free (infile);
	m_free (outfile);
	if (!ok) continue;
    }
    return (TRUE);
}   /*  End Function klincoords   */

static flag do_klincoords (CONST char *infile, CONST char *outfile)
/*  [SUMMARY] 
    <infile> The file to operate on
    <outfile> The output file
    [RETURNS] TRUE on success, else FALSE.
*/
{
    KwcsAstro in_ap = NULL;
    KwcsAstro out_ap;
    iarray in_arr = NULL;
    iarray out_arr;
    flag ok;
    flag galactic = FALSE;
    unsigned int ftype, num_dim, count, dim_h_index, dim_v_index;
    double tmp1, tmp2, out_crval1, out_crval2, out_cdelt1, out_cdelt2;
    double lon_min = TOOBIG;
    double lon_max = -TOOBIG;
    double lat_min = TOOBIG;
    double lat_max = -TOOBIG;
    double CDELT1 = 0.0;
    double CDELT2 = 0.0;
    double CRVAL2 = 0.0;
    CONST dim_desc *in_dim_h, *in_dim_v, *in_dim_lon, *in_dim_lat;
    CONST char *in_projection, *out_projection;
    char txt[STRING_LENGTH], lon[STRING_LENGTH], lat[STRING_LENGTH];
    unsigned long dim_lengths[3];
    double x[4], y[4], lon_coords[4], lat_coords[4];
    double start, end;
    CONST char *dim_names[3];
    extern char module_name[STRING_LENGTH + 1];

    /*convert to angular offset using supplied origin or default to image centre*/
    /*  Read input array  */
    if ( !foreign_read_and_setup (infile, K_CH_MAP_IF_AVAILABLE, FALSE,
				  &ftype, TRUE, 0, K_FLOAT, TRUE,
				  &in_arr, NULL, NULL,
				  TRUE, &in_ap) ) 
    {
	return (FALSE);
    }  
    if (in_ap == NULL)
    {
	fprintf (stderr, "File: \"%s\" has no astronomical projection\n",
		 infile);
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    num_dim = iarray_num_dim (in_arr);
    if ( (num_dim < 2) || (num_dim > 3) )
    {
	fprintf (stderr, "Unsupported number of dimensions: %u\n", num_dim);
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    /*  Construct output grid  */
    if (iarray_num_dim (in_arr) == 3)
    {
	dim_h_index = 2;
	dim_v_index = 1;
	in_dim_h = iarray_get_dim_desc (in_arr, dim_h_index);
	in_dim_v = iarray_get_dim_desc (in_arr, dim_v_index);

	/* JHK Start printing out information from the input cube. */
	fprintf (stderr, "\n\n************** ORIGINAL DATA ***************");

	/* JHK Print out dimension names from dimension descriptor. */
	fprintf (stderr, "\nName of the horizontal dimension is : %s", in_dim_h->name);
	fprintf (stderr, "\nName of the vertical dimention is : %s", in_dim_v->name);
    }
    else
    {
	dim_h_index = 1;
	dim_v_index = 0;
	in_dim_h = iarray_get_dim_desc (in_arr, dim_h_index);
	in_dim_v = iarray_get_dim_desc (in_arr, dim_v_index);
    }
    if ( !wcs_astro_test_if_lon (in_ap, in_dim_h->name) &&
	 !wcs_astro_test_if_lon (in_ap, in_dim_v->name) )
    {
	fprintf (stderr, "No longitude axis found\n");
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    if ( !wcs_astro_test_if_lat (in_ap, in_dim_h->name) &&
	 !wcs_astro_test_if_lat (in_ap, in_dim_v->name) )
    {
	fprintf (stderr, "No latitude axis found\n");
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    if (num_dim == 3) dim_names[0] = iarray_dim_name (in_arr, 0);
    /* Getting CDELT from input. */
    CDELT1 = iarray_get_float (in_arr, "CDELT1");
    CDELT2 = iarray_get_float (in_arr, "CDELT2");
    /* Obtaining CRVAL2 from input header is commented out. */
    /* CRVAL2 = iarray_get_float (in_arr, "CRVAL2"); */
    /* JHK  Use the input origin dec for calculating new CDELT1. */
    CRVAL2 = origin[1];
    /* JHK Print out header info from the input cube. */
    fprintf(stderr, "\n\nCTYPE1 is : %s", iarray_get_named_string (in_arr, "CTYPE1"));
    fprintf(stderr, "\nCTYPE2 is : %s", iarray_get_named_string (in_arr, "CTYPE2"));
    fprintf(stderr, "\nCTYPE3 is : %s", iarray_get_named_string (in_arr, "CTYPE3"));
    fprintf(stderr, "\n\nCDELT1 is : %f", iarray_get_float(in_arr, "CDELT1"));
    fprintf(stderr, "\nCDELT2 is : %f", iarray_get_float(in_arr, "CDELT2"));
    fprintf(stderr, "\nCDELT3 is : %f", iarray_get_float (in_arr, "CDELT3"));
    fprintf(stderr, "\n\nCRVAL1 is : %f", iarray_get_float(in_arr, "CRVAL1"));
    fprintf(stderr, "\nCRVAL2 is : %f", iarray_get_float(in_arr, "CRVAL2"));
    fprintf(stderr, "\nCRVAL3 is : %f", iarray_get_float (in_arr, "CRVAL3"));
    fprintf(stderr, "\n\nCROTA1 Does not Exist");
    fprintf(stderr, "\nCROTA2 Does not Exist");  
    fprintf(stderr, "\n\nCRPIX1 is : %f", iarray_get_float(in_arr, "CRPIX1"));
    fprintf(stderr, "\nCRPIX2 is : %f", iarray_get_float(in_arr, "CRPIX2"));
    fprintf(stderr, "\nCRPIX3 is : %f", iarray_get_float (in_arr, "CRPIX3"));
    /*cdelt:verify input values are nonzero*/
    if (CDELT1 == 0)
    {
	fprintf (stderr, "CDELT1 is zero.\n");
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    if (CDELT2 == 0)
    {
	fprintf (stderr, "CDELT2 is zero.\n");
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    /*  Find the corners of the input image in world co-ordinates  */
    x[0] = 0.0;
    y[0] = 0.0;
    x[1] = in_dim_h->length - 1;
    y[1] = 0.0;
    x[2] = in_dim_h->length - 1;
    y[2] = in_dim_v->length - 1;
    x[3] = 0.0;
    y[3] = in_dim_v->length - 1;
    /* JHK printing out values for the four corners of the image in world coordinates. */
    fprintf (stderr, "\n\nx[0] = %f", x[0]);
    fprintf (stderr, "\ny[0] = %f", y[0]);
    fprintf (stderr, "\nx[1] = %f", x[1]);
    fprintf (stderr, "\ny[1] = %f", y[1]);
    fprintf (stderr, "\nx[2] = %f", x[2]);
    fprintf (stderr, "\ny[2] = %f", y[2]);
    fprintf (stderr, "\nx[3] = %f", x[3]); 
    fprintf (stderr, "\ny[3] = %f", y[3]);
    iarray_get_world_coords(in_arr, dim_h_index, &start, &end);
    iarray_get_world_coords(in_arr, dim_v_index, &start, &end);
    /* transform the world coordinates to the astronomical data. */
    wcs_astro_transform3 (in_ap, FALSE, 4,
			  in_dim_h->name, x, FALSE,
			  in_dim_v->name, y, FALSE,
			  NULL, NULL, FALSE,
			  0, NULL, NULL);
    /*here we munge the input array header*/
    /*set the datastructure dimension names*/
    if (distance == 0.0)
    {
	iarray_set_dim_name (in_arr, dim_h_index, "angle1", TRUE);
	iarray_set_dim_name (in_arr, dim_v_index, "angle2", TRUE);
    }
    else
    {
	/* Set dimenstion name to distance in parsec. */
	iarray_set_dim_name (in_arr, dim_h_index, "distance1 (pc)", TRUE);
	iarray_set_dim_name (in_arr, dim_v_index, "distance2 (pc)", TRUE);
    }
    /*ctype:set to angular offset*/
    if (distance == 0.0)
    {
	/*1 pc = the distance such that the radius of the earths orbit subtends one arc second*/
	/**/
    }
    else
    {
	if (!iarray_put_named_string (in_arr, "CTYPE1", iarray_dim_name(in_arr,dim_h_index)))
	{
	    fprintf (stderr, "Failed to set CTYPE1\n");
	    iarray_dealloc (in_arr);
	    return (FALSE);
	}
	if (!iarray_put_named_string (in_arr, "CTYPE2", iarray_dim_name(in_arr,dim_v_index)))
	{
	    fprintf (stderr, "Failed to set CTYPE2\n");
	    iarray_dealloc (in_arr);
	    return (FALSE);
	}
    }
    /* Setting CRVAL1 and CRVAL2 to zero. */
    if (!iarray_put_double (in_arr, "CRVAL1", 0.0))
    {
	fprintf (stderr, "Failed to set CRVAL1\n");
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    if (!iarray_put_double (in_arr, "CRVAL2", 0.0))
    {
	fprintf (stderr, "Failed to set CRVAL2\n");
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    /* Computing new CDELT1 and CDELT2 */
 
    if (distance == 0)
    {
	CDELT1 = fabs (CDELT1) * cos (CRVAL2 * PION180);
	CDELT2 = fabs (CDELT2);
    }
    else
    {
	/* CDELT1 = distance x tan (Delta RA) x cos (Dec Origin) */
	CDELT1 = distance * 1e6 * tan (CDELT1 * PION180) * cos (CRVAL2 * PION180);

	/* CDELT2 = distance x tan (Delta Dec) */
	CDELT2 = distance * 1e6 * tan (CDELT2 * PION180); 
    } 
    /* Setting CDELT1 and CDELT2 with the computed data. */
    if (!iarray_put_double (in_arr, "CDELT1", CDELT1))
    {
	fprintf (stderr, "Failed to set CDELT1\n");
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    if (!iarray_put_double (in_arr, "CDELT2", CDELT2))
    {
	fprintf (stderr, "Failed to set CDELT2\n");
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    /*crot:set both to pa + 90.0*/
    if (!iarray_put_double (in_arr, "CROTA1", position_angle + 90.0))
    {
	fprintf (stderr, "Failed to set CROTA1\n");
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    if (!iarray_put_double (in_arr, "CROTA2", position_angle + 90.0))
    {
	fprintf (stderr, "Failed to set CROTA2\n");
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    /*crpix:if origin supplied then convert using wcs lib function else use center pixel*/
    if (origin[0] == 0.0 && origin[1] == 0.0)
    {
	/*origin was not specified so we use the center of image as a default*/
	if (!iarray_put_double (in_arr, "CRPIX1", floor((in_dim_h->length)/2)))
	{
	    fprintf (stderr, "Failed to set CRPIX1\n");
	    iarray_dealloc (in_arr);
	    return (FALSE);
	}

	if (!iarray_put_double (in_arr, "CRPIX2", floor((in_dim_v->length)/2)))
	{
	    fprintf (stderr, "Failed to set CRPIX2\n");
	    iarray_dealloc (in_arr);
	    return (FALSE);
	}
    }
    else
    {
	/*origin was specified and so we convert it to pixel coords*/
	wcs_astro_transform3(in_ap,FALSE,1,in_dim_h->name,x,TRUE,in_dim_v->name,y,TRUE,NULL,NULL,FALSE,0,NULL,NULL);
    }

    /* JHK Printing out information for the output cube. */
    fprintf (stderr, "\n\n******************** CHANGED DATA ********************");
    /* JHK Printing out dimension names from dimension descriptor. */
    in_dim_h = iarray_get_dim_desc (in_arr, dim_h_index);
    in_dim_v = iarray_get_dim_desc (in_arr, dim_v_index);
    fprintf (stderr, "\nName of the horizontal dimension is : %s",
	     in_dim_h->name);
    fprintf (stderr, "\nName of the vertical dimension is : %s",
	     in_dim_v->name);
    fprintf (stderr,"\n(For above, set to distanceN (pc) if distance is not 0, set to angleN otherwise.)");

    /* JHK Printing out header information. */
    fprintf(stderr, "\n\nThe new CTYPE1 is : %s", iarray_get_named_string (in_arr, "CTYPE1"));
    fprintf(stderr, "\nThe new CTYPE2 is : %s", iarray_get_named_string (in_arr, "CTYPE2"));
    fprintf(stderr, "\nThe new CTYPE3 is : %s", iarray_get_named_string (in_arr, "CTYPE3"));
    fprintf (stderr, "\n\nThe new CDELT1--distance * TAN(CDELT1) * COS(CRVAL2) : %f", iarray_get_float(in_arr, "CDELT1"));
    fprintf (stderr, "\nThe new CDELT2--distance * tan(CDELT2) : %f", iarray_get_float(in_arr, "CDELT2"));
    fprintf(stderr, "\nThe new CDELT3--unchanged : %f", iarray_get_float (in_arr, "CDELT3"));
    fprintf (stderr, "\n\nThe new CRVAL1--set to 0.0 : %f", iarray_get_float(in_arr, "CRVAL1"));
    fprintf (stderr, "\nThe new CRVAL2--set to 0.0 : %f", iarray_get_float(in_arr, "CRVAL2"));
    fprintf (stderr, "\nThe new CRVAL3--unchanged : %f", iarray_get_float(in_arr, "CRVAL3"));
    fprintf (stderr, "\n\nThe new CROTA1--position angle + 90 : %f", iarray_get_float(in_arr, "CROTA1"));
    fprintf (stderr, "\nThe new CROTA2--position angle + 90 : %f", iarray_get_float(in_arr, "CROTA2"));
    fprintf (stderr, "\n\nThe new CRPIX1 (set to # of dim. coord. / 2 if origin is 0.0 0.0, unchanged otherwise.) : %f", iarray_get_float(in_arr, "CRPIX1"));
    fprintf (stderr, "\nThe new CRPIX2 (set to # of dim. coord. / 2 if origin is 0.0 0.0, unchanged otherwise.) : %f", iarray_get_float(in_arr, "CRPIX2"));
    fprintf (stderr, "\nThe new CRPIX3--unchanged : %f", iarray_get_float(in_arr, "CRPIX3"));
    /* JHK printing out the four corners in astronomical coordinates. */
    fprintf (stderr, "\n\nAfter wcs_astro_transform3");
    fprintf (stderr, "\nx[0] = %f", x[0]);
    fprintf (stderr, "\ny[0] = %f", y[0]);
    fprintf (stderr, "\nx[1] = %f", x[1]);
    fprintf (stderr, "\ny[1] = %f", y[1]);
    fprintf (stderr, "\nx[2] = %f", x[2]);
    fprintf (stderr, "\ny[2] = %f", y[2]);
    fprintf (stderr, "\nx[3] = %f", x[3]);
    fprintf (stderr, "\ny[3] = %f", y[3]);
    iarray_get_world_coords (in_arr, dim_h_index, &start, &end);
    iarray_get_world_coords (in_arr, dim_v_index, &start, &end);
    if ( !ds_remove_named_elements (in_arr->top_pack_desc, in_arr->top_packet,
				    (CONST char **) deletable_keywords) )
    {
	fprintf (stderr, "Failed to remove deletable keywords\n");
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    if (!iarray_write (in_arr, outfile))
    {
	fprintf (stderr, "Failed to write output file\n");
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function do_klincoords  */

static flag copy_header_info (iarray out, iarray in)
/*  [SUMMARY] Copy some header information from one array to another.
    <out> The output array.
    <in> The input array.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int count;
    static char function_name[] = "copy_header_info";

    for (count = 0; keywords[count] != NULL; ++count)
    {
	if ( !iarray_copy_named_element (out, in, keywords[count],
					 FALSE, FALSE, TRUE) )
	{
	    fprintf (stderr, "%s: Failed to copy header keyword %s\n",
		     function_name, keywords[count]);
	    return (FALSE);
	}
    }
    return (TRUE);
}   /*  End Function copy_header_info  */
