/*  kuvlist.c

    Source file for  kuvlist  (print UV header in Karma format file).

    Copyright (C) 2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will print specified UV data in a file.


    Written by      Richard Gooch   14-JUN-2002

    Last updated by Richard Gooch   24-JUN-2002: Check for missing baselines.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_module.h>
#include <karma_iarray.h>
#include <karma_panel.h>
#include <karma_dsxfr.h>
#include <karma_dsrw.h>
#include <karma_pio.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_ch.h>
#include <karma_ex.h>
#include <karma_m.h>
#include <karma_a.h>

#define VERSION "1.0"

STATIC_FUNCTION (flag kuvlist, (CONST char *command, FILE *fp) );
STATIC_FUNCTION (void process_file, (CONST char *infile) );


/*  Private data  */
static int baseline = -1;
static int antenna_a = -1;
static int antenna_b = -1;
static int start_channel = -1;
static int end_channel = -1;


int main (int argc, char **argv)
{
    KControlPanel panel;
    static char function_name[] = "main";

    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
	m_abort (function_name, "control panel");
    }
    panel_add_item (panel, "start_channel", "integer", K_INT, &start_channel,
		    PIA_END);
    panel_add_item (panel, "end_channel", "integer", K_INT, &end_channel,
		    PIA_END);
    panel_add_item (panel, "baseline", "integer", K_INT, &baseline,
		    PIA_END);
    panel_add_item (panel, "antenna_b", "integer", K_INT, &antenna_b,
		    PIA_END);
    panel_add_item (panel, "antenna_a", "integer", K_INT, &antenna_a,
		    PIA_END);
    panel_push_onto_stack (panel);
    panel_set_extra_comment_string (panel, "\nGeneral usage: infile");
    module_run (argc, argv, "kuvlist", VERSION, kuvlist, -1, -1, FALSE);
    return (RV_OK);
}   /*  End Function main   */

static flag kuvlist (CONST char *command, FILE *fp)
{
    char *infile;

    for ( ; command; command = ex_command_skip (command) )
    {
	if ( ( infile = ex_str (command, &command) ) == NULL )
	{
	    fprintf (fp, "Error extracting infile name\n");
	    return (TRUE);
	}
	process_file (infile);
	m_free (infile);
    }
    return (TRUE);
}   /*  End Function kuvlist  */

static void process_file (CONST char *infile)
/*  [SUMMARY] Process an input file.
    <infile> The name of the input file.
    [RETURNS] Nothing.
*/
{
    iarray visdata, bl_to_a, bl_to_b, udata, vdata, wdata;
    unsigned int count_pol, count_time, count_bl;
    multi_array *multi_desc;
    dim_desc *dim_pol, *dim_time, *dim_bl, *dim_channel;
    static char function_name[] = "process_file";

    /*  Read file in whatever format  */
    if ( ( multi_desc =
	   foreign_miriad_read (infile, FALSE, FA_MIRIAD_READ_END) ) == NULL )
    {
	fprintf (stderr, "Error reading\n");
	return;
    }
    if ( ( visdata = iarray_get_from_multi_array (multi_desc, NULL, 4, NULL,
						  NULL) ) == NULL )
    {
	fprintf (stderr, "%s: error getting visdata hypercube\n",
		 function_name);
	ds_dealloc_multi (multi_desc);
	return;
    }
    if (iarray_type (visdata) != K_COMPLEX)
    {
	fprintf (stderr, "Visdata type is not complex\n");
	iarray_dealloc (visdata);
	ds_dealloc_multi (multi_desc);
	return;
    }
    if ( ( bl_to_a = iarray_get_from_multi_array (multi_desc, NULL, 1, NULL,
						  "Antenna A") ) == NULL )
    {
	fprintf (stderr, "%s: error getting BL to Antenna A array\n",
		 function_name);
	iarray_dealloc (visdata);
	ds_dealloc_multi (multi_desc);
	return;
    }
    if ( ( bl_to_b = iarray_get_from_multi_array (multi_desc, NULL, 1, NULL,
						  "Antenna B") ) == NULL )
    {
	fprintf (stderr, "%s: error getting BL to Antenna B array\n",
		 function_name);
	iarray_dealloc (bl_to_a);
	iarray_dealloc (visdata);
	ds_dealloc_multi (multi_desc);
	return;
    }
    if ( ( udata = iarray_get_from_multi_array (multi_desc, NULL, 2, NULL,
						"U Coord") ) == NULL )
    {
	fprintf (stderr, "%s: error getting BL/Time to U coord array\n",
		 function_name);
	iarray_dealloc (bl_to_b);
	iarray_dealloc (bl_to_a);
	iarray_dealloc (visdata);
	ds_dealloc_multi (multi_desc);
	return;
    }
    if ( ( vdata = iarray_get_from_multi_array (multi_desc, NULL, 2, NULL,
						"V Coord") ) == NULL )
    {
	fprintf (stderr, "%s: error getting BL/Time to V coord array\n",
		 function_name);
	iarray_dealloc (udata);
	iarray_dealloc (bl_to_b);
	iarray_dealloc (bl_to_a);
	iarray_dealloc (visdata);
	ds_dealloc_multi (multi_desc);
	return;
    }
    if ( ( wdata = iarray_get_from_multi_array (multi_desc, NULL, 2, NULL,
						"W Coord") ) == NULL )
    {
	fprintf (stderr, "%s: error getting BL/Time to W coord array\n",
		 function_name);
	iarray_dealloc (vdata);
	iarray_dealloc (udata);
	iarray_dealloc (bl_to_b);
	iarray_dealloc (bl_to_a);
	iarray_dealloc (visdata);
	ds_dealloc_multi (multi_desc);
	return;
    }
    ds_dealloc_multi (multi_desc);
    dim_pol = iarray_get_dim_desc (visdata, 0);
    dim_time = iarray_get_dim_desc (visdata, 1);
    dim_bl = iarray_get_dim_desc (visdata, 2);
    dim_channel = iarray_get_dim_desc (visdata, 3);
    for (count_pol = 0; count_pol < dim_pol->length; ++count_pol)
    {
	fprintf (stderr, "Polarisation: %g\n",dim_pol->coordinates[count_pol]);
	for (count_time = 0; count_time < dim_time->length; ++count_time)
	{
	    double mjd;
	    char timebuf[STRING_LENGTH];

	    mjd = dim_time->coordinates[count_time] - 2400000.5;
	    wcs_astro_format_time (timebuf, mjd * 3600.0 * 24.0);
	    fprintf (stderr, "Time: %s\n", timebuf);
	    for (count_bl = 0; count_bl < dim_bl->length; ++count_bl)
	    {
		unsigned int ch_start, ch_stop, count_channel;
		float *vis;

		if ( (baseline >= 0) &&
		     (baseline != (int) dim_bl->coordinates[count_bl]) )
		    continue;
		if ( (antenna_a >= 0) &&
		     (UI1 (bl_to_a, count_bl) != antenna_a) ) continue;
		if ( (antenna_b >= 0) &&
		     (UI1 (bl_to_b, count_bl) != antenna_b) ) continue;
		fprintf ( stderr, "Baseline[%u] %g\tAntenna %u, %u",
			  count_bl, dim_bl->coordinates[count_bl],
			  UI1 (bl_to_a, count_bl), UI1 (bl_to_b, count_bl) );
		if ( DOUBLE_IS_BLANK ( D2 (udata, count_time, count_bl) ) )
		{
		    fprintf (stderr, "\tno data\n");
		    continue;
		}
		putc ('\n', stderr);
		ch_start = (start_channel < 0) ? 0 : start_channel;
		ch_stop = (end_channel < 0) ?
		    dim_channel->length : (end_channel + 1);
		for (count_channel = ch_start; count_channel < ch_stop;
		     ++count_channel)
		{
		    vis = CF4_ptr (visdata, count_pol, count_time, count_bl,
				   count_channel);
		    if ( DOUBLE_IS_VALID (vis[0]) )
			fprintf ( stderr, "Channel: %u R: %f I: %f A: %f\n",
				  count_channel, vis[0], vis[1],
				  sqrt (vis[0] * vis[0] + vis[1] * vis[1]) );
		    else fprintf (stderr, "Channel: %u blank\n",count_channel);
		}
	    }
	}
    }
    iarray_dealloc (wdata);
    iarray_dealloc (vdata);
    iarray_dealloc (udata);
    iarray_dealloc (bl_to_b);
    iarray_dealloc (bl_to_a);
    iarray_dealloc (visdata);
}   /*  End Function process_file  */
