/*  kmodel_shell_xt.c

    Main file for  kmodel-shell  (X11 3D shell modelling tool for Karma).

    Copyright (C) 1997-2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable modelling of an expanding shell in 3D.
    This module runs on an X11 server.


    Written by      Richard Gooch   2-MAR-1997

    Updated by      Richard Gooch   9-MAR-1997: Took account of change to
  <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   22-JUN-1997: Took account of change to
  <xtmisc_init_app_initialise> to double varargs list.

    Updated by      Richard Gooch   5-SEP-1998: Used new comms. setup.

    Updated by      Richard Gooch   17-SEP-1998: Use SimpleMenu class name in
  resources.

    Updated by      Richard Gooch   17-APR-1999: Added beaming power slider.

    Updated by      Richard Gooch   20-APR-1999: Made use of function
  <iarray_create_shm>. Optimised <compute_cbk>: reduced main loop from 4600 ms
  to 1300 ms (PPro 180). Computed cube maximum and attached to header. Threaded
  function <compute_cbk>.

    Updated by      Richard Gooch   16-MAY-1999: Renamed file.

    Updated by      Richard Gooch   8-JUN-1999: Added "Central Pixel" toggle.

    Updated by      Richard Gooch   30-AUG-1999: Added dirname font resource.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Updated by      Richard Gooch   11-MAY-2001: Changed dirname font resource.

    Updated by      Richard Gooch   21-DEC-2001: Included k_version.h directly.

    Last updated by Richard Gooch   22-DEC-2002: Added #include <stdlib.h>


*/
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <errno.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <karma.h>
#include <k_event_codes.h>
#include <k_version.h>
#ifndef X11
#  define X11
#endif
#include <karma_vrender_def.h>
#include <karma_foreign.h>
#include <karma_iarray.h>
#include <karma_canvas.h>
#include <karma_xtmisc.h>
#include <karma_dsxfr.h>
#include <karma_kwin.h>
#include <karma_conn.h>
#include <karma_ds.h>
#include <karma_im.h>
#include <karma_mt.h>
#include <karma_ic.h>
#include <karma_xc.h>
#include <karma_a.h>
#include <karma_m.h>
#include <Xkw/ProfileEdit.h>
#include <Xkw/SimpleSlider.h>
#include <Xkw/ExportMenu.h>
#include <Xkw/Ktoggle.h>


#define VERSION "0.1.2"

#define MIN_VELOCITY    0.0
#define MAX_VELOCITY    100e3
#define NUM_CHANNELS    256
#define NUM_X           256
#define NUM_Y           255    /*  Avoid cache line aliasing  */
#define NUM_TRIG_STEPS  720
#define MAX_THETA       90.0


typedef struct
{
    double sin;
    double cos;
} trig_data_type;

typedef struct
{
    float iscale;
    float rad_to_index, vel_norm_to_index;
    trig_data_type *trig_data;
} info_type;


/*  Private functions  */
STATIC_FUNCTION (void quit_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void compute_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void job_func,
		 (void *pool_info, void *call_info1, void *call_info2,
		  void *call_info3, void *call_info4, void *thread_info) );


/*  Private data  */
String fallback_resources[] =
{
    "KModelShell*ProfileEdit.canvas*background:            black",
    "KModelShell*Command*background:                       grey70",
    "KModelShell*Repeater*background:                      grey70",
    "KModelShell*Ktoggle*background:                       grey80",
    "KModelShell*closeButton*background:                   grey90",
    "KModelShell*ChoiceMenu.background:                    turquoise",
    "KModelShell*ExclusiveMenu.background:                 turquoise",
    "KModelShell*Value*background:                         #d0a0a0",
    "KModelShell*quit*background:                          orange",
    "KModelShell*SimpleSlider.foreground:                  Sea Green",
    "KModelShell*SimpleSlider.borderWidth:                 0",
    "KModelShell*SimpleMenu*font:                          10x20",
    "KModelShell*Filewin.dirname.textSink.font:            6x13",
    "KModelShell*font:                                     9x15bold",
    "KModelShell*borderColor:                              black",
    "KModelShell*background:                               aquamarine",
    "KModelShell*foreground:                               black",
    NULL
};
static XrmOptionDescRec Options[] =
{
    {"-verbose", "*verbose", XrmoptionNoArg, (XPointer) "True"},
};
static Widget main_shell = NULL;
static Widget profile_edit = NULL;
static iarray vel_vs_theta_arr = NULL;
static iarray cube = NULL;
#if 0
static int beamwidth = 10;
#endif
static int off_axis_angle = 0;
static double beaming_power;
static char title_name[STRING_LENGTH] = "Unknown";


int main (int argc, char **argv)
{
    KWorldCanvas wc;
    XtAppContext app_context;
    Widget form, w, export_menu, central_pixel_tgl;
    uaddr dim_lengths[3];
    static char function_name[] = "main";

    /*  Initialise module  */
    im_register_module_name ("kmodel-shell");
    im_register_module_version_date (VERSION);
    im_register_lib_version (KARMA_VERSION);
    /*  Start up Xt  */
    main_shell = xtmisc_init_app_initialise (&app_context, "KModelShell",
					     Options, XtNumber (Options),
					     &argc, argv, fallback_resources,
					     XTMISC_INIT_ATT_COMMS_SETUP, TRUE,
					     XTMISC_INIT_ATT_BECOME_SERVER,
					     TRUE,
					     XTMISC_INIT_ATT_TITLE, title_name,
					     XTMISC_INIT_ATT_END,
					     NULL);
    dsxfr_register_connection_limits (-1, 0);
    xtmisc_set_icon (main_shell, ic_write_kplot_icon);
    form = XtVaCreateManagedWidget ("topForm", formWidgetClass, main_shell,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("exportMenu", exportMenuWidgetClass, form,
				 XkwNsupportSubsets, FALSE,
				 NULL);
    export_menu = w;
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, form,
				 XtNlabel, "Central Pixel",
				 XkwNcrosses, False,
				 XtNfromHoriz, w,
				 NULL);
    central_pixel_tgl = w;
    w = XtVaCreateManagedWidget ("quit", commandWidgetClass, form,
				 XtNlabel, "Quit",
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, quit_cbk, NULL);
#if 0
    w = XtVaCreateManagedWidget ("beamWidthSlider", simpleSliderWidgetClass,
				 form,
				 XtNlabel, "Beamwidth (degrees)",
				 XtNwidth, 300,
				 XtNfromVert, w,
				 XkwNminimum, 1,
				 XkwNmaximum, 60,
				 XkwNmodifier, 1,
				 XtNvalue, beamwidth,
				 XkwNvaluePtr, &beamwidth,
				 NULL);
#endif
    w = XtVaCreateManagedWidget ("angleSlider", simpleSliderWidgetClass,
				 form,
				 XtNlabel, "Off-axis distance (degrees)",
				 XtNwidth, 300,
				 XtNfromVert, w,
				 XkwNminimum, 0,
				 XkwNmaximum, 90,
				 XkwNmodifier, 1,
				 XtNvalue, off_axis_angle,
				 XkwNvaluePtr, &off_axis_angle,
				 NULL);
    w = XtVaCreateManagedWidget ("logSlider", simpleSliderWidgetClass,
				 form,
				 XtNlabel, "Beaming Power",
				 XtNwidth, 300,
				 XtNfromVert, w,
				 XkwNminimum, 0,
				 XkwNmaximum, 100,
				 XkwNmodifier, 1,
				 XtNvalue, 40,
				 XkwNscaledValuePtr, &beaming_power,
				 XkwNscaledFormat, "%g",
				 NULL);
    XkwSimpleSliderSetScale (w, 0.1, 0.0, FALSE, TRUE);
    profile_edit = XtVaCreateManagedWidget ("form", profileEditWidgetClass,
					    form,
					    XtNborderWidth, 0,
					    XtNhorizDistance, 0,
					    XtNvertDistance, 0,
					    XtNfromVert, w,
					    NULL);
    XtAddCallback (profile_edit, XtNcallback, compute_cbk, central_pixel_tgl);
    XtRealizeWidget (main_shell);
    /*  Create array  */
    if ( ( vel_vs_theta_arr = iarray_create_1D (1024, K_FLOAT) ) == NULL )
    {
	m_abort (function_name, "array");
    }
    iarray_set_dim_name (vel_vs_theta_arr, 0, "ANGLE", TRUE);
    iarray_set_world_coords (vel_vs_theta_arr, 0, 0.0, MAX_THETA);
    iarray_set_value_name (vel_vs_theta_arr, "M/S", TRUE);
    XtVaSetValues (profile_edit, XkwNiarray, vel_vs_theta_arr, NULL);
    XtVaGetValues (profile_edit, XkwNworldCanvas, &wc, NULL);
    canvas_set_attributes (wc,
			   CANVAS_ATT_BOTTOM_Y, MIN_VELOCITY,
			   CANVAS_ATT_TOP_Y, MAX_VELOCITY,
			   CANVAS_ATT_END);
    canvas_set_dressing (wc,
			 KCD_ATT_HLABEL,
			 "Angular distance from polar axis",
			 KCD_ATT_VLABEL,
			 "Radial expansion velocity (km.s\\u-\\u1)",
			 KCD_ATT_END);
    dim_lengths[0] = NUM_CHANNELS;
    dim_lengths[1] = NUM_Y;
    dim_lengths[2] = NUM_X;
    if ( (cube  = iarray_create_shm (K_FLOAT, 3, NULL, dim_lengths, NULL,
				     FALSE) ) == NULL )
    {
	m_abort (function_name, "count cube");
    }
    iarray_set_dim_name (cube, 0, "VELO", TRUE);
    iarray_set_world_coords (cube, 0, -MAX_VELOCITY, MAX_VELOCITY);
    XtVaSetValues (export_menu,
		   XkwNiarray, cube,
		   XkwNworldCanvas, wc,
		   NULL);
    XtAppMainLoop (app_context);
    return (RV_OK);
}   /*  End Function main   */

static void quit_cbk (Widget w, XtPointer client_data, XtPointer call_data)

/*  This is the quit button callback.
*/
{
    exit (RV_OK);
}   /*  End Function quit_cbk   */

static void compute_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the compute callback.
*/
{
    KThreadPool pool;
    int phi_num, phi_step, phi;
    unsigned int count, num_threads;
    info_type info;
    float maximum = -TOOBIG;
    Bool central_pixel;
    Widget central_pixel_tgl = client_data;
    dim_desc *angle_dim, *vel_dim;
    static flag first_time = TRUE;
    static trig_data_type trig_data[NUM_TRIG_STEPS];
    static float *maxima = NULL;
    static char function_name[] = "compute_cbk";

    pool = mt_get_shared_pool ();
    num_threads = mt_num_threads (pool);
    if (first_time)
    {
	double theta;

	first_time = FALSE;
	for (count = 0; count < NUM_TRIG_STEPS; ++count)
	{
	    theta = (double) count * 360.0 / (double) NUM_TRIG_STEPS;
	    trig_data[count].sin = sin (theta);
	    trig_data[count].cos = cos (theta);
	}
	if ( ( maxima = m_alloc (num_threads * sizeof *maxima) ) == NULL )
	{
	    m_abort (function_name, "maxima array");
	}
    }
    info.trig_data = trig_data;
    iarray_clear (cube);
    angle_dim = iarray_get_dim_desc (vel_vs_theta_arr, 0);
    info.rad_to_index = (float) (angle_dim->length - 1) * 180.0 /
	angle_dim->last_coord / PI;
    vel_dim = iarray_get_dim_desc (cube, 0);
    info.vel_norm_to_index = (float) (vel_dim->length - 1) /
	(vel_dim->last_coord - vel_dim->first_coord);
    info.iscale = pow (10.0, -beaming_power);
    /*  Loop through "all" points on shell surface. theta is angle in XY plane
	(position angle) and phi is angle from Z axis  */
    phi_num = NUM_TRIG_STEPS / 2;
    phi_step = phi_num / num_threads;
    if (phi_step * num_threads < phi_num) ++phi_step;
    for (count = 0, phi = 0; count < num_threads; ++count)
    {
	int phi_stop_job;

	phi_stop_job = phi + phi_step;
	if (phi_stop_job >= phi_num) phi_stop_job = phi_num;
	maxima[count] = -TOOBIG;
	mt_launch_job (pool, job_func, &info,
		       (void *) phi, (void *) phi_stop_job, maxima + count);
	phi = phi_stop_job;
    }
    mt_wait_for_all_jobs (pool);
    for (count = 0; count < num_threads; ++count)
	if (maxima[count] > maximum) maximum = maxima[count];
    XtVaGetValues (central_pixel_tgl, XtNstate, &central_pixel, NULL);
    if (central_pixel) F3 (cube, NUM_CHANNELS / 2, NUM_Y / 2, NUM_X / 2) = 1.0;
    iarray_set_data_range (cube, 0.0, maximum);
    if (conn_get_num_client_connections ("multi_array") > 0)
	iarray_write (cube, "connections");
}   /*  End Function compute_cbk   */

static void job_func (void *pool_info, void *call_info1, void *call_info2,
		      void *call_info3, void *call_info4, void *thread_info)
/*  [SUMMARY] Perform a job.
    <pool_info> The arbitrary pool information pointer.
    <call_info1> The first arbitrary call information pointer.
    <call_info2> The second arbitrary call information pointer.
    <call_info3> The third arbitrary call information pointer.
    <call_info4> The fourth arbitrary call information pointer.
    <thread_info> A pointer to arbitrary, per thread, information. This
    information is private to the thread.
    [RETURNS] Nothing.
*/
{
    CONST info_type *info = call_info1;
    int phi = (int) call_info2;          /*  phi_start  */
    int phi_stop = (int) call_info3;
    float *maximum = call_info4;
    Kcoord_3d spin_axis;
    float intensity;
    float xscale, yscale, xoffset, yoffset;
    double angle_from_spin_axis;
    trig_data_type *phi_ptr, *theta_ptr;
    dim_desc *angle_dim, *vel_dim;

    spin_axis.x = sin ( (double) off_axis_angle * PION180 );
    spin_axis.y = 0.0;
    spin_axis.z = -cos ( (double) off_axis_angle * PION180 );
    angle_dim = iarray_get_dim_desc (vel_vs_theta_arr, 0);
    vel_dim = iarray_get_dim_desc (cube, 0);
    xscale = NUM_X / 2.1;
    yscale = NUM_Y / 2.1;
    xoffset = NUM_X / 2;
    yoffset = NUM_Y / 2;
    for (phi_ptr = info->trig_data + phi; phi < phi_stop; ++phi, ++phi_ptr)
    {
	Kcoord_3d vec;
	unsigned int theta_count;

	intensity = pow (10.0, fabs (phi_ptr->cos) * beaming_power) *
	    info->iscale;
	for (theta_count = 0, theta_ptr = info->trig_data;
	     theta_count < NUM_TRIG_STEPS;
	     ++theta_count, ++theta_ptr)
	{
	    int x, y, index;
	    float velocity;

	    vec.x = phi_ptr->sin * theta_ptr->cos;
	    vec.y = phi_ptr->sin * theta_ptr->sin;
	    vec.z = phi_ptr->cos;
	    angle_from_spin_axis = fabs (vec.x * spin_axis.x +
					 vec.y * spin_axis.y +
					 vec.z * spin_axis.z);
	    angle_from_spin_axis = acos (angle_from_spin_axis);
	    index = angle_from_spin_axis * info->rad_to_index;
	    if (index < 0) index = 0;
	    else if (index >= angle_dim->length) index = angle_dim->length - 1;
	    velocity = F1 (vel_vs_theta_arr, index) * phi_ptr->cos;
	    x = (int) (vec.x * xscale + xoffset);
	    y = (int) (vec.y * yscale + yoffset);
	    index = (velocity -vel_dim->first_coord) * info->vel_norm_to_index;
	    F3 (cube, index, y, x) = intensity;
	}
	if (intensity > *maximum) *maximum = intensity;
    }
}   /*  End Function job_func  */

#ifdef OS_SunOS
static void dummy ()
/*  [PURPOSE] This routine is needed for SunOS, because of a problem with
    shared archive use. The <<_XtToolkitInitialize>> symbol is defined in
    the Xt shared archive, but it appears the shared archive is not linked for
    this application. Other applications (like <<krenzo>>) have no such
    problems. This is a truly bizzare bug.
*/
{
    XtToolkitInitialize ();
}   /*  End Function dummy  */
#endif
