/*  kmodel_spiral_xt.c

    Main file for  kmodel-spiral  (X11 3D spiral modelling tool for Karma).

    Copyright (C) 1999-2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable modelling of a spiral in 2D.
    This module runs on an X11 server.


    Written by      Richard Gooch   16-MAY-1999: Copied from kmodel_shell_xt.c

    Updated by      Richard Gooch   17-MAY-1999

    Updated by      Richard Gooch   24-MAY-1999: Increased profile array length
  to 2048 entries.

    Updated by      Richard Gooch   30-AUG-1999: Added dirname font resource.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Updated by      Richard Gooch   11-MAY-2001: Changed dirname font resource.

    Updated by      Richard Gooch   21-DEC-2001: Included k_version.h directly.

    Last updated by Richard Gooch   22-DEC-2002: Added #include <stdlib.h>


*/
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <errno.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <karma.h>
#include <k_event_codes.h>
#include <k_version.h>
#ifndef X11
#  define X11
#endif
#include <karma_viewimg.h>
#include <karma_foreign.h>
#include <karma_iarray.h>
#include <karma_canvas.h>
#include <karma_xtmisc.h>
#include <karma_dsxfr.h>
#include <karma_kwin.h>
#include <karma_conn.h>
#include <karma_ds.h>
#include <karma_im.h>
#include <karma_ic.h>
#include <karma_xc.h>
#include <karma_a.h>
#include <karma_m.h>
#include <Xkw/ProfileEdit.h>
#include <Xkw/SimpleSlider.h>
#include <Xkw/ExportMenu.h>
#include <Xkw/ImageDisplay.h>


#define VERSION "0.1.1"

#define IMAGE_SIZE       200



/*  Private functions  */
STATIC_FUNCTION (void quit_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void refresh_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void compute_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );


/*  Private data  */
String fallback_resources[] =
{
    "KModelSpiral.geometry:                                 +0+0",
    "KModelSpiral*Canvas*background:                        black",
    "KModelSpiral*Command*background:                       grey70",
    "KModelSpiral*Repeater*background:                      grey70",
    "KModelSpiral*Ktoggle*background:                       grey80",
    "KModelSpiral*closeButton*background:                   grey90",
    "KModelSpiral*ChoiceMenu.background:                    turquoise",
    "KModelSpiral*ExclusiveMenu.background:                 turquoise",
    "KModelSpiral*Value*background:                         #d0a0a0",
    "KModelSpiral*quit*background:                          orange",
    "KModelSpiral*SimpleSlider.foreground:                  Sea Green",
    "KModelSpiral*SimpleSlider.borderWidth:                 0",
    "KModelSpiral*SimpleMenu*font:                          10x20",
    "KModelSpiral*imageShell.geometry:                      -0+0",
    "KModelSpiral*ImageDisplay.Command.font:                7x13bold",
    "KModelSpiral*Filewin.dirname.textSink.font:            6x13",
    "KModelSpiral*font:                                     9x15bold",
    "KModelSpiral*borderColor:                              black",
    "KModelSpiral*background:                               aquamarine",
    "KModelSpiral*foreground:                               black",
    NULL
};
static XrmOptionDescRec Options[] =
{
    {"-verbose", "*verbose", XrmoptionNoArg, (XPointer) "True"},
};
static Widget main_shell = NULL;
static Widget profile_edit = NULL;
static iarray image = NULL;
static double num_rotations = 1.0;
static char title_name[STRING_LENGTH] = "Unknown";


int main (int argc, char **argv)
{
    KWorldCanvas profile_canvas, pc_canvas, tc_canvas, image_canvas;
    iarray theta_vs_radius_arr;
    XtAppContext app_context;
    Widget form, w, export_menu, rotation_sld, image_shell, image_display;
    static char function_name[] = "main";

    /*  Initialise module  */
    im_register_module_name ("kmodel-spiral");
    im_register_module_version_date (VERSION);
    im_register_lib_version (KARMA_VERSION);
    /*  Start up Xt  */
    main_shell = xtmisc_init_app_initialise (&app_context, "KModelSpiral",
					     Options, XtNumber (Options),
					     &argc, argv, fallback_resources,
					     XTMISC_INIT_ATT_COMMS_SETUP, TRUE,
					     XTMISC_INIT_ATT_BECOME_SERVER,
					     TRUE,
					     XTMISC_INIT_ATT_TITLE, title_name,
					     XTMISC_INIT_ATT_END,
					     NULL);
    dsxfr_register_connection_limits (-1, 0);
    xtmisc_set_icon (main_shell, ic_write_kplot_icon);
    form = XtVaCreateManagedWidget ("topForm", formWidgetClass, main_shell,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("exportMenu", exportMenuWidgetClass, form,
				 XkwNsupportSubsets, FALSE,
				 NULL);
    export_menu = w;
    w = XtVaCreateManagedWidget ("quit", commandWidgetClass, form,
				 XtNlabel, "Quit",
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, quit_cbk, NULL);
    w = XtVaCreateManagedWidget ("rotationSlider", simpleSliderWidgetClass,
				 form,
				 XtNlabel, "Number of Rotations",
				 XtNwidth, 300,
				 XtNfromVert, w,
				 XkwNminimum, 1,
				 XkwNmaximum, 300,
				 XkwNmodifier, 1,
				 XtNvalue, 200,
				 XkwNscaledValuePtr, &num_rotations,
				 XkwNscaledFormat, "%g",
				 NULL);
    XkwSimpleSliderSetScale (w, 1e-2, 0.0, FALSE, TRUE);
    rotation_sld = w;
    profile_edit = XtVaCreateManagedWidget ("form", profileEditWidgetClass,
					    form,
					    XtNborderWidth, 0,
					    XtNhorizDistance, 0,
					    XtNvertDistance, 0,
					    XtNfromVert, w,
					    NULL);
    XtRealizeWidget (main_shell);
    XtVaGetValues (profile_edit, XkwNworldCanvas, &profile_canvas, NULL);
    image_shell = XtVaCreatePopupShell ("imageShell", topLevelShellWidgetClass,
					main_shell,
					XtNtitle, "Image Window",
					NULL);
    image_display = XtVaCreateManagedWidget ("topForm",
					     imageDisplayWidgetClass,
					     image_shell,
					     XtNborderWidth, 0,
					     XkwNcreateFilepopup, FALSE,
					     /*XkwNnumTrackLabels, 3,*/
					     XkwNshowQuitButton, FALSE,
					     NULL);
    XtRealizeWidget (image_shell);
    XtVaGetValues (image_display,
		   XkwNpseudoColourCanvas, &pc_canvas,
		   XkwNtrueColourCanvas, &tc_canvas,
		   NULL);
    image_canvas = (pc_canvas == NULL) ? tc_canvas : pc_canvas;
    if (image_canvas == NULL)
    {
	fprintf (stderr, "PseudoColour or TrueColour canvas not found!\n");
	exit (RV_UNDEF_ERROR);
    }
    XtAddCallback (profile_edit, XtNcallback, compute_cbk, image_canvas);
    /*  Create array  */
    if ( ( theta_vs_radius_arr = iarray_create_1D (2048, K_FLOAT) ) == NULL )
    {
	m_abort (function_name, "array");
    }
    iarray_set_dim_name (theta_vs_radius_arr, 0, "Radius (pixels)", TRUE);
    iarray_set_world_coords (theta_vs_radius_arr, 0,
			     0.0, IMAGE_SIZE / 2 - 1.0);
    iarray_set_value_name (theta_vs_radius_arr, "ANGLE", TRUE);
    XtVaSetValues (profile_edit, XkwNiarray, theta_vs_radius_arr, NULL);
    canvas_set_attributes (profile_canvas,
			   CANVAS_ATT_TOP_Y, 360.0 * num_rotations,
			   CANVAS_ATT_END);
    XtAddCallback (rotation_sld, XkwNvalueChangeCallback, refresh_cbk,
		   profile_canvas);
    if ( (image  = iarray_create_2D (IMAGE_SIZE, IMAGE_SIZE, K_FLOAT) )
	 == NULL )
    {
	m_abort (function_name, "image");
    }
    iarray_set_dim_name (image, 0, "Y (pixels)", TRUE);
    iarray_set_dim_name (image, 1, "X (pixels)", TRUE);
    XtVaSetValues (export_menu,
		   XkwNiarray, image,
		   XkwNworldCanvas, profile_canvas,
		   NULL);
    XtPopup (image_shell, XtGrabNone);
    XtAppMainLoop (app_context);
    return (RV_OK);
}   /*  End Function main   */

static void quit_cbk (Widget w, XtPointer client_data, XtPointer call_data)

/*  This is the quit button callback.
*/
{
    exit (RV_OK);
}   /*  End Function quit_cbk   */

static void refresh_cbk (Widget w, XtPointer client_data, XtPointer call_data)

/*  This is the refresh callback.
*/
{
    KWorldCanvas canvas = client_data;

    canvas_set_attributes (canvas,
			   CANVAS_ATT_TOP_Y, 360.0 * num_rotations,
			   CANVAS_ATT_END);
    canvas_resize (canvas, NULL, TRUE);
}   /*  End Function refresh_cbk   */

static void compute_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the compute callback.
*/
{
    KWorldCanvas canvas = client_data;
    iarray theta_vs_radius_arr = call_data;
    int centre;
    unsigned long count;
    float radius, angle, xo, yo;
    dim_desc *dim;
    static ViewableImage vimage;
    static char function_name[] = "compute_cbk";

    iarray_clear (image);
    centre = IMAGE_SIZE / 2;
    F2 (image, centre, centre) = 1.0;
    dim = iarray_get_dim_desc (theta_vs_radius_arr, 0);
    for (count = 0; count < dim->length; ++count)
    {
	radius = iarray_get_coordinate (theta_vs_radius_arr, 0, count);
	angle = F1 (theta_vs_radius_arr, count);
	xo = radius * cos (angle * PION180);
	yo = radius * sin (angle * PION180);
	F2 (image, centre + (int) yo, centre + (int) xo) = 1.0;
	F2 (image, centre - (int) yo, centre - (int) xo) = 1.0;
    }
    iarray_set_data_range (image, 0.0, 1.0);
    if (vimage == NULL)
    {
	if ( ( vimage = viewimg_create_from_iarray (canvas, image, FALSE) )
	     == NULL )
	{
	    m_abort (function_name, "ViewableImage");
	}
	viewimg_make_active (vimage);
    }
    else viewimg_register_data_change (vimage);
#if 0
    canvas_set_attributes (canvas,
			   CANVAS_ATT_VALUE_MIN, min,
			   CANVAS_ATT_VALUE_MAX, max,
			   CANVAS_ATT_END);
#endif
    if (conn_get_num_client_connections ("multi_array") > 0)
	iarray_write (image, "connections");
}   /*  End Function compute_cbk   */

#ifdef OS_SunOS
static void dummy ()
/*  [PURPOSE] This routine is needed for SunOS, because of a problem with
    shared archive use. The <<_XtToolkitInitialize>> symbol is defined in
    the Xt shared archive, but it appears the shared archive is not linked for
    this application. Other applications (like <<krenzo>>) have no such
    problems. This is a truly bizzare bug.
*/
{
    XtToolkitInitialize ();
}   /*  End Function dummy  */
#endif
