/*  This Karma module will render a 3 dimensional data set into a 2 dimensional
    data set, using a simple perspective algorithm.
    The output is written to a new Karma arrayfile.


    Written by      Richard Gooch   24-MAY-1992

    Updated by      Richard Gooch   29-NOV-1992

    Updated by      Richard Gooch   25-DEC-1992: Added summation/ accumulation
  support.

    Updated by      Richard Gooch   3-JAN-1993: Took acccount of change to
  ds_alloc_data

    Updated by      Richard Gooch   4-JAN-1993: Added event blocking (wait for
  rendering module to finish).

    Updated by      Richard Gooch   27-MAR-1993: Took account of change to
  dsxfr_get_multi  .

    Updated by      Richard Gooch   17-MAY-1993: Took account of change to
  dsxfr_get_multi  .

    Updated by      Richard Gooch   8-AUG-1993: Tidied up by deallocating
  multi_array data structure when finished.

    Updated by      Richard Gooch   27-JUN-1999: Cosmetic changes.

    Updated by      Richard Gooch   30-JUN-1999: Removed <process_object> and
  moved some functionality into <post_process>.

    Last updated by Richard Gooch   22-DEC-2002: Added #include <stdlib.h>


*/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <errno.h>
#include <karma.h>
#include <karma_dsxfr.h>
#include <karma_dsrw.h>
#include <karma_conn.h>
#include <karma_ch.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_s.h>
#include "persp_3d.h"


/*  Private functions  */
static void write_array ();
static void send_finished_notice ();


/*  Private data  */
static multi_array *cumulative = NULL;
static unsigned int buf_array_len = 0;
static char *buffer_array = NULL;


flag render_cube (char *cube, unsigned int inp_lengths[3],
		  double *inp_coordinates[3], unsigned int inp_type,
		  unsigned int inp_stride, float *image,
		  dim_desc **out_dim_desc, flag complex,
		  unsigned int out_stride, double focus[3], double position[3],
		  double vertical[3], double lower_threshold,
		  double upper_threshold)
/*  This routine will render a single cube into an image.
    The cube data must be pointed to by  cube  .
    The lengths of the cube sides must be in the array pointed to by
    inp_lengths  .
    The co-ordinate arrays of the cube sides must be pointed to by
    inp_coordinates  .
    The type of the cube data must be given by  inp_type  .
    The stride (in bytes) between consecutive data points in the cube must be
    given by  inp_stride  .
    The image data must be pointed to by  image  .
    The dimension descriptors for the image must be pointed to by by the array
    of pointers pointed to by  out_dim_desc  .
    If the image is of type K_COMPLEX, then  complex  must be TRUE.
    The stride (in floats) between consecutive data points in the image must be
    given by  out_stride  .
    The focus point must be pointed to by  focus  .
    The eye position must be pointed to by  position  .
    The vertical direction vector must be pointed to by  vertical  .
    If a data point is less than or equal to  lower_threshold  then it is
    ignored.
    If a data point is greater than or equal to  upper_threshold  then it is
    ignored.
    The routine returns TRUE on success, else it returns FALSE.
*/
{
    unsigned int coord_count;
    unsigned int num_coords;
    float coord_z, coord_y;
    unsigned int coord_counters[3];
    float one_on_sq1;
    float one_on_sq2;
    float one_on_sq3;
    float point_v0;
    double direction[3];
    double horizontal[3];
    unsigned int *yoffsets;
    float *fp_ptr;
    float *prep_arrays;
    static char function_name[] = "render_cube";

    if ( (cube == NULL) || (inp_lengths == NULL) || (inp_coordinates == NULL)
	|| (image == NULL) || (out_dim_desc == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    /*  Test for legal stride  */
    if (out_stride != 1)
    {
	fprintf (stderr, "out_stride must be: 1, not: %u\n",out_stride);
	a_prog_bug (function_name);
    }
    /*  Compute vision direction vector  */
    direction[0] = focus[0] - position[0];
    direction[1] = focus[1] - position[1];
    direction[2] = focus[2] - position[2];
    /*  Compute horizontal vector  */
    horizontal[0] = direction[1] * vertical[2] - direction[2] * vertical[1];
    horizontal[1] = direction[2] * vertical[0] - direction[0] * vertical[2];
    horizontal[2] = direction[0] * vertical[1] - direction[1] * vertical[0];
    /*  Compute some square roots  */
    one_on_sq1 = 1.0 / sqrt (horizontal[0] * horizontal[0] +
			     horizontal[1] * horizontal[1] +
			     horizontal[2] * horizontal[2]);
    one_on_sq2 = 1.0 / sqrt (vertical[0] * vertical[0] +
			     vertical[1] * vertical[1] +
			     vertical[2] * vertical[2]);
    one_on_sq3 = 1.0 / sqrt (direction[0] * direction[0] +
			     direction[1] * direction[1] +
			     direction[2] * direction[2]);
    /*  Allocate precomputed data array  */
    if ( ( prep_arrays = m_alloc (4 * sizeof *prep_arrays * inp_lengths[2]) )
	 == NULL )
    {
	m_error_notify (function_name, "prep_arrays");
	return (FALSE);
    }
    fp_ptr = prep_arrays;
    for (coord_count = 0; coord_count < inp_lengths[2];
	 ++coord_count, prep_arrays += 4)
    {
	curr_distance_2d = (direction[0] * inp_coordinates[2][coord_count]);
	point_v0 = inp_coordinates[2][coord_count] - position[0];
	curr_sq_point_v0 = point_v0 * point_v0;
	curr_coord_h = horizontal[0] * point_v0;
	curr_coord_v = vertical[0] * point_v0;
    }
    prep_arrays = fp_ptr;
    /*  Allocate and precompute data used in calculation of 2D distance  */
    if (inp_type != K_BYTE)
    {
	a_func_abort (function_name, "byte only");
	return (FALSE);
    }
    if ( ( yoffsets = m_alloc (sizeof *yoffsets * out_dim_desc[0]->length) )
	 == NULL )
    {
	fprintf (stderr, "y offsets");
	m_error_notify (function_name, "coord_v_array");
	m_free ( (char *) prep_arrays );
	return (FALSE);
    }
    for (coord_count = 0; coord_count < (*out_dim_desc[0]).length;
	 ++coord_count)
    {
	yoffsets[coord_count] = (coord_count * (*out_dim_desc[1]).length *
				 out_stride);
    }
    /*  Initialise image plane  */
    num_coords = (*out_dim_desc[0]).length * (*out_dim_desc[1]).length;
    if (complex)
    {
	num_coords *= 2;
    }
    for (coord_count = 0, fp_ptr = image;
	 coord_count < num_coords;
	 ++coord_count) *fp_ptr++ = TOOBIG;
    /*  Iterate through planes  */
    for (coord_counters[0] = 0; coord_counters[0] < inp_lengths[0];
	 ++coord_counters[0])
    {
	coord_z = inp_coordinates[0][ coord_counters[0] ];
	/*  Iterate through lines  */
	for (coord_counters[1] = 0; coord_counters[1] < inp_lengths[1];
	     ++coord_counters[1])
	{
	    coord_y = inp_coordinates[1][ coord_counters[1] ];
	    /*  Convert values along line  */
	    if ( !project_line_real_byte (cube, coord_z, coord_y,
					  inp_lengths[2], inp_stride,
					  image, out_dim_desc, yoffsets,
					  position, vertical, horizontal,
					  direction,
/*
					  (char) lower_threshold,
					  (char) upper_threshold,
*/
					  -128, 127,
					  one_on_sq1, one_on_sq2, one_on_sq3,
					  prep_arrays) )
	    {
		fprintf (stderr, "Error projecting line\n");
		return (FALSE);
	    }
	    cube += inp_stride * inp_lengths[2];
	    if ( s_check_for_int () )
	    {
		a_func_abort (function_name, "control-C abort");
		m_free ( (char *) prep_arrays );
		return (FALSE);
	    }
	}
    }
    /*  Cube rendered: return OK  */
    m_free (prep_arrays);
    m_free (yoffsets);
    return (TRUE);
}   /*  End Function render_cube  */

flag post_process (CONST multi_array *inp_desc, multi_array *out_desc)
/*  [SUMMARY] Perform post-processing on a multi_array data structure.
    <inp_desc> The input multi_array descriptor.
    <out_desc> The output multi_array_descriptor.
    [RETURNS] TRUE if the array is to be saved/transmitted, else FALSE.
*/
{
    unsigned int array_size;
    packet_desc *top_pack_desc;
    array_desc *arr_desc;
    static char function_name[] = "post_process";

    if (conn_get_num_serv_connections ("persp_3d_images") < 1)
    {
	/*  No need to accumulate images  */
	send_finished_notice ();
	return (TRUE);
    }
    /*  Must accumulate images: first increment the attachment count  */
    ++out_desc->attachments;
    cumulative = out_desc;
    top_pack_desc = out_desc->headers[0];
    arr_desc = (array_desc *) top_pack_desc->element_desc[0];
    array_size = ds_get_array_size (arr_desc);
    array_size *= ds_get_packet_size (arr_desc->packet);
    if (array_size > buf_array_len)
    {
	if (buffer_array != NULL) m_free (buffer_array);
	buf_array_len = 0;
    }
    if ( ( buffer_array = m_alloc (array_size) ) == NULL )
    {
	m_abort (function_name, "buffer array");
    }
    buf_array_len = array_size;
    fprintf (stderr, "%s: returning, waiting for more images...\n",
	     function_name);
    return (FALSE);  /*  Prevent writing out by <dsproc_object>  */
}   /*  End Function post_process  */

flag verify_connect (void **info)
/*  This routine will validate whether it is appropriate to open a
    connection.
    The routine will write any appropriate information to the pointer
    pointed to by  info  .The pointer value written here will be passed
    to the other routines.
    The routine returns TRUE on if the connection should be attempted,
    else it returns FALSE (indicating the connection should be aborted).
    NOTE: Even if this routine is called and returns TRUE, there is no
    guarantee that the connection will be subsequently opened.
*/
{
    if (conn_get_num_client_connections ("multi_array") > 0)
    {
	fprintf (stderr,
		 "Cannot open connection while \"multi_array\" connections open\n");
	return (FALSE);
    }
    fprintf (stderr, "persp_3d_images client connect...\n");
    return (TRUE);
}   /*  End Function verify_connect  */

flag image_open_func (Connection connection, void **info)
/*  This routine will register the opening of a connection.
    The connection will be given by  connection  .
    The routine will write any appropriate information to the pointer
    pointed to by  info  .
    The routine returns TRUE on successful registration,
    else it returns FALSE (indicating the connection should be closed).
    Note that the  close_func  will not be called if this routine returns
    FALSE.
*/
{
    flag *status;
    static char function_name[] = "image_open_func";

    if ( ( status = (flag *) m_alloc (sizeof *status) ) == NULL )
    {
	m_abort (function_name, "status flag");
    }
    *info = (void *) status;
    *status = FALSE;
    fprintf (stderr, "New persp_3d_images incoming connection\n");
    return (TRUE);
}   /*  End Function image_open_func  */

flag image_read_func (Connection connection, void **info)
/*  This routine will read in data from the connection given by  connection
    and will write any appropriate information to the pointer pointed to by
    info  .
    The routine returns TRUE on successful reading,
    else it returns FALSE (indicating the connection should be closed).
    Note that the  close_func  will be called if this routine returns FALSE
*/
{
    flag all_done = TRUE;
    unsigned int count;
    unsigned int num_conn;
    Channel channel;
    Connection conn;
    flag *status;
    float *inp;
    float *cum;
    array_desc *inp_arr_desc;
    array_desc *cum_arr_desc;
    char element[sizeof (char *) + sizeof (unsigned int)];
    extern char *buffer_array;
    extern multi_array *cumulative;
    static char function_name[] = "image_read_func";

    fprintf (stderr, "New persp_3d_images incoming data\n");
    if (cumulative == NULL)
    {
	fprintf (stderr, "Image data came before rendering\n");
	exit (RV_BAD_DATA);
    }
    cum_arr_desc = (array_desc *) cumulative->headers[0]->element_desc[0];
    status = (flag *) *info;
    if (*status)
    {
	/*  Old data not yet used  */
	fprintf (stderr, "New image data came before old data used\n");
	exit (RV_BAD_DATA);
    }
    *status = TRUE;
    channel = conn_get_channel (connection);
    /*  Read array descriptor  */
    if ( ( inp_arr_desc = dsrw_read_array_desc (channel, K_ARRAY) ) == NULL )
    {
	return (FALSE);
    }
    /*  Compare array descriptors  */
    if ( !ds_compare_array_desc (inp_arr_desc, cum_arr_desc, TRUE) )
    {
	fprintf (stderr, "Image array descriptors different!\n");
	exit (RV_BAD_DATA);
    }
    /*  Read array data  */
    *(char **) element = buffer_array;
    *( (unsigned int *) ( element + sizeof (char *) ) ) =
	K_ARRAY_MEM_TYPE_NORMAL;
    if ( !dsrw_read_array (channel, inp_arr_desc, element, TRUE) )
    {
	return (FALSE);
    }
    /*  Accumulate  */
    fprintf (stderr, "%s: adding...\n", function_name);
    for (count = ds_get_array_size (inp_arr_desc),
	 inp = (float *) buffer_array, cum = *(float **) (*cumulative).data[0];
	 count > 0; --count, ++cum, ++inp)
    {
	if (*cum >= TOOBIG)
	{
	    *cum = *inp;
	}
	else
	{
	    if (*inp < TOOBIG)
	    {
		*cum += *inp;
	    }
	}
    }
    /*  Check if all incoming images processed  */
    num_conn = conn_get_num_serv_connections ("persp_3d_images");
    for (count = 0; count < num_conn; ++count)
    {
	if ( ( conn = conn_get_serv_connection ("persp_3d_images", count) )
	     == NULL )
	{
	    fprintf (stderr, "Could not get connection: %u for: \"%s\"\n",
		     count, "persp_3d_images");
	    a_prog_bug (function_name);
	}
	status = (flag *) conn_get_connection_info (conn);
	if (!*status) all_done = FALSE;
    }
    if (all_done)
    {
	fprintf (stderr, "%s: writing...\n", function_name);
	/*  Transmit result and reset statuses  */
	write_array (cumulative);
	cumulative = NULL;
	for (count = 0; count < num_conn; ++count)
	{
	    if ( ( conn = conn_get_serv_connection ("persp_3d_images", count) )
		 == NULL )
	    {
		fprintf (stderr, "Could not get connection: %u for: \"%s\"\n",
			 count, "persp_3d_images");
		a_prog_bug (function_name);
	    }
	    status = (flag *) conn_get_connection_info (conn);
	    *status = FALSE;
	}
    }
    fprintf (stderr, "%s: waiting...\n", function_name);
    return (TRUE);
}   /*  End Function image_read_func  */

static void write_array (multi_array *multi_desc)
/*  This routine will write the array descriptor and the array contained in a
    multi_array data data structure to a disc file or a connection.
    The data structure must be pointed to by  multi_desc  .
    The routine returns nothing.
*/
{
    Connection connection;
    Channel channel;
    char *top_packet;
    packet_desc *top_pack_desc;
    array_desc *arr_desc;
    static char function_name[] = "write_array";

    if (conn_get_num_client_connections ("persp_3d_images") < 1)
    {
	fprintf (stderr, "%s: writing normal...\n", function_name);
	/*  Not special image protocol  */
	if ( !dsxfr_put_multi ("connection[0]", multi_desc) )
	{
	    fprintf (stderr, "Error writing\n");
	}
	send_finished_notice ();
    }
    else
    {
	fprintf (stderr, "%s: writing image only...\n", function_name);
	if ( ( connection =
	       conn_get_client_connection ("persp_3d_images", 0) )
	     == NULL )
	{
	    fprintf (stderr, "Could not get connection: 0 for: \"%s\"\n",
		     "persp_3d_images");
	    a_prog_bug (function_name);
	}
	channel = conn_get_channel (connection);
	top_pack_desc = (*multi_desc).headers[0];
	top_packet = (*multi_desc).data[0];
	arr_desc = (array_desc *) (*top_pack_desc).element_desc[0];
	dsrw_write_array_desc (channel, arr_desc);
	dsrw_write_array (channel, arr_desc, top_packet, TRUE);
	ch_flush (channel);
    }
    ds_dealloc_multi (multi_desc);
}   /*  End Function write_array  */

static void send_finished_notice ()
/*  This routine will notify any "processing_finished" clients that the module
    has finished processing.
    The routine returns nothing.
*/
{
    char buffer = 0;
    unsigned int conn_count;
    unsigned int num_conn;
    Channel channel;
    Connection connection;
    static char function_name[] = "send_finished_notice";

    num_conn = conn_get_num_serv_connections ("processing_finished");
    for (conn_count = 0; conn_count < num_conn; ++conn_count)
    {
	if ( ( connection = conn_get_serv_connection ("processing_finished",
						      conn_count) )
	     == NULL )
	{
	    fprintf (stderr,
		     "Error getting \"processing_finished\" connection: %u\n",
		     conn_count);
	    a_prog_bug (function_name);
	}
	channel = conn_get_channel (connection);
	if (ch_write (channel, &buffer, 1) < 1)
	{
	    fprintf (stderr, "Error writing finished notice\t%s\n",
		     ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	if ( !ch_flush (channel) )
	{
	    fprintf (stderr, "Error flushing channel\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
    }
}   /*  End Function send_finished_notice  */
