/*  compute.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1993-2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This module controls the  xray  rendering module.


    Written by      Richard Gooch   21-JAN-1993

    Updated by      Richard Gooch   27-SEP-1993

    Updated by      Richard Gooch   16-NOV-1993: Took account of change to
  panel_  package in  vximcat  .

    Updated by      Richard Gooch   26-NOV-1993: Added histogram display in
  substance control window.

    Updated by      Richard Gooch   30-JAN-1994: Moved determination of
  whether to freeze controls (based on "processing_finished" clients) from
  load_cube  to  compute  .

    Updated by      Richard Gooch   18-FEB-1994: Created rotation and view
  structures.

    Updated by      Richard Gooch   22-FEB-1994: Added VX orientation display
  support.

    Updated by      Richard Gooch   20-APR-1994: Changed to
  vximcat_3D_cursor_position  protocol and defined new  3D_cursor_position
  protocol.

    Updated by      Richard Gooch   1-AUG-1994: Ported to Xt.

    Updated by      Richard Gooch   2-AUG-1994: Created  transmit_cont_hg_info

    Updated by      Richard Gooch   6-AUG-1994: Added  #include <karma_pio.h>
  which fouled up continuous hot gas write on Sparcs.

    Updated by      Richard Gooch   21-SEP-1994: Added primitive stereo support

    Updated by      Richard Gooch   19-NOV-1994: Added mono hot gas support.

    Updated by      Richard Gooch   23-NOV-1994: Fixed undersampling bug due to
  incorrect extent computation.

    Updated by      Richard Gooch   26-DEC-1994: Added "array_desc_data"
  server protocol support.

    Updated by      Richard Gooch   8-JAN-1995: Took account of change to
  <panel_> package in <raycast_3d>.

    Updated by      Richard Gooch   23-APR-1995: Added support for foreign data
  formats.

    Updated by      Richard Gooch   24-APR-1995: Forced conversion of FITS data
  to floats.

    Updated by      Richard Gooch   28-AUG-1995: Added raycast_mode support.

    Updated by      Richard Gooch   29-AUG-1995: Use global eye separation val.

    Updated by      Richard Gooch   30-AUG-1995: Changed to
  <canvas_set_attributes> to refresh canvas.

    Updated by      Richard Gooch   1-SEP-1995: Stripped <kcmap_>
  initialisation.

    Updated by      Richard Gooch   8-OCT-1995: Destroyed all ViewableImages in
  <compute>.

    Updated by      Richard Gooch   8-FEB-1996: Changed distance units to cm.

    Updated by      Richard Gooch   6-MAY-1996: Added support for
  "hot_gas_three_colour" shader.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   19-JUN-1996: Added astronomical projection
  support.

    Updated by      Richard Gooch   31-JUL-1996: Now use filtered/unfiltered
  float cubes.

    Updated by      Richard Gooch   21-NOV-1996: Dispatch byte cube change
  events through <<new_byte_cube_event_list>>.

    Updated by      Richard Gooch   23-NOV-1996: Moved substance code to
  subs_hg.c

    Updated by      Richard Gooch   7-DEC-1996: Changed subcube sliders to be
  in pixel units, rather than percent.

    Updated by      Richard Gooch   10-DEC-1996: Added code to <compute> to
  safely clear wireframe overlay lists without refreshing canvases.

    Updated by      Richard Gooch   22-DEC-1996: Set world co-ordinates for
  orientation canvas when new file is loaded.

    Updated by      Richard Gooch   27-DEC-1996: No longer load cube, just
  process a multi_array descriptor in renamed function <process_cube>.

    Updated by      Richard Gooch   21-MAR-1997: Cosmetic improvment in
  <compute> function for toggling canvas visibilities.

    Updated by      Richard Gooch   22-APR-1997: Fixed bug in <process_cube>
  where byte cube was being incorrectly deallocated when input cube was not
  byte and not float.

    Updated by      Richard Gooch   25-MAY-1997: No longer manipulate global
  voxel sum offset variable directly.

    Updated by      Richard Gooch   11-JUN-1998: Added call to function
  <register_histogram_change> in <compute_histogram> and removed offset
  histogram canvas update.

    Updated by      Richard Gooch   29-JUN-1998: Removed substance, continuous
  HG and mono HG histogram canvas updates from <compute_histogram>.

    Updated by      Richard Gooch   1-SEP-1998: Now reset scaling information
  if scaling information applied.

    Updated by      Richard Gooch   8-OCT-1998: Move freeze/unfreeze controls
  to main_xt.c. Added subcube validation to <compute>.

    Updated by      Richard Gooch   10-SEP-1999: Fixed bug in <process_cube>
  when loading a smaller cube.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Updated by      Richard Gooch   11-JUL-2000: Removed unnecessary #includes.

    Updated by      Richard Gooch   12-JUL-2000: Added #include <karma_kwin.h>.

    Last updated by Richard Gooch   22-DEC-2002: Stupid compiler warning fix.


*/
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/param.h>
#include <dirent.h>
#include <errno.h>
#include <karma.h>
#include <karma_vrender.h>
#include <karma_overlay.h>
#include <karma_canvas.h>
#include <karma_iarray.h>
#include <karma_kwin.h>
#include <karma_conn.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_a.h>
#include <karma_r.h>
#include <karma_m.h>
#include "xraycast_3d.h"


/*  Private functions  */


/*  Private data  */
static double histogram_array[255];


/*  Public routines follow  */

flag process_cube (multi_array *multi_desc)
/*  [SUMMARY] Process a new cube data structure.
    <multi_desc> The multi_array structure.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    iarray raw_cube;
    uaddr ua_tmp;
    double extent;
    double raw_cube_scale, raw_cube_offset;
    double scale[2], offset[2];
    extern KWorldCanvas orientation_worldcanvas;
    extern iarray byte_cube, filtered_cube, unfiltered_cube;
    extern KwcsAstro astro_projection;
    extern flag byte_cube_valid;
    static flag first_time = TRUE;
    extern i_coord_3d subcube_start, subcube_end;
    static char function_name[] = "process_cube";

    if (first_time)
    {
	first_time = FALSE;
	conn_attempt_connection ("unix", r_get_def_port ("khogd", ":0.0"),
				 "hog_request");
    }
    /*  Since non-signed byte data is converted to float, and FITS does not
	have signed byte data, may as well convert to float now to save
	problems later. Also means we get the scaling information
    */
    byte_cube_valid = FALSE;
    if (byte_cube != NULL)
    {
	iarray_dealloc (byte_cube);
	byte_cube = NULL;
    }
    if (filtered_cube != NULL)
    {
	iarray_dealloc (filtered_cube);
	filtered_cube = NULL;
    }
    if (unfiltered_cube != NULL)
    {
	iarray_dealloc (unfiltered_cube);
	unfiltered_cube = NULL;
    }
    /*  Update astro_projection  */
    if (astro_projection != NULL) wcs_astro_destroy (astro_projection);
    astro_projection = wcs_astro_setup (multi_desc->headers[0],
					multi_desc->data[0]);
    fprintf (stderr, "astro_projection: %p\n", (void *) astro_projection);
    if ( ( raw_cube = iarray_get_from_multi_array (multi_desc, NULL,
						   3, NULL, NULL) )
	== NULL )
    {
	return (FALSE);
    }
    /*  Ensure subcube is valid  */
    if ( subcube_start.x >= iarray_dim_length (raw_cube, 2) )
	subcube_start.x = 0;
    if ( subcube_end.x >= iarray_dim_length (raw_cube, 2) )
	subcube_end.x = iarray_dim_length (raw_cube, 2) - 1;
    if (subcube_end.x <= subcube_start.x)
	subcube_end.x = iarray_dim_length (raw_cube, 2) - 1;
    if ( subcube_start.y >= iarray_dim_length (raw_cube, 1) )
	subcube_start.y = 0;
    if ( subcube_end.y >= iarray_dim_length (raw_cube, 1) )
	subcube_end.y = iarray_dim_length (raw_cube, 1) - 1;
    if (subcube_end.y <= subcube_start.y)
	subcube_end.y = iarray_dim_length (raw_cube, 1) - 1;
    if ( subcube_start.z >= iarray_dim_length (raw_cube, 0) )
	subcube_start.z = 0;
    if ( subcube_end.z >= iarray_dim_length (raw_cube, 0) )
	subcube_end.z = iarray_dim_length (raw_cube, 0) - 1;
    if (subcube_end.z <= subcube_start.z)
	subcube_end.z = iarray_dim_length (raw_cube, 0) - 1;
    /*  Get raw cube scaling information  */
    iarray_get_data_scaling (raw_cube, &raw_cube_scale, &raw_cube_offset);
    /*  Get extent  */
    ua_tmp = iarray_dim_length (raw_cube, 0);
    extent = (ua_tmp * ua_tmp);
    ua_tmp = iarray_dim_length (raw_cube, 1);
    extent += (ua_tmp * ua_tmp);
    ua_tmp = iarray_dim_length (raw_cube, 2);
    extent += (ua_tmp * ua_tmp);
    extent = sqrt (extent);
    extent = ceil (extent / 2.0) - floor (-extent / 2.0);
    /*  Resize orientation canvas  */
    canvas_set_attributes (orientation_worldcanvas,
			   CANVAS_ATT_LEFT_X, -extent * 0.55,
			   CANVAS_ATT_RIGHT_X, extent * 0.55,
			   CANVAS_ATT_BOTTOM_Y, -extent * 0.55,
			   CANVAS_ATT_TOP_Y, extent * 0.55,
			   CANVAS_ATT_END);
    if (iarray_type (raw_cube) == K_BYTE)
    {
	/*  Input is signed bytes: this implies that no clipping or filtering
	    may be done  */
	byte_cube = raw_cube;
	register_new_byte_cube (byte_cube, raw_cube_scale, raw_cube_offset);
	fprintf ( stderr, "Loaded byte cube of size: %lu * %lu * %lu\n",
		  iarray_dim_length (raw_cube, 2),
		  iarray_dim_length (raw_cube, 1),
		  iarray_dim_length (raw_cube, 0) );
	return (TRUE);
    }
    /*  Input is not signed bytes: will need to allocate a byte cube  */
    if ( ( byte_cube = iarray_create_from_template (raw_cube, K_BYTE, TRUE,
						    TRUE, TRUE) ) == NULL )
    {
	m_error_notify (function_name, "byte Intelligent Array");
	iarray_dealloc (raw_cube);
	return (FALSE);
    }
    fprintf ( stderr, "Loaded non-byte cube of size: %lu * %lu * %lu\n",
	      iarray_dim_length (raw_cube, 2),
	      iarray_dim_length (raw_cube, 1),
	      iarray_dim_length (raw_cube, 0) );
    fprintf (stderr,"You will need to select an intensity range in the Range");
    fprintf (stderr, " Selector window\n");
    /*  Have to make sure a float unfiltered array is available  */
    if (iarray_type (raw_cube) == K_FLOAT)
    {
	unfiltered_cube = raw_cube;
	if ( (raw_cube_scale != 1.0) || (raw_cube_offset != 0.0) )
	{
	    /*  Even though this is a float cube, the bloody thing has some
		scaling applied to it: apply the scaling  */
	    fprintf (stderr, "Applying stupid scaling to float cube\n");
	    scale[0] = raw_cube_scale;
	    scale[1] = 1.0;
	    offset[0] = raw_cube_offset;
	    offset[1] = 0.0;
	    iarray_scale_and_offset (unfiltered_cube, raw_cube, scale, offset,
				     FALSE);
	    /*  Strip scaling information from the header  */
	    iarray_set_data_scaling (unfiltered_cube, 1.0, 0.0);
	}
	return (TRUE);
    }
    /*  Input cube is not byte or float: have to create the unfiltered float
	cube and convert the data  */
    if ( ( unfiltered_cube = iarray_create_from_template (raw_cube, K_FLOAT,
							  TRUE, TRUE, TRUE) )
	 == NULL )
    {
	m_error_notify (function_name, "float Intelligent Array");
	iarray_dealloc (raw_cube);
	iarray_dealloc (byte_cube);
	byte_cube = NULL;
	return (FALSE);
    }
    if ( (raw_cube_scale != 1.0) || (raw_cube_offset != 0.0) )
    {
	/*  Apply the scaling  */
	fprintf (stderr, "Applying scaling to cube\n");
    }
    scale[0] = raw_cube_scale;
    scale[1] = 0.0;
    offset[0] = raw_cube_offset;
    offset[1] = 0.0;
    iarray_scale_and_offset (unfiltered_cube, raw_cube, scale, offset, FALSE);
    /*  Strip scaling information from the header  */
    iarray_set_data_scaling (unfiltered_cube, 1.0, 0.0);
    return (TRUE);
}   /*  End Function process_cube  */

void compute ()
/*  This routine will compute.
*/
{
    extern KWorldCanvas visible_worldcanvas;
    extern KWorldCanvas left_image_pc_worldcanvas, right_image_pc_worldcanvas;
    extern KWorldCanvas left_image_rgb_worldcanvas,right_image_rgb_worldcanvas;
    extern KVolumeRenderContext main_vrend_ctx;
    extern KPixCanvas orientation_pixcanvas;
    extern KOverlayList cyclops_olist, left_olist, right_olist;
    extern flag byte_cube_valid;
    extern flag depth_plane_valid, running_movie;
    extern unsigned int raycast_mode;
    extern unsigned int num_frames;
    static char function_name[] = "compute";

    depth_plane_valid = FALSE;
    if (!byte_cube_valid) return;
    kwin_resize (orientation_pixcanvas, TRUE, 0, 0, 0, 0);
    freeze_save_movie_button (TRUE);
    show_animate_control (FALSE);
    if ( !running_movie || (num_frames == 0) )
    {
	/*  Empty out orientation overlay lists without refreshing canvases  */
	if (visible_worldcanvas != NULL)
	{
	    if ( (visible_worldcanvas == left_image_pc_worldcanvas) ||
		 (visible_worldcanvas == right_image_pc_worldcanvas) )
	    {
		kwin_set_attributes (canvas_get_pixcanvas
				     (left_image_pc_worldcanvas),
				     KWIN_ATT_VISIBLE, FALSE,
				     KWIN_ATT_END);
		kwin_set_attributes (canvas_get_pixcanvas
				     (right_image_pc_worldcanvas),
				     KWIN_ATT_VISIBLE, FALSE,
				     KWIN_ATT_END);
	    }
	    else if ( (visible_worldcanvas == left_image_rgb_worldcanvas) ||
		      (visible_worldcanvas == right_image_rgb_worldcanvas) )
	    {
		kwin_set_attributes (canvas_get_pixcanvas
				     (left_image_rgb_worldcanvas),
				     KWIN_ATT_VISIBLE, FALSE,
				     KWIN_ATT_END);
		kwin_set_attributes (canvas_get_pixcanvas
				     (right_image_rgb_worldcanvas),
				     KWIN_ATT_VISIBLE, FALSE,
				     KWIN_ATT_END);
	    }
	    else kwin_set_attributes (canvas_get_pixcanvas
				      (visible_worldcanvas),
				      KWIN_ATT_VISIBLE, FALSE,
				      KWIN_ATT_END);
	}
	overlay_remove_objects (cyclops_olist, 0);
	overlay_remove_objects (left_olist, 0);
	overlay_remove_objects (right_olist, 0);
	if (visible_worldcanvas != NULL)
	{
	    if ( (visible_worldcanvas == left_image_pc_worldcanvas) ||
		 (visible_worldcanvas == right_image_pc_worldcanvas) )
	    {
		kwin_set_attributes (canvas_get_pixcanvas
				     (left_image_pc_worldcanvas),
				     KWIN_ATT_VISIBLE, TRUE,
				     KWIN_ATT_END);
		kwin_set_attributes (canvas_get_pixcanvas
				     (right_image_pc_worldcanvas),
				     KWIN_ATT_VISIBLE, TRUE,
				     KWIN_ATT_END);
	    }
	    else if ( (visible_worldcanvas == left_image_rgb_worldcanvas) ||
		      (visible_worldcanvas == right_image_rgb_worldcanvas) )
	    {
		kwin_set_attributes (canvas_get_pixcanvas
				     (left_image_rgb_worldcanvas),
				     KWIN_ATT_VISIBLE, TRUE,
				     KWIN_ATT_END);
		kwin_set_attributes (canvas_get_pixcanvas
				     (right_image_rgb_worldcanvas),
				     KWIN_ATT_VISIBLE, TRUE,
				     KWIN_ATT_END);
	    }
	    else kwin_set_attributes (canvas_get_pixcanvas
				      (visible_worldcanvas),
				      KWIN_ATT_VISIBLE, TRUE,
				      KWIN_ATT_END);
	}
    }
    switch (raycast_mode)
    {
      case MODE_NORMAL:
	vrender_set_context_attributes (main_vrend_ctx,
					VRENDER_CONTEXT_ATT_SMOOTH_CACHE,FALSE,
					VRENDER_CONTEXT_ATT_END);
	break;
      case MODE_SMOOTH_CACHE:
      case MODE_SMOOTH_DELAYED:
      case MODE_SMOOTH_ALWAYS:
	vrender_set_context_attributes (main_vrend_ctx,
					VRENDER_CONTEXT_ATT_SMOOTH_CACHE, TRUE,
					VRENDER_CONTEXT_ATT_END);
	break;
      default:
	fprintf (stderr, "Illegal raycast_mode: %u\n", raycast_mode);
	a_prog_bug (function_name);
	break;
    }
    render ();
}   /*  End Function compute  */

void compute_histogram (iarray cube)
/*  This routine will compute and transmit the histogram of the cube voxels to
    any  kplot  modules.
    The cube must be given by  cube  .
*/
{
    int count;  /*  Doing signed arithmetic on this one  */
    unsigned long hpeak, hmode;
    unsigned long *array;
    double frac, y_min, y_max;
    double x_arr[255];
    multi_array *multi_desc;
    extern KOverlayList histogram_olist;
    extern flag verbose;
    extern unsigned int histogram_peak;
    extern int histogram_mode;
    extern CONST char *current_shader_name;
    static unsigned long lengths[1] = {255};
    static double minima[1];
    static double maxima[1];
    static CONST char *dim_names[1] = {"Bin"};
    static char function_name[] = "compute_histogram";

    minima[0] = -127;
    maxima[0] = 127;
    if ( ( array = (unsigned long *)
	  ds_easy_alloc_array (&multi_desc, 1, lengths, minima, maxima,
			       dim_names, K_ULONG, "Incidences") )
	== NULL )
    {
	m_error_notify (function_name, "histogram data structure");
	return;
    }
    /*  Compute histogram  */
    hpeak = 0;
    hmode = 0;
    if ( !iarray_compute_histogram (cube, KIMAGE_COMPLEX_CONV_REAL,
				    -127.0, 127.0, 255,
				    array, &hpeak, &hmode) )
    {
	fprintf (stderr, "Error computing histogram\n");
	a_prog_bug (function_name);
    }
    histogram_peak = hpeak;
    histogram_mode = (int) hmode - 127;
    if (verbose) fprintf (stderr, "Computed histogram. Peak: %u\tmode: %d\n",
			  histogram_peak, histogram_mode);
    for (count = 0; count < 255; count++)
    {
	histogram_array[count] = array[count];
	x_arr[count] = count - 127;
    }
    if ( !overlay_remove_objects (histogram_olist, 0) )
    {
	fprintf (stderr, "Error emptying histogram_olist\n");
	a_prog_bug (function_name);
    }
    y_max = histogram_peak;
    frac = log10 (y_max);
    if (frac > 6.0) frac = 6.0;
    y_min = y_max / pow (10.0, frac);
    if (y_min < 1.0) y_min = 1.0;
    if (strcmp (current_shader_name, "hot_gas_continuous") == 0)
	compute_cont_hg_info ();
    else if (strcmp (current_shader_name, "hot_gas_three_colour") == 0)
	compute_threecolour_hg_info ();
    register_histogram_change (y_min, y_max);
    if (overlay_lines (histogram_olist, 255, NULL, x_arr, histogram_array,
		       "white") == 0)
    {
	fprintf (stderr, "Error adding to histogram_olist\n");
	a_prog_bug (function_name);
    }
    ds_dealloc_multi (multi_desc);
}   /*  End Function compute_histogram  */


/*  Private routines follow  */
