/*  cont_hg.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1993-2000  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains the source for the continuous Hot Gas colourmap display.


    Written by      Richard Gooch   2-AUG-1994

    Updated by      Richard Gooch   3-AUG-1994

    Updated by      Richard Gooch   3-OCT-1994: Made use of
  canvas_use_log_scale  .

    Updated by      Richard Gooch   21-FEB-1996: Allow no colourmap wedge.

    Updated by      Richard Gooch   6-MAY-1996: Changed to shared arrays for
  colour and opacity information.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   25-MAY-1997: Made use of
  <transmit_fullcolour_info> function.

    Updated by      Richard Gooch   29-JUN-1998: Moved to callback mechanism to
  update histogram canvas when the histogram changes.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Last updated by Richard Gooch   12-JUL-2000: Added #include <karma_kcmap.h>
  and <karma_kwin.h>.

    $Id: cont_hg.c,v 2.1 2011/07/27 02:50:52 cal103 Exp $

*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_overlay.h>
#include <karma_canvas.h>
#include <karma_kcmap.h>
#include <karma_kwin.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_c.h>
#include "xraycast_3d.h"


/*  Private functions  */
STATIC_FUNCTION (void refresh_colourbar,
		 (KPixCanvas canvas, int width, int height, void **info) );
STATIC_FUNCTION (void cmap_resize_func, (Kcolourmap cmap, void **info) );
STATIC_FUNCTION (flag unit_canvas_position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info, void **f_info) );


/*  Private data  */
static Kcolourmap cont_hg_cmap = NULL;


/*  Public functions follow  */

void setup_cont_hg_canvas (KPixCanvas hist_pixcanvas,
			   KPixCanvas cmap_pixcanvas, Kcolourmap cmap)
/*  [PURPOSE] This routine will create the continuous hot gas control canvases.
    <hist_pixcanvas> The histogram pixel.
    <cmap_pixcanvas> The colourmap pixel canvas. If this is NULL, no colourmap
    is displayed.
    <cmap> The colourmap.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas worldcanvas;
    KWorldCanvas cont_hg_hist_worldcanvas;
    struct win_scale_type win_scale;
    extern KCallbackList histogram_change_event_list;
    /*static char function_name[] = "setup_cont_hg_canvas";*/

    cont_hg_cmap = cmap;
    /*  Setup the histogram canvas  */
    cont_hg_hist_worldcanvas = create_fullcolour_hg_canvas (hist_pixcanvas);
    /*  Create a unit world canvas  */
    canvas_init_win_scale (&win_scale, K_WIN_SCALE_MAGIC_NUMBER);
    win_scale.x_min = 0.0;
    win_scale.x_max = 1.0;
    win_scale.y_min = 0.0;
    win_scale.y_max = 1.0;
    win_scale.z_min = 0.0;
    win_scale.z_max = 1.0;
    win_scale.z_scale = K_INTENSITY_SCALE_LINEAR;
    if ( ( worldcanvas = canvas_create (hist_pixcanvas, cmap, &win_scale) )
	 == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    canvas_register_position_event_func (worldcanvas,
					 unit_canvas_position_func,
					 (void *) cmap);
    c_register_callback (&histogram_change_event_list,
			 generic_histogram_change_callback,
			 cont_hg_hist_worldcanvas,
			 NULL, FALSE, NULL, FALSE, FALSE);
    if (cmap_pixcanvas == NULL) return;
    /*  Setup the colourmap canvas  */
    kwin_register_refresh_func (cmap_pixcanvas, refresh_colourbar,
				(void *) cmap);
    kcmap_register_resize_func (cmap, cmap_resize_func,
				(void *) cmap_pixcanvas);
    /*  Create a unit world canvas  */
    if ( ( worldcanvas = canvas_create (cmap_pixcanvas, cmap, &win_scale) )
	 == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    canvas_register_position_event_func (worldcanvas,
					 unit_canvas_position_func,
					 (void *) cmap);
}   /*  End Function setup_cont_hg_canvas  */

void compute_cont_hg_info ()
/*  This routine will compute the continuous Hot Gas information.
    The alpha value must be given by  alpha  .
    The routine returns nothing.
*/
{
    int i_count, index, tmp;
    unsigned int size;
    float alpha;
    float vscale, opacity;
    float iscale = 1.0 / 65535.0;
    unsigned short *rgb_values;
    extern KVolumeRenderContext main_vrend_ctx;
    extern flag byte_cube_valid;
    extern int cont_hg_alpha_tenths;
    extern int histogram_mode;
    extern int cont_hg_blank_min, cont_hg_blank_max;
    extern float fullcolour_hg_reds[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_greens[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_blues[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_opacities[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    static char function_name[] = "compute_cont_hg_info";

    if (!byte_cube_valid) return;
    alpha = 0.1 * (float) cont_hg_alpha_tenths;
    if ( ( rgb_values = kcmap_get_rgb_values (cont_hg_cmap, &size) ) == NULL )
    {
	return;
    }
    if (size != FULLCOLOUR_HOT_GAS_CMAP_SIZE)
    {
	fprintf (stderr, "Colourmap should be %d entries long\n",
		 FULLCOLOUR_HOT_GAS_CMAP_SIZE);
	a_prog_bug (function_name);
    }
    if (histogram_mode < 0)
    {
	vscale = 1.0 / (float) (127 - histogram_mode);
    }
    else
    {
	vscale = 1.0 / (float) (histogram_mode + 127);
    }
    for (i_count = -127; i_count < 128; ++i_count)
    {
	tmp = i_count + 127;
	index = 3 * tmp;
	fullcolour_hg_reds[tmp] = (float) rgb_values[index] * iscale;
	fullcolour_hg_greens[tmp] = (float) rgb_values[index + 1] * iscale;
	fullcolour_hg_blues[tmp] = (float) rgb_values[index + 2] * iscale;
	if ( (cont_hg_blank_min < i_count) && (i_count < cont_hg_blank_max) )
	{
	    /*  Kill it entirely  */
	    opacity = 0.0;
	}
	else
	{
	    if (i_count < histogram_mode)
	    {
		opacity = vscale * (float) (histogram_mode - i_count);
	    }
	    else
	    {
		opacity = vscale * (float) (i_count - histogram_mode);
	    }
	    opacity = pow (opacity, alpha);
	}
	/*fprintf (stderr, "op: %e\n", opacity);*/
	fullcolour_hg_opacities[tmp] = opacity;
    }
    m_free (rgb_values);
    transmit_fullcolour_info (main_vrend_ctx);
}   /*  End Function compute_cont_hg_info  */


/*  Private routines follow  */

static void refresh_colourbar (canvas, width, height, info)
/*  This routine registers a refresh event consumer for a pixel canvas.
    The canvas is given by  canvas  .
    The width of the canvas is given by  width  .
    The height of the canvas is given by  height  .
    The arbitrary canvas information pointer is pointed to by  info  .
    The routine returns nothing.
*/
KPixCanvas canvas;
int width;
int height;
void **info;
{
    Kcolourmap cmap;
    int x;
    unsigned int num_allocated_pixels;
    unsigned long pixel_value;
    float factor;
    float half = 0.5;
    unsigned long *pixel_values;
    static char function_name[] = "refresh_colourbar";

    if ( (cmap = (Kcolourmap) *info) == NULL )
    {
	fprintf (stderr, "NULL colourmap\n");
	a_prog_bug (function_name);
    }
    num_allocated_pixels = kcmap_get_pixels (cmap, &pixel_values);
    if (num_allocated_pixels > width)
    {
	fprintf (stderr, "Window too narrow to fit colourbar%c\n", BEL);
	return;
    }
    factor = (float) (num_allocated_pixels - 1) / (float) (width - 1);
    for (x = 0; x < width; ++x)
    {
	pixel_value = pixel_values[(int) (factor * (float) x + half)];
	kwin_draw_line (canvas, x, 0, x, height - 1, pixel_value);
    }
}   /*  End Function refresh_colourbar  */

static void cmap_resize_func (cmap, info)
/*  This routine registers a change in the size of a colourmap.
    The colourmap must be given by  cmap  .
    The arbitrary colourmap information pointer is pointed to by  info  .
    The routine returns nothing.
*/
Kcolourmap cmap;
void **info;
{
    KPixCanvas canvas;
    static char function_name[] = "cmap_resize_func";

    if ( (canvas = (KPixCanvas) *info) == NULL )
    {
	fprintf (stderr, "NULL pixel canvas\n");
	a_prog_bug (function_name);
    }
    kwin_resize (canvas, TRUE, 0, 0, 0, 0);
}   /*  End Function cmap_resize_func  */

static flag unit_canvas_position_func (canvas, x, y, event_code, e_info,f_info)
/*  This routine is a position event consumer for a world canvas.
    The canvas is given by  canvas  .
    The horizontal world co-ordinate of the event will be given by  x  .
    The vertical world co-ordinate of the event will be given by  y  .
    The arbitrary event code is given by  event_code  .
    The arbitrary event information is pointed to by  e_info  .
    The arbitrary function information pointer is pointed to by  f_info  .
    The routine returns TRUE if the event was consumed, else it returns
    FALSE indicating that the event is still to be processed.
*/
KWorldCanvas canvas;
double x;
double y;
unsigned int event_code;
void *e_info;
void **f_info;
{
    Kcolourmap cmap;
    extern KPixCanvas cont_hg_hist_pixcanvas;
    static char function_name[] = "unit_canvas_position_func";

    if ( (cmap = (Kcolourmap) *f_info) == NULL )
    {
	fprintf (stderr, "NULL Kcolourmap\n");
	a_prog_bug (function_name);
    }
    if ( (event_code != K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK) &&
	(event_code != K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG) ) return (FALSE);
    kcmap_modify (cmap, x, y, NULL);
    compute_cont_hg_info ();
    kwin_resize (cont_hg_hist_pixcanvas, TRUE, 0, 0, 0, 0);
    return (TRUE);
}   /*  End Function unit_canvas_position_func  */
