/*  filter.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1996  Tom Oosterloo  original adaptive filtering code
    Copyright (C) 1996  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This file contains the code necessary for filtering.


    Written by      Richard Gooch   30-JUL-1996: Copied from Tom Oosterloo.

    Last updated by Richard Gooch   3-SEP-1996: Added continuum subtraction.


*/
#include <stdio.h>
#include <math.h>
#include <sys/param.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <karma_iarray.h>
#include "xraycast_3d.h"

#include "memutil.h"
#include "wavelet.h"

iarray filter_cube (iarray incube, unsigned int filter_algorithm,
		    unsigned int filter_level, unsigned int restore_level,
		    double filter_clip)
/*  [SUMMARY] Filter noise from a cube.
    <incube> The input floating-point cube.
    <filter_level> Range 0 to 3. Default 1.
    <restore_level> Use <<filter_level>> value for now.
    <filter_clip> Default 3.0
    [RETURNS] A filtered cube on success, else NULL.
*/
{
    iarray    wavarray, diff, addarray, tar, mask, noise, outcube;
    int       chan, z, y, x, bits, c, iter, num_iter;
    unsigned int xlen, ylen, zlen;
    float sub;
    double    mean, sigma;
  
    if (incube == NULL) return (NULL);
    xlen = iarray_dim_length (incube, 2);
    ylen = iarray_dim_length (incube, 1);
    zlen = iarray_dim_length (incube, 0);
    outcube = iarray_create_from_template (incube, K_FLOAT, TRUE, TRUE, TRUE);
    switch (filter_algorithm)
    {
      case FILTER_SUBTRACT_CONTINUUM:
	for (y = 0; y < ylen; y++) for (x = 0; x < xlen; x++)
	{
	    mean = 0.0;
	    for (z = 0; z < zlen; z++) mean += F3 (incube, z, y, x);
	    sub = mean / (float) zlen;
	    for (z = 0; z < zlen; z++)
		F3 (outcube, z, y, x) = F3 (incube, z, y, x) - sub;
	}
	break;
      default:
      case FILTER_3D:
	wavarray = iarray_create_4D ((filter_level)+1, 
				     zlen, ylen, xlen, K_FLOAT);
	/*get_fil_clip();*/
      
	iarray_wavelet_trans(incube, SPLINE_ATROUS, filter_level, wavarray, 
			     0, WAVELET_XYZ);
	for (chan = 0; chan <= restore_level; chan++) {
	    mean =  0.0;
	    for (z = 0; z < zlen; z++) {
		for (y = 0; y < 10; y++) {
		    for (x = 0; x < xlen; x++) {
			mean += F4(wavarray, chan, z, y, x);
		    }
		}
	    }
	    mean = mean/10.0/xlen/zlen;
	    sigma = 0.0;
	    for (z = 0; z < zlen; z++) {
		for (y = 0; y < 10; y++) {
		    for (x = 0; x < xlen; x++) {
			sigma += (F4(wavarray, chan, z, y, x)-mean)*
			    (F4(wavarray, chan, z, y, x)-mean);
		    }
		}
	    }
	    sigma = filter_clip *
		sqrt(sigma/(xlen*10.0*zlen-1.0));
	    for (z = 0; z < zlen; z++) {
		for (y = 0; y < ylen; y++) {
		    for (x = 0; x < xlen; x++) {
			if (fabs(F4(wavarray, chan, z, y, x)) < sigma) {
			    F4(wavarray, chan, z, y, x) = 0.0;
			}
		    }
		}
	    }
	}

	wavelet_to_iarray(wavarray, SPLINE_ATROUS, filter_level, outcube, 
			  0);

	iarray_dealloc(wavarray);
	break;
      case FILTER_SIMPLE:
	wavarray = iarray_create_3D((filter_level)+1, ylen, 
				    xlen, K_FLOAT);
	/*get_fil_clip();*/
      
	for (chan = 0; chan < zlen; chan++) {
	    iarray_to_wavelet(incube, SPLINE_ATROUS, filter_level, wavarray, 
			      chan, WAVELET_XY);
	    for (z = 0; z <= restore_level; z++) {
		mean =  0.0;
		for (y = 0; y < 10; y++) {
		    for (x = 0; x < xlen; x++) {
			mean += F3(wavarray, z, y, x);
		    }
		}
		mean = mean/(xlen)/10.0;
		sigma = 0.0;
		for (y = 0; y < 10; y++) {
		    for (x = 0; x < xlen; x++) {
			sigma += (F3(wavarray, z, y, x)-mean)*(F3(wavarray, z, y, x)-mean);
		    }
		}
		sigma = filter_clip*sqrt(sigma/(xlen*10-1.0));
    
		for (y = 0; y < ylen; y++) {
		    for (x = 0; x < xlen; x++) {
			if (fabs(F3(wavarray, z, y, x)) < sigma) {
			    F3(wavarray, z, y, x) = 0.0;
			}
		    }
		}
    
	    }
    
	    wavelet_to_iarray(wavarray, SPLINE_ATROUS, filter_level, outcube, 
			      chan);
    
	}
	iarray_dealloc(wavarray);
	break;
      case  FILTER_MEDMASK_SLOW:
      case  FILTER_MEDMASK_FAST:
	if (filter_algorithm == FILTER_MEDMASK_SLOW) {
	    num_iter = 6;
	} else {
	    num_iter = 2;
	}
        
	diff     = iarray_create_2D(ylen, xlen, K_FLOAT);  
	addarray = iarray_create_2D(ylen, xlen, K_FLOAT);  
	tar      = iarray_create_2D(ylen, xlen, K_FLOAT);  
	wavarray = iarray_create_3D(filter_level+1, ylen, 
                                    xlen, K_FLOAT);
	mask     = iarray_create_2D(ylen, xlen, K_INT);
	noise    = iarray_create_2D(2, filter_level+1, K_FLOAT);

	/*get_fil_clip();*/

	for (chan = 0; chan < zlen; chan++) {
	    for (iter = 0 ; iter < num_iter; iter++) {
		if (iter == 0) {
		    iarray_to_wavelet(incube, SPLINE_ATROUS, filter_level, 
				      wavarray, chan, WAVELET_XY);
		} else {
		    iarray_to_wavelet(outcube, SPLINE_ATROUS, filter_level, 
				      wavarray, chan, WAVELET_XY);
		}

		for (x = 0; x < xlen; x++) {
		    for (y = 0; y < ylen; y++) {
			I2(mask, y, x) = 0;
		    }
		}
 
		for (z = 0; z <= restore_level; z++) {
		    if (iter == 0) {
			mean =  0.0;
			for (x = 0; x < xlen; x++) {
			    for (y = 0; y < 10; y++) {
				mean += F3(wavarray, z, y, x);
			    }
			}
			mean = mean/xlen/10.0;
			sigma = 0.0;
			for (x = 0; x < xlen; x++) {
			    for (y = 0; y < 10; y++) {
				sigma += (F3(wavarray, z, y, x)-mean) * 
				    (F3(wavarray, z, y, x)-mean);
			    }
			}
			sigma = filter_clip*sqrt(sigma/(xlen*10.0-1.0));
			F2(noise, 0, z) = mean;
			F2(noise, 1, z) = sigma;
		    }
		}
  
		for (z = 0; z <= restore_level; z++) {
		    sigma = F2(noise, 1, z);
		    bits = 1 << z; 
		    for (y = 0; y < ylen; y++) {
			for (x = 0; x < xlen; x++) {
			    if (fabs(F3(wavarray, z, y, x)) > sigma) {
				I2(mask, y, x) += bits;
			    }
			}
		    }
		}
  
		mask_filter(&mask);
      
		for (z = 0; z <= restore_level; z++) {
		    bits = 1 << z;
		    for (y = 0; y < ylen; y++) {
			for (x = 0; x < xlen; x++) {
			    c = I2(mask, y, x) & bits;
			    if ( c == 0) {
				F3(wavarray, z, y, x) = 0.0;
			    }
			}
		    }
		}

		wavelet_to_iarray(wavarray, SPLINE_ATROUS, filter_level, tar, 0);
        
		for (y = 0; y < ylen; y++) {
		    for (x = 0; x < xlen; x++) {
			F2(diff, y, x) = F3(incube, chan, y, x) - F2(tar, y, x);
		    }
		}
        
		iarray_to_wavelet(diff, SPLINE_ATROUS, filter_level, wavarray, 
				  0, WAVELET_XY);

		for (z = 0; z <= restore_level; z++) {
		    bits = 1 << z; 
		    for (y = 0; y < ylen; y++) {
			for (x = 0; x < xlen; x++) {
			    c = I2(mask, y, x) & bits;
			    if ( c == 0) {
				F3(wavarray, z, y, x) = 0.0;
			    }
			}
		    }
		}

		wavelet_to_iarray(wavarray, SPLINE_ATROUS, filter_level, addarray, 0);
   
		for (y = 0; y < ylen; y++) {
		    for (x = 0; x < xlen; x++) {
			F3(outcube, chan, y, x) = F2(tar, y, x) + 
			    F2(addarray, y, x);
		    }
		}
	    }
	}
      
	iarray_dealloc(diff);
	iarray_dealloc(addarray);
	iarray_dealloc(tar);
	iarray_dealloc(wavarray);
	iarray_dealloc(mask);
	iarray_dealloc(noise);
	break;
    }
    return (outcube);
}
