/*  filter_xt.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1996-2000  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This file contains the code necessary for controlling the filtering.


    Written by      Richard Gooch   30-JUL-1996

    Updated by      Richard Gooch   31-JUL-1996

    Updated by      Richard Gooch   15-AUG-1996: Added unfilter button.

    Updated by      Richard Gooch   31-DEC-1996: Took account of change to
  to Dataclip widget.

    Updated by      Richard Gooch   31-OCT-1997: Took account of change to
  <XkwDataclipNewArray>.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Last updated by Richard Gooch   26-JAN-2000: Removed unnecessary includes.


*/
#include <stdio.h>
#include <errno.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Dialog.h>
#include <karma.h>
#ifndef X11
#  define X11
#endif
#include <karma_xtmisc.h>
#include <karma_iarray.h>
#include <karma_ex.h>
#include <Xkw/Value.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/Dataclip.h>
#include <Xkw/ImageDisplay.h>
#define NEED_FILTER_ALTERNATIVES
#include "xraycast_3d.h"

#define FILTER_MIN_LEVEL     0
#define FILTER_MAX_LEVEL     3
#define FILTER_DEFAULT_LEVEL 1


EXTERN_FUNCTION (iarray filter_cube,
		 (iarray incube, unsigned int filter_algorithm,
		  unsigned int filter_level, unsigned int restore_level,
		  double filter_clip) );

/*  Private functions  */
STATIC_FUNCTION (void filter_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void unfilter_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void save_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );


/*  Private data  */
static int filter_level = FILTER_DEFAULT_LEVEL;
static int filter_algorithm = FILTER_SIMPLE;
static Widget clip_dlg = NULL;
static Widget save_dlg = NULL;


Widget create_filter_win (Widget main_shell)
/*  [PURPOSE] This routine will create the filter control window.
    <main_shell> The application main shell.
    [RETURNS] The filter control popup.
*/
{
    Widget popup, form;
    Widget btn, sld;
    /*static char function_name[] = "create_filter_win";*/

    popup = XtVaCreatePopupShell ("filterwinpopup",
				  topLevelShellWidgetClass, main_shell,
				  XtNtitle, "Filter Control Window",
				  NULL);
    form = XtVaCreateManagedWidget ("topForm", formWidgetClass, popup,
				    XtNborderWidth, 0,
				    NULL);
    btn = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				   XtNlabel, "Close",
				   NULL);
    XtAddCallback (btn, XtNcallback, xtmisc_popdown_cbk, popup);
    btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				   XtNlabel, "Do Filter",
				   XtNfromHoriz, btn,
				   NULL);
    XtAddCallback (btn, XtNcallback, filter_cbk, NULL);
    btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				   XtNlabel, "Undo Filter",
				   XtNfromHoriz, btn,
				   NULL);
    XtAddCallback (btn, XtNcallback, unfilter_cbk, NULL);
    /*  Create the filter algorithm menu  */
    btn = XtVaCreateManagedWidget ("menuButton", exclusiveMenuWidgetClass,form,
				   XtNmenuName, "filterMenu",
				   XkwNchoiceName, "ALGORITHM",
				   XtNfromVert, btn,
				   XkwNnumItems, NUM_FILTER_ALTERNATIVES,
				   XkwNitemStrings, filter_alternatives,
				   XkwNvaluePtr, &filter_algorithm,
				   NULL);
    sld = XtVaCreateManagedWidget ("slider", valueWidgetClass, form,
				   XtNlabel, "Level",
				   XtNfromVert, btn,
				   XtNorientation, XtorientVertical,
				   XkwNminimum, FILTER_MIN_LEVEL,
				   XkwNmaximum, FILTER_MAX_LEVEL,
				   XkwNmodifier, 1,
				   XtNvalue, FILTER_DEFAULT_LEVEL,
				   XkwNvaluePtr, &filter_level,
				   NULL);
    clip_dlg = XtVaCreateManagedWidget ("filterLevel",  dialogWidgetClass,form,
					XtNlabel, "Filter clip   ",
					XtNfromVert, sld,
					XtNvalue, "3.0",
					NULL);
    btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				   XtNlabel, "Save",
				   XtNfromVert, clip_dlg,
				   NULL);
    XtAddCallback (btn, XtNcallback, save_cbk, NULL);
    save_dlg = XtVaCreateManagedWidget ("saveName",  dialogWidgetClass,form,
					XtNlabel, "Output file",
					XtNfromVert, btn,
					XtNwidth, 200,
					XtNvalue, "filtered",
					NULL);
    XtRealizeWidget (popup);
    return (popup);
}   /*  End Function create_filter_win   */


/*  Private functions follow  */

static void filter_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [SUMMARY] Filter callback.
*/
{
    double filter_clip;
    Bool auto_popdown;
    String image_name;
    char *clip_str;
    CONST char *p;
    char txt[STRING_LENGTH];
    extern iarray filtered_cube, unfiltered_cube;
    extern Widget dataclip_winpopup;
    extern Widget image_display;
    extern char module_name[STRING_LENGTH + 1];
    extern char module_version_date[STRING_LENGTH + 1];
    extern char module_lib_version[STRING_LENGTH + 1];
    extern char karma_library_version[STRING_LENGTH + 1];

    if (unfiltered_cube == NULL) return;
    XtVaGetValues (clip_dlg,
		   XtNvalue, &clip_str,
		   NULL);
    XtVaGetValues (dataclip_winpopup,
		   XkwNautoPopdown, &auto_popdown,
		   NULL);
    XtVaSetValues (dataclip_winpopup,
		   XkwNautoPopdown, FALSE,
		   NULL);
    filter_clip = ex_float (clip_str, &p);
    if (filtered_cube != NULL) iarray_dealloc (filtered_cube);
    XtVaSetValues (dataclip_winpopup,
		   XkwNautoPopdown, auto_popdown,
		   NULL);
    filtered_cube = filter_cube (unfiltered_cube, filter_algorithm,
				 filter_level, filter_level, filter_clip);
    if (filtered_cube == NULL)
    {
	fprintf (stderr, "Error filtering cube\n");
	XkwDataclipNewArray (dataclip_winpopup, unfiltered_cube,
			     TOOBIG, -TOOBIG, TRUE);
    }
    XkwDataclipNewArray (dataclip_winpopup, filtered_cube, TOOBIG, -TOOBIG,
			 TRUE);
    /*  Append a bit of history  */
    XtVaGetValues (image_display,
		   XkwNimageName, &image_name,
		   NULL);
    sprintf (txt, "%s: Module version %s  Karma v%s  compiled with v%s",
             module_name, module_version_date,
             karma_library_version, module_lib_version);
    iarray_append_history_string (filtered_cube, txt, TRUE);
    sprintf (txt, "%s: infile: %s  filter algorithm: %s",
             module_name, image_name, filter_alternatives[filter_algorithm]);
    iarray_append_history_string (filtered_cube, txt, TRUE);

    XBell (XtDisplay (w), -90);
}   /*  End Function filter_cbk   */

static void unfilter_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [SUMMARY] Unfilter callback.
*/
{
    extern iarray filtered_cube, unfiltered_cube;
    extern Widget dataclip_winpopup;

    if (unfiltered_cube == NULL) return;
    if (filtered_cube == NULL) return;
    XkwDataclipNewArray (dataclip_winpopup, unfiltered_cube, TOOBIG, -TOOBIG,
			 TRUE);
    iarray_dealloc (filtered_cube);
    filtered_cube = NULL;
}   /*  End Function unfilter_cbk   */

static void save_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [SUMMARY] Save callback.
*/
{
    char *filename;
    extern iarray filtered_cube;

    if (filtered_cube == NULL) return;
    XtVaGetValues (save_dlg,
		   XtNvalue, &filename,
		   NULL);
    fprintf (stderr, "Saving filtered cube: \"%s\"...", filename);
    iarray_write (filtered_cube, filename);
    fprintf (stderr, "done\n");
    XBell (XtDisplay (w), -90);
}   /*  End Function save_cbk   */
