/*  fullcolour_hg.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1996-2003  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains the source for the generic fullcolour Hot Gas colourmap
  display.


    Written by      Richard Gooch   6-MAY-1996: Copied bits from <cont_hg.h>

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   25-MAY-1997: Created
  <transmit_fullcolour_info> function.

    Last updated by Richard Gooch   21-MAR-2003: Switched to function
  <overlay_create_viewable>.

    $Id: fullcolour_hg.c,v 2.1 2011/07/27 02:39:44 cal103 Exp $

*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <karma.h>
#include <karma_vrender.h>
#include <karma_overlay.h>
#include <karma_canvas.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_p.h>
#include "xraycast_3d.h"


/*  Private functions  */
STATIC_FUNCTION (void plot_worldcanvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale, Kcolourmap cmap,
		  flag cmap_resize, void **info) );


/*  Private data  */


/*  Public functions follow  */

KWorldCanvas create_fullcolour_hg_canvas (KPixCanvas hist_pixcanvas)
/*  [PURPOSE] This routine will create a fullcolour hot gas display canvas.
    <hist_pixcanvas> The histogram pixel canvas.
    [RETURNS] The histogram world canvas object.
*/
{
    KWorldCanvas worldcanvas;
    KWorldCanvas hist_worldcanvas;
    KViewableOverlayList vlist;
    double frac;
    struct win_scale_type win_scale;
    extern KOverlayList histogram_olist;
    extern unsigned int histogram_peak;
    extern float fullcolour_hg_reds[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_greens[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_blues[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_opacities[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    /*static char function_name[] = "setup_fullcolour_hg_canvas";*/

    /*  Setup the histogram canvases  */
    canvas_init_win_scale (&win_scale, K_WIN_SCALE_MAGIC_NUMBER);
    win_scale.x_min = -127.0;
    win_scale.x_max = 127.0;
    if (histogram_peak < 1)
    {
	/*  No histogram data yet  */
	win_scale.y_min = 0.0;
	win_scale.y_max = 1.0;
    }
    else
    {
	/*  Have histogram data: setup logarithmic scale  */
	win_scale.y_max = histogram_peak;
	frac = log10 (win_scale.y_max);
	if (frac > 6.0) frac = 6.0;
	win_scale.y_min = win_scale.y_max / pow (10.0, frac);
	if (win_scale.y_min < 1.0) win_scale.y_min = 1.0;
	win_scale.y_log = TRUE;
    }
    if ( ( hist_worldcanvas = canvas_create (hist_pixcanvas, NULL,
					     &win_scale) ) == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    canvas_use_log_scale (hist_worldcanvas, FALSE, TRUE);
    vlist = overlay_create_viewable (histogram_olist, hist_worldcanvas);
    overlay_set_active (vlist, TRUE, FALSE, TRUE, FALSE, 0);
    /*  Setup the red canvas  */
    win_scale.y_min = 0.0;
    win_scale.y_max = 3.2;
    if ( ( worldcanvas =
	  canvas_create (hist_pixcanvas, NULL, &win_scale) ) == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    canvas_register_refresh_func (worldcanvas,
				  plot_worldcanvas_refresh_func,
				  (void *) fullcolour_hg_reds);
    /*  Setup the green canvas  */
    win_scale.y_min = -1.1;
    win_scale.y_max = 2.1;
    if ( ( worldcanvas =
	  canvas_create (hist_pixcanvas, NULL, &win_scale) ) == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    canvas_register_refresh_func (worldcanvas,
				  plot_worldcanvas_refresh_func,
				  (void *) fullcolour_hg_greens);
    /*  Setup the blue canvas  */
    win_scale.y_min = -2.2;
    win_scale.y_max = 1.0;
    if ( ( worldcanvas =
	  canvas_create (hist_pixcanvas, NULL, &win_scale) ) == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    canvas_register_refresh_func (worldcanvas,
				  plot_worldcanvas_refresh_func,
				  (void *) fullcolour_hg_blues);
    /*  Setup the opacity canvas  */
    win_scale.y_min = 0.0;
    win_scale.y_max = 1.0;
    if ( ( worldcanvas =
	  canvas_create (hist_pixcanvas, NULL, &win_scale) ) == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    canvas_register_refresh_func (worldcanvas,
				  plot_worldcanvas_refresh_func,
				  (void *) fullcolour_hg_opacities);
    return (hist_worldcanvas);
}   /*  End Function setup_fullcolour_hg_canvas  */

void transmit_fullcolour_info (KVolumeRenderContext context)
/*  [SUMMARY] Send Hot Gas Fullcolour info to slaves.
    <context> The volume rendering context.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    char *message, *ptr_red, *ptr_green, *ptr_blue, *ptr_opacity;
    extern float fullcolour_hg_reds[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_greens[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_blues[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_opacities[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    static char function_name[] = "transmit_fullcolour_info";

    message = m_alloc_scratch (FULLCOLOUR_HOT_GAS_CMAP_SIZE * 16,
			       function_name);
    ptr_red = message;
    ptr_green = ptr_red + FULLCOLOUR_HOT_GAS_CMAP_SIZE * 4;
    ptr_blue = ptr_green + FULLCOLOUR_HOT_GAS_CMAP_SIZE * 4;
    ptr_opacity = ptr_blue + FULLCOLOUR_HOT_GAS_CMAP_SIZE * 4;
    for (count = 0; count < FULLCOLOUR_HOT_GAS_CMAP_SIZE;
	 ++count, ptr_red += 4, ptr_green += 4, ptr_blue += 4,ptr_opacity += 4)
    {
	p_write_buf_float (ptr_red, fullcolour_hg_reds[count]);
	p_write_buf_float (ptr_green, fullcolour_hg_greens[count]);
	p_write_buf_float (ptr_blue, fullcolour_hg_blues[count]);
	p_write_buf_float (ptr_opacity, fullcolour_hg_opacities[count]);
    }
    vrender_send_message (context, MtoS_FULLCOLOUR_HOT_GAS_INFO, message,
			  FULLCOLOUR_HOT_GAS_CMAP_SIZE * 16, FALSE);
    m_free_scratch ();
}   /*  End Function transmit_fullcolour_info  */


/*  Private routines follow  */

static void plot_worldcanvas_refresh_func (canvas, width, height, win_scale,
					   cmap, cmap_resize, info)
/*  This routine is a refresh event consumer for a world canvas.
    The canvas is given by  canvas  .
    The width of the canvas in pixels is given by  width  .
    The height of the canvas in pixels is given by  height  .
    The window scaling information is pointed to by  win_scale  .
    The colourmap associated with the canvas is given by  cmap  .
    If the refresh function was called as a result of a colourmap resize
    the value of  cmap_resize  will be TRUE.
    The arbitrary canvas information pointer is pointed to by  info  .
    The routine returns nothing.
*/
KWorldCanvas canvas;
int width;
int height;
struct win_scale_type *win_scale;
Kcolourmap cmap;
flag cmap_resize;
void **info;
{
    unsigned int count;
    unsigned long pixel_value;
    char *colour = NULL;  /*  Initialise to keep compiler happy  */
    float *ptr;
    double tmp_arr[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_reds[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_greens[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_blues[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_opacities[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    static char function_name[] ="fullcolour_hg_plot_worldcanvas_refresh_func";

    ptr = (float *) *info;
    if (ptr == fullcolour_hg_opacities)
    {
	colour = "yellow";
    }
    else if (ptr == fullcolour_hg_reds)
    {
	colour = "red";
    }
    else if (ptr == fullcolour_hg_greens)
    {
	colour = "green";
    }
    else if (ptr == fullcolour_hg_blues)
    {
	colour = "blue";
    }
    else
    {
	(void) fprintf (stderr, "Unknown array\n");
	a_prog_bug (function_name);
    }
    if (canvas_get_colour (canvas, colour, &pixel_value,
			   (unsigned short *) NULL, (unsigned short *) NULL,
			   (unsigned short *) NULL) != TRUE) return;
    for (count = 0; count < FULLCOLOUR_HOT_GAS_CMAP_SIZE; ++count)
    {
	tmp_arr[count] = ptr[count];
    }
    canvas_draw_lines_p (canvas, (double *) NULL, tmp_arr,
			 FULLCOLOUR_HOT_GAS_CMAP_SIZE, pixel_value);
}   /*  End Function plot_worldcanvas_refresh_func  */
