/*
  Routines for memory (de)allocation of arrays, matrices etc.
  From Numerical Recipes in C, 1st edition, Press et al.


  Updates:    16-Nov-94:    change to use karma's m_alloc and m_free.
               2-Dec-94:    added double vector and matrix routines  
*/   


#define MEMUTIL
#include <karma_m.h>

#include "memutil.h"


float *mal_vector(nl, nh)
  int nl, nh;
{
/* allocates 1D float array using Karma m_alloc

  int nl, nh    bounds on array index. array should be used as array[nl,nh]

  Updates   16-Nov-94   created from old code

*/    
  float *v;

  if ((nh-nl+1) <= 0) {
    m_abort("mal_vector", "allocation failure in mal_vector()");
  }
    
  v = (float *) m_alloc((unsigned) (nh-nl+1)*sizeof(float));
  if (!v) {
    m_abort("mal_vector", "allocation failure in mal_vector()");
  }
  return(v-nl);
}

float **mal_matrix(nrl, nrh, ncl, nch)
  int nrl, nrh, ncl, nch;
{
/* allocates 2D float array using Karma m_alloc

  int nrl, nrh, ncl, ncr 
        bounds on array index.  array should be used as array[nrl,nrh][ncl,ncr]

  Updates   15-Nov-94   created from old code

*/    
  int i;
  float **m;

  if ((nrh-nrl+1) <= 0) {
    m_abort("mal_matrix", "allocation failure in mal_matrix()");
  }

  if ((nch-ncl+1) <= 0) {
    m_abort("mal_matrix", "allocation failure in mal_matrix()");
  }

  m = (float **) m_alloc((unsigned) (nrh-nrl+1)*sizeof(float*));
  if (!m) {
    m_abort("mal_matrix", "allocation failure 1 in mal_matrix()");
  }
  m -= nrl;

  for(i = nrl; i <= nrh; i++) {
    m[i] = (float *) m_alloc((unsigned) (nch-ncl+1)*sizeof(float));
    if (!m[i]) {
      m_abort("mal_matrix", "allocation failure 2 in matrix()");
    }
    m[i] -= ncl;
  }
  return(m);
}

double *mal_dvector(nl, nh)
  int nl, nh;
{
/* allocates 1D double  array using Karma m_alloc

  int nl, nh    bounds on array index. array should be used as array[nl,nh]

  Updates    2-Dec-94   created from mal_vector

*/    
  double *v;

  if ((nh-nl+1) <= 0) {
    m_abort("mal_vector", "allocation failure in mal_dvector()");
  }
    
  v = (double *) m_alloc((unsigned) (nh-nl+1)*sizeof(double));
  if (!v) {
    m_abort("mal_vector", "allocation failure in mal_dvector()");
  }
  return(v-nl);
}

double  **mal_dmatrix(nrl, nrh, ncl, nch)
  int nrl, nrh, ncl, nch;
{
/* allocates 2D double  array using Karma m_alloc

  int nrl, nrh, ncl, ncr 
        bounds on array index.  array should be used as array[nrl,nrh][ncl,ncr]

  Updates    2-Dec-94   created from mal_matrix

*/    
  int i;
  double  **m;

  if ((nrh-nrl+1) <= 0) {
    m_abort("mal_matrix", "allocation failure in mal_dmatrix()");
  }

  if ((nch-ncl+1) <= 0) {
    m_abort("mal_matrix", "allocation failure in mal_dmatrix()");
  }

  m = (double **) m_alloc((unsigned) (nrh-nrl+1)*sizeof(double*));
  if (!m) {
    m_abort("mal_matrix", "allocation failure 1 in mal_dmatrix()");
  }
  m -= nrl;

  for(i = nrl; i <= nrh; i++) {
    m[i] = (double *) m_alloc((unsigned) (nch-ncl+1)*sizeof(double));
    if (!m[i]) {
      m_abort("mal_matrix", "allocation failure 2 in dmatrix()");
    }
    m[i] -= ncl;
  }
  return(m);
}

void free_vector(v, nl, nh) 
  float *v;
  int   nl, nh;
{
/*
  Frees 1D array allocated with mal_vector. Uses Karma m_free.
  
  float *v    pointer to array to be freed
  int nl, nh  bounds on array index, as used in mal_vector

  Updates:    16-Nov-94 Created from old code
*/    
  m_free((char *) (v+nl));
}

void free_matrix(m, nrl, nrh, ncl, nch)
  float **m;
  int   nrl, nrh, ncl, nch;
{
/*
  Frees 2D array allocated with mal_matrix. Uses Karma m_free.
  
  float *v    pointer to array to be freed
  int nrl, nrh, ncl, nch  bounds on array index, as used in mal_matrix

  Updates:    16-Nov-94 Created from old code
*/    
  int i;

  for(i = nrh; i >= nrl; i--) {
    m_free((char *) (m[i]+ncl));
  }
  m_free((char *) (m+nrl));
}

void free_dvector(v, nl, nh) 
  double *v;
  int   nl, nh;
{
/*
  Frees 1D array allocated with mal_dvector. Uses Karma m_free.
  
  double  *v    pointer to array to be freed
  int nl, nh    bounds on array index, as used in mal_dvector

  Updates:     2-Dec-94 Created from free_vector
*/    
  m_free((char *) (v+nl));
}

void free_dmatrix(m, nrl, nrh, ncl, nch)
  double **m;
  int    nrl, nrh, ncl, nch;
{
/*
  Frees 2D array allocated with mal_dmatrix. Uses Karma m_free.
  
  double **v    pointer to array to be freed
  int nrl, nrh, ncl, nch  bounds on array index, as used in mal_dmatrix

  Updates:     2-Dec-94 Created from free_matrix
*/    
  int i;

  for(i = nrh; i >= nrl; i--) {
    m_free((char *) (m[i]+ncl));
  }
  m_free((char *) (m+nrl));
}

#undef MEMUTIL
