/*  mono_hg.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1993-2003  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains the source for the monochrome Hot Gas colourmap display.


    Written by      Richard Gooch   19-NOV-1994

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   25-MAY-1997: Transmit info to slaves.

    Updated by      Richard Gooch   29-JUN-1998: Moved to callback mechanism to
  update histogram canvas when the histogram changes.

    Updated by      Richard Gooch   22-DEC-2002: Added #include <stdlib.h>

    Last updated by Richard Gooch   21-MAR-2003: Switched to function
  <overlay_create_viewable>.


*/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_overlay.h>
#include <karma_canvas.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_p.h>
#include <karma_c.h>
#include "xraycast_3d.h"


/*  Private functions  */
STATIC_FUNCTION (void plot_worldcanvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale, Kcolourmap cmap,
		  flag cmap_resize, void **info) );


/*  Public data  */


/*  Private data  */


/*  Public functions follow  */

void setup_mono_hg_canvas (KPixCanvas pixcanvas)
/*  This routine will create the mono hot gas control canvases.
    The pixel canvas must be given by  hist_pixcanvas  .
    The colourmap must be given by  cmap  .
    The routine returns nothing.
*/
{
    KWorldCanvas worldcanvas;
    KWorldCanvas mono_hg_hist_worldcanvas;
    KViewableOverlayList vlist;
    double frac;
    struct win_scale_type win_scale;
    extern KOverlayList histogram_olist;
    extern KCallbackList histogram_change_event_list;
    extern unsigned int histogram_peak;
    extern float mono_hg_intensities[MONO_HOT_GAS_CMAP_SIZE];
    extern float mono_hg_opacities[MONO_HOT_GAS_CMAP_SIZE];
    /*static char function_name[] = "setup_mono_hg_canvas";*/

    /*  Setup the histogram canvases  */
    canvas_init_win_scale (&win_scale, K_WIN_SCALE_MAGIC_NUMBER);
    win_scale.x_min = -127.0;
    win_scale.x_max = 127.0;
    if (histogram_peak < 1)
    {
	/*  No histogram data yet  */
	win_scale.y_min = 0.0;
	win_scale.y_max = 1.0;
    }
    else
    {
	/*  Have histogram data: setup logarithmic scale  */
	win_scale.y_max = histogram_peak;
	frac = log10 (win_scale.y_max);
	if (frac > 6.0) frac = 6.0;
	win_scale.y_min = win_scale.y_max / pow (10.0, frac);
	if (win_scale.y_min < 1.0) win_scale.y_min = 1.0;
	win_scale.y_log = TRUE;
    }
    if ( ( mono_hg_hist_worldcanvas =
	   canvas_create (pixcanvas, NULL, &win_scale) ) == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    canvas_use_log_scale (mono_hg_hist_worldcanvas, FALSE, TRUE);
    vlist = overlay_create_viewable (histogram_olist,mono_hg_hist_worldcanvas);
    overlay_set_active (vlist, TRUE, FALSE, TRUE, FALSE, 0);
    /*  Setup the intensity canvas  */
    win_scale.y_min = 0.0;
    win_scale.y_max = 1.0;
    if ( ( worldcanvas =
	   canvas_create (pixcanvas, NULL, &win_scale) ) == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    canvas_register_refresh_func (worldcanvas,
				  plot_worldcanvas_refresh_func,
				  (void *) mono_hg_intensities);
    /*  Setup the opacity canvas  */
    win_scale.y_min = 0.0;
    win_scale.y_max = 1.0;
    if ( ( worldcanvas =
	  canvas_create (pixcanvas, NULL, &win_scale) ) == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    canvas_register_refresh_func (worldcanvas,
				  plot_worldcanvas_refresh_func,
				  (void *) mono_hg_opacities);
    c_register_callback (&histogram_change_event_list,
			 ( flag (*) () ) compute_mono_hg_info, NULL,
			 NULL, FALSE, NULL, FALSE, FALSE);
    c_register_callback (&histogram_change_event_list,
			 generic_histogram_change_callback,
			 mono_hg_hist_worldcanvas,
			 NULL, FALSE, NULL, FALSE, FALSE);
}   /*  End Function setup_mono_hg_canvas  */

void compute_mono_hg_info ()
/*  This routine will compute the mono Hot Gas information.
    The alpha value must be given by  alpha  .
    The routine returns nothing.
*/
{
    int i_count, index, tmp;
    float Ialpha, Oalpha;
    float vscale, intensity, opacity;
    extern flag byte_cube_valid;
    extern KVolumeRenderContext main_vrend_ctx;
    extern int mono_hg_intensity_alpha_tenths;
    extern int mono_hg_opacity_alpha_tenths;
    extern int histogram_mode;
    extern int mono_hg_blank_min, mono_hg_blank_max;
    extern float mono_hg_intensities[MONO_HOT_GAS_CMAP_SIZE];
    extern float mono_hg_opacities[MONO_HOT_GAS_CMAP_SIZE];
    /*static char function_name[] = "compute_mono_hg_info";*/

    if (!byte_cube_valid) return;
    Ialpha = 0.1 * (float) mono_hg_intensity_alpha_tenths;
    Oalpha = 0.1 * (float) mono_hg_opacity_alpha_tenths;
    if (histogram_mode < 0)
    {
	vscale = 1.0 / (float) (127 - histogram_mode);
    }
    else
    {
	vscale = 1.0 / (float) (histogram_mode + 127);
    }
    for (i_count = -127; i_count < 128; ++i_count)
    {
	tmp = i_count + 127;
	index = 3 * tmp;
	if (i_count < histogram_mode)
	{
	    intensity = vscale * (float) (histogram_mode - i_count);
	}
	else
	{
	    intensity = vscale * (float) (i_count - histogram_mode);
	}
	mono_hg_intensities[tmp] = pow (intensity, Ialpha);
	if ( (mono_hg_blank_min < i_count) && (i_count < mono_hg_blank_max) )
	{
	    /*  Kill it entirely  */
	    opacity = 0.0;
	}
	else
	{
	    if (i_count < histogram_mode)
	    {
		opacity = vscale * (float) (histogram_mode - i_count);
	    }
	    else
	    {
		opacity = vscale * (float) (i_count - histogram_mode);
	    }
	    opacity = pow (opacity, Oalpha);
	}
	mono_hg_opacities[tmp] = opacity;
    }
    transmit_mono_hg_info (main_vrend_ctx);
}   /*  End Function compute_mono_hg_info  */

void transmit_mono_hg_info (KVolumeRenderContext context)
/*  [SUMMARY] Transmit Mono Hot Gas information to slaves.
    <context> The volume rendering context.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    char *message;
    extern float mono_hg_intensities[MONO_HOT_GAS_CMAP_SIZE];
    extern float mono_hg_opacities[MONO_HOT_GAS_CMAP_SIZE];
    static char function_name[] = "transmit_mono_hg_info";

    message = m_alloc_scratch (MONO_HOT_GAS_CMAP_SIZE * 8, function_name);
    for (count = 0; count < MONO_HOT_GAS_CMAP_SIZE; ++count)
    {
	p_write_buf_float (message + count * 4, mono_hg_intensities[count]);
	p_write_buf_float (message + (MONO_HOT_GAS_CMAP_SIZE + count) * 4,
			   mono_hg_opacities[count]);
    }
    vrender_send_message (context, MtoS_MONO_HOT_GAS_INFO, message,
			  MONO_HOT_GAS_CMAP_SIZE * 8, FALSE);
    m_free_scratch ();
}   /*  End Function transmit_mono_hg_info  */


/*  Private routines follow  */

static void plot_worldcanvas_refresh_func (KWorldCanvas canvas,
					   int width, int height,
					   struct win_scale_type *win_scale,
					   Kcolourmap cmap, flag cmap_resize,
					   void **info)
/*  This routine is a refresh event consumer for a world canvas.
    The canvas is given by  canvas  .
    The width of the canvas in pixels is given by  width  .
    The height of the canvas in pixels is given by  height  .
    The window scaling information is pointed to by  win_scale  .
    The colourmap associated with the canvas is given by  cmap  .
    If the refresh function was called as a result of a colourmap resize
    the value of  cmap_resize  will be TRUE.
    The arbitrary canvas information pointer is pointed to by  info  .
    The routine returns nothing.
*/
{
    unsigned int count;
    unsigned long pixel_value;
    char *colour = NULL;  /*  Initialise to keep compiler happy  */
    float *ptr;
    double tmp_arr[MONO_HOT_GAS_CMAP_SIZE];
    extern float mono_hg_intensities[MONO_HOT_GAS_CMAP_SIZE];
    extern float mono_hg_opacities[MONO_HOT_GAS_CMAP_SIZE];
    static char function_name[] = "mono_hg_plot_worldcanvas_refresh_func";

    ptr = (float *) *info;
    if (ptr == mono_hg_opacities)
    {
	colour = "yellow";
    }
    else if (ptr == mono_hg_intensities)
    {
	colour = "red";
    }
    else
    {
	fprintf (stderr, "Unknown array\n");
	a_prog_bug (function_name);
    }
    if (canvas_get_colour (canvas, colour, &pixel_value,
			   (unsigned short *) NULL, (unsigned short *) NULL,
			   (unsigned short *) NULL) != TRUE) return;
    for (count = 0; count < MONO_HOT_GAS_CMAP_SIZE; ++count)
    {
	tmp_arr[count] = ptr[count];
    }
    canvas_draw_lines_p (canvas, (double *) NULL, tmp_arr,
			 MONO_HOT_GAS_CMAP_SIZE, pixel_value);
}   /*  End Function plot_worldcanvas_refresh_func  */
