/*  offset.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1993-2003  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains the source for the offset control display.


    Written by      Richard Gooch   20-FEB-1994

    Updated by      Richard Gooch   20-FEB-1994

    Updated by      Richard Gooch   19-MAY-1994: Took account of change to
  canvas_get_colour  routine.

    Updated by      Richard Gooch   2-AUG-1994: Ported to Xt.

    Updated by      Richard Gooch   3-OCT-1994: Made use of
  canvas_use_log_scale  .

    Updated by      Richard Gooch   8-JAN-1995: Took account of change to
  <panel_> package in <raycast_3d>.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   22-NOV-1996: Moved to
  <offset_position_func>.

    Update by       Richard Gooch   25-MAY-1997: No longer manipulate global
  voxel sum offset variable directly.

    Updated by      Richard Gooch   11-JUN-1998: Generalised interface to
  support voxel threshold control.

    Updated by      Richard Gooch   12-JUL-2000: Added #include <karma_kwin.h>.

    Updated by      Richard Gooch   22-DEC-2002: Added #include <stdlib.h>

    Last updated by Richard Gooch   21-MAR-2003: Switched to function
  <overlay_create_viewable>.


*/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <karma.h>
#include <k_event_codes.h>
#ifndef X11
#  define X11
#endif
#include <karma_overlay.h>
#include <karma_canvas.h>
#include <karma_kwin.h>
#include <karma_a.h>
#include <karma_c.h>
#include "xraycast_3d.h"


/*  Private functions  */
STATIC_FUNCTION (void offset_control_canvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale, Kcolourmap cmap,
		  flag cmap_resize, void **info) );
STATIC_FUNCTION (flag offset_control_canvas_position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info, void **f_info) );


/*  Private data  */


/*  Public functions follow  */

KWorldCanvas setup_offset_control_canvas (KPixCanvas pixcanvas, Display *dpy,
					  void *offset_widget,
					  void (*x_func) (signed char x),
					  int *x_ptr)
/*  This routine will create the offset control canvas.
    The pixel canvas must be given by  pixcanvas  .
    The X display must be pointed to by  dpy  .
    The routine returns nothing.
*/
{
    KWorldCanvas worldcanvas;
    KViewableOverlayList vlist;
    double frac;
    GC gc;
    XGCValues gcvalues;
    struct win_scale_type win_scale;
    extern KOverlayList histogram_olist;
    extern KCallbackList histogram_change_event_list;
    extern unsigned int histogram_peak;
    /*static char function_name[] = "setup_offset_control_canvas";*/

    /*  Get the Graphics Context of the pixel canvas  */
    gc = kwin_get_gc_x (pixcanvas);
    if (XGetGCValues (dpy, gc, GCForeground | GCBackground, &gcvalues) == 0)
    {
	fprintf (stderr, "Error getting GC values\n");
	exit (RV_UNDEF_ERROR);
    }
    /*  Create the offset control world canvas  */
    canvas_init_win_scale (&win_scale, K_WIN_SCALE_MAGIC_NUMBER);
    win_scale.x_min = -127.0;
    win_scale.x_max = 127.0;
    if (histogram_peak < 1)
    {
	/*  No histogram data yet  */
	win_scale.y_min = 0.0;
	win_scale.y_max = 1.0;
    }
    else
    {
	/*  Have histogram data: setup logarithmic scale  */
	win_scale.y_max = histogram_peak;
	frac = log10 (win_scale.y_max);
	if (frac > 6.0) frac = 6.0;
	win_scale.y_min = win_scale.y_max / pow (10.0, frac);
	if (win_scale.y_min < 1.0) win_scale.y_min = 1.0;
	win_scale.y_log = TRUE;
    }
    win_scale.min_sat_pixel = gcvalues.background;
    win_scale.max_sat_pixel = gcvalues.foreground;
    if ( ( worldcanvas = canvas_create (pixcanvas, NULL, &win_scale) )
	 == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    canvas_use_log_scale (worldcanvas, FALSE, TRUE);
    vlist = overlay_create_viewable (histogram_olist, worldcanvas);
    overlay_set_active (vlist, TRUE, FALSE, TRUE, FALSE, 0);
    canvas_register_refresh_func (worldcanvas,
				  offset_control_canvas_refresh_func,
				  x_ptr);
    canvas_register_position_event_func (worldcanvas,
					 offset_position_func,
					 offset_widget);
    canvas_register_position_event_func (worldcanvas,
					 offset_control_canvas_position_func,
					 (void *) x_func);
    c_register_callback (&histogram_change_event_list,
			 generic_histogram_change_callback, worldcanvas,
			 NULL, FALSE, NULL, FALSE, FALSE);
    return (worldcanvas);
}   /*  End Function setup_offset_control_canvas  */


/*  Private routines follow  */

static void offset_control_canvas_refresh_func (KWorldCanvas canvas,
						int width, int height,
						struct win_scale_type *win_scale,
						Kcolourmap cmap,
						flag cmap_resize, void **info)
/*  This routine is a refresh event consumer for a world canvas.
    The canvas is given by  canvas  .
    The width of the canvas in pixels is given by  width  .
    The height of the canvas in pixels is given by  height  .
    The window scaling information is pointed to by  win_scale  .
    The colourmap associated with the canvas is given by  cmap  .
    If the refresh function was called as a result of a colourmap resize
    the value of  cmap_resize  will be TRUE.
    The arbitrary canvas information pointer is pointed to by  info  .
    The routine returns nothing.
*/
{
    unsigned long pixel_value;
    double x, y;
    extern unsigned int histogram_peak;
    /*static char function_name[] = "offset_control_canvas_refresh_func";*/

    if (histogram_peak < 1) return;
    if ( !canvas_get_colour (canvas, "yellow4", &pixel_value,
			     NULL, NULL, NULL) ) return;
    x = *(int *) *info;
    y = histogram_peak;
    canvas_draw_line_p (canvas, x, 0.0, x, y, pixel_value);
}   /*  End Function offset_control_canvas_refresh_func  */

static flag offset_control_canvas_position_func (KWorldCanvas canvas,
						 double x, double y,
						 unsigned int event_code,
						 void *e_info, void **f_info)
/*  This routine is a position event consumer for a world canvas.
    The canvas is given by  canvas  .
    The horizontal world co-ordinate of the event will be given by  x  .
    The vertical world co-ordinate of the event will be given by  y  .
    The arbitrary event code is given by  event_code  .
    The arbitrary event information is pointed to by  e_info  .
    The arbitrary function information pointer is pointed to by  f_info  .
    The routine returns TRUE if the event was consumed, else it returns
    FALSE indicating that the event is still to be processed.
*/
{
    void (*func) (signed char x) = ( void (*) (signed char x) ) *f_info;

    if (event_code != K_CANVAS_EVENT_LEFT_MOUSE_CLICK) return (FALSE);
    (*func) (x);
    kwin_resize (canvas_get_pixcanvas (canvas), TRUE, 0, 0, 0, 0);
    compute ();
    return (TRUE);
}   /*  End Function offset_control_canvas_position_func  */
