/*  orient.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1993-2000  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains the source for the orientation display.


    Written by      Richard Gooch   30-JAN-1994

    Updated by      Richard Gooch   17-FEB-1994

    Updated by      Richard Gooch   20-FEB-1994: Added 3D cursor.

    Updated by      Richard Gooch   19-MAY-1994: Took account of change to
  kwin_get_colour  .

    Updated by      Richard Gooch   17-JUL-1995: Moved over to coord_3d type.

    Updated by      Richard Gooch   27-FEB-1996: Extracted (new) 3-D cursor
  code and make wireframe display control separate from cursor display control.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   26-JUL-1996: Made use of <get_colour>.

    Updated by      Richard Gooch   8-DEC-1996: Created <overlay_orientation>.

    Updated by      Richard Gooch   10-DEC-1996: Made use of
  <register_cursor_func> now that wireframe for rendered images is done using
  overlay lists. This separates the management of the wireframe and cursor.

    Updated by      Richard Gooch   11-DEC-1996: Created
  <orientation_post_render_callback>.

    Updated by      Richard Gooch   22-DEC-1996: No longer fiddle orientation
  canvas world co-ordinates.

    Updated by      Richard Gooch   7-JUL-1999: Added hack to make wireframe
  visible with white background.

    Last updated by Richard Gooch   12-JUL-2000: Added #include <karma_kwin.h>.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_vrender.h>
#include <karma_canvas.h>
#include <karma_iarray.h>
#include <karma_kwin.h>
#include <karma_ds.h>
#include <karma_a.h>
#include <karma_m.h>
#include "xraycast_3d.h"


#define EXTENSION_SIZE 10.0
#define STRING_FONT "9x15bold"


/*  Structure declarations follow  */

typedef struct
{
    KVolumeRenderContext context;
    unsigned int eye_view;
    flag *hide_wireframe;
} wireframe_info;


/*  Private functions  */
STATIC_FUNCTION (void worldcanvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize,
		  void **info, PostScriptPage pspage) );
STATIC_FUNCTION (void overlay_orientation,
		 (KOverlayList olist, KVolumeRenderContext context,
		  unsigned int eye_view) );
STATIC_FUNCTION (void show_orientation,
		 (KWorldCanvas canvas, KVolumeRenderContext context,
		  unsigned int eye_view) );
STATIC_FUNCTION (void register_wireframe_func,
		 (KWorldCanvas canvas, KVolumeRenderContext context,
		  unsigned int eye_view, flag *hide_wireframe_ptr) );


/*  Private data  */


/*  Public functions follow  */

void initialise_orientation_canvas (KPixCanvas pixcanvas)
/*  [PURPOSE] This routine will initialise the orientation canvas.
    <pixcanvas> The pixel canvas.
    [RETURNS] Nothing.
*/
{
    struct win_scale_type win_scale;
    extern KWorldCanvas orientation_worldcanvas;
    extern KVolumeRenderContext main_vrend_ctx;
    static char function_name[] = "initialise_orientation_canvas";

    canvas_init_win_scale (&win_scale, K_WIN_SCALE_MAGIC_NUMBER);
    if ( ( orientation_worldcanvas =
	   canvas_create (pixcanvas, NULL, &win_scale) ) == NULL )
    {
	m_abort (function_name, "world canvas");
    }
    register_wireframe_func (orientation_worldcanvas, main_vrend_ctx,
			     VRENDER_EYE_CHOICE_CYCLOPS, NULL);
    register_cursor_func (orientation_worldcanvas, main_vrend_ctx,
			  VRENDER_EYE_CHOICE_CYCLOPS, FALSE);
}   /*  End Function initialise_orientation_canvas  */

flag orientation_post_render_callback (void *object, void *client1_data,
				       void *call_data, void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    post_render_info *info = (post_render_info *) call_data;
    extern KVolumeRenderContext main_vrend_ctx;
    extern KOverlayList cyclops_olist, left_olist, right_olist;
    extern flag hide_wireframe;

    if (hide_wireframe) return (FALSE);
    if (info->did_stereo)
    {
	overlay_orientation (left_olist, main_vrend_ctx,
			     VRENDER_EYE_CHOICE_LEFT);
	overlay_orientation (right_olist, main_vrend_ctx,
			     VRENDER_EYE_CHOICE_RIGHT);
    }
    else overlay_orientation (cyclops_olist, main_vrend_ctx,
			      VRENDER_EYE_CHOICE_CYCLOPS);
    return (FALSE);
}   /*  End Function orientation_post_render_callback  */

static void worldcanvas_refresh_func (KWorldCanvas canvas,
				      int width, int height,
				      struct win_scale_type *win_scale,
				      Kcolourmap cmap, flag cmap_resize,
				      void **info, PostScriptPage pspage)
/*  [PURPOSE] This routine is a refresh event consumer for a world canvas.
    <canvas> The world canvas being refreshed.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The colourmap associated with the canvas.
    <cmap_resize> TRUE if the refresh function was called as a result of a
    colourmap resize, else FALSE.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    [RETURNS] Nothing.
*/
{
    wireframe_info *w_info;
    extern flag byte_cube_valid;
    /*static char function_name[] = "orientation_canvas_refresh_func";*/

    if (!byte_cube_valid) return;
    w_info = (wireframe_info *) *info;
    if ( (w_info->hide_wireframe == NULL) || !*w_info->hide_wireframe )
    {
	show_orientation (canvas, w_info->context, w_info->eye_view);
    }
}   /*  End Function worldcanvas_refresh_func  */


/*  Private routines follow  */

static void overlay_orientation (KOverlayList olist,
				 KVolumeRenderContext context,
				 unsigned int eye_view)
/*  [PURPOSE] This routine will show the orientation of the cube.
    <olist> The overlay list.
    <context> The volume rendering context.
    <eye_view> The eye which will see the point. Legal values are:
        VRENDER_EYE_CYCLOPS    VRENDER_EYE_LEFT    VRENDER_EYE_RIGHT
    [RETURNS] Nothing.
*/
{
    flag v1, v2, v3;
    Kcoord_3d p1_3d, p2_3d, p3_3d;
    Kcoord_2d p1_2d, p2_2d, p3_2d;
    Kcoord_3d cube_min, cube_max;
    unsigned short red, green, blue;
    unsigned long x_s, y_s, z_s, x_e, y_e, z_e, background;
    double r, g, b;
    char x_str[STRING_LENGTH];
    char y_str[STRING_LENGTH];
    char z_str[STRING_LENGTH];
    char *fg_colour = "white";
    extern flag hide_wireframe;
    extern iarray byte_cube;
    extern KWorldCanvas visible_worldcanvas;

    if (hide_wireframe) return;
    /*  This is a hack to try to make the wireframe visible on paper  */
    kwin_get_attributes (canvas_get_pixcanvas (visible_worldcanvas),
			 KWIN_ATT_BACKGROUND_PIXEL, &background,
			 KWIN_ATT_END);
    kwin_get_pixel_RGB_values (canvas_get_pixcanvas (visible_worldcanvas),
			       &background, &red, &green, &blue, 1);
    r = red;
    g = green;
    b = blue;
    if (sqrt (r * r + g * g + b * b) / 65535.0 >= 0.5) fg_colour = "black";
    vrender_get_context_attributes (context,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_START, &x_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_START, &y_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_START, &z_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_END, &x_e,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_END, &y_e,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_END, &z_e,
				    VRENDER_CONTEXT_ATT_END);
    cube_min.x = x_s;
    cube_min.y = y_s;
    cube_min.z = z_s;
    cube_max.x = x_e;
    cube_max.y = y_e;
    cube_max.z = z_e;
    /*  Get axes names and beautify  */
    if (strncmp (iarray_dim_name (byte_cube, 2), "RA-", 3) == 0)
    {
	strcpy (x_str, "Ra");
    }
    else strcpy (x_str, "x");
    if (strncmp (iarray_dim_name (byte_cube, 1), "DEC-", 4) == 0)
    {
	strcpy (y_str, "Dec");
    }
    else strcpy (y_str, "y");
    if (strncmp (iarray_dim_name (byte_cube, 0), "VELO", 4) == 0)
    {
	strcpy (z_str, "Vel");
    }
    else if (strncmp (iarray_dim_name (byte_cube, 0), "FELO", 4) == 0)
    {
	strcpy (z_str, "Vel");
    }
    else if (strncmp (iarray_dim_name (byte_cube, 0), "FREQ", 4) == 0)
    {
	strcpy (z_str, "Freq");
    }
    else strcpy (z_str, "z");
    /*  Display  x  axis  */
    /*  y = min, z = min  */
    p1_3d.x = cube_min.x;
    p1_3d.y = cube_min.y;
    p1_3d.z = cube_min.z;
    p2_3d.x = cube_max.x;
    p2_3d.y = cube_min.y;
    p2_3d.z = cube_min.z;
    p3_3d.x = cube_max.x + EXTENSION_SIZE;
    p3_3d.y = cube_min.y;
    p3_3d.z = cube_min.z;
    v1 = vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE);
    v2 = vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE);
    v3 = vrender_project_3d (context, eye_view, p3_3d, &p3_2d, TRUE);
    if (v1 && v3)
    {
	/*  Primary x axis and extension is visible  */
	overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
		      OVERLAY_COORD_WORLD, p3_2d.x, p3_2d.y, "turquoise1");
    }
    else if (v1 && v2)
    {
	/*  Primary x axis is visible  */
	overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
		      OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y, "turquoise1");
    }
    else if (v2 && v3)
    {
	/*  Only extension is visible  */
	overlay_line (olist, OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y,
		      OVERLAY_COORD_WORLD, p3_2d.x, p3_2d.y, "turquoise1");
    }
    if ( v3 && (v1 || v2) )
    {
	/*  Extension is visible  */
	overlay_text (olist, x_str, OVERLAY_COORD_WORLD, p3_2d.x, p3_2d.y,
		      "turquoise1", STRING_FONT, FALSE);
    }
    /*  y = max, z = min  */
    p1_3d.x = cube_min.x;
    p1_3d.y = cube_max.y;
    p1_3d.z = cube_min.z;
    p2_3d.x = cube_max.x;
    p2_3d.y = cube_max.y;
    p2_3d.z = cube_min.z;
    if ( vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE)
	 && vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE) )
    {
	overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
		      OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y, fg_colour);
    }
    /*  y = min, z = max  */
    p1_3d.x = cube_min.x;
    p1_3d.y = cube_min.y;
    p1_3d.z = cube_max.z;
    p2_3d.x = cube_max.x;
    p2_3d.y = cube_min.y;
    p2_3d.z = cube_max.z;
    if ( vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE)
	 && vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE) )
    {
	overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
		      OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y, fg_colour);
    }
    /*  y = max, z = max  */
    p1_3d.x = cube_min.x;
    p1_3d.y = cube_max.y;
    p1_3d.z = cube_max.z;
    p2_3d.x = cube_max.x;
    p2_3d.y = cube_max.y;
    p2_3d.z = cube_max.z;
    if ( vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE)
	 && vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE) )
    {
	overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
		      OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y, fg_colour);
    }
    /*  Display  y  axis  */
    /*  x = min, z = min  */
    p1_3d.x = cube_min.x;
    p1_3d.y = cube_min.y;
    p1_3d.z = cube_min.z;
    p2_3d.x = cube_min.x;
    p2_3d.y = cube_max.y;
    p2_3d.z = cube_min.z;
    p3_3d.x = cube_min.x;
    p3_3d.y = cube_max.y + EXTENSION_SIZE;
    p3_3d.z = cube_min.z;
    v1 = vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE);
    v2 = vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE);
    v3 = vrender_project_3d (context, eye_view, p3_3d, &p3_2d, TRUE);
    if (v1 && v3)
    {
	/*  Primary y axis and extension is visible  */
	overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
		      OVERLAY_COORD_WORLD, p3_2d.x, p3_2d.y, "yellow");
    }
    else if (v1 && v2)
    {
	/*  Primary y axis is visible  */
	overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
		      OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y, "yellow");
    }
    else if (v2 && v3)
    {
	/*  Only extension is visible  */
	overlay_line (olist, OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y,
		      OVERLAY_COORD_WORLD, p3_2d.x, p3_2d.y, "yellow");
    }
    if ( v3 && (v1 || v2) )
    {
	/*  Extension is visible  */
	overlay_text (olist, y_str, OVERLAY_COORD_WORLD, p3_2d.x, p3_2d.y,
		      "yellow", STRING_FONT, FALSE);
    }
    /*  x = max, z = min  */
    p1_3d.x = cube_max.x;
    p1_3d.y = cube_min.y;
    p1_3d.z = cube_min.z;
    p2_3d.x = cube_max.x;
    p2_3d.y = cube_max.y;
    p2_3d.z = cube_min.z;
    if ( vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE)
	 && vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE) )
    {
	overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
		      OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y, fg_colour);
    }
    /*  x = min, z = max  */
    p1_3d.x = cube_min.x;
    p1_3d.y = cube_min.y;
    p1_3d.z = cube_max.z;
    p2_3d.x = cube_min.x;
    p2_3d.y = cube_max.y;
    p2_3d.z = cube_max.z;
    if ( vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE)
	 && vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE) )
    {
	overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
		      OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y, fg_colour);
    }
    /*  x = max, z = max  */
    p1_3d.x = cube_max.x;
    p1_3d.y = cube_min.y;
    p1_3d.z = cube_max.z;
    p2_3d.x = cube_max.x;
    p2_3d.y = cube_max.y;
    p2_3d.z = cube_max.z;
    if ( vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE)
	 && vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE) )
    {
	overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
		      OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y, fg_colour);
    }
    /*  Display  z  axis  */
    /*  x = min, y = min  */
    p1_3d.x = cube_min.x;
    p1_3d.y = cube_min.y;
    p1_3d.z = cube_min.z;
    p2_3d.x = cube_min.x;
    p2_3d.y = cube_min.y;
    p2_3d.z = cube_max.z;
    p3_3d.x = cube_min.x;
    p3_3d.y = cube_min.y;
    p3_3d.z = cube_max.z + EXTENSION_SIZE;
    v1 = vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE);
    v2 = vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE);
    v3 = vrender_project_3d (context, eye_view, p3_3d, &p3_2d, TRUE);
    if (v1 && v3)
    {
	/*  Primary z axis and extension is visible  */
	overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
		      OVERLAY_COORD_WORLD, p3_2d.x, p3_2d.y, "red");
    }
    else if (v1 && v2)
    {
	/*  Primary z axis is visible  */
	overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
		      OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y, "red");
    }
    else if (v2 && v3)
    {
	/*  Only extension is visible  */
	overlay_line (olist, OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y,
		      OVERLAY_COORD_WORLD, p3_2d.x, p3_2d.y, "red");
    }
    if ( v3 && (v1 || v2) )
    {
	/*  Extension is visible  */
	overlay_text (olist, z_str, OVERLAY_COORD_WORLD, p3_2d.x, p3_2d.y,
		      "red", STRING_FONT, FALSE);
    }
    /*  x = max, y = min  */
    p1_3d.x = cube_max.x;
    p1_3d.y = cube_min.y;
    p1_3d.z = cube_min.z;
    p2_3d.x = cube_max.x;
    p2_3d.y = cube_min.y;
    p2_3d.z = cube_max.z;
    if ( vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE)
	&& vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE) )
    {
	overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
		      OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y, fg_colour);
    }
    /*  x = min, y = max  */
    p1_3d.x = cube_min.x;
    p1_3d.y = cube_max.y;
    p1_3d.z = cube_min.z;
    p2_3d.x = cube_min.x;
    p2_3d.y = cube_max.y;
    p2_3d.z = cube_max.z;
    if ( vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE)
	&& vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE) )
    {
	overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
		      OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y, fg_colour);
    }
    /*  x = max, y = max  */
    p1_3d.x = cube_max.x;
    p1_3d.y = cube_max.y;
    p1_3d.z = cube_min.z;
    p2_3d.x = cube_max.x;
    p2_3d.y = cube_max.y;
    p2_3d.z = cube_max.z;
    if ( vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE)
	&& vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE) )
    {
	overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
		      OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y, fg_colour);
    }
}   /*  End Function overlay_orientation  */

static void show_orientation (KWorldCanvas canvas,
			      KVolumeRenderContext context,
			      unsigned int eye_view)
/*  [PURPOSE] This routine will show the orientation of the cube.
    <canvas> The canvas.
    <context> The volume rendering context.
    <eye_view> The eye which will see the point. Legal values are:
        VRENDER_EYE_CYCLOPS    VRENDER_EYE_LEFT    VRENDER_EYE_RIGHT
    [RETURNS] Nothing.
*/
{
    flag v1, v2, v3;
    Kcoord_3d p1_3d, p2_3d, p3_3d;
    Kcoord_2d p1_2d, p2_2d, p3_2d;
    Kcoord_3d cube_min, cube_max;
    unsigned long x_s, y_s, z_s, x_e, y_e, z_e;
    unsigned long pixel_value;
    unsigned long pixel_value_white;
    double px, py;
    char x_str[STRING_LENGTH];
    char y_str[STRING_LENGTH];
    char z_str[STRING_LENGTH];
    extern iarray byte_cube;

    vrender_get_context_attributes (context,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_START, &x_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_START, &y_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_START, &z_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_END, &x_e,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_END, &y_e,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_END, &z_e,
				    VRENDER_CONTEXT_ATT_END);
    cube_min.x = x_s;
    cube_min.y = y_s;
    cube_min.z = z_s;
    cube_max.x = x_e;
    cube_max.y = y_e;
    cube_max.z = z_e;
    if ( !get_colour (canvas, "white", &pixel_value_white, NULL) ) return;
    /*  Get axes names and beautify  */
    if (strncmp (iarray_dim_name (byte_cube, 2), "RA-", 3) == 0)
    {
	strcpy (x_str, "Ra");
    }
    else strcpy (x_str, "x");
    if (strncmp (iarray_dim_name (byte_cube, 1), "DEC-", 4) == 0)
    {
	strcpy (y_str, "Dec");
    }
    else strcpy (y_str, "y");
    if (strncmp (iarray_dim_name (byte_cube, 0), "VELO", 4) == 0)
    {
	strcpy (z_str, "Vel");
    }
    else if (strncmp (iarray_dim_name (byte_cube, 0), "FELO", 4) == 0)
    {
	strcpy (z_str, "Vel");
    }
    else if (strncmp (iarray_dim_name (byte_cube, 0), "FREQ", 4) == 0)
    {
	strcpy (z_str, "Freq");
    }
    else strcpy (z_str, "z");
    /*  Display  x  axis  */
    /*  y = min, z = min  */
    if ( !get_colour (canvas, "turquoise1", &pixel_value, NULL) ) return;
    p1_3d.x = cube_min.x;
    p1_3d.y = cube_min.y;
    p1_3d.z = cube_min.z;
    p2_3d.x = cube_max.x;
    p2_3d.y = cube_min.y;
    p2_3d.z = cube_min.z;
    p3_3d.x = cube_max.x + EXTENSION_SIZE;
    p3_3d.y = cube_min.y;
    p3_3d.z = cube_min.z;
    v1 = vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE);
    v2 = vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE);
    v3 = vrender_project_3d (context, eye_view, p3_3d, &p3_2d, TRUE);
    if (v1 && v3)
    {
	/*  Primary x axis and extension is visible  */
	canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y, p3_2d.x, p3_2d.y,
			    pixel_value);
    }
    else if (v1 && v2)
    {
	/*  Primary x axis is visible  */
	canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y, p2_2d.x, p2_2d.y,
			    pixel_value);
    }
    else if (v2 && v3)
    {
	/*  Only extension is visible  */
	canvas_draw_line_p (canvas, p2_2d.x, p2_2d.y, p3_2d.x, p3_2d.y,
			    pixel_value);
    }
    if ( v3 && (v1 || v2) )
    {
	/*  Extension is visible  */
	px = p3_2d.x;
	py = p3_2d.y;
	canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
					   &px, &py, &px, &py);
	kwin_draw_string (canvas_get_pixcanvas (canvas), px, py,
			  x_str, pixel_value, FALSE);
    }
    /*  y = max, z = min  */
    p1_3d.x = cube_min.x;
    p1_3d.y = cube_max.y;
    p1_3d.z = cube_min.z;
    p2_3d.x = cube_max.x;
    p2_3d.y = cube_max.y;
    p2_3d.z = cube_min.z;
    if ( vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE)
	&& vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE) )
    {
	canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y, p2_2d.x, p2_2d.y,
			    pixel_value_white);
    }
    /*  y = min, z = max  */
    p1_3d.x = cube_min.x;
    p1_3d.y = cube_min.y;
    p1_3d.z = cube_max.z;
    p2_3d.x = cube_max.x;
    p2_3d.y = cube_min.y;
    p2_3d.z = cube_max.z;
    if ( vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE)
	&& vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE) )
    {
	canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y, p2_2d.x, p2_2d.y,
			    pixel_value_white);
    }
    /*  y = max, z = max  */
    p1_3d.x = cube_min.x;
    p1_3d.y = cube_max.y;
    p1_3d.z = cube_max.z;
    p2_3d.x = cube_max.x;
    p2_3d.y = cube_max.y;
    p2_3d.z = cube_max.z;
    if ( vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE)
	&& vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE) )
    {
	canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y, p2_2d.x, p2_2d.y,
			    pixel_value_white);
    }
    /*  Display  y  axis  */
    /*  x = min, z = min  */
    if ( !get_colour (canvas, "yellow", &pixel_value, NULL) ) return;
    p1_3d.x = cube_min.x;
    p1_3d.y = cube_min.y;
    p1_3d.z = cube_min.z;
    p2_3d.x = cube_min.x;
    p2_3d.y = cube_max.y;
    p2_3d.z = cube_min.z;
    p3_3d.x = cube_min.x;
    p3_3d.y = cube_max.y + EXTENSION_SIZE;
    p3_3d.z = cube_min.z;
    v1 = vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE);
    v2 = vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE);
    v3 = vrender_project_3d (context, eye_view, p3_3d, &p3_2d, TRUE);
    if (v1 && v3)
    {
	/*  Primary y axis and extension is visible  */
	canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y, p3_2d.x, p3_2d.y,
			    pixel_value);
    }
    else if (v1 && v2)
    {
	/*  Primary y axis is visible  */
	canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y, p2_2d.x, p2_2d.y,
			    pixel_value);
    }
    else if (v2 && v3)
    {
	/*  Only extension is visible  */
	canvas_draw_line_p (canvas, p2_2d.x, p2_2d.y, p3_2d.x, p3_2d.y,
			    pixel_value);
    }
    if ( v3 && (v1 || v2) )
    {
	/*  Extension is visible  */
	px = p3_2d.x;
	py = p3_2d.y;
	canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
					   &px, &py, &px, &py);
	kwin_draw_string (canvas_get_pixcanvas (canvas), px, py,
			  y_str, pixel_value, FALSE);
    }
    /*  x = max, z = min  */
    p1_3d.x = cube_max.x;
    p1_3d.y = cube_min.y;
    p1_3d.z = cube_min.z;
    p2_3d.x = cube_max.x;
    p2_3d.y = cube_max.y;
    p2_3d.z = cube_min.z;
    if ( vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE)
	&& vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE) )
    {
	canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y, p2_2d.x, p2_2d.y,
			    pixel_value_white);
    }
    /*  x = min, z = max  */
    p1_3d.x = cube_min.x;
    p1_3d.y = cube_min.y;
    p1_3d.z = cube_max.z;
    p2_3d.x = cube_min.x;
    p2_3d.y = cube_max.y;
    p2_3d.z = cube_max.z;
    if ( vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE)
	&& vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE) )
    {
	canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y, p2_2d.x, p2_2d.y,
			    pixel_value_white);
    }
    /*  x = max, z = max  */
    p1_3d.x = cube_max.x;
    p1_3d.y = cube_min.y;
    p1_3d.z = cube_max.z;
    p2_3d.x = cube_max.x;
    p2_3d.y = cube_max.y;
    p2_3d.z = cube_max.z;
    if ( vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE)
	&& vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE) )
    {
	canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y, p2_2d.x, p2_2d.y,
			    pixel_value_white);
    }
    /*  Display  z  axis  */
    /*  x = min, y = min  */
    if ( !get_colour (canvas, "red", &pixel_value, NULL) ) return;
    p1_3d.x = cube_min.x;
    p1_3d.y = cube_min.y;
    p1_3d.z = cube_min.z;
    p2_3d.x = cube_min.x;
    p2_3d.y = cube_min.y;
    p2_3d.z = cube_max.z;
    p3_3d.x = cube_min.x;
    p3_3d.y = cube_min.y;
    p3_3d.z = cube_max.z + EXTENSION_SIZE;
    v1 = vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE);
    v2 = vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE);
    v3 = vrender_project_3d (context, eye_view, p3_3d, &p3_2d, TRUE);
    if (v1 && v3)
    {
	/*  Primary z axis and extension is visible  */
	canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y, p3_2d.x, p3_2d.y,
			    pixel_value);
    }
    else if (v1 && v2)
    {
	/*  Primary z axis is visible  */
	canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y, p2_2d.x, p2_2d.y,
			    pixel_value);
    }
    else if (v2 && v3)
    {
	/*  Only extension is visible  */
	canvas_draw_line_p (canvas, p2_2d.x, p2_2d.y, p3_2d.x, p3_2d.y,
			    pixel_value);
    }
    if ( v3 && (v1 || v2) )
    {
	/*  Extension is visible  */
	px = p3_2d.x;
	py = p3_2d.y;
	canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
					   &px, &py, &px, &py);
	kwin_draw_string (canvas_get_pixcanvas (canvas), px, py,
			  z_str, pixel_value, FALSE);
    }
    /*  x = max, y = min  */
    p1_3d.x = cube_max.x;
    p1_3d.y = cube_min.y;
    p1_3d.z = cube_min.z;
    p2_3d.x = cube_max.x;
    p2_3d.y = cube_min.y;
    p2_3d.z = cube_max.z;
    if ( vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE)
	&& vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE) )
    {
	canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y, p2_2d.x, p2_2d.y,
			    pixel_value_white);
    }
    /*  x = min, y = max  */
    p1_3d.x = cube_min.x;
    p1_3d.y = cube_max.y;
    p1_3d.z = cube_min.z;
    p2_3d.x = cube_min.x;
    p2_3d.y = cube_max.y;
    p2_3d.z = cube_max.z;
    if ( vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE)
	&& vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE) )
    {
	canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y, p2_2d.x, p2_2d.y,
			    pixel_value_white);
    }
    /*  x = max, y = max  */
    p1_3d.x = cube_max.x;
    p1_3d.y = cube_max.y;
    p1_3d.z = cube_min.z;
    p2_3d.x = cube_max.x;
    p2_3d.y = cube_max.y;
    p2_3d.z = cube_max.z;
    if ( vrender_project_3d (context, eye_view, p1_3d, &p1_2d, TRUE)
	&& vrender_project_3d (context, eye_view, p2_3d, &p2_2d, TRUE) )
    {
	canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y, p2_2d.x, p2_2d.y,
			    pixel_value_white);
    }
}   /*  End Function show_orientation  */

static void register_wireframe_func (KWorldCanvas canvas,
				     KVolumeRenderContext context,
				     unsigned int eye_view,
				     flag *hide_wireframe_ptr)
/*  [PURPOSE] This routine will register an internal refresh function on a
    worldcanvas which will display the wireframe.
    <canvas> The world canvas. If this is NULL the routine does nothing.
    <context> The volume render context.
    <eye_view> The eye which will see the point. Legal values are:
        VRENDER_EYE_CHOICE_CYCLOPS    VRENDER_EYE_CHOICE_LEFT
	VRENDER_EYE_CHOICE_RIGHT
    <hide_wireframe_ptr> A pointer to a flag indicating whether to disable the
    wireframe or not. If this is NULL the wireframe cannot be disabled.
    [RETURNS] Nothing.
*/
{
    wireframe_info *info;
    static char function_name[] = "register_wireframe_func";

    if (canvas == NULL) return;
    if ( ( info = (wireframe_info *) m_alloc (sizeof *info) ) == NULL )
    {
	m_abort (function_name, "wireframe_info structure");
    }
    info->context = context;
    info->eye_view = eye_view;
    info->hide_wireframe = hide_wireframe_ptr;
    canvas_register_refresh_func (canvas, worldcanvas_refresh_func,
				  (void *) info);
}   /*  End Function register_wireframe_func  */
