/*  ray_display.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1993-2003  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains the source for the ray display.


    Written by      Richard Gooch   24-JUL-1995

    Updated by      Richard Gooch   24-JUL-1995

    Updated by      Richard Gooch   4-MAR-1996: Added opacity display where
  appropriate.

    Updated by      Richard Gooch   12-JUL-2000: Added #include <karma_kwin.h>.

    Updated by      Richard Gooch   22-DEC-2002: Added #include <stdlib.h>

    Last updated by Richard Gooch   21-MAR-2003: Switched to function
  <overlay_create_viewable>.


*/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <karma.h>
#include <k_event_codes.h>
#ifndef X11
#  define X11
#endif
#include <karma_overlay.h>
#include <karma_canvas.h>
#include <karma_kwin.h>
#include <karma_ds.h>
#include <karma_xc.h>
#include <karma_a.h>
#include <karma_m.h>
#include "xraycast_3d.h"


/*  Private functions  */


/*  Private data  */
    KWorldCanvas ray_display_worldcanvas;


/*  Public functions follow  */

void setup_ray_display_canvas (KPixCanvas pixcanvas, Display *display)
/*  [PURPOSE] This routine will create the ray display canvas.
    <pixcanvas> The pixel canvas.
    <display> The X display.
    [RETURNS] Nothing.
*/
{
    KViewableOverlayList vlist;
    GC gc;
    XGCValues gcvalues;
    struct win_scale_type win_scale;
    extern KOverlayList ray_display_olist;
    static char function_name[] = "setup_ray_display_canvas";

#ifdef DUMMY
    kwin_set_attributes (pixcanvas,
			 KWIN_ATT_LINEWIDTH, 2.0,
			 KWIN_ATT_END);
#endif
    /*  Get the Graphics Context of the pixel canvas  */
    gc = kwin_get_gc_x (pixcanvas);
    if (XGetGCValues (display, gc, GCForeground | GCBackground, &gcvalues)
	== 0)
    {
	(void) fprintf (stderr, "Error getting GC values\n");
	exit (RV_UNDEF_ERROR);
    }
    /*  Create the offset control world canvas  */
    canvas_init_win_scale (&win_scale, K_WIN_SCALE_MAGIC_NUMBER);
    win_scale.x_min = 0.0;
    win_scale.x_max = 1.0;
    win_scale.y_min = -128.0;
    win_scale.y_max = 127.0;
    win_scale.min_sat_pixel = gcvalues.background;
    win_scale.max_sat_pixel = gcvalues.foreground;
    if ( ( ray_display_worldcanvas =
	   canvas_create (pixcanvas, NULL, &win_scale) )
	 == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    if ( ( ray_display_olist =
	   overlay_va_create_list (ray_display_worldcanvas, "rayOverlayList",
				   OVERLAY_ATT_END) )
	 == NULL )
    {
	m_abort (function_name, "overlay list");
    }
    overlay_specify_canvas (ray_display_olist, ray_display_worldcanvas);
    vlist = overlay_create_viewable (ray_display_olist,
				     ray_display_worldcanvas);
    overlay_set_active (vlist, TRUE, FALSE, TRUE, FALSE, 0);
}   /*  End Function setup_offset_control_canvas  */

void register_ray (CONST signed char *ray, unsigned int length,
		   unsigned int depth, CONST float *opacities)
/*  [PURPOSE] This routine will register a new ray to be displayed.
    <ray> The ray data.
    <length> The length of the ray.
    <depth> The depth along the ray corresponding to what the user sees.
    <opacities> The opacity values for the voxels. If the first element is
    TOOBIG, the opacity values are not defined.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    double x;
    double x_arr[2048], y_arr[2048];
    extern KOverlayList ray_display_olist;

    if (length > 2048)
    {
	(void) fprintf (stderr, "ray too large to display\n");
	exit (2);
    }
    /*  First display the opacities  */
    if (opacities[0] < TOOBIG)
    {
	for (count = 0; count < length; ++count)
	{
	    x_arr[count] = (double) count / (double) (length - 1);
	    y_arr[count] = opacities[count] * 254.0 - 127.0;
	}
	overlay_lines (ray_display_olist, length, (unsigned int *) NULL,
		       x_arr, y_arr, "yellow");
    }
    /*  Now display the voxel values  */
    for (count = 0; count < length; ++count)
    {
	x_arr[count] = (double) count / (double) (length - 1);
	y_arr[count] = ray[count];
    }
    overlay_lines (ray_display_olist, length, (unsigned int *) NULL,
		   x_arr, y_arr, "white");
    x = (double) depth / (double) (length - 1);
    overlay_line (ray_display_olist,
		  OVERLAY_COORD_RELATIVE, x, 0.0,
		  OVERLAY_COORD_RELATIVE, x, 1.0,
		  "yellow");
}   /*  End Function register_ray  */
