/*  router.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1998-2000  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains the switching fabric for internal communictions.


    Written by      Richard Gooch   1-FEB-1998

    Updated by      Richard Gooch   16-JUN-1998

    Updated by      Richard Gooch   9-SEP-1998: Fixed bug in function
  <generic_subcube_callback>: view information was not being updated.

    Updated by      Richard Gooch   8-OCT-1998: No longer set subcube in
  <register_new_byte_cube> and call <register_subcube_size>.

    Last updated by Richard Gooch   12-JUL-2000: Added #include <karma_kwin.h>.


*/
#include <stdio.h>
#include <math.h>
#include <karma.h>
#include <karma_iarray.h>
#include <karma_kwin.h>
#include <karma_c.h>
#include "xraycast_3d.h"


/*  Public data  */
KCallbackList new_byte_cube_event_list = NULL;
KCallbackList post_render_event_list = NULL;
KCallbackList cursor3d_event_list = NULL;
KCallbackList subcube_event_list = NULL;
KCallbackList rotation_event_list = NULL;
KCallbackList shader_event_list = NULL;
KCallbackList projection_event_list = NULL;
KCallbackList eye_separation_event_list = NULL;
KCallbackList histogram_change_event_list = NULL;


/*  Public routines follow  */


void register_new_byte_cube (iarray cube, double scale, double offset)
/*  [SUMMARY] Register a new byte cube ready for processing.
    <cube> The cube.
    <scale> The scaling factor to be applied to the cube to get real values.
    <offset> The offset to be applied to the scaled cube to get real values.
    [NOTE] Events are sent to <<new_byte_cube_event_list>> and,
    <<histogram_change_event_list>> and <<rotation_event_list>>.
    [RETURNS] Nothing.
*/
{
    extern KCallbackList new_byte_cube_event_list;
    extern flag verbose, byte_cube_valid;
    extern int histogram_mode;
    extern int voxel_sum_position;
    extern double byte_cube_scale, byte_cube_offset;
    extern rotation_specification rotation;

    byte_cube_scale = scale;
    byte_cube_offset = offset;
    iarray_set_data_scaling (cube, byte_cube_scale, byte_cube_offset);
    fprintf (stderr, "byte cube scale: %e  offset: %e\n",
	     byte_cube_scale, byte_cube_offset);
    iarray_write (cube, "connections");
    byte_cube_valid = TRUE;
    /*  General notification of byte cube change  */
    c_call_callbacks (new_byte_cube_event_list, cube);
    register_subcube_size ();
    compute_histogram (cube);
    /*  Produce and transmit offset calculated from histogram  */
    register_voxel_offset (histogram_mode);
    if (verbose) fprintf (stderr, "voxel_sum_position %d\n",
			  voxel_sum_position);
    register_rotation_change (rotation);
    compute ();
}   /*  End Function register_new_byte_cube  */

flag generic_new_cube_callback (void *object, void *client1_data,
				void *call_data, void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    iarray cube = (iarray) call_data;
    KVolumeRenderContext vrend_ctx = object;

    vrender_set_context_attributes (vrend_ctx,
				    VRENDER_CONTEXT_ATT_CUBE, cube,
				    VRENDER_CONTEXT_ATT_END);
    return (FALSE);
}   /*  End Function generic_new_cube_callback  */

void register_subcube_size ()
/*  [SUMMARY] Register a new subcube size.
    [RETURNS] Nothing.
*/
{
    unsigned long sc[6];
    extern flag byte_cube_valid;
    extern i_coord_3d subcube_start;
    extern i_coord_3d subcube_end;

    if (!byte_cube_valid) return;
    sc[0] = subcube_start.x;
    sc[1] = subcube_start.y;
    sc[2] = subcube_start.z;
    sc[3] = subcube_end.x;
    sc[4] = subcube_end.y;
    sc[5] = subcube_end.z;
    c_call_callbacks (subcube_event_list, sc);
}   /*  End Function register_subcube_size  */

flag generic_subcube_callback (void *object, void *client1_data,
			       void *call_data, void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called to change the subcube for a volume render
    context. The view information is also automatically updated.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    KVolumeRenderContext vrend_context = object;
    unsigned long *sc = call_data;
    extern rotation_specification rotation;

    vrender_set_context_attributes (vrend_context,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_START, sc[0],
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_START, sc[1],
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_START, sc[2],
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_END, sc[3],
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_END, sc[4],
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_END, sc[5],
				    VRENDER_CONTEXT_ATT_END);
    generic_rotation_callback (object, NULL, &rotation, NULL);
    return (FALSE);
}   /*  End Function generic_subcube_callback  */

void register_rotation_change (rotation_specification rot)
/*  [SUMMARY] Register the change in rotations (view).
    <rot> The rotations.
    [RETURNS] Nothing.
*/
{
    extern KPixCanvas orientation_pixcanvas;
    extern rotation_specification rotation;

    rotation = rot;
    c_call_callbacks (rotation_event_list, &rot);
    /*  We are called by reset_view() long before the canvas is created  */
    if (orientation_pixcanvas != NULL) kwin_resize (orientation_pixcanvas,
						    TRUE, 0, 0, 0, 0);
}   /*  End Function register_rotation_change  */

flag generic_rotation_callback (void *object, void *client1_data,
				void *call_data, void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called to update the view parameters for a volume
    render context.
    <object> The object information pointer. This must point to a volume render
    context.
    <client1_data> The first client information pointer.
    <call_data> A pointer to the rotation information.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    KVolumeRenderContext vrend_context = (KVolumeRenderContext) object;
    Kcoord_3d vision, subcube_centre;
    unsigned long x_s, y_s, z_s, x_e, y_e, z_e;
    unsigned long ul_val;
    float distance;
    view_specification view;
    rotation_specification *rot = call_data;

    vrender_get_context_attributes (vrend_context,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_START, &x_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_START, &y_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_START, &z_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_END, &x_e,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_END, &y_e,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_END, &z_e,
				    VRENDER_CONTEXT_ATT_END);
    subcube_centre.x = (x_s + x_e) / 2.0;
    subcube_centre.y = (y_s + y_e) / 2.0;
    subcube_centre.z = (z_s + z_e) / 2.0;
    vision.x = 0.0;
    vision.y = 0.0;
    vision.z = 1.0;
    geom_rotate_3d_rpy (&vision, rot->pitch, rot->yaw, rot->roll);
    view.vertical.x = 0.0;
    view.vertical.y = 1.0;
    view.vertical.z = 0.0;
    geom_rotate_3d_rpy (&view.vertical, rot->pitch, rot->yaw, rot->roll);
    /*  Compute distance independent of subcube size
	rho is measured in centimeters
	Hence  real_distance = rho * 10 (mm)
	Assume 1 voxel corresponds to 1 pixel (invalid if image expanded)
	Assume screen is 305 mm across (17 inch MAG) with 960 pixels
	Hence 1 mm is 960 / 305 pixels
	If image size is doubled in x and y then 1 voxel corresponds to
	2 pixels
	*/
    distance = rot->rho * 10.0;   /*  Real distance in mm  */
    distance *= 960.0 / 305.0;        /*  Distance in pixels   */
    /*  TODO: distance /= image_expand_factor  */
    view.position.x = subcube_centre.x - distance * vision.x;
    view.position.y = subcube_centre.y - distance * vision.y;
    view.position.z = subcube_centre.z - distance * vision.z;
    /*  Get distance  */
    ul_val = x_e - x_s + 1;
    distance = (ul_val * ul_val);
    ul_val = y_e - y_s + 1;
    distance += (ul_val * ul_val);
    ul_val = z_e - z_s + 1;
    distance += (ul_val * ul_val);
    distance = sqrt (distance);
    view.focus.x = subcube_centre.x;
    view.focus.y = subcube_centre.y;
    view.focus.z = subcube_centre.z;
    vrender_set_context_attributes (vrend_context,
				    VRENDER_CONTEXT_ATT_VIEW, &view,
				    VRENDER_CONTEXT_ATT_END);
    return (FALSE);
}   /*  End Function generic_rotation_callback  */

void register_shader (CONST char *shader)
/*  [SUMMARY] Register a new shader.
    <shader> The name of the new shader.
    [RETURNS] Nothing.
*/
{
    c_call_callbacks (shader_event_list, (void *) shader);
}   /*  End Function register_shader  */

flag generic_shader_callback (void *object, void *client1_data,
			      void *call_data, void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    KVolumeRenderContext vrend_context = object;
    CONST char *shader = (CONST char *) call_data;

    vrender_set_context_attributes (vrend_context,
				    VRENDER_CONTEXT_ATT_SHADER, shader,
				    VRENDER_CONTEXT_ATT_END);
    return (FALSE);
}   /*  End Function generic_shader_callback  */

void register_projection (unsigned int projection)
/*  [SUMMARY] Register a new projection.
    <projection> The new projection.
    [RETURNS] Nothing.
*/
{
    c_call_callbacks (projection_event_list, (void *) (uaddr) projection);
}   /*  End Function register_projection  */

flag generic_projection_callback (void *object, void *client1_data,
				  void *call_data, void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    KVolumeRenderContext vrend_context = object;
    unsigned int projection = (uaddr) call_data;

    vrender_set_context_attributes (vrend_context,
				    VRENDER_CONTEXT_ATT_PROJECTION, projection,
				    VRENDER_CONTEXT_ATT_END);
    return (FALSE);
}   /*  End Function generic_projection_callback  */

void register_eye_separation (double eye_separation)
/*  [SUMMARY] Register a new eye_separation.
    <eye_separation> The the new eye_separation.
    [RETURNS] Nothing.
*/
{
    c_call_callbacks (eye_separation_event_list, &eye_separation);
}   /*  End Function register_eye_separation  */

flag generic_eye_separation_callback (void *object, void *client1_data,
				      void *call_data, void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    KVolumeRenderContext vrend_context = object;
    double eye_separation = *(double *) call_data;

    vrender_set_context_attributes (vrend_context,
				    VRENDER_CONTEXT_ATT_EYE_SEPARATION,
				    eye_separation,
				    VRENDER_CONTEXT_ATT_END);
    return (FALSE);
}   /*  End Function generic_eye_separation_callback  */

void register_histogram_change (double bottom, double top)
/*  [SUMMARY] Register a change in the data histogram.
    <bottom> The bottom of the vertical scale.
    <top> The top of the vertical scale.
    [RETURNS] Nothing.
*/
{
    histogram_scale data;

    data.bottom = bottom;
    data.top = top;
    c_call_callbacks (histogram_change_event_list, &data);
}   /*  End Function register_histogram_change  */

flag generic_histogram_change_callback (void *object, void *client1_data,
					void *call_data, void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    KWorldCanvas canvas = object;
    histogram_scale *data = call_data;

    canvas_set_attributes (canvas,
			   CANVAS_ATT_Y_MIN, data->bottom,
			   CANVAS_ATT_Y_MAX, data->top,
			   CANVAS_ATT_END);
    kwin_refresh_if_visible (canvas_get_pixcanvas (canvas), TRUE);
    return (FALSE);
}   /*  End Function generic_histogram_change_callback  */

void register_3d_cursor_change (Kcoord_3d new_cursor)
/*  [SUMMARY] Register a change in the 3D cursor position.
    <new_cursor> The cursor position.
    [NOTE] The change event is dispatched to <<cursor3d_event_list>>.
    [RETURNS] Nothing.
*/
{
    c_call_callbacks (cursor3d_event_list, &new_cursor);
}   /*  End Function register_3d_cursor_change  */
