/*
 * sbcustom.c
 *
 * Spaceball Library RS232-C I/O and Clock Routines for UNIX*
 *
 * Version: L4.05a
 * Release Date:	       29 August    1989
 * Latest modification:        14 September 1989
 *
 * Written by John Hilton, Spatial Systems Inc, May 1988
 *
 * Copyright (C) 1988 Spatial Systems Inc. All Rights Reserved.
 * Spaceball is a registered trademark of Spatial Systems Inc.
 *
 * Refer to the Spaceball Application Developer's Reference (Document
 * Number D1005) for more information on the following code.
 *
 * RESTRICTIONS
 *
 * The following code is licensed for use and distribution provided the
 * code is used exclusively in conjunction with a Spaceball.  No portion
 * may be used for other purposes without prior written permission from
 * Spatial Systems Inc.  This notice must accompany any full or partial
 * copies.
 *
 * COMMENTS
 *
 * For Unix Berkeley 4.3 ensure the variable BERKELEY is #defined otherwise
 * Unix System V will be used as default.
 *
 * The Spaceball library interfaces to the operating system only via the
 * routines contained in this file:
 *	SbOpen()	    - open the Spaceball tty
 *	SbClose()	    - close the currently open Spaceball tty
 *	SbInput()	    - read in characters from Spaceball
 *	SbOutput()	    - send characters to Spaceball
 *	SbMilliseconds()    - returns the lowest 16 bits of a millisecond clock
 *
 *
 * IMPLEMENTATION NOTES:
 *
 * System V file locking:
 * To prevent multiple access of a tty line a locking mechanism is used
 * whereby a locking file is created, a write lock placed on it and the pid
 * written into it. The file is deleted when the tty line is closed. If the
 * program dies the write lock is automatically removed by the system thereby
 * allowing another process to place a write lock on the file and use
 * spaceball.
 *
 * XON/XOFF control is implemented for both input and output. To prevent
 * hanging the process if the host somehow misses the XON from Spaceball a
 * call is made to restart output whenever the write() does not accept all
 * the data. Attempts are then made, when possible, until a timeout occurs
 * at which point an error is printed and the output discarded.
 *
 * Multiple Spaceballs are supported by this implementation and activated by
 * compiling with Sb_NUMBER_OF_SPACEBALLS being set to the maximum number
 * required.
 *
 *
 * *UNIX is a registered trademark of AT&T
 */

#include <stdio.h>
#define Sb_ENVIRONMENT_INCLUDES
#include "sblibry.h"
#include <signal.h>
#include <utmp.h>
#if SystemV
#  include <termio.h>
#  if Sb_LOCK_TTY
#    include <stropts.h>
#  endif
#  include <sys/times.h>
#  include <sys/param.h>
#else
#  include <sgtty.h>
#  ifdef OS_Solaris
#    define BSD_COMP
#    include <sys/file.h>
#  endif
#  include <sys/ioctl.h>
#endif
#include <fcntl.h>
#include <errno.h>

#define     debug_fp	    stdout
#define     output_fp	    stdout
#define     error_fp	    stderr

extern int
  errno;

static char
  lock_prefix[] = "/tmp/LCK..",
  errormsg[ 100 ];

#if Sb_LOCK_TTY
static char
  *filenameof();
#endif

static
  myperror();

#if SystemV
static struct termio termio;
#else
#endif

#if !Sb_USING_POLL
static fd_set zero_set;
#   ifndef FD_SET
#   define FD_SET( fd, fdsetp ) \
	(fdsetp)->fds_bits[ (fd) / (sizeof( (fdsetp)->fds_bits[0] ) * 8)] =\
	    1<<((fd) % (sizeof( (fdsetp)->fds_bits[0] ) * 8))
#   endif
#   ifndef FD_ZERO
#   define FD_ZERO( fdsetp ) \
	memset((fdsetp)->fds_bits, '\0', sizeof( (fdsetp)->fds_bits ))
#   endif
#endif

SbOpen( tty_name )
char *tty_name;
{
    {
	static int initialized = FALSE;

	if (!initialized) {
	    char *debug_str;

	    debug_str = (char *) getenv( "DEBUG_SPACEBALL" );
	    if (debug_str != NULL) {
#ifdef Sb_DEBUG
		sscanf( debug_str, " 0x%x", &Sb_debug );
		fprintf( debug_fp, "debugging with 0x%4.4x\n", Sb_debug );
#else
		fprintf( output_fp, "%s\n%s\n",
		    "To debug the Spaceball library recompile with:",
		    "\t#define Sb_DEBUG\n"
		);
#endif
	    }
#if !Sb_USING_POLL
	    FD_ZERO( &SbFdset );
	    FD_ZERO( &zero_set );
#endif

#if Sb_NUMBER_OF_SPACEBALLS > 1
	    /* On the first call initialize SbArray with SbArray[ 0 ] */
	    {
		int i;

		for (i=1; i<Sb_NUMBER_OF_SPACEBALLS; i++) {
		    memcpy( SbArray + i, SbArray, sizeof( *SbArray ));
		    SbArray[ i ].Rotate = SbArray[ i ].DummyRotate;
		    SbArray[ i ].Matrix = SbArray[ i ].DummyMatrix;
		}
	    }
#endif
	    initialized = TRUE;
	}
    }

    if (SbFd > 0) {	/* already open? */
	fprintf( error_fp,
	    "sbcustom: Reopening open Spaceball on \"%s\"\n", tty_name
	);
	SbClose();
    }

#if Sb_LOCK_TTY
    if (lock_ttyline( filenameof( tty_name )) == -1)
	return SbFd = -1;	/* flag not open */
#endif

#ifndef ALLIANT
#    define	Sb_OPEN_MODE	O_RDWR | O_NDELAY
#else
#    define	Sb_OPEN_MODE	O_RDWR
#endif
    if ((SbFd = open( tty_name, Sb_OPEN_MODE)) < 0) {
	sprintf( errormsg, "sbcustom: Unable to open \"%s\"", tty_name );
	myperror( errormsg );
#if Sb_LOCK_TTY
	unlock_ttyline();
#endif
	return SbFd = -1;	/* flag not open */
    }
#if SystemV
    if (ioctl( SbFd, TCGETA, &termio) < 0) {
	sprintf( errormsg, "sbcustom: ioctl( %s, TCGETA, &termio)", tty_name );
	myperror( errormsg );
	close( SbFd );
	return SbFd = -1;	/* flag not open */
    }
#else  /*  SystemV == 0  */
    if (ioctl( SbFd, TIOCGETP, &sgttyb ) < 0) {
	sprintf( errormsg, "sbcustom: ioctl( %s, TIOCGETP, &sgttyb)",
	    tty_name
	);
	myperror( errormsg );
	close( SbFd );
	return SbFd = -1;	/* flag not open */
    }
#endif
#if SystemV
    /*  Clear termio struct first  */
    m_clear ( (char *) &termio, sizeof termio );
    /* set input mode to ignore any errors and enable XON/XOFF flow control
     * for both input but for output ignore XON/XOFF requests from Spaceball.
     * This should not cause a problem as, in practice, Spaceball will
     * only generate an XOFF when resetting - and that is dealt with
     * fine by the library */
    termio.c_iflag = IGNBRK | IGNPAR | IXOFF;

    /* set ouput mode to pass the data untouched */
    termio.c_oflag = 0;

    /* set hardware control to 9600 baud, enable receiver, local line,
     * 8 bits */
    termio.c_cflag = B9600 | CREAD | CLOCAL | CS8;

    /* enable canonical processing to allow the system to assemble the
     * input characters into lines delimited by CR (VEOL in c_cc[]) */
    /* For some reason this fails with Linux   REG  23-MAR-1996 */
#  ifndef OS_Linux
    termio.c_lflag = ICANON;
#  endif

#  if sgi
    /* select line discipline 0 */
    termio.c_line = LDISC0;
#  endif

    /* select VEOL to be CR, VERASE to NUL and VKILL to NUL */
    termio.c_cc[VEOL]   = '\r';
    termio.c_cc[VERASE] = '\0';
    termio.c_cc[VKILL]  = '\0';

    if (ioctl( SbFd, TCSETA, &termio) < 0) {
	sprintf( errormsg, "sbcustom: ioctl( \"%s\", TCGETA, &termio)",
	    tty_name
	);
	myperror( errormsg );
	close( SbFd );
	return SbFd = -1;	/* flag not open */
    }
#else  /*  SystemV == 0  */
    sgttyb.sg_flags = CBREAK | TANDEM;
    sgttyb.sg_erase = sgttyb.sg_kill = -1;
    sgttyb.sg_ispeed = sgttyb.sg_ospeed = B9600;
    if (ioctl( SbFd, TIOCSETP, &sgttyb ) < 0) {
	sprintf( errormsg, "ioctl( \"%s\", TIOCSETP, &sgttyb )", tty_name );
	myperror( errormsg );
	close( SbFd );
	return SbFd = -1;	/* flag not open */
    }
    if (ioctl( SbFd, TIOCEXCL, &sgttyb ) < 0) {
	sprintf( errormsg, "ioctl( \"%s\", TIOCEXCL, &sgttyb )", tty_name );
	myperror( errormsg );
	close( SbFd );
	return SbFd = -1;	/* flag not open */
    }
    if (fcntl( SbFd, F_SETFL, FNDELAY ) < 0) {
	sprintf( errormsg, "fcntl( \"%s\", F_SETFL, FNDELAY )", tty_name );
	myperror( errormsg );
	close( SbFd );
	return SbFd = -1;	/* flag not open */
    }
#endif/*  SystemV  */
#if Sb_USING_POLL
    SbFds[ 0 ].fd = SbFd;
#endif
    return SbFd;
}

SbClose()
{
    if (SbFd < 0) {	/* is device not open? */
	if (SbFd == -1)
	    fprintf( stderr,
		"sbcustom: Attempting to close an unopened Spaceball tty\n"
	    );
	SbFd = -2;	/* only print message once */
	return 0;
    }

    SbSetBallMode( "NN" );
    close( SbFd );
#if Sb_LOCK_TTY
    unlock_ttyline();
#endif
    SbFd = -1;
    return 0;
}

int SbInput( dest, maxlength, timeout )
char *dest;
int maxlength;
SbInt16 timeout;
{
    int numchars;
    SbInt16 time_left = 0;  /* skip first wait for input */
#if !Sb_USING_POLL
    static struct timeval timeval = { 0, 0 };
#endif

    if (SbFd < 0) {	/* is device not open? */
	if (SbFd == -1)
	    fprintf( stderr,
		"sbcustom: Attempting to read from unopened Spaceball tty\n"
	    );
	SbFd = -2;	/* only print message once */
	return 0;
    }

    timeout += SbMilliseconds();
    do {
	/* except for first time, wait for input to clear */
	if (time_left > 0) {
#if Sb_USING_POLL
	    SbFds[0].events = POLLIN;
	    poll( SbFds, 1L, timeout );
#else
	    timeval.tv_usec = time_left * 1000;
	    FD_SET( SbFd, &SbFdset );
	    select( SbFd + 1, &SbFdset, &zero_set, &zero_set, &timeval);
#endif
	}

	numchars = read( SbFd, dest, maxlength );
	if (numchars < 0) {
	    if (errno != EWOULDBLOCK && errno != EINTR)
		myperror( "sbcustom: Reading Spaceball tty" );
	    return 0;
	}
	if (numchars > 0) {
#if sgi
	    /* if flag set remove extra backslashes */
	    if (SbBackslash)
	        numchars = remove_extra_backslashes( dest, numchars );
#else
#if OS_HPUX || OS_SunOS || OS_Solaris || OS_Linux || OS_OSF1 || OS_AIX || OS_IRIX
#else
}}}}}}}}} require some platform!
#endif
#endif
		    
#ifdef Sb_DEBUG
	    if (Sb_debug & 1) {
		fprintf( output_fp,
		    "Received from Spaceball %d characters:\n", numchars
		);
		SbShowPacket( dest, numchars );
		putchar('\n');
	    }
#endif
	    return numchars;
	}
    } while ((time_left = timeout - SbMilliseconds()) > 0);
    return 0;
}

#if sgi
static remove_extra_backslashes( string, length )
char *string;
int length;
{
    char *pstart=string, *pend;
    int remove;

    /* a bug occurs with the current tty settings whereby
     * a string of backslashes is expanded upon input. eg.
     * "\\" becomes "\\\", "\\\" becomes "\\\\\", i.e.
     * for n input backslashes 2n-1 are generated.
     * Lets deal with this here...
     */

    for (; length>0; length--) {

	/* look for first backslash */
	if (*pstart == '\\') {

	    /* find end of backslashes */
	    for (pend=pstart+1; --length>0; pend++)
		if (*pend != '\\')
		    break;

	    /* calculate number to remove */
	    remove = (pend-pstart-1) / 2;

	    if ((remove > 0) && (length > 0))
		memcpy( pend-remove, pend, length );
	    pstart = pend-remove+1;
	} else
	    pstart++;
    }
    return pstart - string;
}
#endif

int SbOutput( data, length )
char *data;
int length;
{
    int num;
    SbInt16 time_left = 0,	/* skip first wait for output */
	timeout = 2000; 	/* 2 second timeout if forced to wait */
#if !Sb_USING_POLL
    static struct timeval timeval = { 0, 0 };
#endif

    if (SbFd < 0) {	/* is device not open? */
	if (SbFd == -1)
	    fprintf( stderr,
		"Attempt to write to unopened Spaceball tty\n"
	    );
	SbFd = -2;	/* only print message once */
	return 0;
    }
#ifdef Sb_DEBUG
    if (Sb_debug & 2) {
	fprintf( output_fp, "Writing %d characters to Spaceball:\n", length );
	SbShowPacket( data, length );
    }
#endif
    timeout += SbMilliseconds();
    do {
	/* except for first time, wait for output to clear */
	if (time_left > 0) {
#if Sb_USING_POLL
	    SbFds[0].events = POLLOUT;
	    poll( SbFds, 1L, timeout );
#else
	    FD_SET( SbFd, &SbFdset );
	    select( SbFd, &zero_set, &SbFdset, &zero_set, &timeval);
#endif
	}

	/* write data */
	if ((num = write( SbFd, data, length )) < 0) {
	    myperror( "sbcustom: Writing Spaceball tty" );
	    return -1;
	}
	if ((length -= num) <= 0)
	    return 0;
	data += num;

	/* restart output if suspended - this avoids
	 * hanging the process if the host believes
	 * it is XOFFed when it may not be */
#if SystemV
	ioctl( SbFd, TCXONC, 1 );
#else
#ifdef TIOCSETN
	ioctl( SbFd, TIOCSETN, &sgttyb );	/* hope this works? */
#endif
#endif

    } while ((time_left = timeout - SbMilliseconds()) > 0);
#ifdef Sb_DEBUG
    if (Sb_debug & 8)
	fprintf( debug_fp, "Only wrote %d of %d characters\n",
	    num, length+num
	);
#endif
    fprintf( error_fp, "sbcustom: Timeout writing to Spaceball tty\n" );
    errno = EWOULDBLOCK;
    return -1;		/* timeout error */
}

SbInt16 SbMilliseconds()
{
#if SystemV && !defined(OS_Solaris)
    struct tms tms;

#  if HZ <= 1000
    return times( &tms ) * (1000 / HZ);
#  else
    return (times( &tms ) * 1000) / HZ;
#  endif

#else
    struct timeval timeval;
    struct timezone timezone;

    gettimeofday( &timeval, &timezone );
    return timeval.tv_sec * 1000 + timeval.tv_usec / 1000;
#endif
}

static myperror( s )
char *s;
{
    extern int sys_nerr;

    if (*s != '\0')
	fprintf( error_fp, "%s: ", s );
    if (errno < sys_nerr)
	fprintf( error_fp, "%s\n", strerror (errno) );
    else
	fprintf( error_fp, "Undefined error number %d\n", errno );
}

#if Sb_LOCK_TTY

static char *filenameof( ttyname )
char *ttyname;
{
    char *p = ttyname;

    while (*ttyname!='\0')
	if (*ttyname++ == '/')
	    p = ttyname;
    return p;
}

static int lock_ttyline( ttyname )
char *ttyname;
{
    struct flock flock;
    struct utmp *utmpp;
    char pid_str[ 8 ];
    int existed;
    int locking_pid = -1;

    SbLock_name = (char *) malloc(
	strlen( lock_prefix ) + strlen( ttyname ) + 1
    );
    strcat( strcpy( SbLock_name, lock_prefix ), ttyname );

    /* open file - note if it already existed */

    do {
	SbLock_fp = fopen( SbLock_name, "r+" );
	existed = (SbLock_fp != NULL);
	if (!existed)
	    if ((SbLock_fp = fopen( SbLock_name, "w" )) == NULL ) {
		sprintf( errormsg, "sbcustom: open( %s, \"w\" )", SbLock_name );
		myperror( errormsg );
	    }
    } while (SbLock_fp == NULL);

    for (;;) {
	do {

	    /* attempt write lock the whole file */

	    flock.l_type = F_WRLCK;
	    flock.l_whence = 0;
	    flock.l_start = 0;
	    flock.l_len = 0;
	    if (fcntl( fileno( SbLock_fp ), F_SETLK, &flock ) != -1) {
#if 1
		if (existed) {
		    fscanf( SbLock_fp, " %7s", pid_str );
		    fprintf( output_fp
"Lock file \"%s\" present but its creator, process #%s,\n\
does not have it write locked - using %s\n",
			SbLock_name, pid_str, ttyname
		    );
		    /* seek to beginning */
		    rewind( SbLock_fp );
		}
#endif
		fprintf( SbLock_fp, "%d\n", getpid());
		fflush( SbLock_fp );
		return 0;	/* lock successful */
	    } else
		fcntl( fileno( SbLock_fp ), F_GETLK, &flock );
	} while (flock.l_type == F_UNLCK);

	/* the file is locked - print message about user */

	setutent();
	while ((utmpp = (struct utmp *) getutent()) != NULL) {
	    if (utmpp->ut_pid == flock.l_pid) {
		fprintf( stderr,
		    "Spaceball currently allocated to %s on %s.\n",
		    utmpp->ut_user, utmpp->ut_line
		);
		fclose ( SbLock_fp );
		return -1;
	    }
	}

	/* locking process may have just died */

	if (flock.l_pid != locking_pid) {
	    locking_pid = flock.l_pid;
	    fprintf( stderr,
"Lock file present and locked but locking process, #%d, is not!\n",
		flock.l_pid
	    );
	} else	/* something is very strange! - assume unavailable */
	    return -1;
    }
}

static unlock_ttyline()
{
    if (SbLock_fp != NULL) {
	if (close( fileno( SbLock_fp )) == -1)
	    myperror( "sbcustom: Closing lock file" );
	SbLock_fp = NULL;
	if (unlink( SbLock_name ) == -1)
	    myperror( "sbcustom: Deleting lock file" );
	free( SbLock_name );
	SbLock_name = NULL;
    }
}

#endif	/* Sb_LOCK_TTY */

