/*  spaceball.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1994-2000  Richard Gooch


    Written by      Richard Gooch   17-FEB-1994

    Updated by      Richard Gooch   20-FEB-1994: Added 3 dimensional cursor
  control.

    Updated by      Richard Gooch   22-FEB-1994: Added VX orientation display
  support.

    Updated by      Richard Gooch   22-FEB-1994: Added VX 3D cursor display
  support.

    Updated by      Richard Gooch   27-FEB-1994: Separated orientation and
  3D cursor transmission to VX.

    Updated by      Richard Gooch   1-MAR-1994: Provided different mapping for
  cursor control and limited cursor movement to inside subcube.

    Updated by      Richard Gooch   20-APR-1994: Added generic 3D cursor
  information information protocol.

    Updated by      Richard Gooch   21-MAY-1994: Added  #include <karma_ch.h>

    Updated by      Richard Gooch   1-AUG-1994: Improved some declarations.

    Updated by      Richard Gooch   6-AUG-1994: Moved to full ASCII function
  declaration syntax. Had to do this because of the confused Dec Alpha compiler

    Updated by      Richard Gooch   31-OCT-1994: Changed from  sun4sol2  to
  sparc_Solaris

    Updated by      Richard Gooch   3-NOV-1994: Switched to OS_ and MACHINE_
  macros for machine determination.

    Updated by      Richard Gooch   18-AUG-1995: Took account of changes to
  <geom_*> routines.

    Updated by      Richard Gooch   27-FEB-1996: Made use of <move_3d_cursor>
  routine.

    Updated by      Richard Gooch   23-MAR-1996: Improved responsiveness of
  3-D cursor by accumulating buffered up translation events and move cursor
  once.

    Updated by      Richard Gooch   19-APR-1996: Check for access to Spaceball
  before calling <SbOpen> to avoid error message.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   25-JUL-1996: Took account of change to
  <move_3d_cursor>.

    Updated by      Richard Gooch   23-AUG-1996: Took account of change to
  <move_3d_cursor>.

    Updated by      Richard Gooch   29-AUG-1996: Moved to <change_3d_cursor>

    Updated by      Richard Gooch   9-MAR-1998: Moved to new function
  <register_rotation_change>.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Last updated by Richard Gooch   26-JAN-2000: Removed unnecessary includes.


*/
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#define     Sb_ENVIRONMENT_INCLUDES
#include    "sblibry.h"
#include <karma.h>
#include <karma_vrender.h>
#include <karma_chx.h>
#include <karma_ch.h>
#include <karma_a.h>
#include <karma_r.h>
#include <karma_m.h>
#include "xraycast_3d.h"


#define SBMODE_TRANSLATE (unsigned int) 0
#define SBMODE_ROTATE (unsigned int) 1
#define CURSOR_MODE_NATURAL (unsigned int) 0
#define CURSOR_MODE_MOUSE (unsigned int) 1


/*  Private functions  */
static flag input_func (/* channel, info */);
static int mypressed (/* keys */);
static int mytranslate (/* period, vec */);
static int myrotate (/* period, vec */);


/*  Private data  */
static Channel sb_channel = NULL;
static unsigned int spaceball_mode = SBMODE_ROTATE;
static unsigned int cursor_mode = CURSOR_MODE_NATURAL;
static flag pending_translate = FALSE;
static Kcoord_3d new_cursor;


/*  Public functions follow  */

flag initialise_spaceball ()
/*  [PURPOSE] This routine will initialise the SpaceBall.
    [RETURNS] TRUE if a SpaceBall was found and initialsed, else FALSE.
*/
{
    int fd;
    CONST char *display_host;
    static flag initialised = FALSE;
    static SbUint16 mypulse[2] = { 20, 20 };
    static char function_name[] = "initialise_spaceball";

    if (initialised)
    {
	fprintf (stderr, "SpaceBall already initialised\n");
	a_prog_bug (function_name);
    }
    initialised = TRUE;
    if ( ( display_host = r_get_host_from_display ( r_getenv ("DISPLAY") ) )
	 == NULL )
    {
	m_abort (function_name, "getting display host");
    }
    if ( (strcmp (display_host, "unix") != 0) &&
	 (strcmp (display_host, "localhost") != 0) ) return (FALSE);
    if (access ("/dev/spaceball", R_OK) != 0) return (FALSE);
    fprintf (stderr, "Opening Spaceball...\n");
    if ( ( fd = SbOpen ("/dev/spaceball") ) < 0)
    {
	return (FALSE);
    }
    fprintf (stderr, "Opened Spaceball\n");
    if (SbReset (TRUE) == -1)
    {
	fprintf (stderr,
		 "Spaceball did not appear to respond on /dev/spaceball - assuming not turned on\n");
	return (FALSE);
    }
    if ( ( sb_channel = ch_attach_to_asynchronous_descriptor (fd) ) == NULL )
    {
	fprintf (stderr, "Error getting channel\n");
	a_prog_bug (function_name);
    }
    if ( !chx_manage (sb_channel, NULL, input_func,
		      ( void (*) () ) NULL, ( flag (*) () ) NULL,
		      ( flag (*) () ) NULL) )
    {
	fprintf (stderr, "Error managing channel\n");
	a_prog_bug (function_name);
    }
    SbSetBallMode( "SSV" );     /* turn translations and rotations on */
    SbFunctions.pressed = mypressed;
    SbFunctions.translate = mytranslate;
    SbFunctions.rotate = myrotate;
    SbPulse[0] = 20;
    SbPulse[1] = 20;
    SbRequestBallData ();
    SbSetPulse (mypulse);
    return (TRUE);
}   /*  End Function initialise_spaceball  */

void spin_sb (float roll, float pitch, float yaw)
/*  This routine will spin the cube, modifying the view specification
    determined by the SpaceBall.
    The roll, pitch and yaw increment must be given by  roll  ,  pitch  and
    yaw  ,respectively.
    The routine returns nothing.
*/
{
    rotation_specification rot;
    extern rotation_specification rotation;

    rot.roll = roll;
    rot.pitch = pitch;
    rot.yaw = yaw;
    rot.rho = rotation.rho;
    register_rotation_change (rot);
}   /*  End Function spin_sb  */

void close_spaceball ()
/*  This routine will close the SpaceBall cleanly.
    The routine returns nothing.
*/
{
    chx_unmanage (sb_channel);
    ch_close (sb_channel);
    SbSetBallMode ("NN");  /* turn off ball */
    SbClose ();
}   /*  End Function close_spaceball  */


/*  Private functions follow  */

static flag input_func (channel, info)
/*  This routine is called when new input occurs on a channel.
    The channel object is given by  channel  .
    An arbitrary pointer may be written to the storage pointed to by  info
    The pointer written here will persist until the channel is unmanaged
    (or a subsequent callback routine changes it).
    The routine returns TRUE if the channel is to remain managed and
    open, else it returns FALSE (indicating that the channel is to be
    unmanaged and closed). This routine MUST NOT unmanage or close the
    channel given by  channel  .
    Note that the  close_func  will be called if this routine returns FALSE
*/
Channel channel;
void **info;
{
    pending_translate = FALSE;
    SbCheckSpaceball ();
    if (!pending_translate) return (TRUE);
    change_3d_cursor (new_cursor, TRUE);
    return (TRUE);
}   /*  End Function input_func  */

static int mypressed (keys)
SbUint16 keys;
{
    SbBeep( "cE" );

    if (keys & 0x01)  /*  Key #1  */
    {
	spaceball_mode = SBMODE_ROTATE;
    }
    if (keys & 0x02)  /*  Key #2  */
    {
	if (spaceball_mode == SBMODE_ROTATE) compute ();
	spaceball_mode = SBMODE_TRANSLATE;
    }
    if (keys & 0x04)  /*  Key #3  */
    {
	cursor_mode = CURSOR_MODE_NATURAL;
    }
    if (keys & 0x08)  /*  Key #4  */
    {
	cursor_mode = CURSOR_MODE_MOUSE;
    }
    if (keys & 0x10)  /*  Key #5  */
    {
    }
    if (keys & 0x20)  /*  Key #6  */
    {
    }
    if (keys & 0x40)  /*  Key #7  */
    {
	reset_view ();
    }
    if (keys & 0x80)  /*  Key #8  */
    {
	SbRezero ();
    }
    if (keys & 0x100)
    {
	/*  Pick button  */
	compute ();
    }
    return (0);
}   /*  End Function mypressed  */

static int mytranslate (period, vec)
SbUint16 period;
SbReal vec[3];
{
    Kcoord_3d direction, hv, tmp3d;
    view_specification view;
    unsigned long x_s, y_s, z_s, x_e, y_e, z_e;
    double mag;
    float factor;
    float offset3d[3];
    extern KVolumeRenderContext main_vrend_ctx;
    extern float spaceball_sensitivity;

    if (spaceball_mode != SBMODE_TRANSLATE) return (0);
    vrender_get_context_attributes (main_vrend_ctx,
				    VRENDER_CONTEXT_ATT_VIEW, &view,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_START, &x_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_START, &y_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_START, &z_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_END, &x_e,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_END, &y_e,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_END, &z_e,
				    VRENDER_CONTEXT_ATT_END);
    factor = spaceball_sensitivity * 10.0;
    switch (cursor_mode)
    {
      case CURSOR_MODE_NATURAL:
	offset3d[0] = vec[0] * factor;
	offset3d[1] = vec[1] * factor;
	offset3d[2] = vec[2] * factor;
	break;
      case CURSOR_MODE_MOUSE:
	offset3d[0] = vec[0] * factor;
	offset3d[1] = vec[2] * factor;
	offset3d[2] = -vec[1] * factor;
	break;
    }
    direction.x = view.focus.x - view.position.x;
    direction.y = view.focus.y - view.position.y;
    direction.z = view.focus.z - view.position.z;
    mag = sqrt (direction.x * direction.x + direction.y * direction.y +
		direction.z * direction.z);
    direction.x /= mag;
    direction.y /= mag;
    direction.z /= mag;
    /*  Apply forward motion  */
    tmp3d.x = new_cursor.x + offset3d[2] * direction.x;
    tmp3d.y = new_cursor.y + offset3d[2] * direction.y;
    tmp3d.z = new_cursor.z + offset3d[2] * direction.z;
    /*  Apply up/down motion  */
    tmp3d.x += offset3d[1] * view.vertical.x;
    tmp3d.y += offset3d[1] * view.vertical.y;
    tmp3d.z += offset3d[1] * view.vertical.z;
    /*  Apply left/right motion  */
    /*  Compute horizontal vector  */
    geom_vector_multiply (&hv, direction, view.vertical);
    tmp3d.x += offset3d[0] * hv.x;
    tmp3d.y += offset3d[0] * hv.y;
    tmp3d.z += offset3d[0] * hv.z;
    /*  Even though <move_3d_cursor> now rejects bad cursors, need to check
	here as well because of the internal queueing  */
    if ( (tmp3d.x >= x_s) && (tmp3d.x <= x_e) ) new_cursor.x = tmp3d.x;
    if ( (tmp3d.y >= y_s) && (tmp3d.y <= y_e) ) new_cursor.y = tmp3d.y;
    if ( (tmp3d.z >= z_s) && (tmp3d.z <= z_e) ) new_cursor.z = tmp3d.z;
    pending_translate = TRUE;
    return (0);
}   /*  End Function mytranslate  */

static int myrotate (period, vec)
SbUint16 period;
SbReal vec[3];
{
    extern float spaceball_sensitivity;

    if (spaceball_mode == SBMODE_ROTATE)
    {
	spin_sb (-vec[2] * spaceball_sensitivity,
		 -vec[0] * spaceball_sensitivity,
		 -vec[1] * spaceball_sensitivity);
    }
    return (0);
}   /*  End Function myrotate  */
