/*  c2doc.c

    Source file for  c2doc  (extract documentation from C code).

    Copyright (C) 1993-1997  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma utility programme will read a C source file and extract
  documentation.


    Written by      Richard Gooch   1993

    Updated by      Richard Gooch   30-MAR-1996: Created this file from the old
  docstyle extractor.

    Updated by      Richard Gooch   6-APR-1996: New output format as per
  Patrick's requests.

    Uupdated by     Richard Gooch   12-APR-1996: Allow multiple source files.

    Updated by      Richard Gooch   15-APR-1996: Supported [SEE ALSO] tag.

    Updated by      Richard Gooch   24-JUL-1996: Supported experimental
  functions.

    Updated by      Richard Gooch   7-DEC-1996: Supported markups in summary
  line.

    Updated by      Richard Gooch   23-MAR-1997: Removed extraneous newline in
  function summary output.

    Last updated by Richard Gooch   24-MAR-1997: Deal with spaces in tags in
  <extract_plain_text> function.


*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <karma.h>


static int process_file (FILE *fp_in, FILE *fp_out);
static char *read_line (char *buffer, int length, FILE *fp);
#if 0
static char *skip_until (FILE *fp, char *string, char *out_line,
			 unsigned int length);
#endif
static void write_string (char *string, FILE *fp);
static void process_function (FILE *in, FILE *out, char *tag);
static void extract_func_name_and_type (char *line, FILE *fp);
static void extract_parameter_list (FILE *in, char *line, FILE *out);
static char *extract_plain_text (FILE *in, char *purpose_ptr,
				 char *line, int length, FILE *out);
static char *extract_parameter (FILE *in, char *start_ptr,
				char *line, int length, FILE *out);

static char indent4[] = "    ";

int main (argc, argv)
int argc;
char *argv[];
{
    int argcount;
    FILE *fp;
    extern int errno;
    /*static char usage_string[] = "Usage:\tc2doc [infile...]";*/

    if (argc == 1)
    {
	if ( !process_file (stdin, stdout) ) exit (1);
	exit (0);
    }
    for (argcount = 1; argcount < argc; ++argcount)
    {
	fprintf (stderr, "%s\n", argv[argcount]);
	if ( ( fp = fopen (argv[argcount], "r") ) == NULL )
	{
	    fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
			    argv[argcount], ERRSTRING);
	    exit (1);
	}
	if ( !process_file (fp, stdout) ) exit (1);
	fclose (fp);
    }
    return (0);
}   /*  End Function main  */

static int process_file (FILE *fp_in, FILE *fp_out)
/*  [SUMMARY] Process a source file.
    <fp_in> The input source file.
    <fp_out> The output documentation file.
    [RETURNS] TRUE on success, else FALSE
*/
{
    char txt[STRING_LENGTH + 1];

    /*  Process lines  */
    /*  Start looking for functions  */
    while (read_line (txt, STRING_LENGTH, fp_in) != NULL)
    {
	if (strstr (txt, "/*PUBLIC_FUNCTION*/") != NULL)
	{
	    /*  Found public function  */
	    process_function (fp_in, fp_out, "$FUNCTION");
	}
	else if (strstr (txt, "/*PROTOTYPE_FUNCTION*/") != NULL)
	{
	    /*  Found prototype function  */
	    process_function (fp_in, fp_out, "$PROTO_FUNC");
	}
	else if (strstr (txt, "/*EXPERIMENTAL_FUNCTION*/") != NULL)
	{
	    /*  Found experimental function  */
	    process_function (fp_in, fp_out, "$EXPERIMENTAL_FUNC");
	}
    }
    return (TRUE);
}   /*  End Function process_file  */

static char *read_line (char *buffer, int length, FILE *fp)
/*  [PURPOSE] This routine will read in a line from a file. The newline
    character is discarded.
    <buffer> The line is written here.
    <length> The length of the buffer.
    <fp> The file to read from.
    [RETURNS] <<buffer>> on success, else NULL.
*/
{
    int len;

    if (fgets (buffer, length, fp) == NULL) return (NULL);
    len = strlen (buffer);
    if (buffer[len - 1] == '\n') buffer[len - 1] = '\0';
    return (buffer);
}   /*  End Function read_line  */

#if 0
static char *skip_until (FILE *fp, char *string, char *out_line,
			 unsigned int length)
/*  [PURPOSE] This routine will read lines from a file until a specified line
    is found.
    <fp> The file to read from.
    <string> The string to search for.
    <out_line> The line containing the string is written here.
    <length> The length of <<out_line>>.
    [RETURNS] A pointer to the start of the found string, else NULL if it was
    not found.
*/
{
    char *ptr;

    while (read_line (out_line, length, fp) != NULL)
    {
	if ( ( ptr = strstr (out_line, string) ) != NULL )
	{
	    /*  Found string  */
	    return (ptr);
	}
    }
    return (NULL);
}   /*  End Function skip_until  */
#endif

static void write_string (char *string, FILE *fp)
/*  [SUMMARY] Write a string to a file.
    [PURPOSE] This routine will write a string to a file. Any embedded markups
    are translated.
    <string> The string to write.
    <fp> The file to write to.
    [RETURNS] Nothing.
*/
{
    int len;
    char *ptr;

    while (*string != '\0')
    {
	if (strncmp (string, "[<", 2) == 0)
	{
	    /*  Found an active link markup  */
	    string += 2;
	    if ( ( ptr = strstr (string, ">]") ) == NULL )
	    {
		fprintf (stderr, "Active link markup not closed\n");
		exit (1);
	    }
	    len = ptr - string;  /*  Length of embedded string  */
	    fputs ("<-", fp);
	    while (len > 0)
	    {
		putc (*string++, fp);
		--len;
	    }
	    string += 2;
	    fputs ("->", fp);
	}
	else putc (*string++, fp);
    }
}   /*  End Function write_string  */

static void process_function (FILE *in, FILE *out, char *tag)
/*  [PURPOSE] This routine will process a function.
    <in> The input file.
    <out> The output file.
    <tag> The tag that should be written to the output file.
    [RETURNS] Nothing.
*/
{
    char *ptr;
    char txt[STRING_LENGTH + 1];
    char summary[STRING_LENGTH + 1];

    /*  Another function to process: read function line. This must contain
	the type of the function and the function name. It may also contain
	some or all of the parameter list  */
    if (read_line (txt, STRING_LENGTH, in) == NULL) exit (1);
    fprintf (out, "\n%-20s", tag);
    extract_func_name_and_type (txt, out);
    fputs ("$PARAMETER_LIST\n", out);
    extract_parameter_list (in, txt, out);
    /*  Read what should be the summary line  */
    /*if (read_line (txt, STRING_LENGTH, in) == NULL) exit (1);*/
    if ( ( ptr = strstr (txt, "[SUMMARY]") ) == NULL )
    {
	fprintf (stderr, "Function has no summary!\n");
	fprintf (stderr, "line: \"%s\"\n", txt);
	exit (1);
    }
    /*  Jump past markup sub-string  */
    while ( !isspace (*ptr) ) ++ptr;
    /*  Skip leading white-space  */
    while ( isspace (*ptr) ) ++ptr;
    strcpy (summary, ptr);
    fprintf (out, "%-20s", "$SUMMARY");
    write_string (summary, out);
    putc ('\n', out);
    if (read_line (txt, STRING_LENGTH, in) == NULL) exit (1);
    if ( ( ptr = strstr (txt, "[PURPOSE]") ) == NULL )
    {
	/*  No purpose: use summary  */
	fprintf (out, "$PURPOSE\n%s", indent4);
	write_string (summary, out);
	putc ('\n', out);
	/*  Advance to next tag  */
	for (ptr = txt; isspace (*ptr); ++ptr);
    }
    else
    {
	fputs ("$PURPOSE\n", out);
	ptr = extract_plain_text (in, ptr, txt, STRING_LENGTH, out);
    }
    /*  ptr should now point to next tag  */
    while (ptr != NULL)
    {
	if (*ptr == '<') ptr = extract_parameter (in, ptr,
						  txt, STRING_LENGTH, out);
	else if (strncmp (ptr, "[NOTE]", 6) == 0)
	{
	    fputs ("$NOTE\n", out);
	    ptr = extract_plain_text (in, ptr, txt, STRING_LENGTH, out);
	}
	else if (strncmp (ptr, "[SEE ALSO]", 10) == 0)
	{
	    fputs ("$SEE_ALSO\n", out);
	    ptr = extract_plain_text (in, ptr, txt, STRING_LENGTH, out);
	}
	else if (strncmp (ptr, "[VARARGS]", 9) == 0)
	{
	    fputs ("$PARAMETER        ... :\n", out);
	    ptr = extract_plain_text (in, ptr, txt, STRING_LENGTH, out);
	}
	else if (strncmp (ptr, "[RETURNS]", 9) == 0)
	{
	    fputs ("$RETURNS\n", out);
	    ptr = extract_plain_text (in, ptr, txt, STRING_LENGTH, out);
	}
	else if (strncmp (ptr, "[MT-LEVEL]", 10) == 0)
	{
	    fputs ("$MT-LEVEL         ", out);
	    /*  Jump past markup sub-string  */
	    while ( !isspace (*ptr) ) ++ptr;
	    /*  Skip leading white-space  */
	    while ( isspace (*ptr) ) ++ptr;
	    fputs (ptr, out);
	    putc ('\n', out);
	    if (read_line (txt, STRING_LENGTH, in) == NULL) exit (1);
	    /*  Skip leading white-space  */
	    for (ptr = txt; isspace (*ptr); ++ptr);
	}
	else
	{
	    fprintf (stderr, "Illegal markup: \"%s\"\n", ptr);
	    exit (1);
	}
    }
    fputs ("$END\n", out);
}   /*  End Function process_function  */

static void extract_func_name_and_type (char *line, FILE *fp)
/*  [PURPOSE] This routine will extract the function name and type from a
    function line. This line must contain the type and name of the function.
    The implicit "int" type is not permitted: it must be explicit.
    <line> The function line.
    <fp> The file to write to.
    [PURPOSE] Nothing.
*/
{
    int name_length, type_length;
    char *ptr;
    char *name_start, *name_end, *type_start, *type_end;
    char name[STRING_LENGTH];
    char type[STRING_LENGTH];

    /*  Search for '(' in function line  */
    if ( ( ptr = strchr (line, '(') ) == NULL )
    {
	/*  There is no parameter list on this line  */
	ptr = line + strlen (line) - 1;
    }
    else
    {
	/*  There is the start of a parameter list on this line  */
	--ptr;
    }
    /*  Search backwards for non white-space  */
    while ( isspace (*ptr) ) --ptr;
    /*  Now should be positioned at the last character of the function name  */
    name_end = ptr;
    /*  Search backwards for white-space or '*'  */
    while ( !isspace (*ptr) && (*ptr != '*') ) --ptr;
    /*  Now should be positioned just before the function name  */
    name_start = ptr + 1;
    /*  Search backwards for non white-space  */
    while ( isspace (*ptr) ) --ptr;
    /*  Now should be positioned at the end of the type  */
    type_end = ptr;
    /*  Search from start of line for non white-space  */
    ptr = line;
    while ( isspace (*ptr) ) ++ptr;
    type_start = ptr;
    name_length = name_end - name_start + 1;
    type_length = type_end - type_start + 1;
    strncpy (name, name_start, name_length);
    name[name_length] = '\0';
    strncpy (type, type_start, type_length);
    type[type_length] = '\0';
    fputs (name, fp);
    putc ('\n', fp);
    fprintf (fp, "%-20s%s\n", "$RETURN_TYPE", type);
}   /*  End Function func_name_and_type  */

static char *extract_plain_text (FILE *in, char *start_ptr,
				 char *line, int length, FILE *out)
/*  [PURPOSE] This routine will extract an item of plain text.
    <in> The input file.
    <start_ptr> A pointer to the start of the markup sub-string.
    <line> A pointer to storage into which lines may be read.
    <length> The length of <<line>>.
    <out> The file to write to.
    [RETURNS] A pointer to the next item of interest. If there are no more
    items for this function, NULL is returned.
*/
{
    int keep_going = TRUE;
    char *ptr;

    /*  Jump past markup sub-string  */
    for (ptr = start_ptr; *ptr != ']'; ++ptr)
    {
	if (*ptr == '\0')
	{
	    fprintf (stderr, "Unterminated tag: \"%s\"\n", start_ptr);
	    return (NULL);
	}
    }
    ++ptr;
    /*  Now process lines, discarding the leading white-space and write  */
    while ( isspace (*ptr) ) ++ptr;
    while (keep_going)
    {
	if ( ( (ptr[0] == '<') && (ptr[1] != '<') ) || 
	     ( (ptr[0] == '[') && (ptr[1] != '<') ) )
	{
	    keep_going = FALSE;
	    continue;
	}
	if (strncmp (ptr, "*/", 2) == 0) return (NULL);
	fputs (indent4, out);
	write_string (ptr, out);
	putc ('\n', out);
	if (read_line (line, length, in) == NULL) exit (1);
	for (ptr = line; isspace (*ptr); ++ptr);
    }
    /*  This is the start of a new item: stop now  */
    return (ptr);
}   /*  End Function extract_plain_text  */

static void extract_parameter_list (FILE *in, char *line, FILE *out)
/*  [PURPOSE] This routine will extract the parameter list.
    <in> The input file.
    <line> The line containing the function name. This may contain the start of
    the parameter list. This will be updated with the first comment line.
    <out> The file to write to.
    [RETURNS] Nothing.
*/
{
    char *ptr;

    fputs (indent4, out);
    if ( ( ptr = strchr (line, '(') ) == NULL )
    {
	/*  Function line does not contain start of parameter list: read a line
	    to get it  */
	if (read_line (line, STRING_LENGTH, in) == NULL) exit (1);
    }
    if ( ( ptr = strchr (line, '(') ) == NULL )
    {
	fprintf (stderr, "No parameter list\n");
	exit (1);
    }
    /*  Now have pointer to start of parameter list  */
    while ( (line[0] != '/') && (line[1] != '*') )
    {
	/*  Write out previous line  */
	fputs (ptr, out);
	putc ('\n', out);
	/*  Have not reached end of parameter list: read and write another line
	 */
	if (read_line (line, STRING_LENGTH, in) == NULL) exit (1);
	ptr = line;
    }
}   /*  End Function extract_parameter_list  */

static char *extract_parameter (FILE *in, char *start_ptr,
				char *line, int length, FILE *out)
/*  [PURPOSE] This routine will extract a single parameter.
    <in> The input file.
    <start_ptr> A pointer to the '<' character.
    <line> A pointer to storage into which lines may be read.
    <length> The length of <<line>>.
    <out> The file to write to.
    [RETURNS] A pointer to the next item of interest.
*/
{
    int keep_going = TRUE;
    char *ptr, *end_ptr;

    ++start_ptr;
    if ( ( end_ptr = strchr (start_ptr, '>') ) == NULL )
    {
	fprintf (stderr, "No closing bracket for parameter\n");
	exit (1);
    }
    ptr = end_ptr + 1;
    *end_ptr = '\0';
    fprintf (out, "%-20s%s :\n", "$PARAMETER", start_ptr);
    /*  Now print the first line, discarding leading white-space  */
    fputs (indent4, out);
    while ( isspace (*ptr) ) ++ptr;
    write_string (ptr, out);
    putc ('\n', out);
    /*  Now process lines, discarding the leading white-space and write  */
    if (read_line (line, length, in) == NULL) exit (1);
    for (ptr = line; isspace (*ptr); ++ptr);
    while (keep_going)
    {
	if ( (ptr[0] == '<') && (ptr[1] != '<') )
	{
	    keep_going = FALSE;
	    continue;
	}
	if ( (ptr[0] == '[') && (ptr[1] != '<') )
	{
	    keep_going = FALSE;
	    continue;
	}
	fputs (indent4, out);
	write_string (ptr, out);
	putc ('\n', out);
	if (read_line (line, length, in) == NULL) exit (1);
	for (ptr = line; isspace (*ptr); ++ptr);
    }
    /*  This is the start of a new item: stop now  */
    return (ptr);
}   /*  End Function extract_parameter  */
