/*  codeview.c

    Source file for  codeview  (code-count viewer).

    Copyright (C) 1994-2000  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This programme will display a code-count file onto the screen.


    Written by      Richard Gooch   14-NOV-1994

    Updated by      Richard Gooch   18-NOV-1994

    Updated by      Richard Gooch   7-DEC-1994: Stripped declaration of  errno
  and added #include <errno.h>

    Updated by      Richard Gooch   4-JAN-1995: Use  XtNameToWidget  to get
  child widgets.

    Updated by      Richard Gooch   9-FEB-1999: Increased canvas size.

    Updated by      Richard Gooch   15-AUG-1999: Made egcs happy.

    Last updated by Richard Gooch   28-JAN-2000: Allow any event to update
  cursor display.


*/
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_canvas.h>
#include <karma_ch.h>
#include <karma_ex.h>
#include <karma_a.h>
#include <karma_m.h>
#include <Xkw/Canvas.h>


/*  Private functions  */
STATIC_FUNCTION (void setup_display, () );
STATIC_FUNCTION (void quit_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (void setup_canvas, (KPixCanvas pixcanvas) );
STATIC_FUNCTION (void worldcanvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale, Kcolourmap cmap,
		  flag cmap_resize, void **info) );
STATIC_FUNCTION (flag position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info, void **f_info,
		  double x_lin, double y_lin) );

typedef struct
{
    char *name;
    int days;
} month;


/*  Private data  */
String fallback_resources[] =
{
    "CodeView*pseudocolourCanvas*background:       black",
    "CodeView*pseudocolourCanvas*foreground:       white",
    "CodeView*Command*background:                  grey70",
    "CodeView*toggle*background:                   grey80",
    "CodeView*closeButton*background:              grey90",
    "CodeView*menuButton*background:               turquoise",
    "CodeView*topForm*quit*background:             orange",
    "CodeView*background:                          aquamarine",
    "CodeView*topForm*trackOutput*font:            8x13bold",
    "CodeView*topForm*zoomMenu*font:               10x20",
    "CodeView*topForm*font:                        9x15bold",
    "CodeView*topForm*zoomMenu*unZoom*foreground:  red",
    "CodeView*font:                                9x15bold",
    NULL
};
static unsigned int num_lines = 0;
static unsigned int num_entries;
static unsigned int vscale = 0;
static double *data = NULL;
static Widget track_label;
static month months[12] =
{
    {"Jan", 31},
    {"Feb", 28},
    {"Mar", 31},
    {"Apr", 30},
    {"May", 31},
    {"Jun", 30},
    {"Jul", 31},
    {"Aug", 31},
    {"Sep", 30},
    {"Oct", 31},
    {"Nov", 30},
    {"Dec", 31}
};


int main (int argc, char **argv)
{
    Channel channel;
    unsigned int tmp1, tmp2, acc;
    unsigned int num_arr, lcount, ecount;
    double last_date = 0.0;
    double days_in_year;
    char *ptr;
    char **entry_ptr;
    char buffer[STRING_LENGTH];
    char cmpbuf[STRING_LENGTH];
    static char usage_string[] = "Usage:\tcodeview infile entries...";
    static char function_name[] = "main";

    if (argc < 3)
    {
	fprintf (stderr, "%s\n", usage_string);
	exit (RV_MISSING_PARAM);
    }
    num_entries = argc - 2;
    if (num_entries > 1)
    {
	/*  Multiple entries: will need to total  */
	num_arr = num_entries + 2;
    }
    else
    {
	num_arr = num_entries + 1;
    }
    entry_ptr = argv + 2;
    if ( ( channel = ch_open_file (argv[1], "r") ) == NULL )
    {
	fprintf (stderr, "Error opening file: \"%s\" for reading\t%s\n",
		 argv[1], ERRSTRING);
	exit (RV_CANNOT_OPEN);
    }
    while ( ch_getl (channel, buffer, STRING_LENGTH) ) ++num_lines;
    if ( !ch_seek (channel, 0) )
    {
	fprintf (stderr, "Error rewinding file: \"%s\"\t%s\n",
		 argv[1], ERRSTRING);
	exit (RV_UNDEF_ERROR);
    }
    if ( ( data = m_alloc (sizeof *data * num_lines * num_arr) ) == NULL )
    {
	m_abort (function_name, "data array");
    }
    m_clear (data, sizeof *data * num_lines * num_arr);
    /*  Read in data  */
    for (lcount = 0; lcount < num_lines; ++lcount)
    {
	if ( !ch_getl (channel, buffer, STRING_LENGTH) )
	{
	    fprintf (stderr, "Error reading\t%s\n", ERRSTRING);
	    exit (RV_READ_ERROR);
	}
	/*  Extract day/year  */
	if (sscanf (buffer, "%u %u", &tmp1, &tmp2) != 2)
	{
	    fprintf (stderr, "Bad day/date\n");
	    exit (RV_BAD_DATA);
	}
	--tmp1;
	days_in_year = (tmp2 % 4 == 0) ? 366.0 : 365.0;
	data[lcount] = (double) tmp2 + (double) tmp1 / days_in_year;
	if (data[lcount] < last_date)
	{
	    fprintf (stderr, "Time index must not decrease!\n");
	    exit (RV_BAD_DATA);
	}
	last_date = data[lcount];
	for (ecount = 0, acc = 0; ecount < num_entries; ++ecount)
	{
	    cmpbuf[0] = '*';
	    cmpbuf[1] = '\0';
	    strcat (cmpbuf, entry_ptr[ecount]);
	    if ( ( ptr = strstr (buffer, cmpbuf) ) == NULL ) continue;
	    if ( ( ptr = strchr (ptr, ' ') ) == NULL ) continue;
	    if (sscanf (++ptr, "%u", &tmp1) != 1) continue;
	    data[(ecount + 1) * num_lines + lcount] = tmp1;
	    acc += tmp1;
	}
	if (acc > vscale) vscale = acc;
	if (num_entries > 1)
	{
	    /*  Need total  */
	    data[(num_entries + 1) * num_lines + lcount] = acc;
	}
    }
    ch_close (channel);
    fprintf (stderr, "vscale: %u\n", vscale);
    setup_display ();
    return (0);
}   /*  End Function main  */


/*  Private routines follow  */

static void setup_display ()
{
    KPixCanvas pixcanvas;
    int argc = 1;
    XtAppContext app_context;
    Widget main_shell, topform;
    Widget quit_btn;
    Widget pseudo_cnvs;
    static char *argv[1] = {"codeview"};

    /*  Start up Xt  */
    main_shell = XtVaAppInitialize (&app_context, "CodeView", NULL, 0,
				    &argc, argv, fallback_resources,
				    NULL, 0);
    topform = XtVaCreateManagedWidget ("topForm", formWidgetClass, main_shell,
				       XtNorientation, XtorientHorizontal,
				       XtNhSpace, 0,
				       XtNborderWidth, 0,
				       NULL);
    quit_btn = XtVaCreateManagedWidget ("quit", commandWidgetClass, topform,
					XtNlabel, "Quit",
					XtNtop, XtChainTop,
					XtNbottom, XtChainTop,
					NULL);
    XtAddCallback (quit_btn, XtNcallback, quit_cbk, NULL);
    track_label = XtVaCreateManagedWidget ("trackOutput", labelWidgetClass,
					   topform,
					   XtNlabel, "Track Output",
					   XtNwidth, 500,
					   XtNfromHoriz, quit_btn,
					   XtNtop, XtChainTop,
					   XtNbottom, XtChainTop,
					   NULL);
    pseudo_cnvs = XtVaCreateManagedWidget ("pseudocolourCanvas",
					   canvasWidgetClass,
					   topform,
					   XtNfromVert, quit_btn,
					   XtNtop, XtChainTop,
					   XtNbottom, XtChainBottom,
					   XtNleft, XtChainLeft,
					   XtNright, XtChainRight,
					   XtNwidth, 800,
					   XtNheight, 400,
					   XkwNsilenceUnconsumed, True,
					   NULL);

    XtRealizeWidget (main_shell);
    XtVaGetValues (pseudo_cnvs,
		   XkwNmonoPixCanvas, &pixcanvas,
		   NULL);
    setup_canvas (pixcanvas);
    XtAppMainLoop (app_context);
}   /*  End Function setup_display  */

static void quit_cbk (w, client_data, call_data)
/*  This is the quit button callback.
*/
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    exit (RV_OK);
}   /*  End Function quit_cbk   */

static void setup_canvas (KPixCanvas pixcanvas)
{
    KWorldCanvas worldcanvas;
    struct win_scale_type win_scale;

    /*  Create the main world canvas (with dummy world min and max)  */
    canvas_init_win_scale (&win_scale, K_WIN_SCALE_MAGIC_NUMBER);
    win_scale.x_min = data[0];
    win_scale.x_max = data[num_lines - 1];
    win_scale.y_min = 0.0;
    win_scale.y_max = vscale;
    win_scale.z_scale = K_INTENSITY_SCALE_LINEAR;
    win_scale.conv_type = CONV1_REAL;
    if ( ( worldcanvas = canvas_create (pixcanvas, (Kcolourmap) NULL,
					&win_scale) )
	== NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    canvas_register_refresh_func (worldcanvas, worldcanvas_refresh_func,
				  (void *) data);
    canvas_register_position_event_func (worldcanvas, position_func,
					 (void *) data);
}   /*  End Function setup_canvas  */

static void worldcanvas_refresh_func (canvas, width, height, win_scale,
				      cmap, cmap_resize, info)
/*  This routine is a refresh event consumer for a world canvas.
    The canvas is given by  canvas  .
    The width of the canvas in pixels is given by  width  .
    The height of the canvas in pixels is given by  height  .
    The window scaling information is pointed to by  win_scale  .
    The colourmap associated with the canvas is given by  cmap  .
    If the refresh function was called as a result of a colourmap resize
    the value of  cmap_resize  will be TRUE.
    The arbitrary canvas information pointer is pointed to by  info  .
    The routine returns nothing.
*/
KWorldCanvas canvas;
int width;
int height;
struct win_scale_type *win_scale;
Kcolourmap cmap;
flag cmap_resize;
void **info;
{
    unsigned int count;
    unsigned long pixel_value;
    double px, py;
    double *data;

    data = (double *) *info;
    if ( !canvas_get_colour (canvas, "blue", &pixel_value,
			     NULL, NULL, NULL) ) return;
    for (count = 0; count < num_lines; ++count)
    {
	px = data[count];
	py = 0.0;
	canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
					   &px, &py, &px, &py);
	kwin_draw_line (canvas_get_pixcanvas (canvas), px, 0.0, px, 3.0,
			pixel_value);
    }
    if ( !canvas_get_colour (canvas, "white", &pixel_value,
			     (unsigned short *) NULL, (unsigned short *) NULL,
			     (unsigned short *) NULL) ) return;
    for (count = 0; count < num_entries; ++count)
    {
	canvas_draw_lines_p (canvas, data, data + (count + 1) * num_lines,
			     num_lines, pixel_value);
    }
    if (num_entries > 1)
    {
	if ( !canvas_get_colour (canvas, "red", &pixel_value,
				 (unsigned short *) NULL,
				 (unsigned short *) NULL,
				 (unsigned short *) NULL) ) return;
	canvas_draw_lines_p (canvas, data,
			     data + (num_entries + 1) * num_lines, num_lines,
			     pixel_value);
    }
}   /*  End Function worldcanvas_refresh_func  */

static flag position_func (KWorldCanvas canvas, double x, double y,
			   unsigned int event_code, void *e_info,
			   void **f_info, double x_lin, double y_lin)
/*  This routine is a position event consumer for a world canvas.
    The canvas is given by  canvas  .
    The horizontal world co-ordinate of the event will be given by  x  .
    The vertical world co-ordinate of the event will be given by  y  .
    These values will have been transformed by the registered transform
    function (see  canvas_register_transform_func  ).
    The arbitrary event code is given by  event_code  .
    The arbitrary event information is pointed to by  e_info  .
    The arbitrary function information pointer is pointed to by  f_info  .
    The linear world co-ordinates (the co-ordinates prior to the transform
    function being called) will be given by  x_lin  and  y_lin  .
    The routine returns TRUE if the event was consumed, else it returns
    FALSE indicating that the event is still to be processed.
*/
{
    int year, day_of_year, first_day, first_year, day;
    int count, mcount, dcount;
    double days_in_year;
    double *data;
    char window_string[STRING_LENGTH];
    /*static char function_name[] = "position_func";*/

    data = (double *) *f_info;
    /*  First calculate first day  */
    first_year = (int) floor (data[0]);
    days_in_year = (first_year % 4 == 0) ? 366.0 : 365.0;
    first_day = (int) floor ( (data[0] - (double) first_year) * days_in_year );
    /*  Now calculate current day  */
    year = (int) floor (x);
    days_in_year = (year % 4 == 0) ? 366.0 : 365.0;
    months[1].days = (year % 4 == 0) ? 29 : 28;
    day_of_year = (int) floor ( (x - (double) year) * days_in_year );
    for (count = 0, mcount = 0, dcount = 0; count < day_of_year; ++count)
    {
	if (++dcount >= months[mcount].days)
	{
	    dcount = 0;
	    ++mcount;
	}
    }
    /*  Now calculate current day relative to first day  */
    day = day_of_year - first_day;
    for (count = first_year; count < year; ++count)
    {
	day += (count % 4 == 0) ? 366 : 365;
    }
    sprintf (window_string,
	     "Day: %d  day_of_year: %d   %d-%s-%d    num_lines: %d",
	     day + 1, day_of_year + 1, dcount + 1, months[mcount].name,
	     year, (int) y);
    XtVaSetValues (track_label, XtNlabel, window_string, NULL);
    return (event_code == K_CANVAS_EVENT_POINTER_MOVE) ? TRUE : FALSE;
}   /*  End Function position_func  */
